/*
 * ProfilerInformation.cpp - module to gather profile information and treat is properly.
 *
 * Copyright (c) 2000, 2001, 2010 - University of Virginia 
 *
 * This file is part of the Memory Error Detection System (MEDS) infrastructure.
 * This file may be used and modified for non-commercial purposes as long as 
 * all copyright, permission, and nonwarranty notices are preserved.  
 * Redistribution is prohibited without prior written consent from the University 
 * of Virginia.
 *
 * Please contact the authors for restrictions applying to commercial use.
 *
 * THIS SOURCE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Author: University of Virginia
 * e-mail: jwd@virginia.com
 * URL   : http://www.cs.virginia.edu/
 *
 * Additional copyrights 2010, 2011 by Zephyr Software LLC
 * e-mail: {clc,jwd}@zephyr-software.com
 * URL   : http://www.zephyr-software.com/
 */

#include <string>
#include <bitset>

#include <cassert>
#include <cstdlib>
#include <ctime>
#include <cstring>

using namespace std;

#include "base/ProfilerInformation.h"
#include "base/SMPProgram.h"

#define SMP_DEBUG_DATA_GRANULARITY 0
#define SMP_VERBOSE_DEBUG_DATA_GRANULARITY 0

#define SMP_GRANULARITY_STATISTICS 1
#define SMP_GRANULARITY_STATISTICS_VERBOSE 0

#if 0
static int MultiProblemCount = 0; // MultiAccess, not really memset()-type insts
#endif
static int IgnoredProfilerCount = 0; // PROF_FIELD annotations with negative sizes
static int SizeMismatchStack = 0;
static int SizeMismatchHeap = 0;
static int SizeMismatchGlobal = 0;

/*
 * ascii-to-longlong conversion; Visual Studio 2005 lacks atoll() library func
 *
 * no error checking; assumes decimal digits
 *
 * efficient conversion: 
 *   start with value = 0
 *   then, starting at first character, repeat the following
 *   until the end of the string:
 *
 *     new value = (10 * (old value)) + decimal value of next character
 *
 */

long long my_atoll(char *instr) {
	long long retval = 0;

	for (; *instr; ++instr) {
		retval = 10*retval + (*instr - '0');
	}
	return retval;
}

#ifdef WIN32
#define atoll my_atoll
#endif

ProfilerInformation::ProfilerInformation(const char* fn, SMPProgram *CurrProg)
	: filename(std::string(fn))
{

	FILE* fin;
	int addr;
	union {int size, type;} size_type_u;
	char type[200];
	char scope[200];
	char remainder[320000];
	int line = 0;
	this->ProfilerAnnotsRead = 0;

	assert(NULL != CurrProg);
	this->MyProg = CurrProg;
	this->MemProfileInfo = new MemoryAccessInfo(CurrProg);
	this->the_map.clear();
	this->indirect_call_map.clear();
	this->constant_info.clear();

	// For debugging, we can add a delay loop so we have time to attach gdb to the
	// running process and set a breakpoint.
#define SMP_DEBUG_DELAY 0
#if SMP_DEBUG_DELAY
	time_t start;
	time_t current;

	time(&start);
	printf("delay for 15 seconds.\n");
	do {
		time(&current);
	} while(difftime(current,start) < 15.0);
#endif

	if (fn[0] == 0)
		return;

	fin = SMP_fopen(fn, "r");
	if (!fin)
	{
		SMP_msg("WARNING: Cannot open strata annotation file %s\n", fn);
		this->the_map.clear();
		return;
	}

	do 
	{
		SMP_fscanf(fin, "%x %d\n", &addr, &(size_type_u.size));

		if (SMP_feof(fin))		// deal with blank lines at the EOF
			break;
		
		SMP_fscanf(fin, "%s%s", type, scope);

		if (strcmp(type, "FUNC") == 0)
		{
			; // no useful info for the SA.
			
		}
		else if (strcmp(type, "MEMORYHOLE") == 0)
		{
			// no useful info for the SA.
		}
		else if (strcmp(type, "LOCALFRAME") == 0)
		{
			// no useful info for the SA.
		}
		else if (strcmp(type, "INARGS") == 0)
		{
			// no useful info for the SA.
		}
		else if (strcmp(type, "OBJECT") == 0)
		{
			// no useful info for the SA.
		}
		else if (strcmp(type, "INSTR") == 0)
		{
			if (strcmp(scope, "INDIRECTCALL")==0)
			{
				int target;
				SMP_fscanf(fin, "%d", &target);

				indirect_call_map[addr].insert(target); 
				this->ProfilerAnnotsRead++;
			}
			else if (strcmp(scope, "PROF_FIELD") == 0) {
#if 1
				// record the PROF_FIELD into the constant info structure so it's output with the other
				// profile information.  Then rewind and parse out the information correctly.
				long pos = SMP_ftell(fin);
				char *ResultStr = SMP_fgets(remainder, sizeof(remainder), fin);
				assert(NULL != ResultStr);
				int ResultCode = SMP_fseek(fin, pos, SEEK_SET);
				assert(0 == ResultCode);
				std::string output_string(remainder);
#if 1
				SMP_snprintf(remainder, sizeof(remainder), "%llx %d %s %s ", (unsigned long long)addr, size_type_u.size, "INSTR", "PROF_FIELD");
#else
				qsnprintf(remainder, sizeof(remainder), "%llx %d %s %s ", (unsigned long long)addr, size_type_u.size, "INSTR", "PROF_FIELD");
#endif
				output_string = string(remainder) + output_string;
				constant_info.push_back(*new std::string(output_string));
#endif			
				this->MemProfileInfo->ProcessMemoryAccessAnnotation(fin, addr);
				this->ProfilerAnnotsRead++;
			}
		}
		else if (strcmp(type, "PTRIMMEDEBP") == 0 || 
			strcmp(type, "PTRIMMEDESP") == 0 || 
			strcmp(type, "PTRIMMEDESP2") == 0 || 
			strcmp(type, "PTRIMMEDABSOLUTE") == 0
			   )
		{
			int the_const;
			int real_const = 0;
			char field[100];
			assert(strcmp(scope, "STACK") == 0 || strcmp(scope, "GLOBAL") == 0);

			/* remaining params are <const> <field> <real_const_if_global> <comment> */ 
			SMP_fscanf(fin, "%d %s", &the_const, field);
			if (strcmp(type,"PTRIMMEDESP2") == 0 || 
				strcmp(type,"PTRIMMEDABSOLUTE") == 0)
				SMP_fscanf(fin, "%x", &real_const);
			else
				real_const = the_const;

			// successfully read in, but, ignoring for now.
			if (strcmp(type, "PTRIMMEDESP2") == 0 || strcmp(type, "PTRIMMEDABSOLUTE") == 0)
			{
				profileAddressingInformation(addr, size_type_u.size,
					type, scope, the_const, field, real_const);
			}
			
		}
		else if (strcmp(type, "DATAREF") == 0)
		{
			// no useful info for the SA.

		}
		else if (strcmp(type, "DEALLOC") == 0)
		{
			// no useful info for the SA.
			assert(strcmp(scope, "STACK") == 0);
			/* remaining params: comment */
		}
		else if (strcmp(type, "PROFILEDNUMERIC") == 0)
		{
			/* profiler generated information */
			addProfileInformation(addr, atoll(scope), 0, 0);
			this->ProfilerAnnotsRead++;
		}
		else if (strcmp(type, "PROFILEDPOINTER") == 0 )
		{
			/* profiler generated information */
			addProfileInformation(addr, 0, atoll(scope), 0);
			this->ProfilerAnnotsRead++;
		}
		else if (strcmp(type, "PROFILEDOTHER") == 0 )
		{
			/* profiler generated information */
			addProfileInformation(addr, 0, 0, atoll(scope));
			this->ProfilerAnnotsRead++;
		}
		else if (strcmp(type, "BLOCK") == 0)
		{
			;  // not processing BLOCK profiling annotation yet
		}
		else
		{
			SMP_msg("ERROR: Unknown annotation type '%s' near line %d\n", type, line);
			assert(0);
		}
	
		int TempChar;
		do {
			TempChar = SMP_fgetc(fin);
		} while ((TempChar != '\n') && (TempChar != EOF));
		line++;
	} while (!SMP_feof(fin));

	int ReturnCode = SMP_fclose(fin);
	if (0 != ReturnCode) {
		SMP_msg("FATAL ERROR: Failed to close annotation file %s in ProfilerInformation. ReturnCode: %d\n", fn, ReturnCode);
		assert(0 == ReturnCode);
	}

	SMP_msg("Successfully loaded annotation file %s\n", fn);
} // end of ProfilerInformation constructor

ProfilerInformation::~ProfilerInformation() {
	delete this->MemProfileInfo;
	FILE* fout = SMP_fopen(filename.c_str(), "a+");
	
	if (!fout) {
		SMP_msg("Cannot open annotations file (%s) for output\n", filename.c_str());		
		assert(fout);
	}

	for (std::list<std::string>::iterator iter=constant_info.begin();
		iter != constant_info.end();
		++iter
	   )
	{
		SMP_fprintf(fout, "%s", (*iter).c_str());
	}

	for (std::map<STARS_ea_t,InstructionInformation*>::iterator iter = the_map.begin();
		iter != the_map.end();
		++iter
	   )
	{
		STARS_ea_t addr=(*iter).first;
		InstructionInformation *ii=(*iter).second;
		if (ii)
		{

			if (ii->getNumericCount())
			{
				SMP_fprintf(fout, "%llx %d %s %lld Profiler generated, summarized by idal\n",
					(unsigned long long) addr, 0, "PROFILEDNUMERIC", ii->getNumericCount());
			}
			if (ii->getPointerCount())
			{
				SMP_fprintf(fout, "%llx %d %s %lld Profiler generated, summarized by idal\n",
					(unsigned long long) addr, 0, "PROFILEDPOINTER", ii->getPointerCount());
			}
			if (ii->getOtherCount())
			{
				SMP_fprintf(fout, "%llx %d %s %lld Profiler generated, summarized by idal\n",
					(unsigned long long) addr, 0, "PROFILEDOTHER", ii->getOtherCount());
			}
			delete ii;
		}
	}

	for (std::map<STARS_ea_t, set<STARS_ea_t> >::iterator iter = indirect_call_map.begin();
		iter != indirect_call_map.end();
		++iter
	   )
	{
		for (std::set<STARS_ea_t>::iterator iter2 = (*iter).second.begin();
			iter2 != (*iter).second.end();
			++iter2
		   )
		{
			STARS_ea_t source=iter->first;
			STARS_ea_t target=*iter2;
			SMP_fprintf(fout, "%llx 0 INSTR INDIRECTCALL %llx Profiler generated, idal summarized\n", (unsigned long long) source, (unsigned long long) target);
		}
	}

	SMP_fclose(fout);
	this->constant_info.clear();
}

void ProfilerInformation::addProfileInformation(STARS_ea_t addr, long long nc, long long pc, long long oc) {
	InstructionInformation *ii = GetInfo(addr);

	if (!ii) {
		ii = new InstructionInformation;
		SetInfo(addr, ii);
	}

	if (nc)
		ii->setNumericCount((int) nc);
	if (pc)
		ii->setPointerCount((int) pc);
	if (oc)
		ii->setOtherCount((int) oc);

//	SMP_msg("Profiled %x to be nc:%lld pc:%lld oc:%lld\n", addr,
//		ii->getNumericCount(),
//		ii->getPointerCount(),
//		ii->getOtherCount());

	return;
}

void ProfilerInformation::profileAddressingInformation
	(
		STARS_ea_t addr,
		int size,
		char *type,
		char *scope, 
		STARS_ea_t the_const,
		char* field, 
		STARS_ea_t real_const
	)
{
	char buffer[1000];
	(void) SMP_snprintf(buffer, sizeof(buffer), "%llx %d %s %s %lu %s %llx Profiler generated, idal summarized\n", 
		(unsigned long long) addr, size, type, scope, (unsigned long) the_const, field, (unsigned long long) real_const); 
	constant_info.push_back(std::string(buffer));
}

// ******************************************************
// Class SMPMemoryAccesses
// ******************************************************

// Constructor.
SMPMemoryAccesses::SMPMemoryAccesses(STARS_ea_t addr) {
	address = addr;
	ByteLimit = 0;
	region = SMP_NO_REGION;
	ProfTypes.clear();
	return;
}

// Meet function over memory access types.
void SMPMemoryAccesses::MeetAccessType(size_t index, SMPAccessType TypeSeen) {
	SMPAccessType CurrentType = this->GetAccessType(index);
	if (SMP_NO_ACCESS == CurrentType) {
		// We move from the lattice top to TypeSeen regardless of what it is.
		this->SetAccessType(index, TypeSeen);
	}
	else if (SMP_DIRECT == CurrentType) {
		// We move from direct to scaled or indirect.
		if (SMP_SCALED <= TypeSeen) {
			this->SetAccessType(index, TypeSeen);
		}
	}
	else if (SMP_SCALED == CurrentType) {
		// lattice bottom; once a scaled access is seen, it stays scaled.
		; // do nothing
	}
	else {
		assert(SMP_INDIRECT == CurrentType);
		// Only change would be if TypeSeen is scaled.
		if (SMP_SCALED == TypeSeen) {
			this->SetAccessType(index, SMP_SCALED);
		}
	}

	return;
} // end of SMPMemoryAccesses::MeetAccessType()

// ******************************************************
// Class MemoryAccessInfo
// ******************************************************

// Constructor.
MemoryAccessInfo::MemoryAccessInfo(SMPProgram *CurrProg) {
	this->MyProg = CurrProg;
	this->MultipleAccessTypeSet.clear();
	this->InstrAccessMap.clear();
	this->InstrObjectMap.clear();
	this->ObjectInstrMap.clear();
	this->InstrTypeMap.clear();
	this->ObjectTypeMap.clear();
	this->TypeAccessMap.clear();
	return;
} // end of MemoryAccessInfo::MemoryAccessInfo()

//Destructor
MemoryAccessInfo::~MemoryAccessInfo() {
	this->InstrAccessMap.clear();
	this->InstrObjectMap.clear();
	this->MultipleAccessTypeSet.clear();
	this->ObjectInstrMap.clear();
	this->InstrTypeMap.clear();
	this->ObjectTypeMap.clear();
	this->TypeAccessMap.clear();
	return;
} // end of MemoryAccessInfo destructor

// DEBUG dump.
void MemoryAccessInfo::Dump(void) {
	SMP_msg("Debug dump of MemoryAccessInfo:\n");
	
	set<STARS_ea_t>::iterator MultiIter;
	SMP_msg("\nMultiAccess set: ");
	int count = 0;
	for (MultiIter = this->MultipleAccessTypeSet.begin(); 
		MultiIter != this->MultipleAccessTypeSet.end();
		++MultiIter) {

		SMP_msg("%lx ", (unsigned long) (*MultiIter));
		++count;
		if (0 == (count % 10))
			SMP_msg("\n");
	}
	SMP_msg("\n");

	map<STARS_ea_t, size_t>::iterator InstType;
	SMP_msg("\nInst to Type map: ");
	count = 0;
	for (InstType = this->InstrTypeMap.begin(); InstType != this->InstrTypeMap.end(); ++InstType) {
		SMP_msg("%lx %zu ", (unsigned long) InstType->first, InstType->second);
		++count;
		if (0 == (count % 8))
			SMP_msg("\n");
	}
	SMP_msg("\n");

	map<STARS_ea_t, size_t>::iterator ObjType;
	SMP_msg("\nObject to Type map: ");
	count = 0;
	for (ObjType = this->ObjectTypeMap.begin(); ObjType != this->ObjectTypeMap.end(); ++ObjType) {
		SMP_msg("%lx %zu ", (unsigned long) ObjType->first, ObjType->second);
		++count;
		if (0 == (count % 8))
			SMP_msg("\n");
	}
	SMP_msg("\n");

#define SMP_DEBUG_TYPE_NUMBER 1175
	if ((0 <= SMP_DEBUG_TYPE_NUMBER) 
		&& (this->TypeAccessMap.size() > SMP_DEBUG_TYPE_NUMBER)) {

		size_t TypeNo = SMP_DEBUG_TYPE_NUMBER;
		SMP_msg("\nAccess map for type %zu\n", TypeNo);
		count = 0;
		for (size_t ByteNo = 0; ByteNo < this->TypeAccessMap.at(TypeNo)->GetLimit(); ++ByteNo) {
			SMPAccessType CurrType = this->TypeAccessMap.at(TypeNo)->GetAccessType(ByteNo);
			SMP_msg("%zu ", ByteNo);
			if (CurrType == SMP_NO_ACCESS)
				SMP_msg("U ");
			else if (CurrType == SMP_DIRECT)
				SMP_msg("D ");
			else if (CurrType == SMP_INDIRECT)
				SMP_msg("I ");
			else {
				assert(CurrType == SMP_SCALED);
				SMP_msg("S ");
			}
			++count;
			if (0 == (count % 10))
				SMP_msg("\n");
		}
		SMP_msg("\n");
	}

	return;
} // end of MemoryAccessInfo::Dump()

// Process the information in the rest of a line that is of type
//  INSTR PROF_FIELD (i.e. a profiler-generated map of memory access
//  types from the instruction at address "addr").
void MemoryAccessInfo::ProcessMemoryAccessAnnotation(FILE *fin, int addr) {

	int CurrRegion = SMP_NO_REGION;
	// Extract the remaining fields of the annotation.
	char NameAddrsSize[STARS_MAXSTR];
	char ObjectName[STARS_MAXSTR];
	STARS_ea_t AllocationAddress;  // == 0 for static; code address otherwise
	STARS_ea_t InstructionAddress; // redundant; == addr
	size_t ObjectSize; // in bytes
	int ProcessID;  // of Profiler process that gathered data
	size_t RedundantObjectSize;  // == ObjectSize
	size_t BitVectorSize; // in bits
	size_t HexGroups; // how many 32-bit hex fields for bit vector

	SMP_fscanf(fin, "%s", NameAddrsSize);
	char *EndPtr;
// #if defined(WIN32) && (IDA_SDK_VERSION < 640)
// Visual Studio has deprecated strtok() and puts out a warning. Disable it
//  because the Visual Studio replacement, strtok_s, is non-standard.
// #pragma warning(disable:4996)
// #elif (IDA_SDK_VERSION >= 640)
	// IDA 6.4 and later won't allow strtok, use thread-safe IDA-provided qstrtok
	// Better idea: Use the new C++11 strtok_s() when we fully convert to C++11 instead of C++0x
	char *saveptr;
#define strtok(STR, DELIM) strtok_r(STR, DELIM, &saveptr);
// #endif
	char *NamePtr = strtok(NameAddrsSize, ":");
	SMP_strncpy(ObjectName, NamePtr, STARS_MAXSTR-1);
	char *TempStr = strtok(NULL, ":");
	AllocationAddress = (STARS_ea_t) strtol(TempStr, &EndPtr, 16);
	TempStr = strtok(NULL, ":");
	InstructionAddress = (STARS_ea_t) strtol(TempStr, &EndPtr, 16);
	TempStr = strtok(NULL, " ");
	signed long SignedObjectSize = strtol(TempStr, &EndPtr, 10);
	assert(addr == ((int) InstructionAddress));

	if (0 > SignedObjectSize) {
		// A negative object size is a code from the profiler indicating
		//  that the allocation address allocates objects of different
		//  sizes. This is similar to having the instruction address
		//  operate on objects of different sizes. We must punt on
		//  this instruction address to be safe.
#if SMP_VERBOSE_DEBUG_DATA_GRANULARITY
		SMP_msg("Ignoring PROF_FIELD annotation with size -1 at %x\n",
			InstructionAddress);
#endif
		this->MultipleAccessTypeSet.insert(InstructionAddress);
		++IgnoredProfilerCount;
		return;
	}

	ObjectSize = (size_t) SignedObjectSize;
	string ObjectString(ObjectName);
	SMP_fscanf(fin, "%d %zu %zu %zu", &ProcessID, &RedundantObjectSize,
		&BitVectorSize, &HexGroups);
	assert(RedundantObjectSize == ObjectSize);
	assert(BitVectorSize == (3 * ObjectSize));
	assert(HexGroups >= (BitVectorSize / 32));

	// If allocation address is 0, we have a static global object.
	// (The exception is the incoming stack when the startup code
	//  gets control of the program. This contains environment data
	//  and should be treated as a special stack frame. The allocation
	//  address is 0 because mmStrata did not yet have control of the
	//  application when the operating system allocated that part of
	//  the stack.)
	//  Get the data address corresponding to the object name.
	//  We will keep the data address in AllocationAddress so that
	//  one variable can be the key for certain maps.
	if (0 == strcmp("sr", ObjectName)) {
		CurrRegion = SMP_STACK;
	}
	else if (0 == AllocationAddress) {
		// static global data is not allocated at an instruction
		map<string, STARS_ea_t>::iterator NameMapIter;
		NameMapIter = this->MyProg->FindGlobalName(ObjectString);
		if (NameMapIter == this->MyProg->GetLastGlobalName()) {
			SMP_msg("WARNING: Cannot find global name %s ; discarding memory access annotation.\n",
				ObjectName);
			return;
		}
		AllocationAddress = NameMapIter->second;
		CurrRegion = SMP_GLOBAL;
	}
	else {
		CurrRegion = SMP_HEAP;
	}

	// Add to ObjectInstrMap.
	map<STARS_ea_t, set<STARS_ea_t> >::iterator ObjInstIter;
	ObjInstIter = this->ObjectInstrMap.find(AllocationAddress);
	if (ObjInstIter == this->ObjectInstrMap.end()) {
		// Not already in map; create map entry.
		pair<map<STARS_ea_t, set<STARS_ea_t> >::iterator, bool> ObjPairIB;
		set<STARS_ea_t> EmptyInstrs;
		pair<STARS_ea_t, set<STARS_ea_t> > TempObjPair(AllocationAddress, EmptyInstrs);
		ObjPairIB = this->ObjectInstrMap.insert(TempObjPair);
		assert(true == ObjPairIB.second);
		ObjInstIter = ObjPairIB.first;
	}
	ObjInstIter->second.insert(InstructionAddress);

	// Add to InstrObjectMap.
	map<STARS_ea_t, set<STARS_ea_t> >::iterator InstObjIter;
	InstObjIter = this->InstrObjectMap.find(InstructionAddress);
	if (InstObjIter == this->InstrObjectMap.end()) {
		// Not already in map; create map entry.
		pair<map<STARS_ea_t, set<STARS_ea_t> >::iterator, bool> InstPairIB;
		set<STARS_ea_t> EmptyObjs;
		pair<STARS_ea_t, set<STARS_ea_t> > TempInstPair(InstructionAddress, EmptyObjs);
		InstPairIB = this->InstrObjectMap.insert(TempInstPair);
		assert(true == InstPairIB.second);
		InstObjIter = InstPairIB.first;
	}
	InstObjIter->second.insert(AllocationAddress);

	// If no entry exists in the InstrAccessMap for this instruction
	//  address, create one and update its access map using the bit
	//  vectors from the annotation.
	// Else, if size conflict with existing address, add the allocation
	//  address to the MultipleAccessTypeSet and do no more work on this
	//  annotation.
	// Else (no size conflict) use the meet function to update the
	//  memory access map for this instruction address.
	map<STARS_ea_t, SMPMemoryAccesses *>::iterator InstrAccessIter;
	InstrAccessIter = this->InstrAccessMap.find(InstructionAddress);
	if (InstrAccessIter != this->InstrAccessMap.end()) {
		// Already have seen this instruction address.
		// Check for size mismatch.
		size_t ExistingObjectSize = InstrAccessIter->second->GetLimit();
		if (ExistingObjectSize != ObjectSize) {
			// Size mismatch.
			this->MultipleAccessTypeSet.insert(InstructionAddress);
#if SMP_GRANULARITY_STATISTICS
			if (CurrRegion == SMP_GLOBAL) {
				++SizeMismatchGlobal;
				SMP_msg("Global size mismatch at inst %lx\n", (unsigned long) InstructionAddress);
			}
			else if (CurrRegion == SMP_STACK) {
				++SizeMismatchStack;
				SMP_msg("Stack size mismatch at inst %lx\n", (unsigned long) InstructionAddress);
			}
			else {
				SMP_msg("Heap size mismatch at inst %lx\n", (unsigned long) InstructionAddress);
				++SizeMismatchHeap;
			}
#endif
			return;
		}
	}
	else {
		// First time we have seen this instruction address.
		SMPMemoryAccesses *TempMemoryAccess = new SMPMemoryAccesses(InstructionAddress);
		pair<STARS_ea_t, SMPMemoryAccesses *> TempPair(InstructionAddress, TempMemoryAccess);
		pair<map<STARS_ea_t, SMPMemoryAccesses *>::iterator, bool> PairIB;
		PairIB = this->InstrAccessMap.insert(TempPair);
		assert(true == PairIB.second);
		InstrAccessIter = PairIB.first;
		assert(InstrAccessIter != this->InstrAccessMap.end());
		InstrAccessIter->second->SetLimit(ObjectSize);
	}

	// At this point, InstrAccessIter points to the right access map entry
	//  and there is no size mismatch. We can update the access map using
	//  the meet function over the new accesses recorded in the bit vector
	//  in the annotation.
	InstrAccessIter->second->SetRegion(CurrRegion);
	size_t BytesProcessed = 0;
	unsigned int CurrHexGroup;
	size_t BitsProcessed = 0; // progresses 0 -> 1 -> 2 -> 0 ...
	bool Direct = false, Scaled = false, Indirect = false; // current byte
	bool HasDirect = false, HasNonDirect = false; // across all bytes
	for (size_t index = 0; index < HexGroups; ++index) {
		// Guard against profiler producing too many HexGroups.
		if (BytesProcessed >= ObjectSize) {
			SMP_msg("ERROR: Too many hex groups in annotation for address %lx\n",
				(unsigned long) InstructionAddress);
			break;
		}

		SMP_fscanf(fin, "%x", &CurrHexGroup);
		if (SMP_feof(fin)) {
			SMP_msg("ERROR: Premature end of line on memory access annotation for address %lx\n",
				(unsigned long) InstructionAddress);
			break;
		}

		for (size_t BitIndex = 0; BitIndex < 32; ++BitIndex) {
			bool CurrAccess = (0 != (0x00000001 & CurrHexGroup));
			if (0 == BitsProcessed)
				Direct = CurrAccess;
			else if (1 == BitsProcessed)
				Scaled = CurrAccess;
			else {
				assert(2 == BitsProcessed);
				Indirect = CurrAccess;
				// Pass results of current 3 bits in priority order
				//  to the meet function.
				if (Scaled) {
					InstrAccessIter->second->MeetAccessType(BytesProcessed, SMP_SCALED);
					HasNonDirect = true;
				}
				else if (Indirect) {
					InstrAccessIter->second->MeetAccessType(BytesProcessed, SMP_INDIRECT);
					HasNonDirect = true;
				}
				else if (Direct) {
					InstrAccessIter->second->MeetAccessType(BytesProcessed, SMP_DIRECT);
					HasDirect = true;
				}
				++BytesProcessed;
				if (BytesProcessed >= ObjectSize)
					break;
			}
			BitsProcessed = (BitsProcessed + 1) % 3;
			CurrHexGroup = CurrHexGroup >> 1;
		}  // end for all bits in current hex group
	} // end for all HexGroups

	// The profiler defines DIRECT to include accessing bytes from the
	//  start of the object. So, if register ebx points to the start
	//  of the object, then [ebx], [ebx+4], [ebx+8] are all DIRECT
	//  accesses. If ebx is an induction variable, it will create
	//  the odd pattern of having DIRECT accesses at the beginning
	//  of the object and INDIRECT accesses later. This combination
	//  should be converted to all INDIRECT accesses, while an
	//  instruction that has all DIRECT accesses should be left alone.
	if (HasDirect && HasNonDirect) {
		// Convert the DIRECT bytes to INDIRECT accesses.
		for (size_t index = 0; index < ObjectSize; ++index) {
			if (SMP_DIRECT == InstrAccessIter->second->GetAccessType(index)) {
				InstrAccessIter->second->SetAccessType(index, SMP_INDIRECT);
			}
		}
	}

	return;
} // end of MemoryAccessInfo::ProcessMemoryAccessAnnotation()

// After all profiler memory access annotations have been processed, infer
//  the fine-grained data structures of the program.
void MemoryAccessInfo::InferDataGranularity(void) {
	// For each entry in MultipleAccessSet, remove all its information
	//  from all other maps.
	set<STARS_ea_t>::iterator MultiIter;
	map<STARS_ea_t, set<STARS_ea_t> >::iterator InstObjIter;
	set<STARS_ea_t>::iterator ObjIter;
	map<STARS_ea_t, set<STARS_ea_t> >::iterator ObjInstIter;

#if SMP_GRANULARITY_STATISTICS
	SMP_msg("MultiAccess instructions: %zu\n", this->MultipleAccessTypeSet.size());
	SMP_msg("Ignored PROF_FIELD with negative size: %d\n", IgnoredProfilerCount);
	SMP_msg("Size mismatch: Global: %d\n", SizeMismatchGlobal);
	SMP_msg("Size mismatch: Stack : %d\n", SizeMismatchStack);
	SMP_msg("Size mismatch: Heap  : %d\n", SizeMismatchHeap);
#endif

	for (MultiIter = this->MultipleAccessTypeSet.begin(); MultiIter != this->MultipleAccessTypeSet.end(); ++MultiIter) {
#if SMP_VERBOSE_DEBUG_DATA_GRANULARITY
		SMP_msg("Processing MultiAccess address %x\n", *MultiIter);
#endif
		InstObjIter = this->InstrObjectMap.find(*MultiIter);
		if (InstObjIter == this->InstrObjectMap.end()) {
			// Instruction had only a -1 size coded PROF_FIELD annotation,
			//  so it is not present in the other maps.
			continue;
		}
		ObjIter = InstObjIter->second.begin();
		while (ObjIter != InstObjIter->second.end()) {
			// For each object accessed from the multi-type code address,
			//  remove all record of the access from the multi-type code addr.
			ObjInstIter = this->ObjectInstrMap.find(*ObjIter);
			assert(ObjInstIter != this->ObjectInstrMap.end());
			// Remove the multi-type access code address from the set
			//  of code addresses that access the object.
			ObjInstIter->second.erase(*MultiIter);
			// If that was the only code address that accessed the object,
			//  we have no granularity info on the object so we can remove
			//  the object from the ObjInstrMap.
			if (ObjInstIter->second.empty()) {
				this->ObjectInstrMap.erase(ObjInstIter);
			}
			++ObjIter;
		} // end while not end of obj addresses
		// We can now remove the entry in the InstrObjectMap, because
		//  this Instr was multi-type access.
		this->InstrObjectMap.erase(InstObjIter);
		// Likewise, we can delete the entry in the InstrAccessMap for
		//  this code address.
		this->InstrAccessMap.erase(*MultiIter);
	} // end for all entries in the MultipleAccessTypeSet

	// We now group together allocation addresses into a vector of
	//  data types. Now that the MultipleAccessTypeSet code addresses
	//  no longer affect the access maps, we can infer that two data
	//  allocation addresses are of the same type if they are accessed
	//  from a common instruction address. This inference will help us
	//  aggregate more profiler info during the final granularity inference
	//  stage.
	size_t TypeIndex = 0; // each data type gets a unique ID
	this->InstrTypeMap.clear();
	this->TypeAccessMap.clear();
	for (InstObjIter = this->InstrObjectMap.begin(); 
		InstObjIter != this->InstrObjectMap.end();
		++InstObjIter) {

		// See if the instruction address has already been associated
		//  with a type number. If not, associate it with a new type #.
		STARS_ea_t InstAddr = InstObjIter->first;
		size_t CurrTypeIndex;
		map<STARS_ea_t, SMPMemoryAccesses*>::iterator InstAccessIter;
		InstAccessIter = this->InstrAccessMap.find(InstAddr);
		assert(InstAccessIter != this->InstrAccessMap.end());
		size_t ObjectSize = InstAccessIter->second->GetLimit();
		size_t TypeSize;
		assert(0 < ObjectSize);
		map<STARS_ea_t, size_t>::iterator InstTypeIter;
		InstTypeIter = this->InstrTypeMap.find(InstAddr);
		if (InstTypeIter != this->InstrTypeMap.end()) {
			// Already associated with a type ID.
			CurrTypeIndex = InstTypeIter->second;
			TypeSize = this->TypeAccessMap.at(CurrTypeIndex)->GetLimit();
		}
		else { // Need to create new type ID
			CurrTypeIndex = TypeIndex++;
			// Allocate and initialize the SMPMemoryAccesses vector entry.
			SMPMemoryAccesses *TempAccesses = new SMPMemoryAccesses(InstAddr);
			TempAccesses->SetLimit(ObjectSize);
			TypeSize = ObjectSize;
			this->TypeAccessMap.push_back(TempAccesses);
			// Associate all instruction addresses that process this data type
			//  with the data type ID.
			this->AssociateInstrWithType(InstAddr, CurrTypeIndex, InstObjIter);
		}
		// Perform the meet function for the access type info associated
		//  with the current instruction.
		if (ObjectSize != TypeSize) {
			SMP_msg("ERROR: Changing type size for type %zu TypeSize %zu InstAddr %lx ObjectSize %zu\n",
				CurrTypeIndex, TypeSize, (unsigned long) InstAddr, ObjectSize);
		}
		else {
			for (size_t ByteIndex = 0; ByteIndex < ObjectSize; ++ByteIndex) {
				this->TypeAccessMap.at(CurrTypeIndex)->MeetAccessType(ByteIndex,
					InstAccessIter->second->GetAccessType(ByteIndex));
			}
		}
	} // end for all instructions in the InstrObjectMap

#if SMP_DEBUG_DATA_GRANULARITY
	this->Dump();
#endif

	return;
} // end of MemoryAccessInfo::InferDataGranularity()

// Associate instruction and data allocation addresses with the data TypeID
//  for the common data type they all reference.
void MemoryAccessInfo::AssociateInstrWithType(STARS_ea_t InstAddr, size_t TypeID,
								map<STARS_ea_t, set<STARS_ea_t> >::iterator InstObjIter) {
	pair<STARS_ea_t, size_t> InstPair(InstAddr, TypeID);
	this->InstrTypeMap.insert(InstPair);
	set<STARS_ea_t>::iterator ObjIter = InstObjIter->second.begin();
	while (ObjIter != InstObjIter->second.end()) {
		// For each data object, if it is not already associated with the
		//  TypeID, associate it and recurse through all unassociated
		//  instruction addresses that access the data object.
		map<STARS_ea_t, size_t>::iterator ObjTypeIter;
		ObjTypeIter = this->ObjectTypeMap.find(*ObjIter);
		if (ObjTypeIter != this->ObjectTypeMap.end()) {
			// Already associated with a type ID #.
			assert(TypeID == ObjTypeIter->second);
		}
		else {
			// Not yet associated with TypeID.
			pair<STARS_ea_t, size_t> ObjPair(*ObjIter, TypeID);
			this->ObjectTypeMap.insert(ObjPair);
			// Find all inst addresses that access this object.
			map<STARS_ea_t, set<STARS_ea_t> >::iterator ObjInstIter;
			ObjInstIter = this->ObjectInstrMap.find(*ObjIter);
			assert(ObjInstIter != this->ObjectInstrMap.end());
			set<STARS_ea_t>::iterator InstIter = ObjInstIter->second.begin();
			while (InstIter != ObjInstIter->second.end()) {
				// Is this inst address not yet associated with TypeID?
				map<STARS_ea_t, size_t>::iterator InstTypeIter;
				InstTypeIter = this->InstrTypeMap.find(*InstIter);
				if (InstTypeIter == this->InstrTypeMap.end()) {
					// Not yet in InstrTypeMap
					pair<STARS_ea_t, size_t> InstPair2(*InstIter, TypeID);
					this->InstrTypeMap.insert(InstPair2);
					// Recurse on new inst address.
					map<STARS_ea_t, set<STARS_ea_t> >::iterator InstObjIter2;
					InstObjIter2 = this->InstrObjectMap.find(*InstIter);
					assert(InstObjIter2 != this->InstrObjectMap.end());
					this->AssociateInstrWithType(*InstIter, TypeID, InstObjIter2);
				}
				else {
					assert(TypeID == InstTypeIter->second);
				}
				++InstIter;
			} // end while all inst addresses in ObjInstIter's set
		} // end if (ObjTypeIter != this->ObjectTypeMap.end()) ... else ...
		++ObjIter;
	} // end while all object addresses in InstObjIter's set
	return;
} // end of MemoryAccessInfo::AssociateInstrWithType()

// For the given instruction address, find the limits of the child region that
//  can be accessed by non-direct (i.e. indirect or scaled) accesses.
// Return false if there are no non-direct accesses through this instruction
//  address.
// Return true if there are non-direct accesses through this instruction,
//  and set ChildOffset/RegionSize to mark the non-direct child region.
bool MemoryAccessInfo::ComputeNonDirectAccessRegion(STARS_ea_t InstAddr, int &ChildOffset, int &RegionSize) {
	bool NonDirect = false;
	bool Scaled = false;
	size_t TypeID;

	map<STARS_ea_t, size_t>::iterator InstTypeIter;
	InstTypeIter = this->InstrTypeMap.find(InstAddr);
	if (InstTypeIter == this->InstrTypeMap.end())  {
		// InstAddr was not hit in the profiling run.
#if SMP_GRANULARITY_STATISTICS_VERBOSE
		SMP_msg("Store not hit in profiling run: %x\n", InstAddr);
#endif
		return false;
	}
	else {
		TypeID = InstTypeIter->second;
	}

	map<STARS_ea_t, SMPMemoryAccesses *>::iterator InstAccessIter;
	InstAccessIter = this->InstrAccessMap.find(InstAddr);
	assert(InstAccessIter != this->InstrAccessMap.end());

	// If the InstAccess was to a stack frame of a caller (not the
	//  stack frame of the function including InstAddr), and the
	//  function including InstAddr has multiple callers, then we
	//  have the dangerous case in which the profiling run only exercised
	//  one path to this function but later application runs could
	//  exercise paths from the other callers to this function. This
	//  would have made InstAddr a multiple-access instruction if the
	//  profiling run had detected it. We want to treat InstAddr
	//  as a multiple-access instruction by not emitting CHILDACCESS
	//  annotations for it.
	// EXAMPLE: Functions A and B call C, but only A calls C in the
	//  profiling run. Each caller passes a ptr to an object within
	//  its stack frame. InstAddr in C uses that pointer to write
	//  a result to memory. Based on the profiling run and the structure
	//  of the stack frame in A, we might emit a CHILDACCESS annotation
	//  restricting accesses from InstAddr to bytes 20-39 inside the
	//  memory referent. During the application run, the path from B to C
	//  is taken, and mmStrata tries to limit InstAddr to accessing
	//  bytes 20-39 within the stack frame of B. But the stack frame
	//  of B might have a completely different structure than the stack
	//  frame of A, so false positives will result. There is no annotation
	//  we can emit right now that works for the two different cases
	//  with two different callers. In fact, we would have to have either
	//  a union of the byte ranges for all callers, or a context sensitive
	//  annotation in which mmStrata checks the calling context, and we don't
	//  have profiler help for either approach, as only one path was hit.
	if (SMP_STACK & InstAccessIter->second->GetRegion()) {
		// Inst accessed a stack frame. Now, see if it accessed more
		//  than one object.
		map<STARS_ea_t, set<STARS_ea_t> >::iterator InstObjIter;
		InstObjIter = this->InstrObjectMap.find(InstAddr);
		assert(InstObjIter != this->InstrObjectMap.end());
		if (1 < InstObjIter->second.size()) {
			// Multiple objects accessed. At least one is a stack
			//  frame. If any object accessed is not a stack frame, then
			//  we have a dangerous coincidence in which a heap or global
			//  object happens to have the same size as a stack frame,
			//  in which case we don't want to emit a CHILDACCESS annotation
			//  as the fine grained structures cannot possibly match.
			//  More likely, multiple stack frames are being accessed,
			//  and their fine grained structures are unlikely to match.
			SMP_msg("No CHILDACCESS annotation for multi-caller frame access at %lx\n",
				(unsigned long) InstAddr);
			return false;
		}
		else {
			// Only accessed one object, which was a stack frame. If it is
			//  a caller's stack frame, and there is more than one potential
			//  caller to the present instruction's function, then we
			//  have the incomplete profiling situation detailed above.
			STARS_ea_t InstFirstEA, InstLastEA;
			STARS_Function_t *InstFunc = SMP_get_func(InstAccessIter->first);
			assert(NULL != InstFunc);
			InstFirstEA = InstFunc->get_startEA();
			InstLastEA = InstFunc->get_endEA();
			STARS_ea_t AllocationAddr = *(InstObjIter->second.begin());
			bool AccessedCallerFrame = 
				((InstFirstEA > AllocationAddr) || (InstLastEA < AllocationAddr));
			if (AccessedCallerFrame) {
				// Accessed stack frame of a caller. Final test is to
				//  see if this callee has multiple possible callers
				//  according to IDA's cross references.
				SMPFunction *InstFunction = this->MyProg->FindFunction(InstFirstEA);
				assert(NULL != InstFunction);
				if (1 < InstFunction->GetNumCallers()) {
					// Multiple callers. Only one caller called this
					//  callee during the profiling run, but others
					//  could call it during any application run.
					SMP_msg("No CHILDACCESS annotation for multi-caller partially profiled frame access at %lx\n",
						(unsigned long) InstAddr);
					return false;
				}
			}
		}
	} // end if SMP_STACK region access

	// Find the first and last non-directly accessed byte offsets from
	//  this particular InstAddr.
	ChildOffset = 0;
	size_t LastNonDirectByte = 0;
	size_t index;
	for (index = 0; index < InstAccessIter->second->GetLimit(); ++index) {
		SMPAccessType CurrAccess = InstAccessIter->second->GetAccessType(index);
		if ((SMP_SCALED == CurrAccess) || (SMP_INDIRECT == CurrAccess)) {
			if (!NonDirect) {
				NonDirect = true;
				ChildOffset = (int) index;
			}
			LastNonDirectByte = index;
			Scaled = Scaled || (SMP_SCALED == CurrAccess);
		}
	}
	if (!NonDirect) {
#if SMP_GRANULARITY_STATISTICS_VERBOSE
		SMP_msg("Store had only direct accesses in profiling run: %x\n", InstAddr);
#endif
		return false;
	}

	// Now, expand the region to include adjacent non-directly accessed bytes
	//  within the data type, using all profiled info from all instruction
	//  addresses that accessed the data type.
	int LowerBound = ChildOffset - 1;
	while (LowerBound >= 0) {
		SMPAccessType CurrAccess = this->TypeAccessMap.at(TypeID)->GetAccessType(LowerBound);
		// Treat SMP_NO_ACCESS (gaps in profiling) as INDIRECT if they
		//  border INDIRECT or SCALED (i.e. coalesce all contiguous regions
		//  that are not SMP_DIRECT access).
		if (SMP_DIRECT == CurrAccess) {
			break;
		}
		--ChildOffset; // update to new lower bound
		--LowerBound;
	}

	index = LastNonDirectByte + 1;
	size_t TypeLimit = this->TypeAccessMap.at(TypeID)->GetLimit();
	while (index < TypeLimit) {
		SMPAccessType CurrAccess = this->TypeAccessMap.at(TypeID)->GetAccessType(index);
		// Treat SMP_NO_ACCESS (gaps in profiling) as INDIRECT if they
		//  border INDIRECT or SCALED (i.e. coalesce all contiguous regions
		//  that are not SMP_DIRECT access).
		if (SMP_DIRECT == CurrAccess) {
			break;
		}
		LastNonDirectByte = index; // update to new upper bound
		++index;
	}

	RegionSize = ((int) (LastNonDirectByte + 1 - ChildOffset));

	// One common difference between profiling runs and real application
	//  runs is that arrays are only partially accessed. If a scaled access
	//  occurs to only 4 bytes or fewer, it could be a sign of such a
	//  profiling run. It would be likely to produce false positives if
	//  we return a scaled region of 4 or fewer bytes. This is one of the
	//  few cases in which we treat scaled and indirect differently.
	if (Scaled && (RegionSize <= 4)) {
#if SMP_GRANULARITY_STATISTICS_VERBOSE
		SMP_msg("SCALED Store to 4 or fewer bytes: %x\n", InstAddr);
#endif
		return false;
	}

	// We only care about non-direct accesses to children, not parent
	//  objects. Return true only if the whole parent was not accessed.
#if SMP_GRANULARITY_STATISTICS_VERBOSE
	if (RegionSize >= (int) TypeLimit)
		SMP_msg("Store accesses entire object at %x\n", InstAddr);
#endif
	return (RegionSize < (int) TypeLimit);
} // end of MemoryAccessInfo::ComputeNonDirectAccessRegion()

