/*
 * SMPBasicBlock.cpp - Class file for basic block analysis.
 *
 * Copyright (c) 2000, 2001, 2010 - University of Virginia 
 *
 * This file is part of the Memory Error Detection System (MEDS) infrastructure.
 * This file may be used and modified for non-commercial purposes as long as 
 * all copyright, permission, and nonwarranty notices are preserved.  
 * Redistribution is prohibited without prior written consent from the University 
 * of Virginia.
 *
 * Please contact the authors for restrictions applying to commercial use.
 *
 * THIS SOURCE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Author: University of Virginia
 * e-mail: jwd@virginia.com
 * URL   : http://www.cs.virginia.edu/
 *
 * Additional copyrights 2010, 2011, 2012, 2013, 2014, 2015 by Zephyr Software LLC
 * e-mail: {clc,jwd}@zephyr-software.com
 * URL   : http://www.zephyr-software.com/
 *
 */

//
// SMPBasicBlock.cpp
//
// This module performs the fundamental basic block level analyses needed for the
//   SMP project (Software Memory Protection).
//

#include <list>
#include <set>
#include <vector>
#include <algorithm>

#include <cstring>
#include <cassert>
#include <cstdint>

#include "interfaces/SMPDBInterface.h"
#include "base/SMPDataFlowAnalysis.h"
#include "base/SMPBasicBlock.h"
#include "base/SMPInstr.h"
#include "base/SMPFunction.h"
#include "base/SMPProgram.h"

using namespace std;

#define SMP_DEBUG_DATAFLOW 0   // verbose output for setting up SSA/LVA data structures
#define SMP_DEBUG_OPTIMIZATIONS 0    // verbose output for type propagation
#define SMP_VERBOSE_POINTER_REFINEMENT 0 // refining POINTER to HEAPPTR, STACKPTR, etc.
#define SMP_VERBOSE_ALIAS_DETECTION 0  // log message every time alias in DU chain is seen
#define STARS_DETECT_OPTIMIZED_RANGE_CHECK 1 // Detect misleading signedness case
#define STARS_MINIMIZE_UNSIGNED_BRANCH_PROPAGATION 1 // stop propagation at possible mixed-mode arithmetic
#define STARS_AGGRESSIVE_BENIGN_POINTER_ARITHMETIC 1 // subtract then add to POINTER can underflow then overflow benignly

// Basic block number 0 is the top of the CFG lattice.
#define SMP_TOP_BLOCK 0 

// Minimum # of instructions before we check a block not in a loop for the possibility of being a hashing arithmetic block.
#define STARS_HASH_BLOCK_MIN_INSTRS 50
// Minimum # of hashing rotates or shifts to the left that identifies a block as a hashing arithmetic block.
#define STARS_HASHING_OPERATIONS_THRESHOLD 4

// *****************************************************************
// Class SMPBasicBlock
// *****************************************************************

// Constructor
SMPBasicBlock::SMPBasicBlock(SMPFunction *Func, list<SMPInstr *>::iterator First, list<SMPInstr *>::iterator Last) {
	this->FirstAddr = (*First)->GetAddr();
	this->SetNumber(SMP_BLOCKNUM_UNINIT);
	this->SetLoopHeaderBlockNumberForTailBlock(SMP_BLOCKNUM_UNINIT);
	this->SetOuterLoopHeaderBlockNumberForDoubleTailBlock(SMP_BLOCKNUM_UNINIT);
	this->MyFunc = Func;
	this->SetOutgoingStackDelta(0);
#if 0
	this->IncomingStackPtrDelta = 0;
#endif
#if 1
	this->SetProcessed(false);

	this->SetReturns(false);
	this->ResetShared();
	this->SetLiveInStale(true);
	this->SetUnreachableBlock(false);
	this->SetFlagsDeadBeforeFirstKill(false);
	this->SetMaybeAliased(false);
#else
	this->booleans1 = 0;
	this->SetLiveInStale(true);
#endif
	this->booleans2 = 0;
	this->booleans3 = 0;

	this->InstVec.clear();
	this->AddrInstMap.clear();
	this->Predecessors.clear();
	this->Successors.clear();
	this->KillSet.clear();
	this->UpExposedSet.clear();
	this->LiveOutSet.clear();
	this->LiveInSet.clear();
	this->DomFrontier.clear();
	this->PhiFunctions.clear();
	this->LocalNames.clear();

	STARS_ea_t InstAddr;

	list<SMPInstr *>::iterator InstIter = First;
	SMPInstr *CurrInst;
	size_t VecIndex = this->InstVec.size() - 1;
	while (InstIter != Last) {
		CurrInst = (*InstIter);
		this->InstVec.push_back(CurrInst);
		VecIndex = this->InstVec.size() - 1;
		InstAddr = CurrInst->GetAddr();
		pair<STARS_ea_t, size_t> MapItem(InstAddr, VecIndex);
		this->AddrInstMap.insert(MapItem);
		++InstIter;
	}

	// Now process the last instruction.
	CurrInst = (*Last);
	this->InstVec.push_back(CurrInst); // Add last instruction
	InstAddr = CurrInst->GetAddr();
	VecIndex = this->InstVec.size() - 1;
	pair<STARS_ea_t, size_t> MapItem2(InstAddr, VecIndex);
	this->AddrInstMap.insert(MapItem2);
	CurrInst->SetTerminatesBlock();

	// Mark first true (non-marker) instruction.
	InstIter = First;
	CurrInst = (*InstIter);
	if (CurrInst->IsMarkerInst()) {
		// This is SSA marker instruction.
		++InstIter; // skip to first real instruction
		CurrInst = (*InstIter);
	}
	CurrInst->SetFirstInBlock();
}

// Get address of first instruction in the block.
STARS_ea_t SMPBasicBlock::GetFirstAddr(void) const {
	return this->FirstAddr;
}

// get address of first real inst, not SSA marker pseudo-inst.
STARS_ea_t SMPBasicBlock::GetFirstNonMarkerAddr(void) const {
	STARS_ea_t FirstInstAddr = this->GetFirstAddr();
	if (STARS_IsSSAMarkerPseudoID(FirstInstAddr)) {
		SMPInstr *FirstRealInst = this->GetInstFromIndex(1);
		FirstInstAddr = FirstRealInst->GetAddr();
	}
	return FirstInstAddr;
} // end of SMPBasicBlock::GetFirstNonMarkerAddr()

// Get address of last instruction in the block.
STARS_ea_t SMPBasicBlock::GetLastAddr(void) const {
	vector<SMPInstr *>::const_reverse_iterator LastInstIter = this->GetRevInstCBegin();
	return (*LastInstIter)->GetAddr();
}

// Six methods to get values from the maps of global reg/SSA to FG info.
//  For global names, see corresponding methods in SMPFunction.
unsigned short SMPBasicBlock::GetDefSignMiscInfo(int DefHashValue) {
	map<int, struct FineGrainedInfo>::iterator MapIter;

	MapIter = this->LocalDefFGInfoBySSA.find(DefHashValue);
	if (MapIter != this->LocalDefFGInfoBySSA.end())
		return MapIter->second.SignMiscInfo;
	else
		return 0;
} // end of SMPBasicBlock::GetDefSignMiscInfo()

unsigned short SMPBasicBlock::GetUseSignMiscInfo(int UseHashValue) {
	map<int, struct FineGrainedInfo>::iterator MapIter;
	pair<map<int, struct FineGrainedInfo>::iterator, bool> MapResult;

	MapIter = this->LocalUseFGInfoBySSA.find(UseHashValue);
	if (MapIter != this->LocalUseFGInfoBySSA.end())
		return MapIter->second.SignMiscInfo;
	else
		return 0;
} // end of SMPBasicBlock::GetUseSignMiscInfo()

unsigned short SMPBasicBlock::GetDefWidthTypeInfo(int DefHashValue) {
	map<int, struct FineGrainedInfo>::iterator MapIter;

	MapIter = this->LocalDefFGInfoBySSA.find(DefHashValue);
	if (MapIter != this->LocalDefFGInfoBySSA.end())
		return MapIter->second.SizeInfo;
	else
		return 0;
} // end of SMPBasicBlock::GetDefWidthTypeInfo()

unsigned short SMPBasicBlock::GetUseWidthTypeInfo(int UseHashValue) {
	map<int, struct FineGrainedInfo>::iterator MapIter;

	MapIter = this->LocalUseFGInfoBySSA.find(UseHashValue);
	if (MapIter != this->LocalUseFGInfoBySSA.end())
		return MapIter->second.SizeInfo;
	else
		return 0;
} // end of SMPBasicBlock::GetUseWidthTypeInfo()

struct FineGrainedInfo SMPBasicBlock::GetDefFGInfo(int DefHashValue) {
	map<int, struct FineGrainedInfo>::iterator MapIter;

	MapIter = this->LocalDefFGInfoBySSA.find(DefHashValue);
	if (MapIter != this->LocalDefFGInfoBySSA.end())
		return MapIter->second;
	else {
		struct FineGrainedInfo EmptyFG;
		EmptyFG.SignMiscInfo = 0;
		EmptyFG.SizeInfo = 0;
		return EmptyFG;
	}
} // end of SMPBasicBlock::GetDefFGInfo()

struct FineGrainedInfo SMPBasicBlock::GetUseFGInfo(int UseHashValue) {
	map<int, struct FineGrainedInfo>::iterator MapIter;

	MapIter = this->LocalUseFGInfoBySSA.find(UseHashValue);
	if (MapIter != this->LocalUseFGInfoBySSA.end())
		return MapIter->second;
	else {
		struct FineGrainedInfo EmptyFG;
		EmptyFG.SignMiscInfo = 0;
		EmptyFG.SizeInfo = 0;
		return EmptyFG;
	}
} // end of SMPBasicBlock::GetUseFGInfo()

bool SMPBasicBlock::IsLocalName(const STARSOpndTypePtr &CurrOp) const {
	return (!this->LocalNames.empty()
		&& (this->LocalNames.end() != this->LocalNames.find(CurrOp)));
}

STARSOpndSetIter SMPBasicBlock::FindLocalName(const STARSOpndTypePtr &FindOp) {
	return this->LocalNames.find(FindOp);
}

// Equality operator for SMPBasicBlock. Key field is address of first instruction.
int SMPBasicBlock::operator==(const SMPBasicBlock &rhs) const {
	if (rhs.GetFirstAddr() != this->GetFirstAddr())
		return 0;
	else
		return 1;
}

bool SMPBasicBlock::IsAddrInBlock(const STARS_ea_t InstAddr) const {
	return (this->GetFunc()->GetBlockNumFromInstAddr(InstAddr) == this->GetNumber());
}

// does VarKill set have any stack operands?
bool SMPBasicBlock::IsAnyStackOpKilled(void) const {
	bool FoundStackOp = false;
	bool UseFP = this->GetFunc()->UsesFramePointer();
	// Search set from the end for speed.
	for (STARSOpndSetConstRevIter KillIter = this->KillSet.crbegin(); KillIter != this->KillSet.crend(); ++KillIter) {
		STARSOpndTypePtr KillOp = (*KillIter);
		if (MDIsStackAccessOpnd(KillOp, UseFP)) {
			FoundStackOp = true;
			break;
		}
	}

	return FoundStackOp;
} // end of SMPBasicBlock::IsAnyStackOpKilled()

// Link to predecessor block.
void SMPBasicBlock::LinkToPred(SMPBasicBlock *Predecessor) {
	this->Predecessors.push_back(Predecessor);
	return;
}

// Link to successor block.
void SMPBasicBlock::LinkToSucc(SMPBasicBlock *Successor) {
	bool inserted = false;
#if 1
	// We will keep successors sorted in address order.
	for (list<SMPBasicBlock *>::iterator SuccIter = this->GetFirstSucc(); SuccIter != this->GetLastSucc(); ++SuccIter) {
		SMPBasicBlock *SuccBlock = (*SuccIter);
		STARS_ea_t SuccStartAddr = SuccBlock->GetFirstAddr();
		if (SuccStartAddr > Successor->GetFirstAddr()) {
			SuccIter = this->Successors.insert(SuccIter, Successor);
			inserted = true;
			break;
		}
	}
#endif
	if (!inserted) { // first successor, or needs to be last
		this->Successors.push_back(Successor);
	}
	return;
}

pair<map<STARSOpndTypePtr, int, LessOp>::iterator, bool> SMPBasicBlock::AddLiveOutSSAMap(const STARSOpndTypePtr &Opnd, int SSANum) {
	pair<STARSOpndTypePtr, int> InsertValue(Opnd, SSANum);
	pair<map<STARSOpndTypePtr, int, LessOp>::iterator, bool> InsertResult = this->LiveOutSSAMap.insert(InsertValue);
	return InsertResult;
} // end of SMPBasicBlock::AddLiveOutSSAMap()

// See if all predecessors have set their ordering number.
bool SMPBasicBlock::AllPredecessorsNumbered(void) {
	list<SMPBasicBlock *>::iterator CurrPred;
	for (CurrPred = this->Predecessors.begin(); CurrPred != this->Predecessors.end(); ++CurrPred) {
		// Don't count current block, in case we have a one-block loop with this block
		//  as its own predecessor.
		if (**CurrPred == *this)
			continue;
		if ((*CurrPred)->GetNumber() == SMP_BLOCKNUM_UNINIT)
			return false;
	}
	return true;
} // end of SMPBasicBlock::AllPredecessorsNumbered()

// See if all predecessors have been processed.
bool SMPBasicBlock::AllPredecessorsProcessed(void) {
	list<SMPBasicBlock *>::iterator CurrPred;
	for (CurrPred = this->Predecessors.begin(); CurrPred != this->Predecessors.end(); ++CurrPred) {
		// Don't count current block, in case we have a one-block loop with this block
		//  as its own predecessor.
		if (**CurrPred == *this)
			continue;
		if (!(*CurrPred)->IsProcessed())
			return false;
	}
	return true;
} // end of SMPBasicBlock::AllPredecessorsProcessed()

// Mark as processed, recurse into successors depth-first.
void SMPBasicBlock::DepthFirstMark(bool MarkIBTBlocks) {
	if (this->IsProcessed()) {
		return;
	}

	this->SetProcessed(true);
	bool SuccessorsCanBeIBTs = false;
	list<SMPBasicBlock *>::iterator CurrSucc;
	if (MarkIBTBlocks) {
		// We need to detect some odd patterns of computed goto macros in libc.
		//  The patterns include blocks with just a prefetch instruction, with
		//  the indirect jump either going to the prefetch block or bypassing it
		//  to go to the fall-through, so both are IBTs (indirect branch targets).
		//  The other pattern is an undefined opcode, ud2, at the beginning of a block,
		//  generally falling right after the indirect jump, indicating that there
		//  is no fall-through to the ud2. Instead, the indirect jump goes to the instruction
		//  after the ud2, which is the true IBT in this case.
		;
	}

	for (CurrSucc = this->Successors.begin(); CurrSucc != this->Successors.end(); ++CurrSucc) {
		SMPBasicBlock *SuccBlock = (*CurrSucc);
		SuccBlock->DepthFirstMark(MarkIBTBlocks); // recurse, depth-first
	}

	return;
} // end of SMPBasicBlock::DepthFirstMark()

// Mark as processed, recurse into successors depth-first, do not follow back edges.
void SMPBasicBlock::DepthFirstMarkNonBackEdgeSuccessors(void) {
	if (this->IsProcessed()) {
		return;
	}

	this->SetProcessed(true);
	list<SMPBasicBlock *>::iterator CurrSucc;
	int CurrBlockNum = this->GetNumber();
	for (CurrSucc = this->Successors.begin(); CurrSucc != this->Successors.end(); ++CurrSucc) {
		SMPBasicBlock *SuccBlock = (*CurrSucc);
		int SuccBlockNum = SuccBlock->GetNumber();
		if (!(this->GetFunc()->DoesBlockDominateBlock(SuccBlockNum, CurrBlockNum))) {
			SuccBlock->DepthFirstMarkNonBackEdgeSuccessors(); // recurse, depth-first
		}
	}

	return;
} // end of SMPBasicBlock::DepthFirstMark()

// Compute predecessor count, excluding loop-back edges. Loop ID must occur first.
size_t SMPBasicBlock::GetNumPredsMinusBackEdges(void) const {
	size_t PredCount = this->GetNumPreds();

	// If this block is a loop header block, reduce the PredCount
	//  by finding loop-back edges.
	if (this->IsLoopHeaderBlock()) {
		int CurrBlockNum = this->GetNumber();
		for (list<SMPBasicBlock *>::const_iterator PredIter = this->GetFirstConstPred(); PredIter != this->GetLastConstPred(); ++PredIter) {
			int PredBlockNum = (*PredIter)->GetNumber();
			if (this->GetFunc()->DoesBlockDominateBlock(CurrBlockNum, PredBlockNum)) {
				assert(0 < PredCount); // Cannot be loop header with back edges as the only predecessors.
				--PredCount;
			}
		}
	}

	return PredCount;
} // end of SMPBasicBlock::GetNumPredsMinusBackEdges()

// Are all instructions in the block no-ops?
bool SMPBasicBlock::AllNops(void) {
	size_t NopCount = 0;
	size_t GoodCount = 0;  // non-nop instructions
	vector<SMPInstr *>::iterator InstIter;
	for (InstIter = this->GetFirstInst(); InstIter != this->GetLastInst(); ++InstIter) {
		SMPInstr *CurrInst = (*InstIter);
		if (CurrInst->IsNop()) {
			++NopCount;
		}
		else {
			++GoodCount;
			break;
		}
	}
	return ((0 == GoodCount) && (0 < NopCount));
} // end of SMPBasicBlock::AllNops()

// Analyze basic block and fill data members.
void SMPBasicBlock::Analyze() {
	vector<SMPInstr *>::reverse_iterator RevInstIter = this->GetRevInstBegin();
	SMPInstr *BackInst = (*RevInstIter);
	if (BackInst->GetDataFlowType() == RETURN) {
		this->SetReturns(true);
	}
} // end of SMPBasicBlock::Analyze()

// DEBUG dump of block
void SMPBasicBlock::Dump(void) {
	SMP_msg("Dump of basic block %d\n", this->GetNumber());
	// Dump dataflow analysis sets and links before dumping instructions.
	list<SMPBasicBlock *>::iterator CurrLink;
	SMP_msg("Predecessors: ");
	if (!this->Predecessors.empty()) {
		for (CurrLink = this->Predecessors.begin(); CurrLink != this->Predecessors.end(); ++CurrLink) {
			SMP_msg("%d ", (*CurrLink)->GetNumber());
		}
	}
	SMP_msg("\n");
	SMP_msg("Successors: ");
	if (!this->Successors.empty()) {
		for (CurrLink = this->Successors.begin(); CurrLink != this->Successors.end(); ++CurrLink) {
			SMP_msg("%d ", (*CurrLink)->GetNumber());
		}
	}
	SMP_msg("\n");

	SMP_msg(" Booleans1 : %x  Booleans2: %x  Booleans3: %x\n", this->booleans1, this->booleans2, this->booleans3);
	if (this->HasIndirectJump())
		SMP_msg("Has indirect jump. ");
	if (this->HasReturn())
		SMP_msg("Has return. ");
	if (this->IsShared())
		SMP_msg("Is shared tail chunk block. ");
	SMP_msg("\n");
	if (this->IsLoopHeaderBlock()) {
		SMP_msg("Header block for loop %d\n", this->GetFunc()->GetLoopNumFromHeaderBlockNum(this->GetNumber()));
	}
	if (this->IsLoopTailBlock()) {
		SMP_msg("Tail block for loop with header block %d\n", this->GetLoopHeaderNumber());
		if (this->IsDoubleLoopTailBlock()) {
			SMP_msg("Double Tail block for outer loop with header block %d\n", this->GetOuterLoopHeaderNumberForDoubleTailBlock());
		}
	}
	
	STARSOpndSetIter SetItem;
	SMP_msg("VarKill set: ");
	if (!this->KillSet.empty()) {
		SMP_msg("size: %zu ", this->KillSet.size());
		for (SetItem = this->KillSet.begin(); SetItem != this->KillSet.end(); ++SetItem) {
			PrintListOperand(*SetItem);
		}
	}
	SMP_msg("\n\n");
	SMP_msg("UpExposed set: ");
	if (!this->UpExposedSet.empty()) {
		for (SetItem = this->GetFirstUpExposed(); SetItem != this->GetLastUpExposed(); ++SetItem) {
			PrintListOperand(*SetItem);
		}
	}
	SMP_msg("\n\n");
	SMP_msg("LiveIn set: ");
	if (!this->LiveInSet.empty()) {
		for (SetItem = this->LiveInSet.begin(); SetItem != this->LiveInSet.end(); ++SetItem) {
			PrintListOperand(*SetItem);
		}
	}
	SMP_msg("\n\n");
	SMP_msg("LiveOut set: ");
	if (!this->LiveOutSet.empty()) {
		for (SetItem = this->LiveOutSet.begin(); SetItem != this->LiveOutSet.end(); ++SetItem) {
			STARSOpndTypePtr LiveOutOp = (*SetItem);
			map<STARSOpndTypePtr, int, LessOp>::const_iterator SSAIter = this->FindLiveOutSSAMap(LiveOutOp);
			assert(SSAIter != this->GetLastLiveOutSSAMap());
			int SSANum = SSAIter->second;
			PrintListOperand(LiveOutOp, SSANum);
		}
	}
	SMP_msg("\n\n");
	SMP_msg("Dominance frontier: ");
	set<int>::iterator DomIter;
	if (!this->DomFrontier.empty()) {
		for (DomIter = this->DomFrontier.begin(); DomIter != this->DomFrontier.end(); ++DomIter) {
			SMP_msg("%d ", *DomIter);
		}
	}
	SMP_msg("\n\n");
	set<SMPPhiFunction, LessPhi>::iterator PhiIter;
	if (!this->PhiFunctions.empty()) {
		for (PhiIter = this->PhiFunctions.begin(); PhiIter != this->PhiFunctions.end(); ++PhiIter) {
			SMP_msg("Phi function for %d : ", PhiIter->GetIndex());
			// Dump out all phi operands
			PhiIter->Dump();
		}
	}
	SMP_msg("\n");

	// Now, dump all the instructions.
	vector<SMPInstr *>::iterator CurrInst;
	for (CurrInst = this->GetFirstInst(); CurrInst != this->GetLastInst(); ++CurrInst) {
		(*CurrInst)->Dump();
	}
	SMP_msg("\n");
	return;
} // end of SMPBasicBlock::Dump()

// Call SMPInstr::EmitSPARKAda() until we hit a control flow inst.
void SMPBasicBlock::EmitSPARKAdaForFallThroughInsts(FILE *BodyFile) {
	STARS_ea_t LastID = this->GetLastAddr();
	for (vector<SMPInstr *>::iterator InstIter = this->GetFirstInst(); InstIter != this->GetLastInst(); ++InstIter) {
		SMPInstr *CurrInst = (*InstIter);
		STARS_ea_t InstAddr = CurrInst->GetAddr();
		bool ControlFlowTerminator = ((InstAddr == LastID) && (DEFAULT != CurrInst->GetDataFlowType()));
		bool Untranslatable = CurrInst->IsMarkerInst() || (!CurrInst->IsAnalyzeable());
		if (!ControlFlowTerminator && (!Untranslatable)) {
			CurrInst->EmitSPARKAda(BodyFile);
		}
		else {
			// Set junk instructions to translated.
			if (Untranslatable) {
				CurrInst->SetSPARKTranslated();
			}
		}
	}
	this->SetProcessed(true);
	return;
} // end of SMPBasicBlock::EmitSPARKAdaForFallThroughInsts()

// Call SMPInstr::EmitSPARKAda() for all insts in block.
void SMPBasicBlock::EmitSPARKAdaForAllInsts(FILE *BodyFile) {
	for (vector<SMPInstr *>::iterator InstIter = this->GetFirstInst(); InstIter != this->GetLastInst(); ++InstIter) {
		SMPInstr *CurrInst = (*InstIter);
		bool Untranslatable = CurrInst->IsMarkerInst() || (!CurrInst->IsAnalyzeable());
		if (!Untranslatable) {
			CurrInst->EmitSPARKAda(BodyFile);
		}
		else {
			// Set junk instructions to translated.
			CurrInst->SetSPARKTranslated();
		}
	}
	this->SetProcessed(true);
	return;
}

// For SingleExpression block, emit SPARK Ada expression
void SMPBasicBlock::EmitSPARKAdaForExpression(FILE *BodyFile) {
	// Identify HeaderBlock of compound conditional, whose compare operands do NOT need
	//  to be traced back further in the header block. The header block has initialized the 
	//  register that is used in the compare or test, so just print that register.
	bool UseFP = this->GetFunc()->UsesFramePointer();
	STARSCFGBlock *CurrCFGBlock = this->GetFunc()->GetCFGBlockByNum((size_t) this->GetNumber());
	bool HeaderBlock = (!CurrCFGBlock->IsCoalesced());

	// We will divide into cases: Expression block can end in signed compare instruction or
	//  unsigned compare (bitwise AND) instruction, and unsigned compare is often to itself
	//  to test for zero or non-zero value.
	vector<SMPInstr *>::const_reverse_iterator RevInstIter = this->GetRevInstCBegin();
	SMPInstr *LastInst = (*RevInstIter);
	++RevInstIter;
	SMPInstr *CompareInst = (*RevInstIter);
	if (!CompareInst->MDIsCompareOrTest()) {
		SMP_fprintf(BodyFile, "ERROR in SingleExpression block\n");
		return;
	}
	SMPRegTransfer *CompareRT = CompareInst->GetRT(0);
	assert(CompareRT->HasRightSubTree());
	CompareRT = CompareRT->GetRightTree();
	SMPoperator CompareOperator = CompareRT->GetOperator();
	bool SignedCompareCase = (SMP_S_COMPARE == CompareOperator);
	bool UnsignedSelfCompareCase = ((SMP_U_COMPARE == CompareOperator) && IsEqOp(CompareRT->GetConstRightOperandNoNorm(), CompareRT->GetConstLeftOperandNoNorm()));

	// Emit the SPARK Ada for the left operand of the comparison.
	STARSOpndTypePtr LeftOp = CompareRT->GetLeftOperand();
	if (LeftOp->IsImmedOp() || HeaderBlock) {
		if (!LeftOp->IsImmedOp()) {
			// Might need to unnormalize reg or memory
			STARSOpndTypePtr LeftOpNoNorm = CloneIfNecessary(LeftOp, UseFP);
			CompareInst->MDGetUnnormalizedOp(LeftOpNoNorm);
			CompareInst->PrintSPARKAdaOperand(LeftOpNoNorm, BodyFile, false, UseFP, true, false);
		}
		else {
			CompareInst->PrintSPARKAdaOperand(LeftOp, BodyFile, false, UseFP, true, false);
		}
	}
	else {
		STARSDefUseIter LeftUseIter;
		LeftUseIter = CompareInst->FindUse(LeftOp);
		if (LeftUseIter != CompareInst->GetLastUse()) {
			(void) this->EmitSPARKAdaForExprUse(BodyFile, RevInstIter, LeftOp, LeftUseIter->GetSSANum());
		}
		else {
			SMP_fprintf(BodyFile, "ERROR in compare operand\n");
		}
	}

	if (SignedCompareCase) {
		// Emit the signed comparison operator based on the conditional branch opcode
		LastInst->MDEmitSPARKAdaExprCompare(BodyFile);
	}
	else {
		// Emit the unsigned comparison operator based on the conditional branch opcode
		LastInst->MDEmitSPARKAdaExprTest(BodyFile);
	}

	// Emit the SPARK Ada for the right operand of the comparison.
	if (UnsignedSelfCompareCase) {
		SMP_fprintf(BodyFile, " 0 "); // AND with self is comparison to zero, set flags
	}
	else {
		STARSOpndTypePtr RightOp = CompareRT->GetRightOperand();
		if (RightOp->IsImmedOp() || HeaderBlock) {
			if (!RightOp->IsImmedOp()) {
				// Might need to unnormalize reg or memory
				STARSOpndTypePtr RightOpNoNorm = CloneIfNecessary(RightOp, UseFP);
				CompareInst->MDGetUnnormalizedOp(RightOpNoNorm);
				CompareInst->PrintSPARKAdaOperand(RightOpNoNorm, BodyFile, false, UseFP, true, false);
			}
			else {
				CompareInst->PrintSPARKAdaOperand(RightOp, BodyFile, false, UseFP, true, false);
			}
		}
		else {
			STARSDefUseIter RightUseIter;
			RightUseIter = CompareInst->FindUse(RightOp);
			if (RightUseIter != CompareInst->GetLastUse()) {
				(void) this->EmitSPARKAdaForExprUse(BodyFile, RevInstIter, RightOp, RightUseIter->GetSSANum());
			}
			else {
				SMP_fprintf(BodyFile, "ERROR in compare operand\n");
			}
		}
	}

	// Mark all instructions in the single-expr block as translated.
	for (vector<SMPInstr *>::iterator InstIter = this->GetFirstInst(); InstIter != this->GetLastInst(); ++InstIter) {
		(*InstIter)->SetSPARKTranslated();
	}
	return;
} // end of SMPBasicBlock::EmitSPARKAdaForExpression()

// Emit SPARK expression for *UseIter at RevInstIter
bool SMPBasicBlock::EmitSPARKAdaForExprUse(FILE *BodyFile, vector<SMPInstr *>::const_reverse_iterator RevInstIter, STARSOpndTypePtr UseOp, int UseSSANum) {
	// Search backwards in this block for DEF of UseIter operand. If not found,
	//  just print UseIter operand. If DEF found, recursively follow and print ultimate DEF expression.
	bool Printed = false;
	bool UseFP = this->GetFunc()->UsesFramePointer();
	vector<SMPInstr *>::const_reverse_iterator NewRevInstIter(RevInstIter);
	vector<SMPInstr *>::const_reverse_iterator FoundInstIter(RevInstIter);
	++NewRevInstIter;
	while (NewRevInstIter != this->GetRevInstCEnd()) {
		SMPInstr *CurrInst = (*NewRevInstIter);
		STARSDefUseIter DefIter = CurrInst->FindDef(UseOp);
		STARS_ea_t CurrAddr = CurrInst->GetAddr();
		if ((DefIter != CurrInst->GetLastDef()) && (!STARS_IsSSAMarkerPseudoID(CurrAddr))) {
			assert(DefIter->GetSSANum() == UseSSANum);
			FoundInstIter = NewRevInstIter;
			STARSOpndTypePtr NewUseOp = nullptr;
			if (CurrInst->IsSimpleCopy(NewUseOp)) {
				// Follow NewUseOp or print it.
				if (NewUseOp->IsImmedOp()) {
					CurrInst->PrintSPARKAdaOperand(NewUseOp, BodyFile, false, false, true, false);
					Printed = true;
				}
				else {
					STARSDefUseIter NewUseIter = CurrInst->FindUse(NewUseOp);
					assert(NewUseIter != CurrInst->GetLastUse());
					Printed = this->EmitSPARKAdaForExprUse(BodyFile, NewRevInstIter, NewUseOp, NewUseIter->GetSSANum());
					assert(Printed);
				}
				break;
			}
			else {
				SMPRegTransfer *DefRT = CurrInst->GetRT(0);
				assert(IsEqOp(DefRT->GetConstLeftOperandNoNorm(), UseOp));
				assert(SMP_ASSIGN == DefRT->GetOperator());
				assert(DefRT->HasRightSubTree());
				DefRT = DefRT->GetRightTree();
				// Recurse into DEF expression.
				string OperatorString;
				bool OperatorIsProcCall = false;
				bool PrefixUnaryOperator = false;
				bool ConstFollows = (!DefRT->HasRightSubTree()) && DefRT->GetConstRightOperandNoNorm()->IsImmedOp();
				PrintSPARKAdaOperator(DefRT->GetOperator(), OperatorString, OperatorIsProcCall, PrefixUnaryOperator, ConstFollows);
				if (OperatorIsProcCall) {
					SMP_fprintf(BodyFile, "ERROR ");
					return true;
				}
				if (PrefixUnaryOperator) {
					SMP_fprintf(BodyFile, "%s(", OperatorString.c_str());
				}
				else {
					SMP_fprintf(BodyFile, " (");
				}
				STARSOpndTypePtr DefLeftOp = DefRT->GetLeftOperand();
				if (DefLeftOp->IsImmedOp()) {
					CurrInst->PrintSPARKAdaOperand(DefLeftOp, BodyFile, false, false, true, false);
					Printed = true;
				}
				else {
					STARSDefUseIter LeftUseIter = CurrInst->FindUse(DefLeftOp);
					assert(LeftUseIter != CurrInst->GetLastUse());
					Printed = this->EmitSPARKAdaForExprUse(BodyFile, NewRevInstIter, DefLeftOp, LeftUseIter->GetSSANum());
					assert(Printed);
				}
				if (PrefixUnaryOperator) {
					SMP_fprintf(BodyFile, ") ");
				}
				else {
					SMP_fprintf(BodyFile, " %s ", OperatorString.c_str());
					const STARSOpndTypePtr DefRightOp = DefRT->GetConstRightOperandNoNorm();
					// NOTE:  Should handle RightTree possibility.
					if (DefRightOp->IsImmedOp()) {
						CurrInst->PrintSPARKAdaOperand(DefRightOp, BodyFile, false, false, true, false);
						Printed = true;
						SMP_fprintf(BodyFile, ") ");
					}
					else {
						STARSDefUseIter RightUseIter = CurrInst->FindUse(DefRightOp);
						assert(RightUseIter != CurrInst->GetLastUse());
						Printed = this->EmitSPARKAdaForExprUse(BodyFile, NewRevInstIter, DefRightOp, RightUseIter->GetSSANum());
						assert(Printed);
						SMP_fprintf(BodyFile, ") ");
					}
				}
			}
		}
		if (Printed)
			break;
		else
			++NewRevInstIter;
	} // end while new reverse iterator has not reached beginning of block

	if (!Printed) {
		// Print the original UseIter; no trace back occurred.
		// NOTE: Need to get unnormalized operand from RTL.
		STARSOpndTypePtr UseNoNormOp = CloneIfNecessary(UseOp, UseFP);
		SMPInstr *CurrInst = (*RevInstIter);
		CurrInst->MDGetUnnormalizedOp(UseNoNormOp);
		(*RevInstIter)->PrintSPARKAdaOperand(UseNoNormOp, BodyFile, false, this->GetFunc()->UsesFramePointer(), true, false);
		Printed = true;
	}

	return Printed;
} // end of SMPBasicBlock::EmitSPARKAdaForExprUse()

// Search successors for fall-through block, if any
list<SMPBasicBlock *>::const_iterator SMPBasicBlock::GetFallThroughSucc(void) const {
	list<SMPBasicBlock *>::const_iterator FallThroughSuccIter = this->GetLastConstSucc();
	vector<SMPInstr *>::const_reverse_iterator LastInstIter = this->GetRevInstCBegin();
	SMPInstr *LastInst = (*LastInstIter);
	SMPitype LastDataFlow = LastInst->GetDataFlowType();

	if ((JUMP != LastDataFlow) && (INDIR_JUMP != LastDataFlow) && (RETURN != LastDataFlow) && (HALT != LastDataFlow) && (!LastInst->IsFixedCallJump())) {
		// Block has fall-through.
		STARS_ea_t FallThroughSuccAddr = STARS_BADADDR;
#if 0
		bool ValidLastInst = ((nullptr != LastInst) && (nullptr != LastInst->GetInstID().GetInstruction()));
		if (ValidLastInst) {
			STARS_InstructionID_t NextInstID = LastInst->GetInstID().GetInstruction()->GetNextInstructionID();
			FallThroughSuccAddr = NextInstID.GetIDWithinFile();
			for (list<SMPBasicBlock *>::const_iterator SuccIter = this->GetFirstConstSucc(); SuccIter != this->GetLastConstSucc(); ++SuccIter) {
				STARS_ea_t FirstSuccAddr = (*SuccIter)->GetFirstAddr();
				if (FirstSuccAddr == FallThroughSuccAddr) { // found fall-through
					FallThroughSuccIter = SuccIter;
					break;
				}
			}
		}
#else
		if (COND_BRANCH == LastDataFlow) {
			STARS_ea_t NonFallThroughAddr = LastInst->GetJumpTarget();
			assert(STARS_BADADDR != NonFallThroughAddr);
			for (list<SMPBasicBlock *>::const_iterator SuccIter = this->GetFirstConstSucc(); SuccIter != this->GetLastConstSucc(); ++SuccIter) {
				SMPBasicBlock *SuccBlock = (*SuccIter);
				if (SuccBlock->GetFirstAddr() != NonFallThroughAddr) {
					FallThroughSuccIter = SuccIter;
					break;
				}
			}
		}
		else {
			assert((DEFAULT == LastDataFlow) || (CALL == LastDataFlow) || (INDIR_CALL == LastDataFlow));
			if (1 >= this->GetNumSuccessors()) 
				FallThroughSuccIter = this->GetFirstConstSucc();
			else { // computed call or indir_call targets
				STARS_ea_t FallThroughAddr = LastInst->GetFallThroughAddr();
				for (list<SMPBasicBlock *>::const_iterator SuccIter = this->GetFirstConstSucc(); SuccIter != this->GetLastConstSucc(); ++SuccIter) {
					SMPBasicBlock *SuccBlock = (*SuccIter);
					if (SuccBlock->GetFirstAddr() == FallThroughAddr) {
						FallThroughSuccIter = SuccIter;
						break;
					}
				}
			}
		}
#endif
	}

	return FallThroughSuccIter;
} // end of SMPBasicBlock::GetFallThroughSucc()

// Search successors for conditional non-fall-through block, if any
list<SMPBasicBlock *>::const_iterator SMPBasicBlock::GetCondNonFallThroughSucc(void) const {
	list<SMPBasicBlock *>::const_iterator BranchTakenSuccIter = this->Successors.cend();
	vector<SMPInstr *>::const_reverse_iterator LastInstIter = this->GetRevInstCBegin();
	SMPInstr *LastInst = (*LastInstIter);
	SMPitype LastDataFlow = LastInst->GetDataFlowType();
	
	if (COND_BRANCH == LastDataFlow) {
		STARS_ea_t NonFallThroughAddr = LastInst->GetJumpTarget();
		assert(STARS_BADADDR != NonFallThroughAddr);
		for (list<SMPBasicBlock *>::const_iterator SuccIter = this->GetFirstConstSucc(); SuccIter != this->GetLastConstSucc(); ++SuccIter) {
			SMPBasicBlock *SuccBlock = (*SuccIter);
			if (SuccBlock->GetFirstAddr() == NonFallThroughAddr) {
				BranchTakenSuccIter = SuccIter;
				break;
			}
		}
	}
	return BranchTakenSuccIter;
} // end of SMPBasicBlock::GetCondNonFallThroughSucc()

// Search successors for conditional successor that is not #NotThisBlockNum
std::list<SMPBasicBlock *>::const_iterator SMPBasicBlock::GetCondOtherSucc(int NotThisBlockNum) const {
	list<SMPBasicBlock *>::const_iterator OtherSuccIter = this->Successors.cend();
	for (list<SMPBasicBlock *>::const_iterator SuccIter = this->GetFirstConstSucc(); SuccIter != this->GetLastConstSucc(); ++SuccIter) {
		SMPBasicBlock *SuccBlock = (*SuccIter);
		if (SuccBlock->GetNumber() != NotThisBlockNum) {
			OtherSuccIter = SuccIter;
			break;
		}
	}

	return OtherSuccIter;
} // end of SMPBasicBlock::GetCondOtherSucc()

// Search successors for conditional successor that is not in loop #LoopNumber
std::list<SMPBasicBlock *>::const_iterator SMPBasicBlock::GetSuccNotInLoop(size_t LoopNumber) const {
	std::list<SMPBasicBlock *>::const_iterator SuccIter;
	for (SuccIter = this->GetFirstConstSucc(); SuccIter != this->GetLastConstSucc(); ++SuccIter) {
		int SuccBlockNum = (*SuccIter)->GetNumber();
		if (!this->GetFunc()->IsBlockInLoop(SuccBlockNum, LoopNumber))
			break;
	}
	return SuccIter;
} // end of SMPBasicBlock::GetSuccNotInLoop()

// Is BlockNum a predecessor block of this block
bool SMPBasicBlock::IsBlockPred(int BlockNum) const {
	bool FoundPred = false;
	for (list<SMPBasicBlock *>::const_iterator PredIter = this->GetFirstConstPred(); PredIter != this->GetLastConstPred(); ++PredIter) {
		int PredBlockNum = (*PredIter)->GetNumber();
		if (PredBlockNum == BlockNum) {
			FoundPred = true;
			break;
		}
	}
	return FoundPred;
}

// Get the enum ControlFlowType for the last inst
ControlFlowType SMPBasicBlock::GetLastInstCFType(void) const {
	STARS_ea_t LastAddr = this->GetLastAddr();
	return this->GetFunc()->GetControlFlowType(LastAddr);
}

// Get the enum SMPitype for the last inst
SMPitype SMPBasicBlock::GetLastInstFlowType(void) const {
	SMPInstr *LastInst = (*(this->GetRevInstCBegin()));
	return LastInst->GetDataFlowType();
}

// block has HeadBlockNum as successor, which is a loop header
bool SMPBasicBlock::HasLoopHeadAsSuccessor(int &HeadBlockNum) const {
	bool FoundLoopHead = false;
	for (list<SMPBasicBlock *>::const_iterator SuccIter = this->GetFirstConstSucc(); SuccIter != this->GetLastConstSucc(); ++SuccIter) {
		SMPBasicBlock *SuccBlock = (*SuccIter);
		if (SuccBlock->IsLoopHeaderBlock()) {
			FoundLoopHead = true;
			HeadBlockNum = SuccBlock->GetNumber();
		}
	}
	return FoundLoopHead;
} // end of SMPBasicBlock::HasLoopHeadAsSuccessor()

// block has loop head successor that ends with INVERTED_LOOP_EXIT
bool SMPBasicBlock::HasLoopHeadWithInvertedExitAsSuccessor(void) const {
	bool OddFlowCase = false;
	int HeadBlockNum = SMP_BLOCKNUM_UNINIT;
	if (this->HasLoopHeadAsSuccessor(HeadBlockNum)) {
		SMPBasicBlock *HeadBlock = this->GetFunc()->GetBlockByNum((size_t) HeadBlockNum);
		ControlFlowType HeadCFType = HeadBlock->GetLastInstCFType();
		OddFlowCase = (INVERTED_LOOP_EXIT == HeadCFType);
	}
	return OddFlowCase;
} // end of SMPBasicBlock::HasLoopHeadWithInvertedExitAsSuccessor()


// Determine if CurrOp is LiveIn; if LiveIn set is stale, recompute.
bool SMPBasicBlock::IsLiveIn(const STARSOpndTypePtr &CurrOp) {
	if (this->IsLiveInStale()) {
		this->ComputeLiveInSet(false);
	}
	return (this->GetLastLiveIn() != this->LiveInSet.find(CurrOp));
} // end of SMPBasicBlock::IsLiveIn()

// Return true if anything already in the KillSet would kill the operand value.
bool SMPBasicBlock::MDAlreadyKilled(const STARSOpndTypePtr &Opnd1) const {
	// We have assembly language operands that can be complex, such as
	//  [ebx + esi*4 + 04h]. If ebx or esi have been killed, then this memory
	//  phrase should be considered killed. We could be even more conservative
	//  with base addresses, declaring an entire array killed whenever its base
	//  address appears in a definition, for example. We will do that if it proves
	//  to be necessary.
	bool FoundInKillSet = this->IsVarKill(Opnd1);
	if (!(Opnd1->IsMemDisplacementOp() || Opnd1->IsMemNoDisplacementOp()))
		// Look in KillSet. These simple types should be found there with
		//  no complicated comparisons needed.
		return FoundInKillSet;

	// If found directly in KillSet, return true. Otherwise, see if any registers
	//  used in the memory addressing expression were killed.
	if (FoundInKillSet)
		return true;

	bool UseFP = this->MyFunc->UsesFramePointer();
	bool StackAccess = MDIsStackAccessOpnd(Opnd1, UseFP);
	bool DirectAccess = false;


	if (StackAccess) {
		DirectAccess = !MDIsIndirectMemoryOpnd(Opnd1, UseFP);
	}
	else if (!(DirectAccess && StackAccess)) {
#if 1
		// We only keep registers and direct stack accesses in the data flow sets.
		//  Commented-out code has const-ness compiler problems, anyway.
		return false;
#else
		// Should we add Opnd1 to the KillSet every time we return true below? **!!**
		int BaseReg;
		int IndexReg;
		uint16_t ScaleFactor;
		STARS_ea_t displacement;

		MDExtractAddressFields(Opnd1, BaseReg, IndexReg, ScaleFactor, displacement);

		vector<SMPInstr *>::const_iterator FirstInst = this->GetFirstInst();
		if (STARS_x86_R_none != BaseReg) {
			STARSOpndTypePtr TempOp = (*FirstInst)->MakeRegOpnd(BaseReg);
			if (this->KillSet.end() != this->KillSet.find(TempOp))
				return true;
		}
		if (STARS_x86_R_none != IndexReg) { // Cannot have ESP index reg in SIB
			STARSOpndTypePtr TempOp = (*FirstInst)->MakeRegOpnd(IndexReg);
			if (this->KillSet.end() != this->KillSet.find(TempOp))
				return true;
		}
#endif
	}
	return false;
} // end of SMPBasicBlock::MDAlreadyKilled()

// Initialize the KilledSet and UpExposedSet for live variable analysis.
void SMPBasicBlock::InitKilledExposed(bool UseFP, bool StackOnly) {
	// Find all upwardly exposed operands and killed operands in this block.
	vector<SMPInstr *>::iterator CurrIter;
	set<DefOrUse, LessDefUse>::iterator CurrUse, CurrDef;
	if (!StackOnly) {
		this->UpExposedSet.clear();
		this->KillSet.clear();
	}

	for (CurrIter = this->GetFirstInst(); CurrIter != this->GetLastInst(); ++CurrIter) {
		SMPInstr *CurrInst = (*CurrIter);

		bool RelevantInst = (!StackOnly || CurrInst->HasDestMemoryOperand() || CurrInst->HasSourceMemoryOperand());
		if (!RelevantInst)
			continue;

		// Dataflow equation for upward exposed variables: If a variable has not been
		//  killed yet in this block, starting from the top of the block, and it is used
		//  in the current instruction, then it is upwardly exposed.
		for (CurrUse = CurrInst->GetFirstUse(); CurrUse != CurrInst->GetLastUse(); ++CurrUse) {
			STARSOpndTypePtr UseOp = CurrUse->GetOp();
			// Only add non-immediate operands that are not already killed in this block.
			//  o_near and o_far operands are code addresses in immediate form, e.g.
			//  call _printf might be call 0x8048040, with o_near = 0x8048040.
			if (MDIsDataFlowOpnd(UseOp, UseFP) && (!(this->MDAlreadyKilled(UseOp)))) {
				if (!StackOnly || MDIsStackAccessOpnd(UseOp, UseFP)) {
					this->AddUpExposed(UseOp);
				}
			}
		}
		// Dataflow equation for killed variables: If a variable is defined in any
		//  instruction in the block, it is killed by this block (i.e. prior definitions
		//  of that variable will not make it through the block).
		for (CurrDef = CurrInst->GetFirstDef(); CurrDef != CurrInst->GetLastDef(); ++CurrDef) {
			STARSOpndTypePtr DefOp = CurrDef->GetOp();
			if (MDIsDataFlowOpnd(DefOp, UseFP)) {
				if (!StackOnly || MDIsStackAccessOpnd(DefOp, UseFP)) {
					this->AddVarKill(DefOp);
				}
			}
		}
	} // end for all instrs in block
	if (!StackOnly)
		this->SetLiveInStale(true);  // Would need to compute LiveInSet for first time
	
	return;
} // end of SMPBasicBlock::InitKilledExposed()

// returns size in bytes
size_t SMPBasicBlock::GetBlockSize(void) {
	vector<SMPInstr *>::iterator InstIter = this->GetFirstInst();
	size_t BlockSize = 0;
	if ((*InstIter)->IsMarkerInst())
		++InstIter;
	while (InstIter != this->GetLastInst()) {
		SMPInstr *CurrInst = (*InstIter);
		BlockSize += CurrInst->GetSize();
		++InstIter;
	}
	return BlockSize;
} // end of SMPBasicBlock::GetBlockSize()

// compute or recompute LiveIn set
void SMPBasicBlock::ComputeLiveInSet(bool StackOnly) {
	// Dataflow equation: A variable is live-in to this block if it
	//  is upwardly exposed from this block, or if it passes through
	//  the block unchanged (i.e. it is not killed and is live out).
	bool UseFP = this->GetFunc()->UsesFramePointer();
	if (!StackOnly) {
		this->LiveInSet.clear();
	}
	STARSOpndSetIter OutIter;
	for (OutIter = this->GetFirstUpExposed(); OutIter != this->GetLastUpExposed(); ++OutIter) {
		STARSOpndTypePtr UpExposedOp = (*OutIter);
		if (!StackOnly || MDIsStackAccessOpnd(UpExposedOp, UseFP)) {
			this->AddLiveIn(UpExposedOp);
		}
	}
	for (OutIter = this->LiveOutSet.begin(); OutIter != this->LiveOutSet.end(); ++OutIter) {
		STARSOpndTypePtr LiveOutOp = (*OutIter);
		if (!this->IsVarKill(LiveOutOp)) { // Found live out but not killed
			if (!StackOnly || MDIsStackAccessOpnd(LiveOutOp, UseFP)) {
				this->AddLiveIn(LiveOutOp);
			}
		}
	}
	if (0 == this->GetNumber()) { // First block has marker inst with LiveIn DEFs.
		// LiveIn DEFs in marker inst are also VarKill for this block. That makes the "LiveOut but not VarKill"
		//  equation invalid, because the MarkerInst DEFs are special cases of VarKill.
		vector<SMPInstr *>::iterator MarkerIter = this->GetFirstInst();
		SMPInstr *MarkerInst = (*MarkerIter);
		assert(MarkerInst->IsMarkerInst());
		for (STARSDefUseIter DefIter = MarkerInst->GetFirstDef(); DefIter != MarkerInst->GetLastDef(); ++DefIter) {
			this->AddLiveIn(DefIter->GetOp());
		}
	}
	this->SetLiveInStale(false);
	return;
} // end of SMPBasicBlock::ComputeLiveInSet()

// Return an iterator for the beginning of the LiveInSet. If the set is stale,
//  recompute it first.
STARSOpndSetIter SMPBasicBlock::GetFirstLiveIn(void) {
	if (this->IsLiveInStale()) {
		this->ComputeLiveInSet(false);
	}
	return this->LiveInSet.begin();
} // end of SMPBasicBlock::GetFirstLiveIn()

// Get termination iterator marker for the LiveIn set, for use by predecessors.
STARSOpndSetIter SMPBasicBlock::GetLastLiveIn(void) {
	// Does not matter if it is stale or not; end marker is the same
	return this->LiveInSet.end();
}

// Get iterator for the start of the LiveOut set.
STARSOpndSetIter SMPBasicBlock::GetFirstLiveOut(void) {
	return this->LiveOutSet.begin();
}

// Get termination iterator marker for the LiveOut set.
STARSOpndSetIter SMPBasicBlock::GetLastLiveOut(void) {
	return this->LiveOutSet.end();
}

// Get iterator for the start of the VarKill set.
STARSOpndSetIter SMPBasicBlock::GetFirstVarKill(void) {
	return this->KillSet.begin();
}

// Get termination iterator marker for the VarKill set.
STARSOpndSetIter SMPBasicBlock::GetLastVarKill(void) {
	return this->KillSet.end();
}

// Get iterator for the start of the UpExposed set.
STARSOpndSetIter SMPBasicBlock::GetFirstUpExposed(void) {
	return this->UpExposedSet.begin();
}

// Get termination iterator marker for the UpExposed set.
STARSOpndSetIter SMPBasicBlock::GetLastUpExposed(void) {
	return this->UpExposedSet.end();
}

// Get iterator for the start of the LocalNames set.
STARSOpndSetIter SMPBasicBlock::GetFirstLocalName(void) {
	return this->LocalNames.begin();
}

// Get termination iterator marker for the LocalNames set.
STARSOpndSetIter SMPBasicBlock::GetLastLocalName(void) {
	return this->LocalNames.end();
}

// Get iterator for the start of the DomFrontier set.
set<int>::iterator SMPBasicBlock::GetFirstDomFrontier(void) {
	return this->DomFrontier.begin();
}

// Get termination iterator marker for the DomFrontier set.
set<int>::iterator SMPBasicBlock::GetLastDomFrontier(void) {
	return this->DomFrontier.end();
}

// Get iterator for first Phi function.
set<SMPPhiFunction, LessPhi>::iterator SMPBasicBlock::GetFirstPhi(void) {
	return this->PhiFunctions.begin();
}

// Get termination iterator marker for Phi functions set.
set<SMPPhiFunction, LessPhi>::iterator SMPBasicBlock::GetLastPhi(void) {
	return this->PhiFunctions.end();
}

// Find the Phi function that matches the given STARSOpndType.
set<SMPPhiFunction, LessPhi>::iterator SMPBasicBlock::FindPhi(const STARSOpndTypePtr &FindOp) {
	set<SMPPhiFunction, LessPhi>::iterator PhiIter;
	STARSOpndSet::value_compare OpComp = this->LiveOutSet.value_comp(); // LessOp()
	for (PhiIter = this->GetFirstPhi(); PhiIter != this->GetLastPhi(); ++PhiIter) {
		STARSOpndTypePtr PhiOp = PhiIter->GetPhiRef(0).GetOp();
		if (OpComp(FindOp, PhiOp) || OpComp(PhiOp, FindOp)) {
			// One is less than the other.
			; // do nothing
		}
		else {
			// They are equal.
			return PhiIter;
		}
	}
	// If we exit the loop, we did not find it.
	return this->PhiFunctions.end();
} // end of SMPBasicBlock::FindPhi()

// Get a pointer to the instruction at address InstAddr.
//  Will assert if it does not find the instruction.
SMPInstr *SMPBasicBlock::FindInstr(STARS_ea_t InstAddr) {
	SMPInstr *CurrInst;
	map<STARS_ea_t, size_t>::iterator MapEntry;

	MapEntry = this->AddrInstMap.find(InstAddr);
	if (MapEntry == this->AddrInstMap.end()) {
		SMP_msg("FATAL ERROR: Failure on AddrInstMap. Size: %zu Addr: %lx \n",
			this->AddrInstMap.size(), (unsigned long) InstAddr);
		SMP_msg("First addr in block: %lx \n", (unsigned long) this->GetFirstAddr());
	}
	assert(MapEntry != this->AddrInstMap.end());
	CurrInst = this->GetInstFromIndex(MapEntry->second);

	return CurrInst;
} // end of SMPBasicBlock::FindInstr()

// Get an iterator to the instruction at address InstAddr.
//  Will assert if it does not find the instruction.
vector<SMPInstr *>::iterator SMPBasicBlock::GetInstIterFromAddr(STARS_ea_t InstAddr) {
#if 1
	size_t VecIndex = this->GetIndexFromInstAddr(InstAddr);
	assert(VecIndex < this->InstVec.size());
	vector<SMPInstr *>::iterator InstIter(this->InstVec.begin() + VecIndex);
#else

	vector<SMPInstr *>::iterator InstIter;
	for (InstIter = this->GetFirstInst(); InstIter != this->GetLastInst(); ++InstIter) {
		if (InstAddr == (*InstIter)->GetAddr()) {
			break;
		}
	}
	if (InstIter == this->GetLastInst()) {
		SMP_msg("FATAL ERROR: Failure to find Instr. Addr: %x \n", InstAddr);
		SMP_msg("First addr in block: %x \n", this->GetFirstAddr());
	}
#endif
	assert(InstIter != this->GetLastInst());

	return InstIter;
} // end of SMPBasicBlock::GetInstIterFromAddr()

std::vector<SMPInstr *>::const_iterator SMPBasicBlock::GetInstConstIterFromAddr(STARS_ea_t InstAddr) const {
	size_t VecIndex = this->GetIndexFromInstAddr(InstAddr);
	vector<SMPInstr *>::const_iterator InstIter(this->InstVec.begin() + VecIndex);
	return InstIter;
}

// Get the InstVec index for the instruction at InstAddr. Assert if not found.
size_t SMPBasicBlock::GetIndexFromInstAddr(STARS_ea_t InstAddr) const {
	SMPInstr *CurrInst = NULL;
	STARS_ea_t CurrAddr = 0;
	size_t VecIndex = 0;
	size_t UpperLimit = this->InstVec.size();
	if (STARS_IsSSAMarkerPseudoID(InstAddr)) {
		// SSA Marker has to be the first inst.
		VecIndex = 0;
		CurrAddr = this->InstVec[0]->GetAddr();
	}

	else if (global_STARS_program->IsIDAProDriverMode() // cannot binary search on IRDB instruction IDs
		&& (UpperLimit > 10)) { // big enough for binary search
		size_t LowerLimit = 0;
		do {
			VecIndex = (LowerLimit + UpperLimit) / 2;
			assert(VecIndex < this->InstVec.size());
			CurrInst = this->InstVec[VecIndex];
			CurrAddr = CurrInst->GetAddr();
			bool MarkerInst = CurrInst->IsMarkerInst();
			if (MarkerInst) {
				LowerLimit = VecIndex + 1;
			}
			else if (CurrAddr > InstAddr) { // search lower half
				UpperLimit = VecIndex - 1;
			}
			else if (CurrAddr < InstAddr) { // search upper half
				LowerLimit = VecIndex + 1;
			}
		} while ((CurrAddr != InstAddr) && (LowerLimit <= UpperLimit));
	}
	else { // small enough for linear search, or IRDB build so we have to do linear search
		for (VecIndex = 0; VecIndex < UpperLimit; ++VecIndex) {
			CurrInst = this->InstVec[VecIndex];
			CurrAddr = CurrInst->GetAddr();
			if (CurrAddr == InstAddr) {
				break;
			}
		}
	}
	assert((CurrAddr == InstAddr) && (VecIndex < this->InstVec.size()));
	return VecIndex;
} // end of SMPBasicBlock::GetIndexFromInstAddr()

// Find the compare or test that sets the flags used in COND_BRANCH at end of block, else NULL; If decrement found instead, fill DecAddr
//  Set the BranchOperator to correspond to the type of COND_BRANCH opcode.
SMPInstr * SMPBasicBlock::FindBranchCompareOrTest(SMPoperator &BranchOperator, STARS_ea_t &DecAddr) const {
	SMPInstr *CompareOrTestInst = nullptr;
	vector<SMPInstr *>::const_reverse_iterator InstIter = this->GetRevInstCBegin();
	SMPInstr *BranchInst = (*InstIter);
	// NOTE: Handle loop opcodes later.
	if ((COND_BRANCH == BranchInst->GetDataFlowType()) && (!BranchInst->MDIsLoopInstr())) {
		BranchOperator = BranchInst->GetCondBranchOperator();
		// Get the SSA number of the flags regs USE in the branch.
		STARSOpndTypePtr FlagsOp = BranchInst->MakeRegOpnd(MD_FLAGS_REG);
		STARSDefUseIter FlagsUseIter = BranchInst->FindUse(FlagsOp);
		assert(FlagsUseIter != BranchInst->GetLastUse());
		int FlagsSSANum = FlagsUseIter->GetSSANum();
		// Make sure this method is not called before SSA form is computed.
		if (SMP_SSA_UNINIT != FlagsSSANum) {
			++InstIter;
			while (InstIter != this->GetRevInstCEnd()) {
				SMPInstr *CurrInst = (*InstIter);
				bool DecFlag = CurrInst->IsDecrementRTL() && (!CurrInst->MDIsLoadEffectiveAddressInstr());
				if (CurrInst->MDIsCompareOrTest() || DecFlag) {
					STARSDefUseIter FlagsDefIter = CurrInst->FindDef(FlagsOp);
					assert(FlagsDefIter != CurrInst->GetLastDef());
					if (FlagsDefIter->GetSSANum() == FlagsSSANum) {
						if (DecFlag) {
							// Leave CompareOrTestInst set to nullptr. Update DecAddr argument.
							DecAddr = CurrInst->GetAddr();
							SMP_msg("INFO: Found loop BIV decrement rather than compare or test inst at %llx in func %s\n",
								(uint64_t) DecAddr, this->GetFunc()->GetFuncName());
						}
						else {
							CompareOrTestInst = CurrInst;
							DecAddr = STARS_BADADDR;
						}
					}
					break;
				}
				else {
					++InstIter;
				}
			}
		}
	}
	return CompareOrTestInst;
} // end of SMPBasicBlock::FindBranchCompareOrTest()

// The predecessor order is used in SSA phi functions. Get the ordinal number
//  of a given block number in the predecessor list.
int SMPBasicBlock::GetPredPosition(int BlockNum) {
	int counter = 0;
	list<SMPBasicBlock *>::iterator PredIter;
	for (PredIter = this->GetFirstPred(); PredIter != this->GetLastPred(); ++PredIter) {
		if (BlockNum == (*PredIter)->GetNumber()) {
			return counter;
		}
		++counter;
	}
	return -1;
}

// Set the DEF iterator from the DefAddr and DefOp. Assert if not found.
set<DefOrUse, LessDefUse>::iterator SMPBasicBlock::GetGlobalDefIterFromDefAddr(const STARSOpndTypePtr &DefOp, STARS_ea_t DefAddr) {
	map<STARS_ea_t, size_t>::iterator MapIter = this->AddrInstMap.find(DefAddr);
	assert(MapIter != this->AddrInstMap.end());
	SMPInstr *DefInst = this->GetInstFromIndex(MapIter->second);
	set<DefOrUse, LessDefUse>::iterator DefIter = DefInst->FindDef(DefOp);
	assert(DefIter != DefInst->GetLastDef());
	return DefIter;
} // end of SMPBasicBlock::GetGlobalDefIterFromDefAddr()

// retrieve from LocalDefAddrBySSA or return BADADDR
STARS_ea_t SMPBasicBlock::GetLocalDefAddrForRegHash(int RegHashValue) {
	STARS_ea_t DefAddr = STARS_BADADDR; // STARS_BADADDR means we did not find it
	map<int, STARS_ea_t>::iterator MapResult = this->LocalDefAddrBySSA.find(RegHashValue);
	if (MapResult != this->LocalDefAddrBySSA.end()) { // Found it.
		DefAddr = (STARS_ea_t) MapResult->second;
	}
	return DefAddr;
} // end of SMPBasicBlock::GetLocalDefAddrForRegHash()

// Given a USE operand and the address of its instr, return DEF addr using SSA chains or Phi func.
//  If DEF is in a Phi function, we return the block number, which (hopefully) never conflicts with instruction
//   addresses.
STARS_ea_t SMPBasicBlock::GetDefAddrFromUseAddr(const STARSOpndTypePtr &UseOp, STARS_ea_t InstAddr, int SSANum, bool LocalName) {
	STARS_ea_t DefAddr = STARS_BADADDR;
	if (0 > SSANum)
		return DefAddr;

	bool DataFlowOp = MDIsDataFlowOpnd(UseOp, this->GetFunc()->UsesFramePointer());
	bool PhiUse = (STARS_IsBlockNumPseudoID(InstAddr));

	if (DataFlowOp && (!LocalName) && (!PhiUse)) {
		// Global DEF for this SSANum must be in the Phi functions or within a block.
		DefAddr = this->MyFunc->GetGlobalDefAddr(UseOp, SSANum); // only works on registers and stack locations
		if (DefAddr == STARS_BADADDR) { // Could not find it anywhere.
			this->GetFunc()->Dump();
			SMP_msg("ERROR: Failure in GetDefAddrFromUseAddr(): InstAddr %lx SSANum %d\n",
					(unsigned long) InstAddr, SSANum);
			SMP_msg(" LocalName: %d UseOp.reg: %d\n", LocalName, UseOp->GetReg());
			assert(DefAddr != STARS_BADADDR); // kablooey!
		}
	}
	else if (LocalName || PhiUse) {
		if (PhiUse) {
			// Need to find DEF for a PhiUse.
			DefAddr = this->GetFunc()->GetGlobalDefAddr(UseOp, SSANum);
			if (STARS_BADADDR == DefAddr) { // Need to search Phi functions.
				// Rare case in which a Phi USE was traced only to a Phi DEF in an earlier block.
				DefAddr = (STARS_ea_t) this->GetFunc()->GetBlockNumForPhiDef(UseOp, SSANum);
				assert(STARS_BADADDR != DefAddr);
				DefAddr += STARS_PSEUDO_ID_MIN;
			}
		}
		else {
			vector<SMPInstr *>::iterator InstIter = this->GetInstIterFromAddr(InstAddr);
			if (this->IsVarKill(UseOp)) { // Cannot find a DEF in block if not in VarKill set for this block
				if (UseOp->IsRegOp()) {
					int HashedName = HashGlobalNameAndSSA(UseOp, SSANum);
					DefAddr = this->GetLocalDefAddrForRegHash(HashedName);
				}
				else {
#if 0
					vector<SMPInstr *>::reverse_iterator RevInstIter(InstIter);
#else
					vector<SMPInstr *>::reverse_iterator RevInstIter = this->GetRevInstBegin();
#endif
					while (RevInstIter != this->GetRevInstEnd()) {
						SMPInstr *CurrInst = (*RevInstIter);
						STARS_ea_t CurrAddr = CurrInst->GetAddr();
						set<DefOrUse, LessDefUse>::iterator DefIter = CurrInst->FindDef(UseOp);
						if ((DefIter != CurrInst->GetLastDef()) && (SSANum == DefIter->GetSSANum())) {
							DefAddr = CurrAddr;
							break;
						}
						++RevInstIter;
					}
				}
			}
			if (DefAddr == STARS_BADADDR) { // not found yet
				// Must be a global name defined in the Phi functions or LiveIn and UpExposed to the block, or a non-stack mem access.
				bool UseFP = this->GetFunc()->UsesFramePointer();
				if (LocalName && MDIsDataFlowOpnd(UseOp, UseFP)) {
					// Should have found DefAddr by now. Why were we called in the first place?
					SMP_msg("\nERROR in GetDefAddrFromUseAddr() for UseAddr: %llx\n", (unsigned long long) InstAddr);
					SMP_msg("UseOp: "); PrintOperand(UseOp); SMP_msg("\n");
					SMP_msg("LocalName: %d  IsDataFlowOpnd: %d\n", LocalName, MDIsDataFlowOpnd(UseOp, UseFP));
					SMP_msg("UseFP: %d LiveIn: %d Killed: %d\n", UseFP, this->IsLiveIn(UseOp), this->IsVarKill(UseOp));
#if 0				// Just leave DefAddr set to STARS_BADADDR and return.
					assert(!LocalName || (!MDIsDataFlowOpnd(UseOp, UseFP)));
#endif

				}
				else {
					set<SMPPhiFunction, LessPhi>::iterator PhiIter = this->FindPhi(UseOp);
					if (PhiIter != this->GetLastPhi()) {
						// Found a Phi function for UseOp; does it DEF our SSANum?
						int PhiDefSSANum = PhiIter->GetDefSSANum();
						if (PhiDefSSANum == SSANum) {
							DefAddr = this->GetNumber() + STARS_PSEUDO_ID_MIN;  // BlockNum is signal that the Def has no real addr, is in Phi
						}
					}
					else {
						DefAddr = STARS_LIVEIN_PSEUDO_ID; // pseudo-def-addr for LiveIn, not in Phi, UpExposed
						SMP_msg("INFO: STARS_LIVEIN_PSEUDO_ID returned as DefAddr for UseAddr %llx SSANum: %d UseOp: ", (uint64_t) InstAddr, SSANum);
						PrintOperand(UseOp);
						SMP_msg("\n");
					}
				}
			}
		}
	}

	return DefAddr;
} // end of SMPBasicBlock::GetDefAddrFromUseAddr()

// trace into predecessors if necessary
STARS_ea_t SMPBasicBlock::GetUltimateDefAddr(const STARSOpndTypePtr &UseOp, STARS_ea_t InstAddr, int SSANum, bool LocalName, bool RecurseOnMoves, bool TraceStackMoves, STARSOpndTypePtr &DefMoveOp) {
	STARS_ea_t DefAddr = this->GetDefAddrFromUseAddr(UseOp, InstAddr, SSANum, LocalName);
	bool UseFP = this->GetFunc()->UsesFramePointer();
	bool LiveInAddr = STARS_IsLiveInPseudoID(DefAddr);
	if (LiveInAddr) { // UpExposed or pass-through case
		// We need to find the predecessor block that DEFed the UseOp/SSANum.
		assert(!LocalName); // also cannot be a register; it is a global name stack var
		assert(this->IsLiveIn(UseOp));

#if 1
		size_t NumBlocks = this->GetFunc()->GetNumBlocks();
		bool FoundDEF = false;
		STARSDefUseIter DefIter;
		// Check special case of DEF in marker instruction at beginning of function.
		if (SSANum == 0) {
			SMPInstr *FirstInst = *(this->GetFunc()->GetFirstInstIter());
			if (FirstInst->IsMarkerInst()) {
				DefIter = FirstInst->FindDef(UseOp);
				if (DefIter != FirstInst->GetLastDef()) {
					assert(0 == DefIter->GetSSANum());
					DefAddr = FirstInst->GetAddr();
					FoundDEF = true;
				}
			}
		}

		for (size_t BlockIndex = 0; (BlockIndex < NumBlocks) && (!FoundDEF); ++BlockIndex) {
			SMPBasicBlock *CurrBlock = this->GetFunc()->GetBlockByNum(BlockIndex);
			// Don't bother to look inside block insts unless UseOp is killed and LiveOut
			if (CurrBlock->IsLiveOut(UseOp)) {
				// Can we match the SSANum in an inst DEF or a Phi DEF?
				set<SMPPhiFunction, LessPhi>::iterator PhiIter = CurrBlock->FindPhi(UseOp);
				if (PhiIter != CurrBlock->GetLastPhi()) {
					if (SSANum == PhiIter->GetDefSSANum()) {
						DefAddr = (STARS_ea_t) BlockIndex + STARS_PSEUDO_ID_MIN;
						FoundDEF = true;
						break;
					}
				}
				if (CurrBlock->IsVarKill(UseOp)) {
					vector<SMPInstr *>::reverse_iterator RevInstIter;
					for (RevInstIter = CurrBlock->GetRevInstBegin(); RevInstIter != CurrBlock->GetRevInstEnd(); ++RevInstIter) {
						SMPInstr *CurrInst = (*RevInstIter);
						STARS_ea_t CurrAddr = CurrInst->GetAddr();
						if (CurrInst->HasDestMemoryOperand() || CurrInst->MDIsPushInstr() || RecurseOnMoves) {
							DefIter = CurrInst->FindDef(UseOp);
							if (DefIter != CurrInst->GetLastDef()) {
								// Because we are using a reverse iterator, the first DEF we encounter either
								//  matches the SSANum or else this block is irrelevant.
								if (SSANum == DefIter->GetSSANum()) {
									DefAddr = CurrAddr;
									FoundDEF = true;
									if (RecurseOnMoves) {
										bool NewDefMoveOp = false;
										if (CurrInst->MDIsMoveInstr()) {
											DefMoveOp = CurrInst->GetMoveSource();
											DefMoveOp = CloneIfNecessary(DefMoveOp, this->GetFunc()->UsesFramePointer());
											CanonicalizeOpnd(DefMoveOp);
											NewDefMoveOp = true;
										}
										else if (CurrInst->MDIsLoadEffectiveAddressInstr()) {
											DefMoveOp = CurrInst->GetLeaMemUseOp();
											NewDefMoveOp = true;
											if (!MDIsIndexedMemoryAccess(DefMoveOp)) { // lea is just a move
												STARS_regnum_t LeaAddrReg = DefMoveOp->GetReg();
												LeaAddrReg = MDCanonicalizeSubReg(LeaAddrReg);
												DefMoveOp = CurrInst->MakeRegOpnd(LeaAddrReg); // convert to RegOp
											}
										}
										// We recurse if the DefMoveOp is a non-stack-ptr-reg, including the LeaMemUseOp being
										//  a single register (e.g. base but no index: lea ebx,[edx] is just a move in disguise)
										bool ValidOperand = NewDefMoveOp && ((DefMoveOp->IsRegOp() && (!MDIsStackPtrReg(DefMoveOp->GetReg(), UseFP)))
											|| (TraceStackMoves && MDIsDirectStackAccessOpnd(DefMoveOp, UseFP)));
										if (RecurseOnMoves && ValidOperand) {
											STARSDefUseIter UseIter = CurrInst->FindUse(DefMoveOp);
											assert(UseIter != CurrInst->GetLastUse());
											SMPBasicBlock *NewBlock = CurrInst->GetBlock();
											bool LocMoveName = NewBlock->IsLocalName(DefMoveOp);
											DefAddr = NewBlock->GetUltimateDefAddr(DefMoveOp, DefAddr, UseIter->GetSSANum(), LocMoveName, true, TraceStackMoves, DefMoveOp);
										}
									}
								}
								break; // found it or did not; break regardless
							}
						}
					} // end for all insts in CurrBlock in reverse order
				} // end if VarKill in CurrBlock
			} // end if LiveOut from CurrBlock
		} // end for all blocks by BlockIndex && !FoundDEF
#else
		list<SMPBasicBlock *>::iterator PredIter;
		bool FoundDefiningBlock = false;
		for (PredIter = this->GetFirstPred(); PredIter != this->GetLastPred(); ++PredIter) {
			SMPBasicBlock *PredBlock = (*PredIter);
			if (!PredBlock->IsProcessed() && PredBlock->IsLiveOut(UseOp)) {
				// Found possible defining block. At least one predecessor should have UseOp LiveOut.
				// We can recurse, but we need a pseudo UseAddr. We
				//  will use the last instruction in PredBlock as the pseudo UseAddr, but first
				//  we need to check to see if it is the DefAddr.
				vector<SMPInstr *>::reverse_iterator RevInstIter = PredBlock->GetRevInstBegin();
				SMPInstr *LastPredInst = (*RevInstIter);
				STARS_ea_t LastPredAddr = LastPredInst->GetAddr();
				if (LastPredInst->FindDef(UseOp) != LastPredInst->GetLastDef()) {
					DefAddr = LastPredAddr;
					PredBlock->SetProcessed(true);
					FoundDefiningBlock = true;
				}
				else { // We can use LastPredInst addr as pseudo-UseAddr.
					DefAddr = PredBlock->GetUltimateDefAddr(UseOp, LastPredAddr, SSANum, LocalName);
				}
				if (STARS_BADADDR != DefAddr) {
					break;
				}
			}
		}
#endif
	} // end if LiveIn
	else if (RecurseOnMoves) {
		if (!(STARS_IsBlockNumPseudoID(DefAddr) || STARS_IsSSAMarkerPseudoID(DefAddr))) { // not Phi or SSA Marker DEF, real inst found
			SMPInstr *CurrInst = this->GetFunc()->GetInstFromAddr(DefAddr);
			assert(NULL != CurrInst);
			bool NewDefMoveOp = false;
#if 0
			if (CurrInst->MDIsMoveInstr()) {
				DefMoveOp = CurrInst->GetMoveSource();
				DefMoveOp = CloneIfNecessary(DefMoveOp, this->GetFunc()->UsesFramePointer());
				CanonicalizeOpnd(DefMoveOp);
				NewDefMoveOp = true;
			}
			else if (CurrInst->MDIsLoadEffectiveAddressInstr()) {
				DefMoveOp = CurrInst->GetLeaMemUseOp();
				NewDefMoveOp = true;
				if (!MDIsIndexedMemoryAccess(DefMoveOp)) { // lea is just a move
					STARS_regnum_t LeaAddrReg = DefMoveOp->GetReg();
					LeaAddrReg = MDCanonicalizeSubReg(LeaAddrReg);
					DefMoveOp = CurrInst->MakeRegOpnd(LeaAddrReg); // convert to RegOp
				}
			}
#else
			NewDefMoveOp = CurrInst->IsSimpleCopy(DefMoveOp);
#endif
			// We recurse if the DefMoveOp is a non-stack-ptr-reg, including the LeaMemUseOp being
			//  a single register (e.g. base but no index: lea ebx,[edx] is just a move in disguise)
			CanonicalizeOpnd(DefMoveOp);
			bool ValidOperand = NewDefMoveOp && ((DefMoveOp->IsRegOp() && (!MDIsStackPtrReg(DefMoveOp->GetReg(), UseFP)))
				|| (TraceStackMoves && MDIsDirectStackAccessOpnd(DefMoveOp, UseFP)));
			if (ValidOperand) {
				STARSDefUseIter UseIter = CurrInst->FindUse(DefMoveOp);
				assert(UseIter != CurrInst->GetLastUse());
				SMPBasicBlock *NewBlock = CurrInst->GetBlock();
				bool LocMoveName = NewBlock->IsLocalName(DefMoveOp);
				DefAddr = NewBlock->GetUltimateDefAddr(DefMoveOp, DefAddr, UseIter->GetSSANum(), LocMoveName, true, TraceStackMoves, DefMoveOp);
			}
		}
	}
	assert(STARS_BADADDR != DefAddr);
	return DefAddr;
} // end of SMPBasicBlock::GetUltimateDefAddr()

// STUB: Trace UseOp through move and Phi defs to some defining operation.
STARS_ea_t SMPBasicBlock::GetUltimateOperandSource(STARS_ea_t UseAddr, const STARSOpndTypePtr &UseOp, int UseSSANum) {
	bool LocalName = this->IsLocalName(UseOp);
	STARS_ea_t DefiningAddr = this->GetDefAddrFromUseAddr(UseOp, UseAddr, UseSSANum, LocalName);
	SMPBasicBlock *CurrBlock;

	// We have four possibilities for DefiningAddr:
	//  1: STARS_BADADDR means we cannot trace the operand.
	//  2: An instruction address, in which case we recurse if inst is a move.
	//  3: Our block number, which means a Phi DEF, so we need to recurse into our predecessors.
	//  4: Our block's first addr minus one, indicating upexposed in this block, need to recurse into predecessors.


	return DefiningAddr;
} // end of SMPBasicBlock::GetUltimateOperandSource()

// Trace back to const or InArg with min const value; follow Phi USEs to before-loop locations only.
bool SMPBasicBlock::GetUltimateInitValue(const STARSOpndTypePtr &UseOp, STARS_ea_t InstAddr, int SSANum, bool LocalName, STARSOpndTypePtr &DefMoveOp, STARS_uval_t &InitValue) {
	bool success = false;
	// First, see if SCCP resolved the operand to a constant.
	if (UseOp->IsRegOp()) {
		int RegHashValue = HashGlobalNameAndSSA(UseOp, SSANum);
		STARSSCCPMapIter SCCPIter;
		success = this->GetFunc()->FindSCCPConstIter(this, RegHashValue, LocalName, SCCPIter);
		if (success && ((*SCCPIter).second.ConstType == STARS_CONST_HAS_VALUE)) {
			InitValue = (*SCCPIter).second.ConstValue;
			return success;
		}
	}

	STARS_ea_t DefAddr = this->GetDefAddrFromUseAddr(UseOp, InstAddr, SSANum, LocalName);
	bool UseFP = this->GetFunc()->UsesFramePointer();
	bool LiveInAddr = STARS_IsLiveInPseudoID(DefAddr);
	if (LiveInAddr) { // UpExposed or pass-through case
		return success; // !!!!****!!!! What about the InitValue?
	} // end if LiveIn
	else {
		if (!(STARS_IsBlockNumPseudoID(DefAddr) || STARS_IsSSAMarkerPseudoID(DefAddr))) { // not Phi or SSA Marker DEF, real inst found
			SMPInstr *CurrInst = this->GetFunc()->GetInstFromAddr(DefAddr);
			assert(NULL != CurrInst);
			STARS_uval_t ConstValue;
			bool NewDefMoveOp = CurrInst->IsSimpleOrExtendedCopy(DefMoveOp);
			if (NewDefMoveOp && DefMoveOp->IsImmedOp()) {
				InitValue = DefMoveOp->GetImmedValue();
				success = true;
			}
			else if (NewDefMoveOp) { // Not ImmedOp, but we can follow simple copy chain
				// We recurse if the DefMoveOp is a non-stack-ptr-reg, including the LeaMemUseOp being
				//  a single register (e.g. base but no index: lea ebx,[edx] is just a move in disguise)
				CanonicalizeOpnd(DefMoveOp);
				bool ValidOperand = NewDefMoveOp && ((DefMoveOp->IsRegOp() && (!MDIsStackPtrReg(DefMoveOp->GetReg(), UseFP)))
					|| MDIsDirectStackAccessOpnd(DefMoveOp, UseFP));
				if (ValidOperand) {
					STARSDefUseIter UseIter = CurrInst->FindUse(DefMoveOp);
					assert(UseIter != CurrInst->GetLastUse());
					SMPBasicBlock *NewBlock = CurrInst->GetBlock();
					bool LocMoveName = NewBlock->IsLocalName(DefMoveOp);
					success = NewBlock->GetUltimateInitValue(DefMoveOp, DefAddr, UseIter->GetSSANum(), LocMoveName, DefMoveOp, InitValue);
				}
				else if (DefMoveOp->IsStaticMemOp() && (!MDIsIndirectMemoryOpnd(DefMoveOp, false))) {
					// We have a global var accessed without index register.
					//  See if we have observed constant writes to the global var.
					STARS_ea_t GlobalAddr = DefMoveOp->GetAddr();
					STARS_uval_t ImmedVal;
					bool found = this->GetFunc()->GetProg()->FindGlobalMinValue(GlobalAddr, ImmedVal);
					if (found) {
						InitValue = ImmedVal;
						success = true;
					}
				}
			}
			else if (CurrInst->IsAdditionOfConstant(DefMoveOp, ConstValue)) {
				bool ValidOperand = ((DefMoveOp->IsRegOp() && (!MDIsStackPtrReg(DefMoveOp->GetReg(), UseFP)))
					|| MDIsDirectStackAccessOpnd(DefMoveOp, UseFP));
				if (ValidOperand) {
					CanonicalizeOpnd(DefMoveOp);
					STARSDefUseIter UseIter = CurrInst->FindUse(DefMoveOp);
					assert(UseIter != CurrInst->GetLastUse());
					SMPBasicBlock *NewBlock = CurrInst->GetBlock();
					bool LocMoveName = NewBlock->IsLocalName(DefMoveOp);
					success = NewBlock->GetUltimateInitValue(DefMoveOp, DefAddr, UseIter->GetSSANum(), LocMoveName, DefMoveOp, InitValue);
					if (success)
						InitValue += ConstValue;
				}
			}
			else if (CurrInst->IsSubtractionOfConstant(DefMoveOp, ConstValue)) {
				bool ValidOperand = ((DefMoveOp->IsRegOp() && (!MDIsStackPtrReg(DefMoveOp->GetReg(), UseFP)))
					|| MDIsDirectStackAccessOpnd(DefMoveOp, UseFP));
				if (ValidOperand) {
					CanonicalizeOpnd(DefMoveOp);
					STARSDefUseIter UseIter = CurrInst->FindUse(DefMoveOp);
					assert(UseIter != CurrInst->GetLastUse());
					SMPBasicBlock *NewBlock = CurrInst->GetBlock();
					bool LocMoveName = NewBlock->IsLocalName(DefMoveOp);
					success = NewBlock->GetUltimateInitValue(DefMoveOp, DefAddr, UseIter->GetSSANum(), LocMoveName, DefMoveOp, InitValue);
					if (success)
						InitValue -= ConstValue;
				}
			}
		}
		else if (STARS_IsSSAMarkerPseudoID(DefAddr)) {
			// Traced to InArg; see if we have ConstInArgValues for this InArg.
			assert((0 == SSANum) || this->GetFunc()->GetBlockByNum(0)->IsLoopHeaderBlock());
			size_t ArgPos;
			if (UseOp->IsRegOp() && global_STARS_program->GetArgRegPosition(UseOp->GetReg(), ArgPos)) {
				if ((ArgPos < this->GetFunc()->GetIncomingArgCount()) && this->GetFunc()->DoesInArgHaveConstValues(ArgPos)) {
					InitValue = this->GetFunc()->GetFirstConstInArgValue(ArgPos);
					success = true;
				}
			}
		}
		else { // Must be Phi Function.
			// If loop header block, get incoming value to the loop and recurse.
			int PhiBlockNum = STARS_GetBlockNumFromPseudoID(DefAddr);
			SMPBasicBlock *PhiBlock = this->GetFunc()->GetBlockByNum(PhiBlockNum);
			assert(nullptr != PhiBlock);
			if (PhiBlock->IsLoopHeaderBlock()) {
				int IncomingSSANum = PhiBlock->GetLoopLiveInSSANum(UseOp);
				if (0 <= IncomingSSANum) {
					success = PhiBlock->GetUltimateInitValue(UseOp, DefAddr, IncomingSSANum, LocalName, DefMoveOp, InitValue);
				}
			}
			else { // if-else phi function; could spawn separate tracers and
				// accumulate a set of values in the future.
				success = false;
			}
		}
	}

	return success;
} // end of SMPBasicBlock::GetUltimateInitValue()

// Does DefOp get written out in truncated form (lower bits only)?
bool SMPBasicBlock::IsOpDestTruncatedWrite(const STARSOpndTypePtr &DefOp, int DefSSANum, STARS_ea_t DefAddr) { 
	vector<SMPInstr *>::iterator InstIter;
	SMPInstr *LastUseInst = NULL;
	set<DefOrUse, LessDefUse>::iterator DefIter;
	bool RegisterOp = (DefOp->IsRegOp());
	bool LastUseIsTruncatedWrite = false;
	STARS_ea_t LastAddr = STARS_BADADDR;
	
	if (!RegisterOp || (DefSSANum == SMP_SSA_UNINIT))
		return false;

	InstIter = this->GetInstIterFromAddr(DefAddr);
	++InstIter;
	while (InstIter != this->GetLastInst()) {
		LastUseInst = (*InstIter);
		if (LastUseInst->FindUse(DefOp) != LastUseInst->GetLastUse()) {
			LastAddr = LastUseInst->GetAddr();
		}
		if (LastUseInst->FindDef(DefOp) != LastUseInst->GetLastDef()) {
			// re-DEF
			break;
		}
		++InstIter;
	}

	if (STARS_BADADDR != LastAddr) {
		InstIter = this->GetInstIterFromAddr(LastAddr);
		LastUseInst = (*InstIter);
		if (LastUseInst->MDDoublesWidth()) {
			LastUseIsTruncatedWrite = true;
		}
		else if (LastUseInst->MDIsReducedWidthMove()) {
			// If DefOp is used as an address reg, it is not a truncated write of DefOp;
			//  otherwise, it is.
			LastUseIsTruncatedWrite = LastUseInst->IsNonAddressReg(DefOp);
		}
		if (!LastUseIsTruncatedWrite && (NULL != LastUseInst)) {
			// See if we can recurse through a regular move.
			if (LastUseInst->MDIsMoveInstr()) {
				DefIter = LastUseInst->GetFirstNonFlagsDef();
				assert(DefIter != LastUseInst->GetLastDef());
				STARSOpndTypePtr CopyDefOp = DefIter->GetOp();
				bool UseFP = this->GetFunc()->UsesFramePointer();
				if (MDIsDataFlowOpnd(CopyDefOp, UseFP)) {
					int CopyDefSSANum = DefIter->GetSSANum();
					LastUseIsTruncatedWrite = this->IsOpDestTruncatedWrite(CopyDefOp, CopyDefSSANum, LastAddr);
				}
			}
		}
	}

	return LastUseIsTruncatedWrite;
} // end of SMPBasicBlock::IsOpDestTruncatedWrite()

// Update the LiveOut set for the block.
// Return true if it changed, false otherwise.
bool SMPBasicBlock::UpdateLiveOut(bool StackOnly) {
	bool changed = false;
	bool UseFP = this->GetFunc()->UsesFramePointer();
	STARSOpndSet OldLiveOut(this->LiveOutSet); // save copy of old LiveOutSet
	if (!StackOnly) {
		this->LiveOutSet.clear();  // Clear it and rebuild it
	}
	// Dataflow equation for LiveOutSet: If a variable is live-in for any successor
	//  block, it is live out for this block.
	list<SMPBasicBlock *>::iterator SuccIter;
	STARSOpndSetIter InSuccIter;
	for (SuccIter = this->Successors.begin(); SuccIter != this->Successors.end(); ++SuccIter) {
		for (InSuccIter = (*SuccIter)->GetFirstLiveIn(); InSuccIter != (*SuccIter)->GetLastLiveIn(); ++InSuccIter) {
			STARSOpndTypePtr SuccLiveInOp = (*InSuccIter);
			if (!StackOnly || MDIsStackAccessOpnd(SuccLiveInOp, UseFP)) {
				this->AddLiveOut(SuccLiveInOp);
			}
		}
	}

	// Only remaining question: Did the LiveOutSet change?
	// Short cut: If the set cardinality changed, then the set changed.
	if (this->LiveOutSet.size() != OldLiveOut.size()) {
		changed = true;
	}
	else if (!StackOnly) { // Same # of elements; move through in lockstep and compare.
		STARSOpndSetIter NewIter = this->LiveOutSet.begin();
		STARSOpndSetIter OldIter = OldLiveOut.begin();
		STARSOpndSet::value_compare OpComp = OldLiveOut.value_comp(); // LessOp()
		while (OldIter != OldLiveOut.end()) { // both iters terminate simultaneously
			if (OpComp(*OldIter, *NewIter) || OpComp(*NewIter, *OldIter)) {
				changed = true;
				break;
			}
			++OldIter;
			++NewIter;
		}
	}

	if (!StackOnly && changed)
		this->SetLiveInStale(true);

	OldLiveOut.clear();
	return changed;
} // end of SMPBasicBlock::UpdateLiveOut()

// Remove stack ops starting at LVASetIter; return true if any removed; helper for RecomputeStackLVA
bool SMPBasicBlock::RemoveStackOpsFromLVASets(void) {
	bool StackOpRemoved = false;
	bool UseFP = this->GetFunc()->UsesFramePointer();

	STARSOpndSetIter LVAIter = this->GetFirstLiveIn();
	STARSOpndSetIter TempIter;
	while (LVAIter != this->GetLastLiveIn()) {
		if (MDIsStackAccessOpnd((*LVAIter), UseFP)) {
			TempIter = LVAIter; // work around buggy STL implementations
			++TempIter;
			LVAIter = this->EraseLiveIn(LVAIter);
			LVAIter = TempIter;
			StackOpRemoved = true;
		}
		else {
			++LVAIter;
		}
	}

	LVAIter = this->GetFirstLiveOut();
	while (LVAIter != this->GetLastLiveOut()) {
		if (MDIsStackAccessOpnd((*LVAIter), UseFP)) {
			TempIter = LVAIter; // work around buggy STL implementations
			++TempIter;
			LVAIter = this->EraseLiveOut(LVAIter);
			LVAIter = TempIter;
			StackOpRemoved = true;
		}
		else {
			++LVAIter;
		}
	}

	LVAIter = this->GetFirstUpExposed();
	while (LVAIter != this->GetLastUpExposed()) {
		if (MDIsStackAccessOpnd((*LVAIter), UseFP)) {
			TempIter = LVAIter; // work around buggy STL implementations
			++TempIter;
			LVAIter = this->EraseUpExposed(LVAIter);
			LVAIter = TempIter;
			StackOpRemoved = true;
		}
		else {
			++LVAIter;
		}
	}

	LVAIter = this->GetFirstVarKill();
	while (LVAIter != this->GetLastVarKill()) {
		if (MDIsStackAccessOpnd((*LVAIter), UseFP)) {
			TempIter = LVAIter; // work around buggy STL implementations
			++TempIter;
			LVAIter = this->EraseVarKill(LVAIter);
			LVAIter = TempIter;
			StackOpRemoved = true;
		}
		else {
			++LVAIter;
		}
	}

	return StackOpRemoved;
} // end of SMPBasicBlock::RemoveStackOpsFromSet()

#if 0
// Set the SSA numbers for the DEFs in the LiveOut set.
void SMPBasicBlock::SetLiveOutSSANums(void) {
	size_t LiveOutCount = this->LiveOutSet.size();
	size_t LiveOutDefsMarked = 0;
	vector<SMPInstr *>::reverse_iterator RevInstIter;

	for (RevInstIter = this->GetRevInstBegin(); (RevInstIter != this->GetRevInstEnd()) &&  (LiveOutDefsMarked < LiveOutCount); ++RevInstIter) {
		set<DefOrUse, LessDefUse>::iterator DefIter;
		SMPInstr *CurrInst = (*RevInstIter);
		for (DefIter = CurrInst->GetFirstDef(); DefIter != CurrInst->GetLastDef(); ++DefIter) {
			STARSOpndTypePtr DefOp = DefIter->GetOp();
			set<DefOp, LessOp>::iterator LiveOutIter = this->LiveOutSet.find(DefOp);
			if (LiveOutIter != this->GetLastLiveOut()) { // found DefOp in LiveOut set
				;
			}
		}
	}
	return;
} // end of SMPBasicBlock::SetLiveOutSSANums()
#endif

// Compute the ReachesIn set as the union of the ReachesOut sets of all predecessor blocks.
// Return TRUE if the ReachesIn set was updated, false if it gained no new elements.
bool SMPBasicBlock::ComputeReachesInSet(void) {
	size_t OldSize = this->ReachesInSet.size();
	size_t NewSize; // If NewSize becomes different from OldSize, then ReachesInSet changed.
	list<SMPBasicBlock *>::iterator PredIter;

	for (PredIter = this->GetFirstPred(); PredIter != this->GetLastPred(); ++PredIter) {
		SMPBasicBlock *CurrPred = *PredIter;
		this->ReachesInSet.insert(CurrPred->GetFirstReachesOut(), CurrPred->GetLastReachesOut());
	}

	NewSize = this->ReachesInSet.size();
	return (NewSize != OldSize);
} // end of SMPBasicBlock::ComputeReachesInSet()

// Compute the ReachesOut set according to its dataflow formula: ReachesOut = ((ReachesIn minus VarKill) union DownExposedDefs).
bool SMPBasicBlock::ComputeReachesOutSet(void) {
	size_t OldSize = this->ReachesOutSet.size();
	size_t NewSize; // If NewSize becomes different from OldSize, then ReachesOutSet changed.

	// First step: Add all ReachesIn defns that are not killed.
	set<pair<STARSOpndTypePtr, STARS_ea_t>, LessDefinition>::iterator ReachesInIter;
	for (ReachesInIter = this->GetFirstReachesIn(); ReachesInIter != this->GetLastReachesIn(); ++ReachesInIter) {
		STARSOpndTypePtr InOp = ReachesInIter->first;
		if (!(this->IsVarKill(InOp))) {
			this->AddReachesOut(*ReachesInIter);
		}
	}

	// Second step: Add DEDefns.
	set<pair<STARSOpndTypePtr, STARS_ea_t>, LessDefinition>::iterator DEDefnIter;
	for (DEDefnIter = this->GetFirstDownExposedDefn(); DEDefnIter != this->GetLastDownExposedDefn(); ++DEDefnIter) {
		this->AddReachesOut(*DEDefnIter);
	}

	NewSize = this->ReachesOutSet.size();
	this->ResetReachesOutStale();
	return (NewSize != OldSize);
} // end of SMPBasicBlock::ComputeReachesOutSet()

// Add DefOp, remove previous defs of DefOp that now do not reach the end of the block.
void SMPBasicBlock::UpdateDownExposedDefs(STARSOpndTypePtr DefOp, STARS_ea_t InstAddr) {
	pair<STARSOpndTypePtr, STARS_ea_t> NewDefn(DefOp, InstAddr);

	// First, remove any definition of DefOp that precedes the current definition.
	STARSDefinitionSet::iterator DEDefnIter = this->GetFirstDownExposedDefn();
	while (DEDefnIter != this->GetLastDownExposedDefn()) {
		STARSOpndTypePtr OldOp = DEDefnIter->first;
		if (IsEqOpIgnoreBitwidth(OldOp, DefOp)) {
			STARS_ea_t OldAddr = DEDefnIter->second;
			if (OldAddr == InstAddr) { // nothing is changing
				return;
			}
			else {
				(void) this->DownExposedDefnSet.erase(DEDefnIter);
				this->SetReachesOutStale();
				break; // save time; should never be more than one defn per DefOp in this set, unlike ReachesIn & ReachesOut
			}
		}
		else {
			++DEDefnIter;
		}
	}

#if 1
	// Only purpose of the DownExposedDefnSet is to track stack pointer copies, so save time.
	if (!this->GetFunc()->IsDefnInStackPtrCopySet(DefOp, InstAddr))
		return;
#endif

	// Next, insert the new definition.
	pair<STARSDefinitionSet::iterator, bool> InsertResult = this->DownExposedDefnSet.insert(NewDefn);
	assert(InsertResult.second);
	this->SetReachesOutStale();
	return;
} // end of SMPBasicBlock::UpdateDownExposedDefs()

// Evaluate constants for all instructions as part of SCCP algorithm at SMPFunction level.
void SMPBasicBlock::SCCPEvaluateConstants(enum STARSBranchConst &BranchEval, list<pair<int, int> > &CFGWorkList, list<pair<int, int> > &SSAWorkList) {
	// for (each instruction i in block) do
	//    if (i is an assignment) then
	//       EvaluateAssign(i)
	//    else if (i is a conditional branch) then
	//       EvaluateConditional(i)
	//    endif
	// endfor

	BranchEval = STARS_BRANCH_UNKNOWN;

	// for (each instruction i in block) do
	vector<SMPInstr *>::iterator InstIter;
	for (InstIter = this->GetFirstInst(); InstIter != this->GetLastInst(); ++InstIter) {
		SMPInstr *CurrInst = (*InstIter);
	//    if (i is an assignment) then
	//       EvaluateAssign(i)
		SMPitype DataFlowType = CurrInst->GetDataFlowType();
		if (DEFAULT == DataFlowType) {
			CurrInst->SCCPEvaluateAssignment(SSAWorkList);
		}
	//    else if (i is a conditional branch) then
	//       EvaluateConditional(i)
		else if (COND_BRANCH == DataFlowType) {
			CurrInst->SCCPEvaluateCondBranch(BranchEval);
		}
		// We need to augment the published algorithm to handle interprocedural DEFs, which
		//  appear as DEFs in call instructions.
		else if ((CALL == DataFlowType) || (INDIR_CALL == DataFlowType)) {
			for (set<DefOrUse, LessDefUse>::iterator DefIter = CurrInst->GetFirstDef(); DefIter != CurrInst->GetLastDef(); ++DefIter) {
				int DefSSANum = DefIter->GetSSANum();
				STARSOpndTypePtr DefOp = DefIter->GetOp();
				assert(DefOp->IsRegOp());
				int DefHashValue = HashGlobalNameAndSSA(DefOp, DefSSANum);
				bool LocalName = CurrInst->GetBlock()->IsLocalName(DefOp);
				struct STARS_SCCP_Const_Struct CallConstStruct;
				CallConstStruct.ConstType = STARS_CONST_BOTTOM;
				map<int, struct STARS_SCCP_Const_Struct>::iterator NewConstIter;
				if (LocalName) {
					NewConstIter = this->InsertLocalConstValue(DefHashValue, CallConstStruct);
				}
				else {
					NewConstIter = this->GetFunc()->InsertGlobalConstValue(DefHashValue, CallConstStruct);
				}
				// Propagate along SSA edges.
				int BlockNum = CurrInst->GetBlock()->GetNumber();
				pair<int, int> SSAEdge(BlockNum, DefHashValue);
				SSAWorkList.push_back(SSAEdge);
			}
		}
	//    endif
	// endfor
	}
	this->SetSCCPVisited(true);

	// Handle successors, based on BranchEval.
	this->SCCPHandleSuccessors(BranchEval, CFGWorkList);

	return;
} // end of SMPBasicBlock::SCCPEvaluateConstants()

// Propagate GlobalOp/DefSSANum const value
void SMPBasicBlock::SCCPGlobalPropagationHelper(const STARSOpndTypePtr &GlobalOp, int DefSSANum, vector<STARSBitSet> ExecutedEdgeBitSet, list<pair<int, int> > &CFGWorkList, list<pair<int, int> > &SSAWorkList) {
	// Three cases: 
	// (1) GlobalOp/DefSSANum is a Phi USE, in which case we re-eval the Phi Function and return.
	// (2) GlobalOp/DefSSANum is upward exposed, in which case we find the UpExposed USEs and re-eval those statements and recurse if LiveOut.
	// (3) GlobalOp/DefSSANum is LiveOut (either pass-through case or DEFed in this block) and we recurse.
	bool FoundReDEF = true;
	enum STARSBranchConst BranchEval;
	BranchEval = STARS_BRANCH_UNKNOWN;
	this->SetProcessed(true); // prevent infinite recursion
	set<SMPPhiFunction, LessPhi>::iterator PhiIter = this->FindPhi(GlobalOp);
	if (PhiIter != this->GetLastPhi()) { // Case 1.
		int PhiDefSSANum = PhiIter->GetDefSSANum();
		if (PhiDefSSANum != DefSSANum) { // DefSSANum should be found among the Phi USEs
			this->GetFunc()->EvaluateAllPhiConstants(this->GetNumber(), ExecutedEdgeBitSet, SSAWorkList);
		}
	}
	else if (this->IsUpExposed(GlobalOp)) { // case 2
		FoundReDEF = false;
		vector<SMPInstr *>::iterator InstIter;
		for (InstIter = this->GetFirstInst(); InstIter != this->GetLastInst(); ++InstIter) {
			SMPInstr *CurrInst = (*InstIter);
			set<DefOrUse, LessDefUse>::iterator UseIter = CurrInst->FindUse(GlobalOp);
			if (UseIter != CurrInst->GetLastUse()) { // operand is USEd; check SSANum
				int UseSSANum = UseIter->GetSSANum();
				if (UseSSANum == DefSSANum) {
					SMPitype DataFlowType = CurrInst->GetDataFlowType();
					if (DEFAULT == DataFlowType) {
						CurrInst->SCCPEvaluateAssignment(SSAWorkList);
					}
					else if (COND_BRANCH == DataFlowType) {
						CurrInst->SCCPEvaluateCondBranch(BranchEval);
						this->SCCPHandleSuccessors(BranchEval, CFGWorkList);
					}
				}
			}
			set<DefOrUse, LessDefUse>::iterator DefIter = CurrInst->FindDef(GlobalOp);
			if (DefIter != CurrInst->GetLastDef()) { // check for re-DEF
				int NewDefSSANum = DefIter->GetSSANum();
				if (NewDefSSANum > DefSSANum) { // re-DEF; we can stop searching for USEs of DefOp/DefSSANum in this block
					FoundReDEF = true;
					break;
				}
			}
		} // end for all insts in block
	}
	else if (this->IsLiveOut(GlobalOp)) { // case 3
		// Not UpExposed, but LiveOut; not re-DEF in block
		FoundReDEF = false;
	}
	if (!FoundReDEF) { // common recursion for case 2 and case 3
		list<SMPBasicBlock *>::iterator SuccIter;
		for (SuccIter = this->GetFirstSucc(); SuccIter != this->GetLastSucc(); ++SuccIter) {
			SMPBasicBlock *SuccBlock = (*SuccIter);
			if ((!SuccBlock->IsProcessed()) && SuccBlock->IsSCCPVisited() && SuccBlock->IsLiveIn(GlobalOp)) {
				SuccBlock->SCCPGlobalPropagationHelper(GlobalOp, DefSSANum, ExecutedEdgeBitSet, CFGWorkList, SSAWorkList);
			}
		}
	}

	return;
} // end of SMPBasicBlock::SCCPGlobalPropagationHelper()

// Based on BranchEval, push successors onto CFGWorkList
void SMPBasicBlock::SCCPHandleSuccessors(enum STARSBranchConst &BranchEval, list<pair<int, int> > &CFGWorkList) {
	//  Put block successors on CFGWorkList, based on conditional branch evaluation if any
	bool TakeDistantSucc = (BranchEval != STARS_BRANCH_NEVER_TAKEN);
	bool TakeFallThroughSucc = (BranchEval != STARS_BRANCH_ALWAYS_TAKEN);
	list<SMPBasicBlock *>::const_iterator SuccBlockIter = this->GetFirstConstSucc();
	list<SMPBasicBlock *>::const_iterator FallThroughSuccIter = this->GetFallThroughSucc();
	while (SuccBlockIter != this->GetLastSucc()) { // not the return block; has a successor
		if (((SuccBlockIter == FallThroughSuccIter) && (TakeFallThroughSucc)) 
			|| ((SuccBlockIter != FallThroughSuccIter) && (TakeDistantSucc))) {
			SMPBasicBlock *SuccBlock = (*SuccBlockIter);
			if (!(SuccBlock->IsUnreachableBlock())) {  // might have call 0 inst but otherwise look like it is reachable
				int NextBlockNum = SuccBlock->GetNumber();
				pair<int, int> TempEdge(this->GetNumber(), NextBlockNum);
				CFGWorkList.push_back(TempEdge);
			}
		}
		++SuccBlockIter;
	}
	return;
} // end of SMPBasicBlock::SCCPHandleSuccessors()

// Patch binary to make this block into nops.
void SMPBasicBlock::SCCPNullifyUnreachableBlock(void) {
	SMP_msg("INFO: SCCP: Removing unreachable block at %lx, patching in NOPs\n", 
		(unsigned long) this->GetFirstAddr());

	// Remove this block from the predecessors list of its successors.
	list<SMPBasicBlock *>::iterator SuccIter;
	STARS_ea_t TempAddr = this->GetFirstAddr();
	for (SuccIter = this->GetFirstSucc(); SuccIter != this->GetLastSucc(); ++SuccIter) {
		(*SuccIter)->ErasePred(TempAddr);
	}

#if 1
	// Remove this block from the successors list of its predecessors.
	list<SMPBasicBlock *>::iterator PredIter;
	for (PredIter = this->GetFirstPred(); PredIter != this->GetLastPred(); ++PredIter) {
		(*PredIter)->EraseSucc(TempAddr);
	}
#endif

	// Patch all bytes in the database for this block into nops.
	size_t NumBytes = this->GetBlockSize();
	STARS_ea_t CurrAddr = this->GetFirstAddr();
	for (size_t ByteIndex = 0; ByteIndex < NumBytes; ++ByteIndex) {
		bool success = global_stars_interface->STARS_patch_byte(CurrAddr, MD_BINARY_NOP_INSTRUCTION);
		// If success == false, that just means the byte already had that value in the 
		//  database, so nothing was changed. Cannot assert that success should be true.
		//  assert(success);
		++CurrAddr;
	}

	return;
} // end of SMPBasicBlock::SCCPNullifyUnreachableBlock()

// Gather statistics on SCCP effectiveness
void SMPBasicBlock::SCCPGatherStatistics(void) {
#if STARS_SCCP_GATHER_STATISTICS
	vector<SMPInstr *>::iterator InstIter;
	bool ConstArgFound = false;
	bool ArgWriteFound = false;
	for (InstIter = this->GetFirstInst(); InstIter != this->GetLastInst(); ++InstIter) {
		SMPInstr *CurrInst = (*InstIter);
		SMPitype DataFlowType = CurrInst->GetDataFlowType();
		if (CurrInst->HasDestMemoryOperand()) { // NOTE: Use MDIsArgumentPass() instead?
			STARSOpndTypePtr MoveSourceOp = CloneIfSubwordReg(CurrInst->GetMoveSource());
			STARSOpndTypePtr DEFMemOp = CurrInst->GetMemDef();
			if ((nullptr != MoveSourceOp) && (! MoveSourceOp->IsVoidOp()) && this->GetFunc()->IsInOutgoingArgsRegion(DEFMemOp)) {
				++SCCPOutgoingArgWriteCount;
				ArgWriteFound = true;
				if (MoveSourceOp->IsImmedOp()) {
					++SCCPConstantOutgoingArgWriteCount;
					ConstArgFound = true;
				}
				else if (MoveSourceOp->IsRegOp()) {
					CanonicalizeOpnd(MoveSourceOp);
					set<DefOrUse, LessDefUse>::iterator UseIter = CurrInst->FindUse(MoveSourceOp);
					assert(UseIter != CurrInst->GetLastUse());
					int UseSSANum = UseIter->GetSSANum();
					bool LocalName = this->IsLocalName(MoveSourceOp);
					int UseHashIndex = HashGlobalNameAndSSA(MoveSourceOp, UseSSANum);
					STARSSCCPMapIter ConstIter;
					ConstArgFound = this->GetFunc()->FindSCCPConstIter(this, UseHashIndex, LocalName, ConstIter);
					if (ConstArgFound) { // found an SCCP const entry
						++SCCPConstantOutgoingArgWriteCount;
					}
				}
				// NOTE: Deal with read-only memory lookup. !!!!****!!!!
			}
		}
		else if ((DataFlowType == CALL) || (DataFlowType == INDIR_CALL)) {
			// Need to start gathering data for the next call; reset counters.
			//  Can have multiple calls in one basic block.
			if (ArgWriteFound) {
				++SCCPFuncsWithArgWriteCount;
				if (ConstArgFound) {
					++SCCPFuncsWithConstantArgWriteCount;
				}
			}
			ArgWriteFound = false;
			ConstArgFound = false;
		}
	} // end for all instructions
	if (ArgWriteFound) {
		++SCCPFuncsWithArgWriteCount;
		if (ConstArgFound) {
			++SCCPFuncsWithConstantArgWriteCount;
		}
	}
#endif
	return;
} // end of SMPBasicBlock::SCCPGatherStatistics()

// Insert RPO number block into the dominance frontier set.
void SMPBasicBlock::AddToDomFrontier(int block) {
	this->DomFrontier.insert(block);
	return;
} // end of SMPBasicBlock::AddToDomFrontier()

// Fill the set of non-global names used in this block. Set local indices that can be
//  used later to index into a vector of def-use chains.
void SMPBasicBlock::SetLocalNames(void) {
	vector<SMPInstr *>::iterator InstIter;
	size_t LocalIndex = 0;
	this->LocalNames.clear(); // in case we are recomputing SSA
#if SMP_DEBUG_DATAFLOW
	SMP_msg("Entered SetLocalNames.\n");
#endif
	set<DefOrUse, LessDefUse>::iterator CurrDef;
	for (InstIter = this->GetFirstInst(); InstIter != this->GetLastInst(); ++InstIter) {
		SMPInstr *CurrInst = (*InstIter);
		for (CurrDef = CurrInst->GetFirstDef(); CurrDef != CurrInst->GetLastDef(); ++CurrDef) {
			STARSOpndTypePtr DefOp = CurrDef->GetOp();
			if (!(this->MyFunc->IsGlobalName(DefOp))) {
				// Not in global names set
				if (this->LocalNames.end() == this->LocalNames.find(DefOp)) {
					// Not yet in LocalNames, so add it.
					SetGlobalIndex(DefOp, LocalIndex);
					this->LocalNames.insert(DefOp);
					++LocalIndex;
					if (DefOp->IsRegOp()) {
						STARS_regnum_t RegNum = DefOp->GetReg();
						if (global_STARS_program->IsCallerSavedReg(RegNum)) {
							this->GetFunc()->SetCallerSavedLocalReg(RegNum);
						}
					}
				}
			}
		}
	}
	return;
} // end of SMPBasicBlock::SetLocalNames()

// Create local DEF-USE chains and renumber all references to all names in LocalNames.
void SMPBasicBlock::SSALocalRenumber(void) {
#if SMP_DEBUG_DATAFLOW
	bool DebugFlag = false;
	DebugFlag |= (0 == strcmp(".init_proc", this->MyFunc->GetFuncName()));
#endif
	size_t NumLocals = this->LocalNames.size();
	vector<int> SSAIndex;
	// Initialize SSAIndex and DUChain values for each local name.
	STARSOpndSetIter NameIter;
#if SMP_DEBUG_DATAFLOW
	if (DebugFlag)
		SMP_msg("LocalNames size: %zu\n", NumLocals);
#endif

	for (NameIter = this->LocalNames.begin(); NameIter != this->LocalNames.end(); ++NameIter) {
		SSAIndex.push_back(-1); // init SSA indices to -1; first DEF will make it 0
	}

	// Iterate through all instructions in the block. For each DEF of a local name,
	//  set a new SSA index and start a new DU chain. For each USE of a local name,
	//  set the current SSA index for that name and add the instruction to the DU
	//  chain.
	vector<SMPInstr *>::iterator InstIter;
	set<DefOrUse, LessDefUse>::iterator CurrUse, CurrDef;
	STARSOpndSetIter UseNameIter, DefNameIter;
	for (InstIter = this->GetFirstInst(); InstIter != this->GetLastInst(); ++InstIter) {
		SMPInstr *CurrInst = (*InstIter);
		STARS_ea_t InstAddr = CurrInst->GetAddr(); // for debugging
		size_t NameIndex;
		// USEs get referenced logically before DEFs, so start with them.
		CurrUse = CurrInst->GetFirstUse();
		while (CurrUse != CurrInst->GetLastUse()) {
			STARSOpndTypePtr UseOp = CurrUse->GetOp();
			UseNameIter = this->LocalNames.find(UseOp);
			if (UseNameIter != this->LocalNames.end()) {
				// Found USE of a local name.
				NameIndex = ExtractGlobalIndex(*UseNameIter);
				if (NameIndex > SSAIndex.size())
					SMP_msg("FATAL ERROR: NameIndex %zu out of range in SSALocalRenumber.\n", NameIndex);
				int SSANum = SSAIndex.at(NameIndex);
				// Update the SSA subscript in the DEF-USE list for the instruction.
#if SMP_DEBUG_DATAFLOW
				if (DebugFlag) {
					SMP_msg("Setting SSANum to %d for local NameIndex %d\n", SSANum, NameIndex);
					SMP_msg("Addr %x USE operand: ", CurrInst->GetAddr()); 
					PrintOneOperand(UseOp, 0, -1);
					SMP_msg("\n");
				}
#endif
				CurrUse = CurrInst->SetUseSSA(UseOp, SSANum);
			}
			++CurrUse;
		} // end for all USEs in the instruction.
		// Now do the DEFs.
		CurrDef = CurrInst->GetFirstDef();
		while (CurrDef != CurrInst->GetLastDef()) {
			STARSOpndTypePtr DefOp = CurrDef->GetOp();
			DefNameIter = this->LocalNames.find(DefOp);
			if (DefNameIter != this->LocalNames.end()) {
				// Found DEF of a local name.
				NameIndex = ExtractGlobalIndex(*DefNameIter);
				if (NameIndex > SSAIndex.size())
					SMP_msg("ERROR: DEF NameIndex %zu out of range in SSALocalRenumber.\n", NameIndex);
				++SSAIndex[NameIndex]; // move up to next SSA subscript number
				int SSANum = SSAIndex.at(NameIndex);
				// Put new SSA number into DEF list entry in the instruction.
				SetGlobalIndex(DefOp, NameIndex);
				CurrDef = CurrInst->SetDefSSA(DefOp, SSANum);
				// Track the DefAddr for local register operands.
				if (DefOp->IsRegOp()) {
					int DefHashValue = HashGlobalNameAndSSA(DefOp, SSANum);
					pair<int, STARS_ea_t> DefMapEntry(DefHashValue, InstAddr);
					pair<map<int, STARS_ea_t>::iterator, bool> MapReturnValue = this->LocalDefAddrBySSA.insert(DefMapEntry);
					assert(MapReturnValue.second);
				}
			}
			++CurrDef;
		} // end for all DEFs in the instruction
	} // end for all instructions in the block

	// Keep track of largest SSA number used, for container allocations.
	for (size_t index = 0; index < this->LocalNames.size(); ++index) {
		this->GetFunc()->UpdateLocalMaxSSANum(SSAIndex[index]);
	}

#if SMP_DEBUG_DATAFLOW
	if (DebugFlag)
		SMP_msg("Exiting SSALocalRenumber()\n");
#endif
	return;
} // end of SMPBasicBlock::SSALocalRenumber()

// Determine if block is loop tail, also set loop header flag for dominating back-edge successor.
// Does DefOp/DefSSANum control loop-back or loop-exit branches?
bool SMPBasicBlock::DoesDefControlLoopFlow(const STARSOpndTypePtr &DefOp, int DefSSANum, STARS_ea_t DefAddr) {
	bool FoundLoopControl = false;

	// First, determine if the current block ends with a loop exit or a loop-back.
	vector<SMPInstr *>::reverse_iterator RevInstIter = this->GetRevInstBegin();
	SMPInstr *LastInst = (*RevInstIter);
	SMPitype FlowType = LastInst->GetDataFlowType();
	if (FlowType == COND_BRANCH) {
		STARS_ea_t LastAddr = LastInst->GetAddr();
		ControlFlowType LastBranchType = this->GetFunc()->GetControlFlowType(LastAddr);
		if ((LastBranchType != LOOP_EXIT) && (LastBranchType != LOOP_BACK)) {
			return false;
		}
	}
	else {
		return false; // no conditional branch being controlled in this block.
	}

	// Now, see if DefOp+DefSSANum is USEd in the compare or test.
	++RevInstIter;
	while (RevInstIter != this->GetRevInstEnd()) {
		SMPInstr *CurrInst = (*RevInstIter);
		if (CurrInst->MDIsCompareOrTest()) {
			set<DefOrUse, LessDefUse>::iterator UseIter = CurrInst->FindUse(DefOp);
			if (UseIter != CurrInst->GetLastUse()) {
				// DefOp is USEd in compare or test. Check SSA number.
				FoundLoopControl = (UseIter->GetSSANum() == DefSSANum);
			}
			break;
		}
		++RevInstIter;
	}

	return FoundLoopControl;
} // end of SMPBasicBlock::DoesDefControlLoopFlow()

// Free memory for reaching defs sets after they are no longer needed.
void SMPBasicBlock::FreeReachingDefsMemory(void) {
	this->ReachesInSet.clear();
	this->ReachesOutSet.clear();
	this->DownExposedDefnSet.clear();
	return;
} // end of SMPBasicBlock::FreeReachingDefsMemory()

// Free SSA data structures that are no longer needed when all SSA numbers have
//  been recorded in DEFs and USEs.
void SMPBasicBlock::FreeSSAMemory(void) {
	this->DomFrontier.clear();
#if SMP_SHRINK_TO_FIT
	std::set<int>(this->DomFrontier).swap(this->DomFrontier);
#endif
	return;
} // end of SMPBasicBlock::FreeSSAMemory()

// After type inference, before annotation emission, free memory that is no
//  longer needed (i.e. after loop 4 in SMPProgram::Analyze()).
void SMPBasicBlock::FreeUnusedMemory4(void) {
#if 0  // PhiFunctions, LiveOutSet, Successors and AddrInstMap are used in SMPInstr::EmitIntegerErrorAnnotations()
	this->LiveInSet.clear();
	this->KillSet.clear();
	this->LiveOutSet.clear();
	this->UpExposedSet.clear();
	this->Predecessors.clear();
	this->Successors.clear();
	this->PhiFunctions.clear();
	this->AddrInstMap.clear();
#endif
#if SMP_SHRINK_TO_FIT
#if 0
	STARSOpndSet(this->LiveInSet).swap(this->LiveInSet);
	STARSOpndSet(this->UpExposedSet).swap(this->UpExposedSet);
	STARSOpndSet(this->KillSet).swap(this->KillSet);
	STARSOpndSet(this->LiveOutSet).swap(this->LiveOutSet);
	std::set<SMPPhiFunction, LessPhi>(this->PhiFunctions).swap(this->PhiFunctions);
	std::map<STARS_ea_t, size_t>(this->AddrInstMap).swap(this->AddrInstMap);
	std::list<SMPBasicBlock *>(this->Predecessors).swap(this->Predecessors);
	std::list<SMPBasicBlock *>(this->Successors).swap(this->Successors);
#endif
#endif
	return;
} // end of SMPBasicBlock::FreeUnusedMemory4()

#if STARS_BUILD_DEF_USE_CHAINS

// Find the local Def-Use chain index for the given DefOp with given SSA index.
unsigned int SMPBasicBlock::GetLocalDUIndex(STARSOpndTypePtr DefOp, int SSANum) {
	unsigned int RegIndex = 0;
	STARSOpndSetIter NameIter;

	NameIter = this->LocalNames.find(DefOp);
	if (NameIter == this->LocalNames.end()) {
		SMP_msg("ERROR: local DEF name not in LocalNames: ");
		PrintOperand(DefOp);
		SMP_msg("\n");
		assert(NameIter != this->LocalNames.end()); // abort
	}
	else {
		// DefOp is in the LocalNames, as expected
		RegIndex = ExtractGlobalIndex(*NameIter);
		assert(0 <= SSANum);
	}
	return RegIndex;
} // end of SMPBasicBlock::GetLocalDUIndex()

// Find the global Def-Use chain index for the given DefOp with given address.
unsigned int SMPBasicBlock::GetGlobalDUIndex(STARSOpndTypePtr DefOp, STARS_ea_t DefAddr) {
	unsigned int RegIndex = 0;
	STARSOpndSetIter NameIter;

	NameIter = this->MyFunc->FindGlobalName(DefOp);
	if (NameIter == this->MyFunc->GetLastGlobalName()) {
		SMP_msg("ERROR: global DEF name at %lx not in GlobalNames: ", (unsigned long) DefAddr);
		PrintOperand(DefOp);
		SMP_msg("\n");
		assert(NameIter != this->MyFunc->GetLastGlobalName()); // abort
	}
	else {
		// DefOp is in the GlobalNames, as expected
		RegIndex = ExtractGlobalIndex(*NameIter);
	}
	return RegIndex;
} // end of SMPBasicBlock::GetGlobalDUIndex()

#endif

// MemDefOp starting at InstIter has DU chain overlapping an indirect mem write
bool SMPBasicBlock::HasUseAfterIndWrite(const STARSOpndTypePtr &MemDefOp, vector<SMPInstr *>::iterator InstIter, bool &IndWriteSeen, bool &ReDef) {
	IndWriteSeen = false;
	ReDef = false;

	bool FoundIndirectWrite = false;
	set<DefOrUse, LessDefUse>::iterator UseIter;
	while (InstIter != this->GetLastInst()) {
		SMPInstr *TempInst = (*InstIter);
		if (TempInst->FindDef(MemDefOp) != TempInst->GetLastDef()) {
			// Re-DEF before indirect write.
			ReDef = true;
			return false;
		}
		else if (FoundIndirectWrite) {
			UseIter = TempInst->FindUse(MemDefOp);
			if (UseIter != TempInst->GetLastUse()) {
				// USEd after indirect write.
				return true;
			}
		}
		else if (TempInst->HasIndirectMemoryWrite()) {
			// Indirect write before re-DEF.
			FoundIndirectWrite = true;
			IndWriteSeen = true;
		}
		++InstIter;
	}
	// No re-DEF, no use after indirect write.
	return false;
} // end of SMPBasicBlock::HasUseAfterIndWrite()

// dead at InstIter?
bool SMPBasicBlock::IsMemOpDead(const STARSOpndTypePtr &MemDefOp, std::vector<SMPInstr *>::iterator InstIter, int SSANum) {
	bool DeadOp = false;
	bool Referenced = false;
	// Avoid iterating through instructions if we have a global MemDefOp that
	//  is the pass-through case (LiveIn, LiveOut, not VarKill) because it
	//  cannot possibly be dead.
	if (this->IsLiveOut(MemDefOp) && (!this->IsVarKill(MemDefOp))) {
		assert(this->IsLiveIn(MemDefOp));
		return false;
	}

	set<DefOrUse, LessDefUse>::iterator UseIter;
	while (InstIter != this->GetLastInst()) {
		SMPInstr *TempInst = (*InstIter);
		UseIter = TempInst->FindUse(MemDefOp);
		if (UseIter != TempInst->GetLastUse()) {
			Referenced = true;
			DeadOp = (UseIter->GetSSANum() != SSANum);
			// USEd after indirect write at incoming InstIter; not dead if SSA match.
			break;
		}
		else if (TempInst->FindDef(MemDefOp) != TempInst->GetLastDef()) {
			// Re-DEF before USE.
			Referenced = true;
			DeadOp = true;
			break;
		}
		++InstIter;
	}

	if (!Referenced) {
		DeadOp = (!this->IsLiveOut(MemDefOp));
	}
	return DeadOp;
} // end of SMPBasicBlock::IsLocalMemOpDead()

// If branch at block end is signed/unsigned, propagate to operands that set flags before it.
void SMPBasicBlock::MarkBranchSignedness(void) {
	unsigned short SignMask;
	set<DefOrUse, LessDefUse>::iterator UseIter;
	STARS_ea_t DefAddr;  // for flags USE in branch
	int SSANum;    // for flags USE in branch
	bool LocalFlags;  // is flags register a local name?
	vector<SMPInstr *>::reverse_iterator InstRevIter;

	InstRevIter = this->GetRevInstBegin(); // Any conditional branch would be last instruction
	SMPInstr *LastInst = (*InstRevIter);
	if (LastInst->MDIsUnsignedBranch()) {
		if (this->IsBranchSignednessUnreliable()) {
			return;
		}
		SignMask = FG_MASK_UNSIGNED;
	}
	else if (LastInst->MDIsSignedBranch()) {
		SignMask = FG_MASK_SIGNED;
	}
	else
		return; // Last instruction is not a signed or unsigned branch.

	// Find the flags USE.
	STARSOpndTypePtr UseOp = LastInst->MakeRegOpnd(X86_FLAGS_REG); // set up a dummy op for searching
	UseIter = LastInst->FindUse(UseOp);
	assert(UseIter != LastInst->GetLastUse());
	UseOp = UseIter->GetOp(); // get full info in all fields
	SSANum = UseIter->GetSSANum();
	LocalFlags = this->IsLocalName(UseOp);

	DefAddr = this->GetDefAddrFromUseAddr(UseOp, LastInst->GetAddr(), SSANum, LocalFlags);
	// Pass DefAddr to recursive helper function.
	this->PropagateBranchSignedness(DefAddr, UseOp, SignMask);

	return;
} // end of SMPBasicBlock::MarkBranchSignedness()

// For the DefAddr of the SearchOp, propagate SignMask to all DEFs and USEs in the
//  instruction or Phi function indicated by DefAddr.
// Currently, the DefAddr initially will be an instruction that defined the flags that were used in a signed or
//  unsigned branch instruction, and the SearchOp will be the flags register. Upon recursion, the DefAddr
//  and SearchOp will change as we backtrack up the signedness propagation chain.
// We recurse on all USEs until we encounter memory operands, also terminating the recursion as
//  soon as we see an instruction other than a move, compare, or test.
void SMPBasicBlock::PropagateBranchSignedness(STARS_ea_t DefAddr, const STARSOpndTypePtr &SearchOp, unsigned short SignMask) {
	SMPInstr *DefInst;
	SMPBasicBlock *DefBlock = NULL;
	set<DefOrUse, LessDefUse>::iterator UseIter;
	set<DefOrUse, LessDefUse>::iterator DefIter;
	int DefHashValue, UseHashValue;
	STARS_ea_t PhiUseDefAddr; // for a Phi USE, where was it DEFed?
	STARS_ea_t UseDefAddr; // for an instruction USE, where was it DEFed?
	bool LocalDef = this->IsLocalName(SearchOp);

	if (! SearchOp->IsRegOp())
		return;  // Limit to registers for now

	if (STARS_IsBlockNumPseudoID(DefAddr)) { // found in Phi DEF; DefAddr is block #
		// We need to propagate the SignMask to the Phi DEF, and then to the Phi USEs, and
		//  then we need to propagate the SignMask to the DEF of each Phi USE. It is possible
		//  that a Phi USE has its DEF in another block's Phi function, so this might recurse.
		size_t BlockNum = STARS_GetBlockNumFromPseudoID(DefAddr);
		if (SearchOp->IsMemOp())
			return;  // end recursion
		DefBlock = this->MyFunc->GetBlockByNum(BlockNum);
		set<SMPPhiFunction, LessPhi>::iterator PhiIter = DefBlock->FindPhi(SearchOp);
		assert(PhiIter != DefBlock->GetLastPhi());
		int SSANum = PhiIter->GetDefSSANum();
		DefHashValue = HashGlobalNameAndSSA(SearchOp, SSANum);
		this->MyFunc->UpdateDefSignMiscInfo(DefHashValue, SignMask);

		size_t UseIndex, UseLimit;
		UseLimit = PhiIter->GetPhiListSize();
		for (UseIndex = 0; UseIndex < UseLimit; ++UseIndex) {
			int UseSSANum = PhiIter->GetUseSSANum(UseIndex);
			UseHashValue = HashGlobalNameAndSSA(SearchOp, UseSSANum);
			// Avoid propagation and recursion if no change will be made.
			unsigned short OldSignInfo = this->MyFunc->GetUseSignMiscInfo(UseHashValue);
			if (OldSignInfo != (OldSignInfo | SignMask)) { // sign info will change
				this->MyFunc->UpdateUseSignMiscInfo(UseHashValue, SignMask);
				PhiUseDefAddr = this->MyFunc->GetGlobalDefAddr(SearchOp, UseSSANum);
				assert(STARS_BADADDR != PhiUseDefAddr);
				if (STARS_IsBlockNumPseudoID(PhiUseDefAddr)) {
					// DEF of current Phi USE is in another Phi function.
					DefBlock = this->MyFunc->GetBlockByNum(STARS_GetBlockNumFromPseudoID(PhiUseDefAddr));
				}
				else { // DEF is in an instruction.
					DefBlock = this->MyFunc->GetBlockFromInstAddr(PhiUseDefAddr);
				}
				if (PhiUseDefAddr != DefAddr) { // Don't recurse infinitely
					DefBlock->PropagateBranchSignedness(PhiUseDefAddr, SearchOp, SignMask); // recurse
				}
			}
		}
	}
	else { // DEF must be in an instruction
		if (LocalDef) { // DEF must be in a local instruction
			DefInst = this->FindInstr(DefAddr);
			DefBlock = this;
		}
		else { // DEF might be in any block in the function
			DefBlock = this->MyFunc->GetBlockFromInstAddr(DefAddr);
			// DefBlock must be good, else GetBlockFromInstAddr() would assert.
			DefInst = DefBlock->FindInstr(DefAddr);
		}

		// We sometimes have a compiler idiom in which a comparison is made between a signed char and
		//  an ASCII value, and then an unsigned branch is taken. We don't want to propagate UNSIGNED
		//  in this case, and we are unsure what signedness is truly present, so we just return.
		if ((FG_MASK_UNSIGNED == SignMask) && (DefInst->MDComparesImmedASCII())) {
			return;
		}

		// If the DEF is from a call instruction, we only want to propagate the sign to the DEF
		//  itself and stop the recursion. The other register USEs and DEFs of the call are unrelated
		//  to the SearchOp DEF, unlike arithmetic instructions.
		if (DefInst->GetDataFlowType() >= JUMP) {
			DefIter = DefInst->FindDef(SearchOp);
			assert(DefIter != DefInst->GetLastDef());
			DefHashValue = HashGlobalNameAndSSA(SearchOp, DefIter->GetSSANum());
			if (LocalDef) {
				this->UpdateDefSignMiscInfo(DefHashValue, SignMask);
			}
			else {
				this->MyFunc->UpdateDefSignMiscInfo(DefHashValue, SignMask);
			}
			return;
		}

		// Now we should have a good DefInst for the Inst that DEFs SearchOp.
		//  Loop through the DEFs and USEs in this Inst and OR in the SignMask for register operands.
		for (DefIter = DefInst->GetFirstDef(); DefIter != DefInst->GetLastDef(); ++DefIter) {
			STARSOpndTypePtr DefOp = DefIter->GetOp();
			if (DefOp->IsRegOp()) {
				DefHashValue = HashGlobalNameAndSSA(DefOp, DefIter->GetSSANum());
				if ((LocalDef && this->IsLocalName(DefOp))
					|| (!LocalDef && DefBlock->IsLocalName(DefOp))) {
					// DefOp is local to a block, must use block-local maps.
					if (LocalDef) { // still within current block
						this->UpdateDefSignMiscInfo(DefHashValue, SignMask);
					}
					else { // DEF in another block; use DefBlock
						DefBlock->UpdateDefSignMiscInfo(DefHashValue, SignMask);
					}
				}
				else { // DefOp is global, use SMPFunction global maps.
					this->MyFunc->UpdateDefSignMiscInfo(DefHashValue, SignMask);
				}
			}
		} // end for all DEFs

		// We don't want to propagate to USEs of instructions that load from non-stack memory. All of the
		//  uses are either memory locations or just addressing registers.
		bool ReadsMemory = DefInst->HasSourceMemoryOperand();
		if (ReadsMemory) {
			if (DefInst->HasIndirectMemoryRead()
				|| (!(DefInst->IsLoadFromStack()) && (DefInst->GetOptType() == 3))) {
				// Either a load that is not from the stack, or an indirect access to memory.
				return; // No USEs we can propagate through
			}
		}

#if STARS_MINIMIZE_UNSIGNED_BRANCH_PROPAGATION
		// Now we propagate to the USEs of the instruction. However, we need to beware of
		//  propagating too much UNSIGNEDNESS, as we sometimes have mixed-mode arithmetic
		//  and lea instructions can add SIGNED and UNSIGNED to produce UNSIGNED, e.g.:
		//   lea edi,[eax+ebx]
		//  Just because EDI is UNSIGNED does not mean that both EAX and EBX are UNSIGNED.
		//   Perhaps EAX is UNSIGNED and EBX is SIGNED, or vice versa.
		if (DefInst->MDIsLoadEffectiveAddressInstr()) {
			// Se if we have more than one non-flags register USE.
			size_t RegUseCount = 0;
			for (UseIter = DefInst->GetFirstUse(); UseIter != DefInst->GetLastUse(); ++UseIter) {
				STARSOpndTypePtr UseOp = UseIter->GetOp();
				if ((UseOp->IsRegOp()) && ((!ReadsMemory) || (DefInst->IsNonAddressReg(UseOp)))) { // don't want addressing registers
					if (!(UseOp->MatchesReg(X86_FLAGS_REG))) {
						++RegUseCount;
					}
				}
			}
			if (RegUseCount > 1) {
				return;
			}
		}
#endif

		// A USE in the instruction that DEFed the flags used in our original condition branch
		//  could be local to the block containing the instruction, else it is a global name.
		//  We always want to update the sign info of the USE. Propagation to the DEF of that USE
		//  will happen only if the instruction is a simple move, compare, or test (for now).
		//  NOTE: We might try to propagate to DEFs of registers by arithmetic instructions later.
#if 1
		bool PropagateThroughUses = DefInst->MDPropagateUseUpFromBranch();
#else
		bool PropagateThroughUses = false;
#endif
		STARS_ea_t UseDefAddr = STARS_BADADDR;
		for (UseIter = DefInst->GetFirstUse(); UseIter != DefInst->GetLastUse(); ++UseIter) {
			STARSOpndTypePtr UseOp = UseIter->GetOp();
			if (UseOp->IsRegOp() && ((!ReadsMemory) || (DefInst->IsNonAddressReg(UseOp)))) { // don't want addressing registers
#if 0  // USEs should already be canonicalized
				CanonicalizeOpnd(UseOp);
#endif
				if (UseOp->MatchesReg(X86_FLAGS_REG)) { 
					// don't need to propagate to flags after initial call to this method
					continue; // save time
				}
				int UseSSANum = UseIter->GetSSANum();
				bool LocalUseOp;
				if (LocalDef) {
					LocalUseOp = this->IsLocalName(UseOp);
				}
				else {
					LocalUseOp = DefBlock->IsLocalName(UseOp);
				}
				UseHashValue = HashGlobalNameAndSSA(UseOp, UseSSANum);
				if (LocalDef) {
					if (LocalUseOp) {// USE must be in a local instruction
						this->UpdateUseSignMiscInfo(UseHashValue, SignMask);
					}
					else { // USE is global name
						this->MyFunc->UpdateUseSignMiscInfo(UseHashValue, SignMask);
					}
				}
				else if (LocalUseOp) { // USE is local to DEF block
					DefBlock->UpdateUseSignMiscInfo(UseHashValue, SignMask);
				}
				else { // USE is global name
					this->MyFunc->UpdateUseSignMiscInfo(UseHashValue, SignMask);
				}
				if (PropagateThroughUses) {
					// For this opcode, we propagate through register USEs to their DEFs.
					if (LocalDef) {
						UseDefAddr = this->GetDefAddrFromUseAddr(UseOp, DefInst->GetAddr(), UseSSANum, LocalUseOp);
						this->PropagateBranchSignedness(UseDefAddr, UseOp, SignMask); // recurse
					}
					else { // recurse through the DEF block
						UseDefAddr = DefBlock->GetDefAddrFromUseAddr(UseOp, DefInst->GetAddr(), UseSSANum, LocalUseOp);
						DefBlock->PropagateBranchSignedness(UseDefAddr, UseOp, SignMask); // recurse
					}
				}
			}
			else if (MDIsDirectStackAccessOpnd(UseOp, this->MyFunc->UsesFramePointer())) {
				// Arithmetic or load using the stack as a source operand. We need to update
				//  our stack frame FG info sign inferences, in case we have some
				//  code that never does signed/unsigned loads from this stack location but uses it
				//  as a source operand for arithmetic or regular loads.
				if (PropagateThroughUses && this->GetFunc()->HasGoodFGStackTable() 
					&& (!this->GetFunc()->IsInOutgoingArgsRegion(UseOp))) {
					struct FineGrainedInfo StackFG;
					bool success = this->GetFunc()->MDGetFGStackLocInfo(DefInst->GetAddr(), UseOp, StackFG);
					assert(success);
					if (0 == (StackFG.SignMiscInfo & FG_MASK_SIGNEDNESS_BITS)) {
						// Stack operand has no signedness info. Time to give it some.
						StackFG.SignMiscInfo |= SignMask;
						success = this->GetFunc()->MDUpdateFGStackLocInfo(DefInst->GetAddr(), UseOp, StackFG);
						assert(success);
						// success => an update occurred
					}
				}
			}
		} // end for all USEs
	} // end if (Phi DEF) else ...
	return;
} // end of SMPBasicBlock::PropagateBranchSignedness()

// propagate DEF signedness to USE if USE has no signedness info.
bool SMPBasicBlock::PropagateDEFSignedness(void) {
	bool changed = false;
	map<int, struct FineGrainedInfo>::iterator UseFGIter, DefFGIter;
	for (UseFGIter = this->LocalUseFGInfoBySSA.begin(); UseFGIter != this->LocalUseFGInfoBySSA.end(); ++UseFGIter) {
		struct FineGrainedInfo UseFG = UseFGIter->second;
		if (0 == (UseFG.SignMiscInfo & FG_MASK_SIGNEDNESS_BITS)) {
			// No signedness info. Propagate any signedness info from DEF.
			int UseHashValue = UseFGIter->first;
			unsigned short DefSignMask = this->GetDefSignMiscInfo(UseHashValue);
			DefSignMask &= FG_MASK_SIGNEDNESS_BITS;
			if (0 != DefSignMask) {
				// DEF has signedness info.
				UseFGIter->second.SignMiscInfo |= DefSignMask;
				changed = true;
			}
		}
	}
	// See if we have DEF signedness info for DEFs with no corresponding USE map entries.
	for (DefFGIter = this->LocalDefFGInfoBySSA.begin(); DefFGIter != this->LocalDefFGInfoBySSA.end(); ++DefFGIter) {
		struct FineGrainedInfo DefFG = DefFGIter->second;
		unsigned short DefSignMask = (DefFG.SignMiscInfo & FG_MASK_SIGNEDNESS_BITS);
		if (0 != DefSignMask) {
			// Has signedness info. See if USE has no entry.
			int DefHashValue = DefFGIter->first;
			UseFGIter = this->LocalUseFGInfoBySSA.find(DefHashValue);
			if (UseFGIter == this->LocalUseFGInfoBySSA.end()) {
				this->UpdateUseSignMiscInfo(DefHashValue, DefSignMask);
				changed = true;
			}
		}
	}
	return changed;
} // end of SMPBasicBlock::PropagateDEFSignedness()

// backtrack from CallAddr and mark unsigned args based on bitset
void SMPBasicBlock::MarkUnsignedArgs(STARS_ea_t CallAddr, unsigned int ArgPosBits, bool IsMallocCall) {
	vector<SMPInstr *>::reverse_iterator RevInstIter;
	unsigned int UnsignedArgCount = CountBitsSet(ArgPosBits);
	unsigned int UnsignedArgsFound = 0;
	bool FoundCallAddr = false;

	for (RevInstIter = this->GetRevInstBegin(); RevInstIter != this->GetRevInstEnd(); ++RevInstIter) {
		SMPInstr *CurrInst = (*RevInstIter);
		STARS_ea_t InstAddr = CurrInst->GetAddr();
		if (InstAddr == CallAddr) {
			FoundCallAddr = true;
		}
		else if (FoundCallAddr) {
			SMPitype FlowType = CurrInst->GetDataFlowType();
			if ((FlowType == CALL) || (FlowType == INDIR_CALL)) {
				// We have reached a call before the one we care about. Done.
				break;
			}
			size_t ArgumentNumber;
			if (CurrInst->MDIsArgumentPass(ArgumentNumber)) {
				if (0 != (ArgPosBits & (1 << ArgumentNumber))) {
					if (IsMallocCall) {
						assert(1 == UnsignedArgCount);
						// Associate CallAddr with argument pass addr for later analyses.
						pair<STARS_ea_t, STARS_ea_t> CallArgAddrPair(CallAddr, InstAddr);
						this->GetFunc()->AddMallocCallWithInArg(CallArgAddrPair);
					}
					// ArgumentNumber matches a bit set in ArgPosBits.
					CurrInst->SetUnsignedArg();
					++UnsignedArgsFound;
					if (UnsignedArgsFound >= UnsignedArgCount) {
						break; // found them all
					}
				}
			}
		}
	}

	return;
} // end of SMPBasicBlock::MarkUnsignedArgs()

// backtrack from CallAddr and mark POINTER args based on bitset
void SMPBasicBlock::MarkPointerArgs(STARS_ea_t CallAddr, unsigned int ArgPosBits, bool IsBufferLibFunc) {
	vector<SMPInstr *>::reverse_iterator RevInstIter;
	unsigned int PointerArgCount = CountBitsSet(ArgPosBits);
	unsigned int PointerArgsFound = 0;
	bool FoundCallAddr = false;

	for (RevInstIter = this->GetRevInstBegin(); RevInstIter != this->GetRevInstEnd(); ++RevInstIter) {
		SMPInstr *CurrInst = (*RevInstIter);
		STARS_ea_t InstAddr = CurrInst->GetAddr();
		if (InstAddr == CallAddr) {
			FoundCallAddr = true;
		}
		else if (FoundCallAddr) {
			SMPitype FlowType = CurrInst->GetDataFlowType();
			if ((FlowType == CALL) || (FlowType == INDIR_CALL)) {
				// We have reached a call before the one we care about. Done.
				break;
			}
			size_t ArgumentNumber;
			if (CurrInst->MDIsArgumentPass(ArgumentNumber)) {
				if (0 != (ArgPosBits & (1 << ArgumentNumber))) {
					// ArgumentNumber matches a bit set in ArgPosBits.
					set<DefOrUse, LessDefUse>::iterator DefIter = CurrInst->GetFirstNonFlagsDef();
					assert(DefIter != CurrInst->GetLastDef());
					STARSOpndTypePtr DefOp = DefIter->GetOp();
					DefIter = CurrInst->SetDefType(DefOp, POINTER);
					assert(DefIter != CurrInst->GetLastDef());
					bool IsMemOp = IsMemOperand(DefOp);
					if (IsBufferLibFunc) {
						// Associate CallAddr with argument pass addr for later analyses.
						pair<STARS_ea_t, STARS_ea_t> CallArgAddrPair(CallAddr, InstAddr);
						this->GetFunc()->AddBufferCallWithInArg(CallArgAddrPair);
					}
					if (this->IsLocalName(DefOp)) {
						this->PropagateLocalDefType(DefOp, POINTER, CurrInst->GetAddr(), DefIter->GetSSANum(), IsMemOp, false);
					}
					else {
						this->GetFunc()->ResetProcessedBlocks();
						this->PropagateGlobalDefType(DefOp, POINTER, DefIter->GetSSANum(), IsMemOp, false, true);
					}
					++PointerArgsFound;
					if (PointerArgsFound >= PointerArgCount) {
						break; // found them all
					}
				}
			}
		}
	}

	return;
} // end of SMPBasicBlock::MarkPointerArgs()

// Find the stack target of a call to memset() and the size in bytes of the memset() region.
//  If the memset() target is not on the stack, return false. If the 
//  size of the memset() region is not a constant, we also return false.
bool SMPBasicBlock::AnalyzeMemSet(STARS_ea_t MemSetAddr, STARSOpndTypePtr &MemSetTarget, size_t &MemSetSize, int &StackOffset, bool &FPRelativeTarget) {
	set<DefOrUse, LessDefUse>::iterator UseIter;
	STARSOpndTypePtr DefOp = nullptr, UseOp = nullptr, UltimateSourceOp = nullptr;
	STARS_ea_t InstAddr;
	int UseSSANum;
	bool FoundTarget = false, FoundSize = false;
	bool UseFP = this->GetFunc()->UsesFramePointer();
	bool DummyFPRelative = false; // don't care about this for the size argument to memset(), which is o_imm anyway
	vector<SMPInstr *>::reverse_iterator InstRevIter;
	STARS_sval_t CurrentDelta;

	STARS_ea_t FirstAddr = this->GetFirstAddr();
	MemSetTarget = nullptr;
	MemSetSize = 0;

	// The MemSetTarget will be moved into the first memset() arg and the size will be moved
	//  into the third arg. The value being copied into each byte will be moved into
	//  the second arg, which we do not care about. We will move backwards through the block until
	//  we find the MemSetAddr, then we will trace the def-use SSA chain back to its 
	//  source for each argument to memset() that we care about.
	InstRevIter = this->GetRevInstBegin(); 
	SMPInstr *CurrInst = (*InstRevIter);
	InstAddr = CurrInst->GetAddr();
	while ((InstAddr != MemSetAddr) && (InstRevIter != this->GetRevInstEnd())) {
		++InstRevIter; // move backwards through block by incrementing the reverse iterator.
		CurrInst = (*InstRevIter);
		InstAddr = CurrInst->GetAddr();
	}
	if (InstAddr != MemSetAddr) {
		SMP_msg("ERROR: Memset not found at MemSetAddr %lx\n", (unsigned long) MemSetAddr);
		return false;
	}

	// Keep iterating backwards from MemSetAddr, searching for DEFs of the 1st and 3rd args.
	++InstRevIter; // move backwards through block by incrementing the reverse iterator.
	size_t ArgumentNumber;
	while (InstRevIter != this->GetRevInstEnd()) {
		CurrInst = (*InstRevIter);
		InstAddr =  CurrInst->GetAddr();
		// Limit to simple Move arg passes for now.
		if (CurrInst->MDIsArgumentPass(ArgumentNumber)) {
			if (0 == ArgumentNumber) {
				if (CurrInst->MDIsLoadEffectiveAddressInstr()) {
					UseOp = CurrInst->GetLeaMemUseOp();
					if (MDIsDirectStackAccessOpnd(UseOp, UseFP)) {
						int BaseReg, IndexReg;
						uint16_t ScaleFactor;
						STARS_ea_t offset;
						MDExtractAddressFields(UseOp, BaseReg, IndexReg, ScaleFactor, offset);
						if ((STARS_x86_R_none != IndexReg) || (0 != ScaleFactor)) {
							break; // cannot make use of complex stack loc expression
						}
						else {
							MemSetTarget = UseOp;
							StackOffset = (int) offset; // NOTE: this is a normalized offset
							FoundTarget =  true; // success on argument #0
							FPRelativeTarget = (BaseReg == MD_FRAME_POINTER_REG);
							if (FoundSize) break;  // nothing left to find
							++InstRevIter;
							continue; // done with argument #0 inst
						}
					}
					else {
						break; // only do simple lea case
					}
				}
				else {
					UseOp = CloneIfNecessary(CurrInst->GetMoveSource(), UseFP);
				}
				assert(nullptr != UseOp);
				if (UseOp->IsRegOp()) {
					CanonicalizeOpnd(UseOp);
				}
				if (UseOp->IsVoidOp()) {
					SMP_msg("ERROR: No move source at %lx within AnalyzeMemSet().\n", (unsigned long) InstAddr);
					break;
				}
				else {
					UseIter = CurrInst->FindUse(UseOp);
					assert(UseIter != CurrInst->GetLastUse());
					UseSSANum = UseIter->GetSSANum();
					if (CurrInst->TraceUltimateMoveSource(UseOp, UseSSANum, UltimateSourceOp, FPRelativeTarget)) {
						if (MDIsDirectStackAccessOpnd(UltimateSourceOp, UseFP)) {
							// memset() has a stack location as its target. Extract info from
							//  UltimateSourceOp.
							int BaseReg, IndexReg;
							uint16_t ScaleFactor;
							STARS_ea_t offset;
							MDExtractAddressFields(UltimateSourceOp, BaseReg, IndexReg, ScaleFactor, offset);
							if ((STARS_x86_R_none != IndexReg) || (0 != ScaleFactor)) {
								break; // cannot make use of complex stack loc expression
							}
							else {
								MemSetTarget = UltimateSourceOp;
								StackOffset = (int) offset; // NOTE: this is a normalized offset
								FoundTarget =  true; // success on argument #1
								if (FoundSize) break;  // nothing left to find
							}
						}
						else {
							// Found ultimate source of 1st argument, but not a stack operand.
							break; // return false
						}
					}
					else { // could not find ultimate source of 1st argument
						break;  // time to return false
					}
				}
			}
			else if (2 == ArgumentNumber) {
				UseOp = CloneIfNecessary(CurrInst->GetMoveSource(), UseFP);
				assert(nullptr != UseOp);
				if (UseOp->IsRegOp()) {
					CanonicalizeOpnd(UseOp);
				}
				if (UseOp->IsVoidOp()) {
					// NOTE: We could drop the requirement for a move inst and trace arg #2
					//  back through arithmetic if SCCP can find constants.
					SMP_msg("WARNING: No move size at %llx within AnalyzeMemSet().\n",
							(unsigned long long) InstAddr);
					break;
				}
				else {
					UseIter = CurrInst->FindUse(UseOp);
					assert(UseIter != CurrInst->GetLastUse());
					UseSSANum = UseIter->GetSSANum();
					if (CurrInst->TraceUltimateMoveSource(UseOp, UseSSANum, UltimateSourceOp, DummyFPRelative)) {
						if (! UltimateSourceOp->IsImmedOp()) {
							break; // return false if cannot find an immediate value for 3rd argument
						}
						else {
							uintptr_t ImmVal = UltimateSourceOp->GetImmedValue();
							MemSetSize = (size_t) ImmVal;
							FoundSize = true;
							if (FoundTarget) break;  // nothing left to find
						}
					}
				}
			} // end if ((0 == ArgumentNumber) || (2 == ArgumentNumber))
		} // end if argument pass

		++InstRevIter; // move backwards through block by incrementing the reverse iterator.
	} // end while we have not reached the block Instrs.rend()

	return (FoundTarget && FoundSize);
} // end of SMPBasicBlock::AnalyzeMemset()

// For the newly defined type DefType for DefOp at instruction DefAddr, propagate the
//  type to all USEs in the local SSA chain for the DEF. If any USE types change,
// return true.
bool SMPBasicBlock::PropagateLocalDefType(const STARSOpndTypePtr &DefOp, SMPOperandType DefType, STARS_ea_t DefAddr, int SSANum, bool IsMemOp, bool PointerOverride) {
	bool changed = false;
	bool SafeFunc = (this->MyFunc) && this->MyFunc->IsSafe();
	bool ChangeNumericToPointer = (PointerOverride && (IsDataPtr(DefType) || IsCodePtr(DefType)));
#if SMP_DEBUG_OPTIMIZATIONS
	bool DebugFlag = (0 == strcmp("memset", this->MyFunc->GetFuncName()));

	if (DebugFlag) {
		SMP_msg("PropagateLocalDefType for DefType %d at %lx ",
			DefType, (unsigned long) DefAddr);
		PrintOperand(DefOp);
		SMP_msg("\n");
	}
#endif

#if STARS_PROPAGATE_STACKMEMTYPE_UNTIL_INDIRECT
	bool UseFP = this->GetFunc()->UsesFramePointer();
	bool StackMemOp = IsMemOp && MDIsDirectStackAccessOpnd(DefOp, UseFP);
	if (IsMemOp && (!StackMemOp))
		return false;
#else
	// When we can traverse all paths of the global DEF-USE chains for this
	//  DefOp, we can be sure there is no aliasing. Until then, we must
	//  be more conservative on global MemOp propagation than for locals.
	if (IsMemOp && (!SafeFunc))
		return false;
#endif

	set<DefOrUse, LessDefUse>::iterator CurrUse;
	vector<SMPInstr *>::iterator InstIter = this->GetInstIterFromAddr(DefAddr);

	// Go through the instructions in the block and find the instructions that have USEs of DefOp/SSANum.
	//  If any USE in the chain has a type other than DefType, change it to DefType and set changed to true.
	++InstIter;
	while (InstIter != this->GetLastInst()) {
		SMPInstr *CurrInst = (*InstIter);
		++InstIter;
		STARS_ea_t CurrAddr = CurrInst->GetAddr();
#if SMP_VERBOSE_ALIAS_DETECTION
		if (IsMemOp && (CurrInst->HasIndirectMemoryWrite())) {
			SMP_msg("CAUTION: Indirect memory write in MemOp chain.\n");
			SMP_msg("CAUTION: at %lx ChainStart: %lx ChainEnd: %lx \n",
				(unsigned long) CurrAddr, (unsigned long) StartAddr, (unsigned long) LastAddr);
#if 0
			if (!SafeFunc) {
				SMP_msg("Local type propagation terminated in unsafe func %s\n",
					this->MyFunc->GetFuncName());
				break;
			}
#endif
		}
#endif
		CurrUse = CurrInst->FindUse(DefOp);
		if (CurrUse != CurrInst->GetLastUse()) { // found a USE in the chain
			uint16_t InstOpcode = CurrInst->GetIDAOpcode();
			SMPOperandType UseType = CurrUse->GetType();
			if ((UseType == UNINIT) || (ChangeNumericToPointer && IsEqType(UseType, NUMERIC))) {
				assert(SSANum == CurrUse->GetSSANum());
				CurrUse = CurrInst->SetUseType(DefOp, DefType);
				assert(CurrUse != CurrInst->GetLastUse()); // found USE
				changed = true;
			}
			else if (DefType != UseType) {
				// See lengthy explanation of propagation rules in PropagateGlobalDefType().
				if (IsRefinedDataPtr(DefType) && (IsEqType(POINTER, UseType))) {
#if SMP_VERBOSE_POINTER_REFINEMENT
					SMP_msg("Refining POINTER to %d at %lx %s for USE:", DefType,
						(unsigned long) CurrInst->GetAddr(), CurrInst->GetDisasm());
					CurrUse->Dump();
					SMP_msg("\n");
#endif
					CurrUse = CurrInst->SetUseType(DefOp, DefType);
					assert(CurrUse != CurrInst->GetLastUse()); // found USE
					changed = true;
				}
				else if (IsDataPtr(DefType) && IsNumeric(UseType)
						&& ((InstOpcode == STARS_NN_adc) || (InstOpcode == STARS_NN_sbb))) {
					SMP_msg("ADC/SBB: Refining NUMERIC to %d at %lx %s for USE:", DefType,
						(unsigned long) CurrInst->GetAddr(), CurrInst->GetDisasm());
					CurrUse->Dump();
					SMP_msg("\n");
					CurrUse = CurrInst->SetUseType(DefOp, DefType);
					assert(CurrUse != CurrInst->GetLastUse()); // found USE
					changed = true;
				}
				else {
					;
#if SMP_DEBUG_OPTIMIZATIONS
					SMP_msg("WARNING: Not overriding type %d to %d at %lx %s for USE:", UseType,
						DefType, (unsigned long) CurrInst->GetAddr(), CurrInst->GetDisasm());
					CurrUse->Dump();
					SMP_msg("\n");
#endif
				}
			}
		}
		if (CurrInst->FindDef(DefOp) != CurrInst->GetLastDef()) {
			// Instruction redefines DefOp. Terminate.
			break;
		}
#if STARS_PROPAGATE_STACKMEMTYPE_UNTIL_INDIRECT
		if (StackMemOp) {
			// Stop propagating when we see an indirect memory write.
			if (CurrInst->HasIndirectMemoryWrite()) {
				return changed;
			}
		}
#endif
	} // end for all instructions

	return changed;
} // end of SMPBasicBlock::PropagateLocalDefType()

// Propagate the DefType of the DEF in DefOp with SSANum to all USEs. DefOp is a global
//  name, so the propagation will have to recurse into successor blocks, check their
//  phi functions for occurrences of DefOp, etc.
// Return true if any USEs have their types changed.
bool SMPBasicBlock::PropagateGlobalDefType(const STARSOpndTypePtr &DefOp, SMPOperandType DefType, int SSANum, bool IsMemOp, bool PointerOverride, bool FirstBlock) {
	bool changed = false;
	bool FoundPhiDef = false;
	bool SafeFunc = (this->MyFunc) && this->MyFunc->IsSafe();
	bool ChangeNumericToPointer = (PointerOverride && (IsDataPtr(DefType) || IsCodePtr(DefType)));
#if SMP_DEBUG_OPTIMIZATIONS
	bool DebugFlag = (0 == strcmp("__strtod_internal", this->MyFunc->GetFuncName()));
#endif
	vector<SMPInstr *>::iterator InstIter;
	set<DefOrUse, LessDefUse>::iterator CurrUse;
	set<DefOrUse, LessDefUse>::iterator CurrDef;
	SMPOperandType UseType;

	// Use the Processed flag to avoid infinite recursion.
	if (this->IsProcessed())
		return false;
	else
		this->SetProcessed(true);

	bool UseFP = this->GetFunc()->UsesFramePointer();
	if ((STACKPTR == DefType) && (DefOp->IsRegOp()) && (MDIsStackPtrReg(DefOp->GetReg(), UseFP))) {
		// We should not need to propagate STACKPTR type for the stack and frame pointer registers. Save time.
		return false;
	}

#if SMP_DEBUG_OPTIMIZATIONS
	if (DebugFlag) {
		SMP_msg("PropagateGlobalDefType: DefType %d SSANum %d ", DefType, SSANum);
		PrintOperand(DefOp);
		SMP_msg("\n");
	}
#endif

#if STARS_PROPAGATE_STACKMEMTYPE_UNTIL_INDIRECT
	bool StackMemOp = IsMemOp && MDIsDirectStackAccessOpnd(DefOp, UseFP);
	if (IsMemOp && (!StackMemOp))
		return false;
#else
	// When we can traverse all paths of the global DEF-USE chains for this
	//  DefOp, we can be sure there is no aliasing. Until then, we must
	//  be more conservative on global MemOp propagation than for locals.
	if (IsMemOp && (!SafeFunc))
		return false;
#endif

	// This might be a recursive invocation, so the DefOp might be in the Phi functions.
	//  NOTE: DefOp/SSANum could also be in the Phi functions as a USE even though it is
	//  DEFed later in the block:
	//  RAX44 := Phi(RAX43, RAX45);
	//  RAX45 := RAX44 + 1;
	//  [use RAX45 later in the block]
	//  So, we need to set DefType into the RAX45 Phi USE, and also into later USEs of RAX45
	//   in moves and arithmetic in the block, and then recurse into successors.
	set<SMPPhiFunction, LessPhi>::iterator CurrPhi = this->FindPhi(DefOp);
	if (CurrPhi != this->GetLastPhi()) { // Found a corresponding phi function
#if SMP_DEBUG_OPTIMIZATIONS
		if (DebugFlag) SMP_msg("Found Phi for DefOp\n");
#endif
		// If SSANum matches the DEF in the Phi function, then we are being
		//  called by InferPhiDefType() and we should proceed to process
		//  the instructions in the block.
		if (SSANum != CurrPhi->GetDefSSANum()) { // not DEF, so find USE
#if SMP_DEBUG_OPTIMIZATIONS
			if (DebugFlag) SMP_msg("Phi was a USE\n");
#endif
			// Now, find the SSANum entry in the phi arguments (USEs)
			bool FoundUse = false;
			size_t PhiIndex;
			for (PhiIndex = 0; PhiIndex < CurrPhi->GetPhiListSize(); ++PhiIndex) {
				if (SSANum == CurrPhi->GetUseSSANum(PhiIndex))  {
					FoundUse = true;
#if SMP_DEBUG_OPTIMIZATIONS
					if (DebugFlag) SMP_msg("Found the phi USE at index %zu\n", PhiIndex);
#endif
					SMPOperandType CurrType = CurrPhi->GetUseType(PhiIndex);
					if ((UNINIT == CurrType) || (ChangeNumericToPointer && IsEqType(CurrType, NUMERIC))) {
						changed = this->SetPhiUseType(DefOp, PhiIndex, DefType);
						assert(changed);
#if SMP_DEBUG_OPTIMIZATIONS
						if (DebugFlag) SMP_msg("Changed the phi USE type\n");
#endif
					}
#if 1
					break; // **!!** COULD BE MULTIPLE USES HERE
#endif
				}
			} // end for all Phi USEs
			if (!FoundUse) {
				// Not actually a problem. E.g. if we find an inst:  add eax,4 that
				//  uses eax SSA #7 and DEFs eax SSA #8, we will call this function
				//  to propagate the type of SSA #8. There could be a phi function
				//  at the top of this block that DEFs eax SSA #7 from earlier USEs
				//  of eax. We will not always find eax SSA #8 in that phi function, so we
				//  pass down to the code below to process all instructions.
				;
#if 0
				SMP_msg("ERROR: Failed to find Phi use of SSA %d in block %d for: ", SSANum,
					this->GetNumber());
				PrintOperand(DefOp);
				SMP_msg("\n");
#endif
			}
			else if (!FirstBlock) // see lengthy RAX45 example in comment above; continue on first block, not recursion.
				return changed;  // end the recursion; phi function does not permit DefOp
								//   with SSANum to be referenced on this path again
		} // end if DefOp and SSANum are not the phi function DEF
		else {
			FoundPhiDef = true;
		}
	} // end if we found a phi function for DefOp

	// If not in the Phi functions, it could be in the instructions for the block.
	bool FoundUse = false;
	for (InstIter = this->GetFirstInst(); InstIter != this->GetLastInst(); ++InstIter) {
		SMPInstr *CurrInst = (*InstIter);
		STARS_ea_t InstAddr = CurrInst->GetAddr();
#if 0
		if (IsMemOp && (CurrInst->HasIndirectMemoryWrite())) {
			if (!SafeFunc) {
				SMP_msg("Global type propagation terminated in unsafe func %s\n",
					this->MyFunc->GetFuncName());
				return changed;
			}
		}
#endif
		CurrUse = CurrInst->FindUse(DefOp);
		if ((CurrUse == CurrInst->GetLastUse()) || (SSANum != CurrUse->GetSSANum())) {
			if (FoundUse) {
				// If we have already gone past the DEF and at least one use,
				//  another DEF will indicate that the original DEF-USE chain is
				//  finished. In that case, we can save a lot of time by not processing
				//  any more instructions and recursing into successor blocks, because
				//  the original DEF is now dead.
				CurrDef = CurrInst->FindDef(DefOp);
				if ((CurrDef != CurrInst->GetLastDef()) && (SSANum != CurrDef->GetSSANum())) {
					return changed;
				}
				else {
					continue;
				}
			}
			else {
				continue;
			}
		}
		// If we reach this point, we have found a use with the same SSA number.
		FoundUse = true;
		UseType = CurrUse->GetType();
		uint16_t InstOpcode = CurrInst->GetIDAOpcode();
#if SMP_DEBUG_OPTIMIZATIONS
		if (DebugFlag) {
			SMP_msg("Found USE with same SSANum at %lx : %s\n", 
				(unsigned long) CurrInst->GetAddr(), CurrInst->GetDisasm());
		}
#endif
		if ((UNINIT == UseType) || (ChangeNumericToPointer && IsEqType(UseType, NUMERIC))) {
			CurrUse = CurrInst->SetUseType(DefOp, DefType);
			assert(CurrUse != CurrInst->GetLastUse()); // found USE
			changed = true;
		}
		else if (DefType != UseType) {
			// Weird. USE already has a type, but it is not the DEF type.
			//  If we are refining a POINTER to a particular pointer type, fine.
			//  We don't want to change NUMERIC to something else, however,
			//  because it is possible to perform inherently NUMERIC operations
			//  on pointer types, e.g. def a pointer, use it as pointer, then
			//  perform hash function computations on it. Those computations
			//  (shift, xor, etc.) should truly remain NUMERIC.
			//
			// EXCEPTION: We need to propagate pointer types to USEs within
			//  subtract-with-borrow and add-with-carry instructions. This is
			//  because of phase ordering issues. E.g. sbb ecx,3 could have
			//  the DEF type inferred to be NUMERIC based on its later uses,
			//  but we still need to know if the USE of ecx was POINTER. This
			//  is a special case in which the sbb or adc instruction changes
			//  the type of the register. We need to observe this type change
			//  when we emit annotations, so that we emit an annotation to
			//  force the DEF metadata to NUMERIC rather than an annotation
			//  that squashes the instruction and leaves ecx with type POINTER.
			//  SMPInstr::InferOperatorType() will propagate NUMERIC from
			//  the SMP_SUBTRACT_BORROW operator to register ecx if ecx still
			//  had type UNINIT. We need to reset the USE of ecx to POINTER
			//  if we later discover that ecx came into the instruction with
			//  type POINTER.
			if (IsRefinedDataPtr(DefType) && (IsEqType(POINTER, UseType))) {
#if SMP_VERBOSE_POINTER_REFINEMENT
				SMP_msg("Refining POINTER to %d at %lx %s for USE:", DefType,
					(unsigned long) CurrInst->GetAddr(), CurrInst->GetDisasm());
				CurrUse->Dump();
				SMP_msg("\n");
#endif
				CurrUse = CurrInst->SetUseType(DefOp, DefType);
				assert(CurrUse != CurrInst->GetLastUse()); // found USE
				changed = true;
			}
			else if (IsDataPtr(DefType) && IsNumeric(UseType)
					&& ((InstOpcode == STARS_NN_adc) || (InstOpcode == STARS_NN_sbb))) {
				SMP_msg("ADC/SBB: Refining NUMERIC to %d at %lx %s for USE:", DefType,
					(unsigned long) CurrInst->GetAddr(), CurrInst->GetDisasm());
				CurrUse->Dump();
				SMP_msg("\n");
				CurrUse = CurrInst->SetUseType(DefOp, DefType);
				assert(CurrUse != CurrInst->GetLastUse()); // found USE
				changed = true;
			}
			else {
				;
#if SMP_DEBUG_OPTIMIZATIONS
				SMP_msg("WARNING: Not overriding type %d to %d at %x %s for USE:", UseType,
					DefType, CurrInst->GetAddr(), CurrInst->GetDisasm());
				CurrUse->Dump();
				SMP_msg("\n");
#endif
			}
		}
#if STARS_PROPAGATE_STACKMEMTYPE_UNTIL_INDIRECT
		if (StackMemOp) {
			// Stop propagating when we see an indirect memory write.
			if (CurrInst->HasIndirectMemoryWrite()) {
				return changed;
			}
		}
#endif
	} // end for all instructions in the block

	// We recurse into all successor blocks looking for more USEs. We know that there
	//  cannot be more USEs if the global name DefOp is not LiveOut from this block, so
	//  we can just return in that case.
	if (this->LiveOutSet.end() == LiveOutSet.find(DefOp))  // not LiveOut
		return changed;

	// Only recurse into successors that have DefOp in their LiveIn set. The others
	//  will not have Phi functions for DefOp, because we constructed fully pruned SSA
	//  form, nor will they have USEs of this SSANum.
	list<SMPBasicBlock *>::iterator SuccIter;
	for (SuccIter = this->GetFirstSucc(); SuccIter != this->GetLastSucc(); ++SuccIter) {
		if ((*SuccIter)->IsLiveIn(DefOp)) {
			changed |= (*SuccIter)->PropagateGlobalDefType(DefOp, DefType, SSANum, IsMemOp, PointerOverride, false);
		}
	}

	return changed;
} // end of SMPBasicBlock::PropagateGlobalDefType()

// Propagate the metadata Status for UseOp/SSANum to its local DEF.
// Return true if successful.
bool SMPBasicBlock::PropagateLocalMetadata(const STARSOpndTypePtr &UseOp, SMPMetadataType Status, int SSANum, STARS_ea_t UseAddr) {
	bool changed = false;
	STARSOpndSetIter NameIter;
	NameIter = this->FindLocalName(UseOp);
	if (this->GetLastLocalName() == NameIter) {
		return false;
	}
	assert(0 <= SSANum);

	STARS_ea_t DefAddr = this->GetDefAddrFromUseAddr(UseOp, UseAddr, SSANum, true);
	if ((DefAddr == STARS_BADADDR) || STARS_IsLiveInPseudoID(DefAddr)) {
		return false;
	}

	// Find the instruction at DefAddr
	vector<SMPInstr *>::iterator InstIter = this->GetInstIterFromAddr(DefAddr);
	SMPInstr *CurrInst = (*InstIter);
	set<DefOrUse, LessDefUse>::iterator CurrDef, CurrUse, NextUse;

	CurrDef = CurrInst->FindDef(UseOp);
	assert(CurrDef != CurrInst->GetLastDef());
	assert(SSANum == CurrDef->GetSSANum());
	// Set the new metadata status.
	if (Status != CurrDef->GetMetadataStatus()) {
		CurrDef = CurrInst->SetDefMetadata(UseOp, Status);
		changed = (CurrDef != CurrInst->GetLastDef());

		// If source operand was memory, we have two cases.
		//  (1) The instruction could be a load, in which
		//  case we should simply terminate the
		//  propagation, because the prior DEF of a memory
		//  location is always considered live metadata
		//  already, and we do not want to propagate liveness
		//  to the address regs in the USE list.
		//  EXCEPTION: For safe funcs, we propagate liveness
		//   for stack locations.
		//  (2) We could have an arithmetic operation such
		//  as reg := reg arithop memsrc. In this case, we
		//  still do not want to propagate through the memsrc,
		//  but the register is both DEF and USE and we need
		//  to propagate through the register.
		if (CurrInst->HasSourceMemoryOperand()) {
			if (this->MyFunc->IsSafe()) {
				bool UseFP = this->MyFunc->UsesFramePointer();
				STARSOpndTypePtr MemSrcOp = CurrInst->MDGetMemUseOp();
				assert(! MemSrcOp->IsVoidOp());
				if (MDIsStackAccessOpnd(MemSrcOp, UseFP)) {
					// We have a SafeFunc stack access. This is
					//  the EXCEPTION case where we want to
					//  propagate metadata liveness for a memory
					//  location.
					CurrUse = CurrInst->FindUse(MemSrcOp);
					assert(CurrUse != CurrInst->GetLastUse());
					if (this->MyFunc->IsGlobalName(MemSrcOp)) {
						changed |= this->MyFunc->PropagateGlobalMetadata(MemSrcOp,
							Status, CurrUse->GetSSANum(), DefAddr);
					}
					else {
						changed |= this->PropagateLocalMetadata(MemSrcOp,
							Status, CurrUse->GetSSANum(), DefAddr);
					}
				} // end if stack access operand
			} // end if SafeFunc
			if (3 == CurrInst->GetOptType()) { // move inst
				return changed; // address regs are not live metadata
			}
			else if ((5 == CurrInst->GetOptType())
				|| (STARS_NN_and == CurrInst->GetIDAOpcode())
				|| (STARS_NN_or == CurrInst->GetIDAOpcode())
				|| (STARS_NN_xor == CurrInst->GetIDAOpcode())) {
				// add, subtract, and, or with memsrc
				// Find the DEF reg in the USE list.
				CurrUse = CurrInst->FindUse(UseOp);
				assert(CurrUse != CurrInst->GetLastUse());
				changed |= this->PropagateLocalMetadata(UseOp,
					Status, CurrUse->GetSSANum(), DefAddr);
				return changed;
			}
		} // end if memory source

		// Now, propagate the metadata status to all the
		//  non-memory, non-flags-reg, non-special-reg 
		//  (i.e. regular registers) USEs.
		CurrUse = CurrInst->GetFirstUse();
		while (CurrUse != CurrInst->GetLastUse()) {
			NextUse = CurrUse;
			++NextUse;
			STARSOpndTypePtr UseOp2 = CurrUse->GetOp();
			int UseSSANum = CurrUse->GetSSANum();
			// NOTE: **!!** To be less conservative, we
			//  should propagate more for exchange category
			//  instructions.
			if ((UseOp2->IsRegOp()) && (!MDIsStackOrFramePointerReg(UseOp2, this->MyFunc->UsesFramePointer()))
				&& (!UseOp2->MatchesReg(X86_FLAGS_REG))) {

				if (this->MyFunc->IsGlobalName(UseOp2)) {
					changed |= this->MyFunc->PropagateGlobalMetadata(UseOp2, Status, UseSSANum, DefAddr);
				}
				else {
					changed |= this->PropagateLocalMetadata(UseOp2, Status, UseSSANum, DefAddr);
				}
			}
			CurrUse = NextUse;
		} // end while all USEs
	} // end if Status != CurrDef status

	return changed;
} // end of SMPBasicBlock::PropagateLocalMetadata()

// For the UNINIT type DEF DefOp at instruction DefAddr, see if all its USEs have
//  a single type. If so, set the DEF to that type and return true, else return false.
bool SMPBasicBlock::InferLocalDefType(const STARSOpndTypePtr &DefOp, unsigned int LocIndex, STARS_ea_t DefAddr) {
	bool changed = false;
#if SMP_DEBUG_OPTIMIZATIONS
	bool DebugFlag = (0 == strcmp("weightadj", this->MyFunc->GetFuncName()));

	if (DebugFlag) {
		SMP_msg("InferLocalDefType at %lx ", (unsigned long) DefAddr);
		PrintOperand(DefOp);
		SMP_msg("\n");
	}
#endif

	vector<SMPInstr *>::iterator InstIter;
	vector<SMPInstr *>::iterator DefInstIter = this->GetInstIterFromAddr(DefAddr);
	set<DefOrUse, LessDefUse>::iterator CurrDef;
	SMPInstr *DefInst = (*DefInstIter);
	CurrDef = DefInst->FindDef(DefOp);
	if (CurrDef == DefInst->GetLastDef()) {
		SMP_msg("ERROR: DEF at %lx not found in DEF list by InferLocalDefType for DefOp: ", (unsigned long) DefAddr);
		PrintOperand(DefOp);
		SMP_msg("\n");
		return false;
	}
	int SSANum = CurrDef->GetSSANum();
	assert(0 <= SSANum);
	set<DefOrUse, LessDefUse>::iterator CurrUse;
	// Go through the instructions in the block and find the instructions between DefAddr
	//  and the last inst that have USEs of DefOp/SSANum. If all USEs in the chain have
	//  a single type (other than UNINIT), change the DEF type to match the USE type
	//  and set changed to true.
	bool FirstUseSeen = false;
	bool FoundUninit = false;
	bool FoundNumeric = false;
	bool FoundCodePtr = false;
	bool FoundPointer = false;
	bool FoundUnknown = false;
	SMPOperandType DefType = CurrDef->GetType();
	SMPOperandType PtrType = UNINIT;
	SMPOperandType UseType = UNINIT;
	InstIter = DefInstIter;
	++InstIter;
	for ( ; InstIter != this->GetLastInst(); ++InstIter) {
		SMPInstr *CurrInst = (*InstIter);
		STARS_ea_t CurrAddr = CurrInst->GetAddr();
		CurrUse = CurrInst->FindUse(DefOp);
		if (CurrUse != CurrInst->GetLastUse()) { // found a USE in the chain
			if (!FirstUseSeen)
				FirstUseSeen = true;
			UseType = CurrUse->GetType();
			if (UNINIT == UseType)
				FoundUninit = true;
			else if (IsCodePtr(UseType))
				FoundCodePtr = true;   // CODEPTR takes precedence over NUMERIC
			else if (IsNumeric(UseType))
				FoundNumeric = true;
			else if (IsUnknown(UseType))
				FoundUnknown = true;
			else if (IsDataPtr(UseType)) {
				if (FoundPointer) {
					if (IsNotEqType(PtrType, UseType)) {
						;
#if SMP_DEBUG_OPTIMIZATIONS
						SMP_msg("WARNING: Differing ptr types in local chain:");
						SMP_msg(" Prev: %d Current: %d %s\n", PtrType, UseType,
							CurrInst->GetDisasm());
#endif
					}
					PtrType = POINTER;
				}
				else {
					FoundPointer = true;
					PtrType = UseType;
				}
			}
		}
		if (CurrInst->FindDef(DefOp) != CurrInst->GetLastDef()) {
			// Redefined. Chain is finished.
			break;
		}
	} // end for all instructions

	if (FirstUseSeen) {
		// See if we have a consistent type.
		// If we see any definite POINTER uses, we must set the DEF
		//  to type POINTER or a refinement of it.
		if (FoundPointer)
			UseType = PtrType;
		else if (FoundCodePtr && !FoundUnknown)
			UseType = CODEPTR;
		else if (FoundNumeric && !FoundUninit && !FoundUnknown)
			UseType = NUMERIC;
		else
			return false; // no POINTER, but no consistent type

		if (!IsEqType(DefType, UseType)) {
#if SMP_DEBUG_OPTIMIZATIONS
			if (DebugFlag) SMP_msg("Inferring local DEF of type %d\n", UseType);
#endif
			CurrDef = DefInst->SetDefType(DefOp, UseType);
			changed = true;
			if ((FoundPointer || FoundCodePtr) && FoundUninit) {
				// We will propagate PtrType to the UNINIT elements of
				//  the DEF-USE chain.
				bool IsMemOp = DefOp->IsMemOp();
				changed |= this->PropagateLocalDefType(DefOp, UseType, DefAddr,
					CurrDef->GetSSANum(), IsMemOp);
			}
		}
	}

	return changed;
} // end of SMPBasicBlock::InferLocalDefType()

void SMPBasicBlock::InferGlobalDefType(const STARSOpndTypePtr &DefOp, int SSANum, STARS_ea_t DefAddr, bool &FoundNumeric, bool &FoundCodePtr, bool &FoundPointer,
	bool &FoundUnknown, bool &FoundUninit, SMPOperandType &PtrType) {

	// Avoid infinite recursion
	if (this->IsProcessed())
		return;
	else
		this->SetProcessed(true);

	bool DefEscapes = true;
	vector<SMPInstr *>::iterator InstIter;

	assert(0 <= SSANum);
	set<DefOrUse, LessDefUse>::iterator CurrUse, CurrDef;
	// Go through all instructions in the block and find the instructions
	//  that have USEs of DefOp with SSANum. Record the results in the
	//  boolean in-out arguments.
	SMPOperandType UseType = UNINIT;
	
	if (STARS_BADADDR == DefAddr) { // DEF is USEd (and thus re-DEFed) in a Phi function
		DefEscapes = false;
		set<SMPPhiFunction, LessPhi>::iterator UsePhi;
		UsePhi = this->FindPhi(DefOp);
		if (UsePhi != this->GetLastPhi()) {
			// Found phi function for DefOp. See if we can find a USE
			//  with USE SSANum corresponding to our DEF SSANum.
			for (size_t PhiIndex = 0; PhiIndex < UsePhi->GetPhiListSize(); ++PhiIndex) {
				if (UsePhi->GetUseSSANum(PhiIndex) == SSANum) {
					// We have a Phi USE that matches our DEF.
					UseType = UsePhi->GetUseType(PhiIndex);
					FoundNumeric |= (IsNumeric(UseType));
					FoundCodePtr |= (IsCodePtr(UseType));
					FoundUnknown |= (IsUnknown(UseType));
					FoundUninit |= (IsEqType(UNINIT, UseType));
					if (IsDataPtr(UseType)) {
						if (FoundPointer) {
							if (IsNotEqType(PtrType, UseType)) {
								PtrType = POINTER;
							}
						}
						else {
							FoundPointer = true;
							PtrType = UseType;
						}
					}
					break;
				} // end if matched SSA #
			} // end for all Phi USEs
		} // end if found matching Phi function for DefOp
	}
	else { // DEF is in an instruction in a previous block  !!!!****!!!! Why not current block? A: Only called when LiveIn
		for (InstIter = this->GetFirstInst(); DefEscapes && InstIter != this->GetLastInst(); ++InstIter) {
			SMPInstr *CurrInst = (*InstIter);
			CurrDef = CurrInst->FindDef(DefOp);
			if (CurrDef != CurrInst->GetLastDef()) {
				// Found re-DEF of DefOp.
				DefEscapes = false;
			}
			CurrUse = CurrInst->FindUse(DefOp);
			if (CurrUse != CurrInst->GetLastUse()) { // found a USE of DefOp
				if (CurrUse->GetSSANum() == SSANum) { // matched SSA number
					UseType = CurrUse->GetType();
					FoundNumeric |= (IsNumeric(UseType));
					FoundCodePtr |= (IsCodePtr(UseType));
					FoundUnknown |= (IsUnknown(UseType));
					FoundUninit |= (IsEqType(UNINIT, UseType));
					if (IsDataPtr(UseType)) {
						if (FoundPointer) {
							if (IsNotEqType(PtrType, UseType)) {
								// Have seen two different pointer types/subtypes
								PtrType = POINTER;
							}
						}
						else {
							FoundPointer = true;
							PtrType = UseType;
						}
					}
				} // end if matched SSA #
			} // end if found a USE of DefOp
		} // end for all instructions
	}

	if (DefEscapes) { // did not find re-def
		DefEscapes = this->IsLiveOut(DefOp);
	}

	if (DefEscapes) { // Need to recurse into successor blocks
		list<SMPBasicBlock *>::iterator SuccIter;
		STARS_ea_t TempAddr;
		set<SMPPhiFunction, LessPhi>::iterator PhiIter;
		for (SuccIter = this->GetFirstSucc(); SuccIter != this->GetLastSucc(); ++SuccIter) {
			SMPBasicBlock *CurrBlock = (*SuccIter);
			PhiIter = CurrBlock->FindPhi(DefOp);
			TempAddr = DefAddr;
			if (PhiIter != CurrBlock->GetLastPhi()) {
				TempAddr = STARS_BADADDR;  // signals that DefOp will get re-DEFed in a Phi function.
			}
			else if (STARS_BADADDR == TempAddr) { // was STARS_BADADDR coming in to this function
				// We don't want to pass STARS_BADADDR down the recursion chain, because it will be interpreted
				//  by each successor block to mean that DefOp was a Phi USE that got re-DEFed in a Phi function
				//  within itself. Pass the dummy address that indicates LiveIn to the block.
				TempAddr = STARS_LIVEIN_PSEUDO_ID;
			}
			// Should we screen the recursive call below using CurrBlock->IsLiveIn(DefOp) for speed?  !!!!****!!!!
			CurrBlock->InferGlobalDefType(DefOp, SSANum, TempAddr, FoundNumeric, FoundCodePtr, FoundPointer, FoundUnknown, FoundUninit, PtrType);
		}
	}

	return;
} // end of SMPBasicBlock::InferGlobalDefType()


// Iterate over all phi functions, calling InferPhiDefType() for any Phi DEF whose
//  type is still UNINIT.
// Return TRUE if any changes are made.
bool SMPBasicBlock::InferAllPhiDefTypes(void) {
	bool changed = false;
	bool NewChange;
	set<SMPPhiFunction, LessPhi>::iterator CurrPhi = this->GetFirstPhi();
	while (CurrPhi != this->GetLastPhi()) {
		if (UNINIT == CurrPhi->GetDefType()) {
			CurrPhi = this->InferPhiDefType(CurrPhi, NewChange);
			changed = (changed || NewChange);
			if (NewChange) {
				this->MyFunc->IncTypedPhiDefs();
			}
			else {
				this->MyFunc->IncUntypedPhiDefs();
			}
		}
		else {
			this->MyFunc->IncTypedPhiDefs();
		}
		++CurrPhi;
	}
	return changed;
}  // end of SMPBasicBlock::InferAllPhiDefTypes()

// If all USEs in a Phi function are set in agreement with each other, then infer
//   the Phi DEF type is the same as the USEs type, set it, propagate it,
//   and return true.
// Otherwise, if all USEs of the Phi DEF in the function have a common type,
//  infer the Phi DEF type from them and return true.
set<SMPPhiFunction, LessPhi>::iterator SMPBasicBlock::InferPhiDefType(set<SMPPhiFunction, LessPhi>::iterator DefPhi, bool &changed) {
	changed = false;
	bool SkipToBackInference = false;
	bool PropagateChange = false;
	set<SMPPhiFunction, LessPhi>::iterator ReturnPhi = DefPhi;
	SMPOperandType MeetType = UNINIT;
	STARSOpndTypePtr DefOp = DefPhi->GetAnyOp();
	int SSANum = DefPhi->GetDefSSANum();
	bool SafeFunc = this->MyFunc->IsSafe();
	bool IndirectOp = MDIsIndirectMemoryOpnd(DefOp, this->MyFunc->UsesFramePointer());
	bool UnsafeOp = (IndirectOp || ((! DefOp->IsRegOp()) && (!SafeFunc)));
	  // Relax the UnsafeOp definition with alias analysis?

	assert(UNINIT == DefPhi->GetDefType());
	for (size_t index = 0; index < DefPhi->GetPhiListSize(); ++index) {
		if (IsEqType(UNINIT, DefPhi->GetUseType(index))) {
			SkipToBackInference = true;
			break;
		}
	}

	if (!SkipToBackInference) {
		// Must have seen all non-UNINIT types if we reach this point.
		MeetType = DefPhi->ConditionalMeetType(this); // DefPhi can be stale after this call
		if (IsNotEqType(UNKNOWN, MeetType)) {
			// Don't infer UNKNOWN until conditional type propagation stage.
			ReturnPhi = this->SetPhiDefType(DefOp, MeetType);
			changed = true;
			PropagateChange = true;
		}
		else {
			ReturnPhi = this->FindPhi(DefOp); // in case DefPhi went stale, causing ReturnPhi to go stale.
		}
	}
	else if (!UnsafeOp) { // try back inference from USEs of Phi DEF
		SMPOperandType DefType;
		DefType = this->MyFunc->InferGlobalDefType(DefOp, SSANum, this, false, STARS_BADADDR);
		if (IsNotEqType(UNINIT, DefType)) {
			// Don't set changed, because there is no downward propagation
			//  to do. Instead, propagate only to UNINIT Phi USEs.
			size_t limit = DefPhi->GetPhiListSize();
			set<size_t> UseUpdates;
			UseUpdates.clear();
			for (size_t index = 0; index < limit; ++index) {
				if (UNINIT == DefPhi->GetUseType(index)) {
					UseUpdates.insert(index); // keep list needing update
				}
			}
			set<size_t>::iterator UpIter;
			for (UpIter = UseUpdates.begin(); UpIter != UseUpdates.end(); ++UpIter) {
				this->SetPhiUseType(DefOp, *UpIter, DefType);
				// NOTE: DefPhi iterator is stale after this update
			}
			ReturnPhi = this->SetPhiDefType(DefOp, DefType);
			// We don't set PropagateChange to true for the NUMERIC back inference case,
			//  because there is no propagating to do. However, for POINTER types, we can
			//  do an aggressive back inference, in which a mixture of UNINIT and POINTER
			//  implies POINTER, and then we need to propagate to get those UNINIT uses changed
			//  to POINTER.
			if (IsDataPtr(DefType) || IsCodePtr(DefType)) {
				PropagateChange = true;
				MeetType = DefType;
			}
			changed = true;
			UseUpdates.clear();
		}
		else {
			ReturnPhi = this->FindPhi(DefOp); // in case DefPhi went stale, causing ReturnPhi to go stale.
		}
	}

	// Propagate down to USEs of this DEF.
	if (PropagateChange && !UnsafeOp) {
		bool IsMemOp = DefOp->IsMemOp();
		this->MyFunc->ResetProcessedBlocks();
		changed |= this->PropagateGlobalDefType(DefOp, MeetType, SSANum, IsMemOp, false);
	}
	return ReturnPhi;
} // end of SMPBasicBlock::InferPhiDefType()

// Return true if UseIter should not be added to the DEFsToSearch set for IsMemWriteUnsafe()
bool SMPBasicBlock::PrescreenUnsafeAddrReg(const STARSDefUseIter UseIter, const STARS_ea_t InstAddr) {
	STARSOpndTypePtr UseOp = UseIter->GetOp();
	bool LocalName = this->IsLocalName(UseOp);
	STARS_ea_t UseDefAddr = this->GetDefAddrFromUseAddr(UseOp, InstAddr, UseIter->GetSSANum(), LocalName);
	bool NonInstAddr = (STARS_BADADDR == UseDefAddr) || STARS_IsLiveInPseudoID(UseDefAddr) || STARS_IsBlockNumPseudoID(UseDefAddr);
	if (!NonInstAddr) {
		if (STARS_IsSSAMarkerPseudoID(UseDefAddr))
			return true;
		else {
			SMPInstr *DefInst = this->GetFunc()->GetInstFromAddr(UseDefAddr);
			int DefBlockNum = DefInst->GetBlock()->GetNumber();
			if (this->GetFunc()->IsBlockInAnyLoop(DefBlockNum))
				return true;
		}
	}
	return false; // no simple way to screen out; return false
} // end of SMPBasicBlock::PrescreenUnsafeAddrReg()

// Are any DEFs in search set of unsafe-write origin?
bool SMPBasicBlock::IsMemWriteUnsafe(STARS_ea_t InstID, bool FirstBlock, STARSDefUseSet &DEFsToSearch) {
	// A mem write is considered unsafe if it is indirect and:
	//  1. The address expression is loop-variant.
	//  2. The address expression is dependent on an incoming arg.
	// Condition 1 will start conservatively as "depends on any value computed inside a loop."
	// Condition 2 will become less conservative in the future with interprocedural analysis.
	if (this->IsProcessed())
		return false;
	this->SetProcessed(true);

	if (FirstBlock) { // no recursion yet
		bool UnsafeIndirectWrite = false;
		SMPInstr *CurrInst = this->GetFunc()->GetInstFromAddr(InstID);
		bool LoopingWrite = CurrInst->MDHasAnyRepeatPrefix();
		int BaseReg, IndexReg;
		uint16_t ScaleFactor;
		STARS_ea_t offset;
		STARSOpndTypePtr MemOp = CurrInst->GetMemDef();
		MDExtractAddressFields(MemOp, BaseReg, IndexReg, ScaleFactor, offset);
		bool EIPRelativeAccess = ((STARS_x86_R_ip == BaseReg) || (STARS_x86_R_ip == IndexReg));
		if (MemOp->IsStaticMemOp() || EIPRelativeAccess) {
			UnsafeIndirectWrite = true;
		}
		else {
			if (!LoopingWrite && (ScaleFactor == 0) && (!((BaseReg != STARS_x86_R_none) && (IndexReg != STARS_x86_R_none)))) {
				// Non-scaled, base or index (implied by !StaticMemOp()) but not both.
				STARS_regnum_t AddrReg = (STARS_regnum_t) BaseReg;
				if (IndexReg != STARS_x86_R_none) {
					AddrReg = (STARS_regnum_t) IndexReg;
				}
				AddrReg = MDCanonicalizeSubReg(AddrReg);
				STARSOpndTypePtr AddrRegOp = CurrInst->MakeRegOpnd(AddrReg);
				STARSDefUseIter UseIter = CurrInst->FindUse(AddrRegOp);
				assert(UseIter != CurrInst->GetLastUse());
				bool SafeAddrReg = this->GetFunc()->IsAddressRegSafe(AddrRegOp, InstID, UseIter->GetSSANum());
				UnsafeIndirectWrite = (!SafeAddrReg);
			}
			else {
				UnsafeIndirectWrite = true;
			}
		}
		if (!UnsafeIndirectWrite) { // analyzed safe
			return false;
		}
	}

	vector<SMPInstr *>::iterator InstIter;
	if (FirstBlock) { // starting with InstID, possibly in the middle of the block.
		if (this->GetFunc()->IsBlockInAnyLoop(this->BlockNum)) {
			return true;
		}
		InstIter = this->GetInstIterFromAddr(InstID);
	}
	else {
		InstIter = this->InstVec.end();
	}

	while (InstIter != this->GetFirstInst()) {
		--InstIter;
		// Search for DEFs of any item in DEFsToSearch, which are the operands that influence the value of the
		//  original mem-write addressing expression.
		SMPInstr *CurrInst = (*InstIter);
		STARS_ea_t InstAddr = CurrInst->GetAddr();
		for (STARSDefUseIter DefIter = CurrInst->GetFirstDef(); DefIter != CurrInst->GetLastDef(); ++DefIter) {
			STARSDefUseIter FindIter = DEFsToSearch.find(*DefIter);
			if (FindIter != DEFsToSearch.end()) { // CurrInst has DEF of operand that influences mem-write addressing expr.
				if (this->GetFunc()->IsBlockInAnyLoop(this->BlockNum)) {
					return true; // conservative; *DefIter might not actually be loop-variant
				}
				if (CurrInst->IsMarkerInst()) { // live-in to func, whether in-arg on stack or live-in reg
					return true; // conservative, until we get interprocedural analysis
				}
				// Add all uses to DEFsToSearch set, unless they can be shown unsafe quickly.
				for (STARSDefUseIter UseIter = CurrInst->GetFirstUse(); UseIter != CurrInst->GetLastUse(); ++UseIter) {
					STARSOpndTypePtr UseOp = UseIter->GetOp();
					bool SearchableOpnd = (UseOp->IsRegOp() || UseOp->IsMemOp());
					if (SearchableOpnd && (!UseOp->MatchesReg(MD_FLAGS_REG))) { // exclude flags reg
						bool Prescreened = this->PrescreenUnsafeAddrReg(UseIter, InstAddr);
						if (Prescreened)
							return true;
						// Cannot quickly say USE is unsafe, so add it to search set.
						pair<STARSDefUseIter, bool> InsertResult = DEFsToSearch.insert(*UseIter);
					}
				} // end for all USEs in CurrInst
			}
		}
	}
	// We did not return true yet, so we need to search Phi functions and recurse through predecessor blocks.
	for (STARSDefUseIter SearchIter = DEFsToSearch.begin(); SearchIter != DEFsToSearch.end(); ++SearchIter) {
		STARSOpndTypePtr SearchOp = SearchIter->GetOp();
		set<SMPPhiFunction, LessPhi>::iterator PhiIter = this->FindPhi(SearchOp);
		if (PhiIter != this->GetLastPhi()) { // Found SearchOp, but must match SSA number
			if (PhiIter->GetDefSSANum() == SearchIter->GetSSANum()) {
				// All Phi USEs now influence the ultimate mem-write addressing expression.
				for (size_t PhiIndex = 0; PhiIndex < PhiIter->GetPhiListSize(); ++PhiIndex) {
					DefOrUse TempUse(SearchOp, PhiIter->GetUseType(PhiIndex), PhiIter->GetUseSSANum(PhiIndex));
					pair<STARSDefUseIter, bool> InsertResult = DEFsToSearch.insert(TempUse);
				}
			}
		}
	}

	bool UnsafeWrite = false;
	for (list<SMPBasicBlock *>::iterator PredIter = this->GetFirstPred(); PredIter != this->GetLastPred(); ++PredIter) {
		UnsafeWrite = (*PredIter)->IsMemWriteUnsafe(InstID, false, DEFsToSearch);
		if (UnsafeWrite) {
			break;  // don't waste time
		}
	}
	return UnsafeWrite;
} // end of SMPBasicBlock::IsMemWriteUnsafe()

// Track Phi USEs of PhiMemOp back and mark DEFs as MaybeAliased
void SMPBasicBlock::MarkAllPhiUsesAliased(STARSOpndTypePtr &PhiMemOp) {
	PhiSetIter PhiDefIter = this->FindPhi(PhiMemOp);
	assert(PhiDefIter != this->GetLastPhi());
	if (this->IsSCCPVisited()) {
		// already processed this block
		return;
	}
	else {
		// Mark block as visited, avoiding infinite looping.
		this->SetSCCPVisited(true);
	}

	// Each Phi USE must be set to MayAlias status.
	size_t PhiUseLimit = PhiDefIter->GetPhiListSize();
	PhiDefIter = this->SetPhiDefMaybeAliased(PhiMemOp);
	for (size_t UseIndex = 0; UseIndex < PhiUseLimit; ++UseIndex) {
		int PhiUseSSANum = PhiDefIter->GetUseSSANum(UseIndex);
		if (0 > PhiUseSSANum) {
			SMP_msg("ERROR: MarkAllPhiUsesAliased() cannot trace uninitialized Phi USE at block %d with first inst at %p\n",
				this->GetNumber(), this->GetFirstAddr());
			continue;
		}
		STARS_ea_t UseDefAddr = this->GetFunc()->GetGlobalDefAddr(PhiMemOp, PhiUseSSANum);
		assert(STARS_BADADDR != UseDefAddr);
		if (STARS_IsBlockNumPseudoID(UseDefAddr)) {
			// Phi USE traced back to Phi DEF; recurse
			size_t DefBlockNum = STARS_GetBlockNumFromPseudoID(UseDefAddr);
			SMPBasicBlock *DefBlock = this->GetFunc()->GetBlockByNum(DefBlockNum);
			assert(nullptr != DefBlock);
			if (DefBlockNum == (size_t) this->GetNumber()) {
				// Infinite recursion? How can Phi USE be DEFed in the same Phi?
				SMP_msg("ERROR: Phi USE is DEFed in same Phi function, block %d, func %s \n",
					DefBlockNum, this->GetFunc()->GetFuncName());
				SMP_msg("PhiMemOp: ");
				PrintOperand(PhiMemOp);
				SMP_msg("  Phi USESSANum: %d\n", PhiUseSSANum);
			}
			else {
				DefBlock->MarkAllPhiUsesAliased(PhiMemOp);
			}
		}
		else {
			SMPInstr *DefInst = this->GetFunc()->GetInstFromAddr(UseDefAddr);
			assert(nullptr != DefInst);
			STARSDefUseIter DefIter = DefInst->SetDefIndWrite(PhiMemOp, true);
		}
	}

	return;
} // end of SMPBasicBlock::MarkAllPhiUsesAliased()

// Trace LiveInOp; don't know SSANum; is it already marked MaybeAliased
STARS_ea_t SMPBasicBlock::TraceLiveInOpToDefAddr(const STARSOpndTypePtr &LiveInOp, bool &MaybeAliased) {
	STARS_ea_t DefAddr = STARS_BADADDR;
	MaybeAliased = false;
	// LiveIn to entry block gets DEFed in the SSA marker inst, not in predecessors.
	if (0 == this->GetNumber())
		DefAddr = STARS_SSA_MARKER_PSEUDO_ID;

	list<SMPBasicBlock *>::const_iterator PredIter;
	this->SetProcessed(true);
	for (PredIter = this->GetFirstConstPred(); PredIter != this->GetLastConstPred(); ++PredIter) {
		SMPBasicBlock *PredBlock = (*PredIter);
		if (PredBlock->IsLiveOut(LiveInOp)) {
			// Look up the SSANum for the LiveOut operand (same as LiveInOp in current block).
			map<STARSOpndTypePtr, int, LessOp>::const_iterator MapIter = PredBlock->FindLiveOutSSAMap(LiveInOp);
			assert(MapIter != PredBlock->GetLastLiveOutSSAMap());
			int LiveOutSSANum = MapIter->second;

			// Find the DefAddr using the SSA number.
			DefAddr = PredBlock->GetFunc()->GetGlobalDefAddr(LiveInOp, LiveOutSSANum);
			if (STARS_IsBlockNumPseudoID(DefAddr)) {
				size_t DefBlockNum = STARS_GetBlockNumFromPseudoID(DefAddr);
				SMPBasicBlock *DefBlock = this->GetFunc()->GetBlockByNum(DefBlockNum);
				PhiSetIter DefPhiIter = DefBlock->FindPhi(LiveInOp);
				assert(DefPhiIter != DefBlock->GetLastPhi());
				MaybeAliased = DefPhiIter->GetDefMaybeAliased();
			}
			else { // inst DefAddr
				SMPInstr *DefInst = this->GetFunc()->GetInstFromAddr(DefAddr);
				STARSDefUseIter DefIter = DefInst->FindDef(LiveInOp);
				assert(DefIter != DefInst->GetLastDef());
				MaybeAliased = DefIter->HasIndirectWrite();
			}

			break;
		}
	} // end for each predecessor block

	return DefAddr;
} // end of SMPBasicBlock::TraceLiveInOpToDefAddr()

// Get SSANum for CurrOp incoming value (pre-loop) to loop header.
//  Return -1 if not a loop header, or more than one pre-loop incoming value.
int SMPBasicBlock::GetLoopLiveInSSANum(const STARSOpndTypePtr &CurrOp) const {
	int SSANum = SMP_SSA_UNINIT;

	if (this->IsLoopHeaderBlock()) {
		int LoopNum = this->GetFunc()->GetLoopNumFromHeaderBlockNum(this->GetNumber());
		assert(0 <= LoopNum);
		list<SMPBasicBlock *>::const_iterator PredIter;
		for (PredIter = this->GetFirstConstPred(); PredIter != this->GetLastConstPred(); ++PredIter) {
			SMPBasicBlock *PredBlock = (*PredIter);
			// Want predecessor that is not in this loop, and has LiveOut CurrOp.
			if (PredBlock->IsLiveOut(CurrOp) && (!this->GetFunc()->IsBlockInLoop(PredBlock->GetNumber(), (size_t) LoopNum))) {
				// Look up the SSANum for the LiveOut operand (same as LiveInOp in current block).
				map<STARSOpndTypePtr, int, LessOp>::const_iterator MapIter = PredBlock->FindLiveOutSSAMap(CurrOp);
				assert(MapIter != PredBlock->GetLastLiveOutSSAMap());
				if (SMP_SSA_UNINIT == SSANum) { // First LiveIn/LiveOut match
					SSANum = MapIter->second;
				}
				else {
					// Not the first match; problem
					SSANum = SMP_SSA_UNINIT;
					break;
				}
			}
		} // end for each predecessor block
	}

	return SSANum;
} // end of SMPBasicBlock::GetLiveInSSANum()


// Is GlobalOp used anywhere in this block as a LiveIn value, including as a Phi USE?
bool SMPBasicBlock::IsGlobalNameUsedLiveIn(const STARSOpndTypePtr &GlobalOp) {
	bool FoundInLiveInSet = this->IsLiveIn(GlobalOp);
	bool OpUsed = FoundInLiveInSet;
	if (FoundInLiveInSet) {
		// exclude the pass-through case, in which the operand is LiveIn, LiveOut, but never killed or used.
		bool FoundInUpExposedSet = (this->GetLastUpExposed() != this->UpExposedSet.find(GlobalOp));
		OpUsed = FoundInUpExposedSet;
		if (!OpUsed) {
			// See if it is used as a Phi USE and is otherwise just passed through the block.
			set<SMPPhiFunction, LessPhi>::iterator PhiIter = this->FindPhi(GlobalOp);
			OpUsed = (PhiIter != this->GetLastPhi());
		}
	}

	return OpUsed;
} // end of SMPBasicBlock::IsGlobalNameUsed()

// Extension of IsRegDead for Global names.  Check and see if it overlaps with a DEF-USE chain
bool SMPBasicBlock::IsGlobalRegDead(STARS_ea_t InstAddr, const STARSOpndTypePtr &Operand, unsigned int RegIndex) {
	bool FoundInLiveInSet = this->IsLiveIn(Operand);
	bool FoundInLiveOutSet = this->IsLiveOut(Operand);
	bool FoundInKillSet = this->IsVarKill(Operand);
#if SMP_DEBUG_OPTIMIZATIONS
	bool DebugFlag = (0x804837b == InstAddr);

	if (DebugFlag) SMP_msg("A1");
#endif

	// If it is Live-In and Live-Out and not in the kill set,
	// it is live the whole time (either pass-through case or use-only case).
	if (FoundInLiveInSet && FoundInLiveOutSet && !FoundInKillSet)
		return false;

	// EXCEPTION CASE: Could be a pass-through case, in which the reg is callee-saved
	//  but is never DEFed in the function, so there was no need to save and restore
	//  it. In that case it is live through the function.
	vector<SMPInstr *>::iterator InstIter = this->GetInstIterFromAddr(InstAddr);
	SMPInstr *SaveInst = (*InstIter);
	STARS_regnum_t RegNo = Operand->GetReg();
	STARSOpndTypePtr RegOp = SaveInst->MakeRegOpnd(RegNo);
	bool PassThroughCase = global_STARS_program->IsCalleeSavedReg(RegNo) && (!this->GetFunc()->IsVarKill(RegOp));
	if (PassThroughCase)
		return false;

#if SMP_DEBUG_OPTIMIZATIONS
	if (DebugFlag) SMP_msg("B");
#endif

	// If it is not Live-In and not in the Kill set it is dead the whole time
	if (!FoundInLiveInSet && !FoundInKillSet) {
		// Must not have USEs, else it would be LiveIn. So, no DEFs, no USEs, not LiveIn: Dead.
		assert(!FoundInLiveOutSet);
		return true;
	}


	while (InstIter != this->GetLastInst()) {
		SMPInstr *CurrInst = (*InstIter);
		// Case 1: Not dead if reg is USE before it is a DEF.
		if (CurrInst->FindUse(Operand) != CurrInst->GetLastUse()) {
			return false;
		}
		// Case 2: Dead if DEF before USE.
		if (CurrInst->FindDef(Operand) != CurrInst->GetLastDef()) {
			return true;
		}
		++InstIter;
	}
	// Case 3: Reached end of block with no DEFs or USEs >=InstAddr. This is a global name,
	//  so it is dead if and only if it is not LiveOut. Does not really matter
	//  if it was DEFed or USEd before InstAddr.
	return (!FoundInLiveOutSet);

} // end of SMPBasicBlock::IsGlobalRegDead()

// If no DEF-USE chains overlap the instrumentation point for InstAddr (which logically
//  falls between the USEs of the instruction at InstAddr and its DEFs), for register RegIndex,
//  then it is safe for the instrumentation to use that register without saving and
//  restoring its value. Return true in this case, false if there is DEF-USE overlap.
bool SMPBasicBlock::IsRegDead(STARS_ea_t InstAddr, STARS_regnum_t RegNo) {
	// See if any DEF-USE chains overlap InstAddr's USEs.
	vector<SMPInstr *>::iterator InstIter = this->GetInstIterFromAddr(InstAddr);
	STARSOpndTypePtr RegOp = (*InstIter)->MakeRegOpnd(RegNo);

	while (InstIter != this->GetLastInst()) {
		SMPInstr *CurrInst = (*InstIter);
		// Case 1: Not dead if reg is USE before it is a DEF.
		if (CurrInst->FindUse(RegOp) != CurrInst->GetLastUse()) {
			return false;
		}
		// Case 2: Dead if DEF before USE.
		if (CurrInst->FindDef(RegOp) != CurrInst->GetLastDef()) {
			return true;
		}
		++InstIter;
	}
	// Case 3: Reached end of block with no DEFs or USEs. This is a local name,
	//  so it was dead at the original InstAddr.
	// EXCEPTION: Could be a pass-through case, in which the reg is callee-saved
	//  but is never DEFed in the function, so there was no need to save and restore
	//  it. In that case it is live through the function.
	bool PassThroughCase = false;
	if (this->GetFunc()->GetNumBlocks() == 1) {
		// If only one block, local names are like global names in other funcs, and 
		//  a lack of kills in this one block means that the name passed through unchanged.
		PassThroughCase = global_STARS_program->IsCalleeSavedReg(RegNo) && (!this->IsVarKill(RegOp));
	}
	return !PassThroughCase;
} // end of SMPBasicBlock::IsRegDead()

// Mark the registers that are dead for each instruction in the block.
void SMPBasicBlock::MarkDeadRegs(void) {
#if SMP_DEBUG_OPTIMIZATIONS
	bool DebugFlag = (0 == strcmp("_int_malloc", this->MyFunc->GetFuncName()));
#endif
	STARSOpndSetIter NameIter;
	STARSOpndSetIter FlagNameIter;
	vector<SMPInstr *>::iterator InstIter;
	STARSOpndTypePtr FlagsOp = (*(this->GetFirstInst()))->MakeRegOpnd(X86_FLAGS_REG);
	bool GlobalFlags = this->MyFunc->IsGlobalName(FlagsOp);
	if (GlobalFlags) {
#if 0
		this->SetFlagsDeadAfterLastUse((this->LiveOutSet.find(FlagsOp) == this->LiveOutSet.end()));
#endif
		this->SetFlagsDeadBeforeFirstKill((!this->IsUpExposed(FlagsOp)) && (!this->IsVarKill(FlagsOp)));
		FlagNameIter = this->MyFunc->FindGlobalName(FlagsOp);
	}
	else {
		FlagNameIter = this->LocalNames.find(FlagsOp);
	}

	// Special case: caller-saved Reg is local in other blocks, but not referenced in this block.
	//  That implies it is dead for all insts in this block and, being local, cannot be needed LiveIn to another block.
	list<STARS_regnum_t> CallerSavedLocalsList;
	list<STARS_regnum_t>::const_iterator CallerSavedIter;
	SMPInstr *FirstInst = (*(this->GetFirstInst()));
	for (CallerSavedIter = global_STARS_program->GetFirstCallerSavedReg(); CallerSavedIter != global_STARS_program->GetLastCallerSavedReg(); ++CallerSavedIter) {
		STARS_regnum_t RegNum = *CallerSavedIter;
		if (this->GetFunc()->IsCallerSavedLocalReg((size_t)RegNum)) {
			// Caller-saved reg is a local name in this function.
			STARSOpndTypePtr LocalRegOp = FirstInst->MakeRegOpnd(RegNum);
			if (!this->IsVarKill(LocalRegOp)) {
				// If a local name is not killed in this block, it is not used in this block
				CallerSavedLocalsList.push_back(RegNum);
			}
		}
	}

	for (InstIter = this->GetFirstInst(); InstIter != this->GetLastInst(); ++InstIter) {
		SMPInstr *CurrInst = (*InstIter);
		if (CurrInst->IsMarkerInst())
			continue;
		STARS_ea_t InstAddr = CurrInst->GetAddr();
#if SMP_DEBUG_OPTIMIZATIONS
		DebugFlag |= (0x804dafc == InstAddr);
		if (DebugFlag) {
			CurrInst->PrintOperands();
			SMP_msg("\n");
		}
#endif
		// First, put EFLAGS at beginning of string if it is dead.
		if (!GlobalFlags) {
			if (FlagNameIter != this->LocalNames.end()) {
				// EFLAGS is in the LocalNames
				STARS_regnum_t RegNo = (*FlagNameIter)->GetReg();
				if (this->IsRegDead(InstAddr, RegNo)) {
					CurrInst->SetRegDead((size_t) MD_FLAGS_REG);
				}
			}
			else { // EFLAGS are not locally used, but not global either; they are dead here
				CurrInst->SetRegDead((size_t) MD_FLAGS_REG);
			}
		}
		else {
			assert(FlagNameIter != this->MyFunc->GetLastGlobalName());
			unsigned int RegIndex = ExtractGlobalIndex(*FlagNameIter);
			bool newCatch = this->IsGlobalRegDead(InstAddr, FlagsOp, RegIndex);
			if (newCatch)
				CurrInst->SetRegDead((size_t) MD_FLAGS_REG);
		}

		// Now, process all other local regs and skip EFLAGS.
		for (NameIter = this->LocalNames.begin(); NameIter != this->LocalNames.end(); ++NameIter) {
			if (!(*NameIter)->IsRegOp())
				continue;
			// We never want to consider ESP to be available for instrumentation.
			if ((*NameIter)->MatchesReg(MD_STACK_POINTER_REG))
				continue;
			// Until we analyze interprocedural register use, it is safest to not
			//  use EBP for instrumentation. This could change with more analysis.
			if ((*NameIter)->MatchesReg(MD_FRAME_POINTER_REG))
				continue;

			STARS_regnum_t RegNo = (*NameIter)->GetReg();
			if (RegNo == MD_FLAGS_REG) {
				//  We moved EFLAGS to the beginning, per annotations spec
				continue;
			}
			if (this->IsRegDead(InstAddr, RegNo)) {
				CurrInst->SetRegDead((size_t) RegNo);
			}
		} // end for all local names

		// Now, process all other global regs and skip EFLAGS.
		//SMP_msg(" Done.  Analyzing global regs...");
		size_t NumGlobals = this->MyFunc->NumGlobalNames();
		if (NumGlobals > 0) {
			for (NameIter = this->MyFunc->GetFirstGlobalName(); NameIter != this->MyFunc->GetLastGlobalName(); ++NameIter) {
				//SMP_msg(" i ");

				//SMP_msg("%d", (*NameIter)->GetType());
				if (!(*NameIter)->IsRegOp())
					continue;
				//SMP_msg(" j ");
				// We never want to consider ESP to be available for instrumentation.
				if ((*NameIter)->MatchesReg(MD_STACK_POINTER_REG))
					continue;
				//SMP_msg(" k ");
				// Until we analyze interprocedural register use, it is safest to not
				//  use EBP for instrumentation. This could change with more analysis.
				if ((*NameIter)->MatchesReg(MD_FRAME_POINTER_REG))
					continue;

				//SMP_msg(" l ");
				unsigned int RegIndex = ExtractGlobalIndex(*NameIter);
				STARS_regnum_t RegNum = (*NameIter)->GetReg();
				if (RegNum == X86_FLAGS_REG) {
					//  We moved EFLAGS to the beginning, per annotations spec
					continue;
				}
				//SMP_msg(" m ");
				STARSOpndTypePtr RegOp = CurrInst->MakeRegOpnd(RegNum);
				//SMP_msg("%u", RegNum);
				if (this->IsGlobalRegDead(InstAddr, RegOp, RegIndex)) {
					CurrInst->SetRegDead((size_t) RegNum);
				}
			} // end for all global names
		}

		// Another case: caller-saved Reg is local in other blocks, but not referenced in this block.
		//  That implies it is dead in this block and, being local, cannot be needed LiveIn to another block.
		list<STARS_regnum_t>::const_iterator CallerSavedRegIter;
		for (CallerSavedRegIter = CallerSavedLocalsList.cbegin(); CallerSavedRegIter != CallerSavedLocalsList.cend(); ++CallerSavedRegIter) {
			STARS_regnum_t RegNum = *CallerSavedRegIter;
			CurrInst->SetRegDead((size_t) RegNum);
		}
		//SMP_msg("Done.  Next instruction.\n");

	} // end for all instructions
#if SMP_DEBUG_OPTIMIZATIONS
	if (DebugFlag)
		SMP_msg("Exiting MarkDeadRegs()\n");
#endif
	return;
} // end of SMPBasicBlock::MarkDeadRegs()

// Does DefOp at DefAddr never get used?
bool SMPBasicBlock::IsDefDead(STARS_ea_t DefAddr, const STARSOpndTypePtr &DefOp) {
	vector<SMPInstr *>::iterator InstIter = this->GetInstIterFromAddr(DefAddr);

	++InstIter; // start with next instruction
	while (InstIter != this->GetLastInst()) {
		SMPInstr *CurrInst = (*InstIter);
		// Case 1: Not dead if reg is USE before it is a DEF.
		if (CurrInst->FindUse(DefOp) != CurrInst->GetLastUse()) {
			return false;
		}
		// Case 2: Dead if re-DEF before USE.
		if (CurrInst->FindDef(DefOp) != CurrInst->GetLastDef()) {
			return true;
		}
		++InstIter;
	}
	// Case 3: Reached end of block with no DEFs or USEs. If this is a local name,
	//  it was dead at the original DefAddr, else a global is dead if it is not LiveOut.
	STARSOpndSetIter LocalIter = this->LocalNames.find(DefOp);
	bool LocalName = (LocalIter != this->LocalNames.end());
	return (LocalName || (!(this->IsLiveOut(DefOp))));
} // end of SMPBasicBlock::IsDefDead()

// Return true and produce list if any flags are live out of inst at InstAddr.
//  NOTE: Assumes deadregs have been computed for all registers.
bool SMPBasicBlock::AreFlagsLiveAfterInst(STARS_ea_t InstAddr, set<int> &LiveFlagRegs) {
	vector<SMPInstr *>::iterator InstIter = this->GetInstIterFromAddr(InstAddr);
	STARSOpndTypePtr FlagsOp = (*InstIter)->MakeRegOpnd(MD_FLAGS_REG);

	++InstIter; // start with next instruction
	while (InstIter != this->GetLastInst()) {
		SMPInstr *CurrInst = (*InstIter);
		if (CurrInst->IsIncomingRegDead(MD_FLAGS_REG)) {
			return false;
		}

		// Go until we find an instruction that USEs or re-DEFs the flags.
		bool UsesFlags = CurrInst->IsUsesFlags();
		bool DefsFlags = CurrInst->IsDefsFlags();
		if (DefsFlags) { // re-DEF
			// See which flags are used.
			if (!UsesFlags) { // not USEd, re-DEFed, so they were dead.
				SMP_msg("ERROR: Flags should have been marked DEAD at %lx\n", (unsigned long) CurrInst->GetAddr());
				return false;
			}
			else {
				CurrInst->MDGetFlagsUsed(LiveFlagRegs);
				return true;
			}
		}
		else { // not re-DEF, might be USEd
			if (UsesFlags) {
				CurrInst->MDGetFlagsUsed(LiveFlagRegs); // accumulate and go on to next inst
			}
		}
		++InstIter;
	}
	// Reached end of block with no re-DEFs. If flags reg is a local name, we are done.
	//  If global name, we are done if not LiveOut flags, else we conservatively mark
	//  all key flags as live.
	STARSOpndSetIter LocalIter = this->LocalNames.find(FlagsOp);
	bool LocalName = (LocalIter != this->LocalNames.end());

	if (LocalName || (!this->IsLiveOut(FlagsOp))) {
		return (!(LiveFlagRegs.empty())); // assumes it was empty on entry
	}
	else {
		pair<set<int>::iterator, bool> InsertResult;
		InsertResult = LiveFlagRegs.insert(MD_CARRY_FLAG);
		InsertResult = LiveFlagRegs.insert(MD_ZERO_FLAG);
		InsertResult = LiveFlagRegs.insert(MD_SIGN_FLAG);
		InsertResult = LiveFlagRegs.insert(MD_OVERFLOW_FLAG);
		InsertResult = LiveFlagRegs.insert(MD_PARITY_FLAG);
	}
	return true;
} // end of SMPBasicBlock::AreFlagsLiveAfterInst()

// Return true if DefOp at DefAddr is used in or redefined by addition; pass back AdditionAddr
bool SMPBasicBlock::IsDefInvolvedInAddition(STARS_ea_t DefAddr, const STARSOpndTypePtr &DefOp, STARS_ea_t &AdditionAddr) {
	STARSOpndSetIter LocalIter = this->LocalNames.find(DefOp);
	bool UseFP = this->GetFunc()->UsesFramePointer();
	vector<SMPInstr *>::iterator UseInstIter;

	if (!MDIsDataFlowOpnd(DefOp, UseFP)) {
		return false;
	}

	UseInstIter = this->GetInstIterFromAddr(DefAddr);
	++UseInstIter;
	while (UseInstIter != this->GetLastInst()) {
		SMPInstr *UseInst = (*UseInstIter);
		AdditionAddr = UseInst->GetAddr();
		if (UseInst->FindUse(DefOp) != UseInst->GetLastUse()) {
			if (UseInst->MDIsAddition()) {
				// hash function value is accumulating in another DEF via addition.
				return true;
			}
			else if (UseInst->MDIsMoveInstr() || UseInst->MDIsUnderflowingOpcode() || UseInst->MDIsLoadEffectiveAddressInstr()) {
				// Copied to another location via move or subtraction; could reach addition in that location.
				set<DefOrUse, LessDefUse>::iterator CopyDefIter = UseInst->GetFirstNonFlagsDef();
				STARSOpndTypePtr CopyDefOp = CopyDefIter->GetOp();
				STARS_ea_t CopyAdditionAddr;
				if (this->IsDefInvolvedInAddition(UseInst->GetAddr(), CopyDefOp, CopyAdditionAddr)) {
					AdditionAddr = CopyAdditionAddr;
					return true;
				}
			}
		}
		if (UseInst->FindDef(DefOp) != UseInst->GetLastDef()) {
			// re-DEF
			return (UseInst->MDIsAddition());
		}
		++UseInstIter;
	}
	return false; // reached end of block without returning true or false above

} // end of SMPBasicBlock::IsDefInvolvedInAddition()

// Does a copy of UseOp get shifted right in this block?
bool SMPBasicBlock::IsUseCopiedAndShiftedRight(const STARSOpndTypePtr &UseOp, STARS_ea_t UseAddr) {
	SMPInstr *UseInst = this->FindInstr(UseAddr);
	STARSOpndTypePtr SearchOp = UseOp;
	set<DefOrUse, LessDefUse>::iterator UseIter = UseInst->FindUse(SearchOp);
	if (UseIter == UseInst->GetLastUse()) {
		return false;
	}

	vector<SMPInstr *>::iterator InstIter;
	bool FoundCopy = false;
	bool FoundShiftRight = false;
	STARSOpndTypePtr CopyOperand = nullptr;
	int CopySSANum = -1;
	for (InstIter = this->GetFirstInst(); InstIter != this->GetLastInst(); ++InstIter) {
		SMPInstr *CurrInst = (*InstIter);
		if (!FoundCopy) {
			if (CurrInst->MDIsMoveInstr()) {
				if (IsEqOpIgnoreBitwidth(SearchOp, CurrInst->GetMoveSource())) {
					FoundCopy = true;
					set<DefOrUse, LessDefUse>::iterator DefIter = CurrInst->GetFirstNonFlagsDef();
					CopyOperand = DefIter->GetOp();
					CopySSANum = DefIter->GetSSANum();
				}
			}
		}
		else { // already found the copy
			if (CurrInst->MDIsShiftRight()) {
				STARSOpndTypePtr ShiftedOperand = nullptr, ShiftCounterOperand = nullptr;
				CurrInst->GetShiftOperands(ShiftedOperand, ShiftCounterOperand);
				if (IsEqOpIgnoreBitwidth(CopyOperand, ShiftedOperand)) {
					set<DefOrUse, LessDefUse>::iterator ShiftUseIter = CurrInst->FindUse(CopyOperand);
					assert(ShiftUseIter != CurrInst->GetLastUse());
					if (CopySSANum == ShiftUseIter->GetSSANum()) {
						FoundShiftRight = true;
						break;
					}
				}
			}
		}
	}

	return (FoundCopy && FoundShiftRight);
} // end of SMPBasicBlock::IsUseCopiedAndShiftedRight()

// Does value of DefOp/DefSSANum reach block end in any computation of any DEF at all?
bool SMPBasicBlock::DoesDefReachBlockEnd(STARS_ea_t DefAddr, const STARSOpndTypePtr &DefOp, int DefSSANum, set<int> &NonEscapingRegisterHashes) {
	STARSOpndSetIter LocalIter = this->LocalNames.find(DefOp);
	bool LocalName = (LocalIter != this->LocalNames.end());
	bool DoesEscape = false;
	bool ReDEF = false;
	bool UseFP = this->GetFunc()->UsesFramePointer();
	// set<int> NonEscapingRegisterHashes; // memoization optimization: set of register/SSA# hashes that do not reach end of block

	// If we encounter indirect memory operands of any kind, we cannot track them and must
	//  assume they are loop-variant and reach outside this block. Ditto for non-reg,
	//  non-stack operands.
	if (MDIsIndirectMemoryOpnd(DefOp, UseFP) || (!MDIsDataFlowOpnd(DefOp, UseFP))) {
		return true;
	}

	vector<SMPInstr *>::iterator InstIter = this->GetInstIterFromAddr(DefAddr);
	++InstIter;
	while (InstIter != this->GetLastInst()) {
		SMPInstr *NextInst = (*InstIter);
		if (NextInst->FindUse(DefOp) != NextInst->GetLastUse()) { // NextInst uses DefOp
			set<DefOrUse, LessDefUse>::iterator NextDefIter = NextInst->GetFirstNonFlagsDef();
			if (NextDefIter != NextInst->GetLastDef()) {
				// Instruction does something besides define the flags. That does not prove that
				//  our DefOp argument is transferred in any way to the next DEF, because our DefOp
				//  could just be used as an addressing register in NextInst.
				if (NextInst->OperandTransfersValueToDef(DefOp)) {
					STARSOpndTypePtr NextDefOp = NextDefIter->GetOp();
					int NextDefSSANum = NextDefIter->GetSSANum();
					bool RegNextDef = (NextDefOp->IsRegOp());
					int NextDefHashValue;
					if (RegNextDef) {
						NextDefHashValue = HashGlobalNameAndSSA(NextDefOp, NextDefSSANum);
						if (NonEscapingRegisterHashes.find(NextDefHashValue) != NonEscapingRegisterHashes.end()) {
							// We have already determined that NextDef does not reach the end of the block.
							++InstIter;
							continue;
						}
					}
					STARS_ea_t NextDefAddr = NextInst->GetAddr();
					if (this->DoesDefReachBlockEnd(NextDefAddr, NextDefOp, NextDefSSANum, NonEscapingRegisterHashes)) {
						DoesEscape = true;
						break;
					}
					else if (RegNextDef) {
						// Memoize this result to save time reanalyzing it for other def-use chains.
						pair<set<int>::iterator, bool> InsertResult;
						InsertResult = NonEscapingRegisterHashes.insert(NextDefHashValue);
						assert(InsertResult.second);
					}
				}
			}
		} // end if FindUse() succeeds
		if (NextInst->FindDef(DefOp) != NextInst->GetLastDef()) {
			// Re-DEF.
			ReDEF = true;
			break;
		}
		++InstIter;
	}
	if (!(DoesEscape || LocalName || ReDEF)) {
		// Did not find an escape, but we have a global name that reaches the end of the block
		//  without being redefined. If it is LiveOut, it escapes.
		DoesEscape = this->IsLiveOut(DefOp);
	}

	NonEscapingRegisterHashes.clear();

	return DoesEscape;
} // end of SMPBasicBlock::DoesDefReachBlockEnd()

// DefOp/DefSSANum is only used in LoopNum (outside of DefAddr) as address reg or Phi USE
bool SMPBasicBlock::IsDefOnlyUsedAsAddressReg(STARS_ea_t DefAddr, const STARSOpndTypePtr &DefOp, std::size_t LoopNum, std::size_t &AddrUseCounter, SMPOperandType &MemType) {
	bool OnlyAddressReg = false;
	bool ReDefSeen = false;

	if (this->IsProcessed() || (!this->GetFunc()->IsBlockInLoop(this->GetNumber(), LoopNum))) {
		// Must have already passed the test, or is an irrelevant block outside the loop.
		return true;
	}

	if (DefOp->IsRegOp()) {
		OnlyAddressReg = true;
		bool FoundDefAddrInst = false;
		vector<SMPInstr *>::iterator InstIter;
		STARSOpndTypePtr SearchOp = DefOp;
		set<DefOrUse, LessDefUse>::iterator UseIter;
		for (InstIter = this->GetFirstInst(); InstIter != this->GetLastInst(); ++InstIter) {
			SMPInstr *CurrInst = (*InstIter);
			STARS_ea_t InstAddr = CurrInst->GetAddr();
			if (CurrInst->IsMarkerInst())
				continue;
			if (InstAddr != DefAddr) { // Skip induction variable re-definition.
				UseIter = CurrInst->FindUse(SearchOp);
				if (UseIter != CurrInst->GetLastUse()) {
					// Found a USE of DefOp not at DefAddr.
					if (CurrInst->HasDestMemoryOperand() || CurrInst->HasSourceMemoryOperand()) {
						STARSOpndTypePtr MemOp = CurrInst->GetMemUse();
						if ((nullptr == MemOp) || MemOp->IsVoidOp()) {
							MemOp = CurrInst->GetMemDef();
						}
						int BaseReg, IndexReg;
						uint16_t ScaleFactor;
						STARS_ea_t offset;
						MDExtractAddressFields(MemOp, BaseReg, IndexReg, ScaleFactor, offset);
						if ((BaseReg != DefOp->GetReg()) && (IndexReg != DefOp->GetReg())) {
							OnlyAddressReg = false; // USEd somehow, but not as an address reg.
						}
						else {
							++AddrUseCounter;
						}
					}
					else {
						// Definitely not an addressing reg, as there are no memory operands.
						OnlyAddressReg = false;
					}
					if (!OnlyAddressReg)
						break;
				}
#if 1
				else if ((FoundDefAddrInst || (STARS_LIVEIN_PSEUDO_ID == DefAddr)) && (CurrInst->FindDef(SearchOp) != CurrInst->GetLastDef())) {
					// SearchOp is DEFed but not USEd in CurrInst. That means some re-DEF
					//  that does not depend on the previous value, e.g. SearchOp := OtherOp,
					//  in a simple move. No need to search the rest of the block, as there
					//  will be no more USEs of the original SearchOp.
					ReDefSeen = true;
					break;
				}
#endif
			}
			else {
				FoundDefAddrInst = true;
			}
		} // end for all insts
		// If we passed all instructions in block with no problems, recurse into successors.
		this->SetProcessed(true);
		if (OnlyAddressReg && (!ReDefSeen)) {
			list<SMPBasicBlock *>::iterator SuccIter;
			STARS_ea_t PseudoDefAddr; // To identify re-DEF instructions in successor blocks.
			for (SuccIter = this->GetFirstSucc(); SuccIter != this->GetLastSucc(); ++SuccIter) {
				PseudoDefAddr = STARS_LIVEIN_PSEUDO_ID; // any DEF without USE will be seen as a re-DEF in successor.
				if (!(*SuccIter)->IsDefOnlyUsedAsAddressReg(PseudoDefAddr, DefOp, LoopNum, AddrUseCounter, MemType)) {
					OnlyAddressReg = false;
					break;
				}
			}
		}
	}

	if (OnlyAddressReg) {
		if (0 < AddrUseCounter) {
			// NOTE: Future work: extract pointer type of base register above.
			MemType = POINTER;
		}
	}

	return OnlyAddressReg;
} // end of SMPBasicBlock::IsDefOnlyUsedAsAddressReg()


// Is DefOp/DefAddr used in a loop back edge compare-and-branch?
bool SMPBasicBlock::IsDefUsedInLoopCompareAndBranch(STARS_ea_t DefAddr, const STARSOpndTypePtr &DefOp, int DefSSANum, bool &Signed) {
	bool UsedInCompareAndBranch = false;
	bool ReDefSeen = false;
	bool ReDefInvalidates = false;

	if (this->IsProcessed()) {
		// Must have already passed the test.
		return false;
	}

	bool CompareSeen = false;
	vector<SMPInstr *>::iterator InstIter;
	STARSOpndTypePtr SearchOp = CloneIfNecessary(DefOp, this->GetFunc()->UsesFramePointer());
	CanonicalizeOpnd(SearchOp);

	if (STARS_IsLiveInPseudoID(DefAddr)) { // PseudoDefAddr case; look in Phi functions
		set<SMPPhiFunction, LessPhi>::iterator PhiIter = this->FindPhi(SearchOp);
		if (PhiIter != this->GetLastPhi()) {
			DefSSANum = PhiIter->GetDefSSANum();
		}
	}

	set<DefOrUse, LessDefUse>::iterator UseIter;
	bool FoundDefAddrInst = false;
	for (InstIter = this->GetFirstInst(); InstIter != this->GetLastInst(); ++InstIter) {
		SMPInstr *CurrInst = (*InstIter);
		STARS_ea_t InstAddr = CurrInst->GetAddr();
		if (CurrInst->IsMarkerInst())
			continue;
		if (InstAddr == DefAddr) { // Don't waste time before DefAddr.
			FoundDefAddrInst = true;
		}
		else if (!FoundDefAddrInst) {
			UseIter = CurrInst->FindUse(SearchOp);
			bool SearchOpUsed = false;
			if (UseIter != CurrInst->GetLastUse()) {
				// Found a USE of DefOp not at DefAddr.
				if (UseIter->GetSSANum() == DefSSANum) { // SSA numbers match
					SearchOpUsed = true;
					if (CurrInst->MDIsCompareOrTest()) { 
						CompareSeen = true;
						if (!(CurrInst->MDComparesImmedASCII())) { // compare with reliable signedness
							// Does the Flags register DEF reach a branch at the end of this block?
							STARSOpndTypePtr FlagsRegOp = CurrInst->MakeRegOpnd(MD_FLAGS_REG);
							set<DefOrUse, LessDefUse>::iterator DefIter = CurrInst->FindDef(FlagsRegOp);
							assert(DefIter != CurrInst->GetLastDef());
							int FlagSSANum = DefIter->GetSSANum();
							vector<SMPInstr *>::reverse_iterator RevInstIter = this->GetRevInstBegin();
							SMPInstr *LastInst = (*RevInstIter);
							if (COND_BRANCH == LastInst->GetDataFlowType() && (LastInst->MDIsZeroFlagCondBranch())) {
								// Block ends with a conditional branch where a signedness error could be significant.
								set<DefOrUse, LessDefUse>::iterator UseIter2 = LastInst->FindUse(FlagsRegOp);
								assert(UseIter2 != LastInst->GetLastUse());
								if ((UseIter2->GetSSANum() == FlagSSANum) && (this->IsLoopTailBlock() || this->IsLoopHeaderBlock())) { 
									// same flags op, loop back edge
									int LoopNumber = this->GetFunc()->GetLoopNumFromTestBlockNum(this->GetNumber());
									if (LoopNumber >= 0) { // Current block is test and branch block for a top- or bottom-testing loop.
										UsedInCompareAndBranch = true;
										Signed = LastInst->MDIsSignedBranch();
										break;
									}
								}
							}
						}
						break; // stop after first compare
					}
					else if (CurrInst->MDIsMoveInstr()) { // DefOp is copied to another operand.
						// NOTE: We could track two DEFs, or test which one is LiveOut.
						set<DefOrUse, LessDefUse>::iterator MoveDefIter = CurrInst->GetFirstNonFlagsDef();
						STARSOpndTypePtr MoveDefOp = MoveDefIter->GetOp();
						UsedInCompareAndBranch = this->IsDefUsedInLoopCompareAndBranch(InstAddr, MoveDefOp, MoveDefIter->GetSSANum(), Signed);
						if (UsedInCompareAndBranch) {
							break;
						}
					}
				} // end if USE SSANum matches
				set<DefOrUse, LessDefUse>::iterator DefIter2 = CurrInst->FindDef(SearchOp);
				if ((DefIter2 != CurrInst->GetLastDef()) && SearchOpUsed && (!CurrInst->MDIsMoveInstr())) {
					// SearchOp is USEd and re-DEFed in CurrInst, but not in form mov eax,[eax]. Recurse.
					ReDefSeen = true;
					if (CurrInst->MDDestroysSignBit()) {
						ReDefInvalidates = true;
					}
					else {
						DefSSANum = DefIter2->GetSSANum();
						DefAddr = InstAddr;
					}
					break;
				}
			}
			if (!ReDefSeen) {
				// See if we have a re-DEF, whether or not SearchOp was USEd.
				set<DefOrUse, LessDefUse>::iterator DefIter3 = CurrInst->FindDef(SearchOp);
				if (DefIter3 != CurrInst->GetLastDef()) {
					// SearchOp is re-DEFed in CurrInst. Stop recursion.
					ReDefInvalidates = true;
					break;
				}
			}
		}
	} // end for all insts
	if (ReDefSeen && !ReDefInvalidates) { // recurse on new DefSSANum.
		UsedInCompareAndBranch = this->IsDefUsedInLoopCompareAndBranch(DefAddr, DefOp, DefSSANum, Signed);
	}

	// If we passed all instructions in block with no problems, recurse into successors.
	this->SetProcessed(true);

	if (!UsedInCompareAndBranch && (!CompareSeen) && !ReDefInvalidates) {
		list<SMPBasicBlock *>::iterator SuccIter;
		STARS_ea_t PseudoDefAddr; // To identify re-DEF instructions in successor blocks.
		for (SuccIter = this->GetFirstSucc(); SuccIter != this->GetLastSucc(); ++SuccIter) {
			if ((*SuccIter)->IsLiveIn(SearchOp)) {
				PseudoDefAddr = STARS_LIVEIN_PSEUDO_ID; // any DEF without USE will be seen as a re-DEF in successor.
				if ((*SuccIter)->IsDefUsedInLoopCompareAndBranch(PseudoDefAddr, SearchOp, DefSSANum, Signed)) {
					UsedInCompareAndBranch = true;
					break;
				}
			}
		}
	}

	return UsedInCompareAndBranch;
} // end of SMPBasicBlock::IsDefUsedInLoopCompareAndBranch()

// Return true so caller can SetSingleExpression() if block is only an expr for a compare and branch.
// This is useful for detecting compound conditionals in languages with short-circuit evaluation,
//  which we are attempting to translate into SPARK Ada.
bool SMPBasicBlock::FindSingleExpression(void) {
	bool FoundSingleExpr = false;
	vector<SMPInstr *>::reverse_iterator RevInstIter = this->GetRevInstBegin();
	SMPInstr *LastInst = (*RevInstIter);
	SMPitype FlowType = LastInst->GetDataFlowType();
	if (FlowType == COND_BRANCH) {
		// We are looking for expressions used in conditional branches, without any extraneous instructions
		//  that cannot be expressed in SPARK-Ada expressions.
		//  So, the block should not store results other than in the registers used for the expression.
		//  If the original C code were:  if (((x * 4) < (z * 3)) || ((x * 5) > (y * 3))) ...
		//  then we will have a block for the first test that might look like:
		//  mov eax, x
		//  sal eax, 2    ; x * 4
		//  mov ebx, z
		//  mov edx, ebx  ; make copy of z
		//  sal ebx, 1    ; z * 2
		//  add ebx, edx  ; z * 3
		//  cmp eax, ebx
		//  jl  if-then-body
		//
		//  Then we fall through to the second test if we failed the first test:
		//  mov eax, x
		//  mov edx, eax  ; make copy of x
		//  sal eax, 2    ; x * 4
		//  add eax, edx  ; x * 5
		//  mov ebx, y
		//  mov edx, ebx  ; copy of y
		//  sal ebx, 1    ; y * 2
		//  add ebx, edx  ; y * 3
		//  cmp eax, ebx
		//  jle skip-then-body  ; invert condition from > to <=
		//
		//  Then we fall through to if-then-body, having failed the first test and passed the second.
		//  From these patterns, we can see several things:
		//  1. All of the computations in the second test relate to the compare and branch at the end.
		//  2. The basic block for the first test could have preceding instructions that are not shown,
		//     which do not form part of the first test, so not all of that block's computations can be
		//     traced through the SSA chains to the compare instruction.
		//  3. Note that instruction scheduling could interleave the EAX and EBX computation chains above,
		//     so we must follow SSA chains and not assume the simplest sequences.
		// Therefore, if we can trace each DEF in a block to the uses of the compare or test instruction
		//  and its dependent conditional branch, then the block is the kind of block that could be nothing
		//  but a short-circuited test that is part of a larger if-then test expression. Final determination
		//  of whether the block should be translated as a short-circuit evaluation will be made later based on
		//  the context within the CFG, e.g. do consecutive blocks lead to the if-then-body and skip-then-body
		//  labels above. Short-circuited AND tests produce a different CFG pattern than short-circuited OR tests, etc.
		//  Recognition of these patterns happens at the SMPFunction level.
		//
		// We will analyze the block by walking backwards from the conditional jump and compare/test. As we see SSA
		//  names that definitely are part of defining the operands of the compare/test, we will add them to a set.
		//  Any prior instructions that define names in that set will have their USEs added to the set. If we encounter
		//  a DEF that is not in the set of SSA names, we have failed. If we reach the beginning of the block without
		//  failure, we have succeeded.
		set<DefOrUse, LessDefUse>::iterator UseIter = LastInst->GetFirstUse();
		assert(UseIter != LastInst->GetLastUse()); // COND_BRANCH uses at least one flag register
		STARSOpndTypePtr UseOp = UseIter->GetOp();
		assert(UseOp->IsRegOp()); // flag reg
		set<int> RelevantRegDEFs;
		pair<set<int>::iterator, bool> RegInsertResult;
		int UseSSANum = UseIter->GetSSANum();
		int UseSSAIndex = HashGlobalNameAndSSA(UseOp, UseSSANum);
		RegInsertResult = RelevantRegDEFs.insert(UseSSAIndex); // whatever DEFs the flags register is relevant
		assert(RegInsertResult.second);

		// Now, reverse iterate through the instructions and see if each DEF is USEd by an already relevant DEF.
		++RevInstIter; // move up from the COND_BRANCH
		bool UseFP = this->GetFunc()->UsesFramePointer();
		while (RevInstIter != this->GetRevInstEnd()) {
			SMPInstr *CurrInst = (*RevInstIter);
			++RevInstIter;
			for (set<DefOrUse, LessDefUse>::iterator DefIter = CurrInst->GetFirstDef(); DefIter != CurrInst->GetLastDef(); ++DefIter) {
				STARSOpndTypePtr DefOp = DefIter->GetOp();
				if (!DefOp->IsRegOp()) { // cannot translate as pure expression if it stores to memory
					return false;
				}
				FoundSingleExpr = true; // until we see a problem
				if ((!CurrInst->MDIsCompareOrTest()) && (DefOp->MatchesReg(MD_FLAGS_REG))) {
					// Not every DEF of flags reg must be relevant. Only the compare or test inst
					//  necessarily has a relevant flags DEF.
					continue;  // on to next DEF
				}
				int DefSSANum = DefIter->GetSSANum();
				int DefSSAIndex = HashGlobalNameAndSSA(DefOp, DefSSANum);
				set<int>::iterator RegIter = RelevantRegDEFs.find(DefSSAIndex);
				if (RegIter == RelevantRegDEFs.end()) { // DEF is not relevant
					return false;
				}
				else { // DEF is relevant; all USEs now become relevant
					for (UseIter = CurrInst->GetFirstUse(); UseIter != CurrInst->GetLastUse(); ++UseIter) {
						UseOp = UseIter->GetOp();
						// We can USE operands beside registers, but we will return false if we ever see a DEF
						//  of memory, so we do not attempt to track memory operands.
						if (UseOp->IsRegOp()) {
							UseSSANum = UseIter->GetSSANum();
							UseSSAIndex = HashGlobalNameAndSSA(UseOp, UseSSANum);
							RegInsertResult = RelevantRegDEFs.insert(UseSSAIndex); // any DEF of a relevant USE is a relevant DEF
						}
					} // end for all USEs
				}
			} // end for all DEFs
		} // end while reverse inst iteration
	} // end if COND_BRANCH for LastInst

	return FoundSingleExpr;
} // end of SMPBasicBlock::FindSingleExpression()

// erase() block starting at FirstAddr from Preds list
void SMPBasicBlock::ErasePred(STARS_ea_t FirstBlockAddr) {
	list<SMPBasicBlock *>::iterator PredIter;
	PredIter = this->Predecessors.begin();
	while (PredIter != this->Predecessors.end()) {
		if ((*PredIter)->GetFirstAddr() == FirstBlockAddr) {
			PredIter = this->Predecessors.erase(PredIter);
			break; // can only appear once
		}
		else
			++PredIter;
	}
	return;
} // end of SMPBasicBlock::ErasePred()

// erase() block starting at FirstAddr from Successors list
void SMPBasicBlock::EraseSucc(STARS_ea_t FirstBlockAddr) {
	list<SMPBasicBlock *>::iterator SuccIter;
	SuccIter = this->Successors.begin();
	while (SuccIter != this->Successors.end()) {
		if ((*SuccIter)->GetFirstAddr() == FirstBlockAddr) {
			SuccIter = this->Successors.erase(SuccIter);
			break; // can only appear once
		}
		else
			++SuccIter;
	}
	return;
} // end of SMPBasicBlock::EraseSucc()

STARSOpndSetIter SMPBasicBlock::EraseLiveIn(STARSOpndSetIter LiveInIter) {
	return this->LiveInSet.erase(LiveInIter);
}

STARSOpndSetIter SMPBasicBlock::EraseLiveOut(STARSOpndSetIter LiveOutIter) {
	return this->LiveOutSet.erase(LiveOutIter);
}

STARSOpndSetIter SMPBasicBlock::EraseUpExposed(STARSOpndSetIter UpExposedIter) {
	return this->UpExposedSet.erase(UpExposedIter);
}

STARSOpndSetIter SMPBasicBlock::EraseVarKill(STARSOpndSetIter VarKillIter) {
	return this->LiveInSet.erase(VarKillIter);
}

// Add ReachInDefn to the ReachesIn set. If it was not already present,
//  then update the ReachesOutStale flag.
void SMPBasicBlock::AddReachesIn(pair<STARSOpndTypePtr, STARS_ea_t> ReachInDefn) { 
	pair<STARSDefinitionSet::iterator, bool> InsertResult;
	InsertResult = ReachesInSet.insert(ReachInDefn);
	if (InsertResult.second) {
		// New element added; ReachesOut set might be stale.
		this->SetReachesOutStale();
	}
} // end of SMPBasicBlock::AddReachesIn()

// Add a new definition to the ReachesOut set. If it was not already
//  present in the ReachesOut set, then propagate it to the ReachesIn sets
//  of all successor blocks. NOTE: The policy here is to keep the ReachesIn
//  sets up to date at all times.
void SMPBasicBlock::AddReachesOut(pair<STARSOpndTypePtr, STARS_ea_t> ReachOutDefn) { 
	pair<STARSDefinitionSet::iterator, bool> InsertResult;
	InsertResult = ReachesOutSet.insert(ReachOutDefn); 
	if (InsertResult.second) {
		// Insertion was of a new element; propagate to ReachesIn of all successors.
		list<SMPBasicBlock *>::iterator SuccIter;
		for (SuccIter = this->GetFirstSucc(); SuccIter != this->GetLastSucc(); ++SuccIter) {
			(*SuccIter)->AddReachesIn(ReachOutDefn);
		}
	}
	return;
} // end of SMPBasicBlock::AddReachesOut()

// Erase a phi function from the list.
bool SMPBasicBlock::ErasePhi(const STARSOpndTypePtr &OldOp) {
	set<SMPPhiFunction, LessPhi>::iterator PhiIter;
	PhiIter = this->FindPhi(OldOp);
	if (PhiIter == this->PhiFunctions.end())
		return false; // did not find, cannot erase
	this->PhiFunctions.erase(PhiIter);
	return true;
} // end of SMPBasicBlock::ErasePhi()

// Add a phi function to the list of phi functions entering this block.
// If phi function for this global name already existed in the block,
//   return false because no new phi function was added; else return true.
bool SMPBasicBlock::AddPhi(SMPPhiFunction NewPhi) {
	if (this->PhiFunctions.end() == this->PhiFunctions.find(NewPhi)) {
		this->PhiFunctions.insert(NewPhi);
		return true;
	}
	else
		return false;
} // end of SMPBasicBlock::AddPhi()

// Change the SSA number of the Phi function DEF corresponding to DefOp.
bool SMPBasicBlock::SetPhiDefSSA(const STARSOpndTypePtr &DefOp, int SSANum) {
	bool updated = false;
	set<SMPPhiFunction, LessPhi>::iterator CurrPhi;
	CurrPhi = this->FindPhi(DefOp);
	if (CurrPhi == this->PhiFunctions.end()) { // did not find
		SMP_msg("ERROR: SetPhiDefSSA for block %d did not find DefOp: ", this->GetNumber());
		PrintOperand(DefOp);
		SMP_msg("\n");
	}
	else { // found it
		// To change a set item, we must change a copy, remove the original item
		//  from the set, and then insert the modified copy of the item.
		SMPPhiFunction TempPhi = (*CurrPhi);
		TempPhi.SetSSADef(SSANum);
		bool erased = this->ErasePhi(DefOp);
		assert(erased);
		updated = this->AddPhi(TempPhi);
		assert(updated);
	}
	return updated;
} // end of SMPBasicBlock::SetPhiDefSSA()

// Change the SSA number of the Phi function USE corresponding to DefOp and index.
bool SMPBasicBlock::SetPhiUseSSA(const STARSOpndTypePtr &DefOp, std::size_t index, int SSANum) {
	bool updated = false;
	set<SMPPhiFunction, LessPhi>::iterator CurrPhi;
	CurrPhi = this->FindPhi(DefOp);
	if (CurrPhi == this->PhiFunctions.end()) { // did not find
		SMP_msg("ERROR: SetPhiUseSSA for block %d did not find DefOp: ", this->GetNumber());
		PrintOperand(DefOp);
		SMP_msg("\n");
	}
	else { // found it
		// To change a set item, we must change a copy, remove the original item
		//  from the set, and then insert the modified copy of the item.
		SMPPhiFunction TempPhi = (*CurrPhi);
		TempPhi.SetSSARef(index, SSANum);
		bool erased = this->ErasePhi(DefOp);
		assert(erased);
		updated = this->AddPhi(TempPhi);
		assert(updated);
	}
	return updated;
} // end of SMPBasicBlock::SetPhiUseSSA()

// Change the SMP type of the Phi function DEF corresponding to DefOp.
set<SMPPhiFunction, LessPhi>::iterator SMPBasicBlock::SetPhiDefType(const STARSOpndTypePtr &DefOp, SMPOperandType Type) {
	bool updated = false;
	set<SMPPhiFunction, LessPhi>::iterator CurrPhi;
	CurrPhi = this->FindPhi(DefOp);
	if (CurrPhi == this->PhiFunctions.end()) { // did not find
		SMP_msg("ERROR: SetPhiDefType for block %d did not find DefOp: ", this->GetNumber());
		PrintOperand(DefOp);
		SMP_msg("\n");
	}
	else { // found it
		// To change a set item, we must change a copy, remove the original item
		//  from the set, and then insert the modified copy of the item.
		SMPPhiFunction TempPhi = (*CurrPhi);
		const SMPInstr *FirstInst = (*(this->GetFirstInst()));
		TempPhi.SetDefType(Type, FirstInst);
		bool erased = this->ErasePhi(DefOp);
		assert(erased);
		pair<set<SMPPhiFunction, LessPhi>::iterator, bool> InsertResult;
		InsertResult = this->PhiFunctions.insert(TempPhi);
		CurrPhi = InsertResult.first;
		updated = InsertResult.second;
		assert(updated);
	}
	return CurrPhi;
} // end of SMPBasicBlock::SetPhiDefType()

// Change the SMP type of the Phi function USE corresponding to DefOp and index.
bool SMPBasicBlock::SetPhiUseType(const STARSOpndTypePtr &DefOp, std::size_t index, SMPOperandType Type) {
	bool updated = false;
	set<SMPPhiFunction, LessPhi>::iterator CurrPhi;
	CurrPhi = this->FindPhi(DefOp);
	if (CurrPhi == this->PhiFunctions.end()) { // did not find
		SMP_msg("ERROR: SetPhiUseType for block %d did not find DefOp: ", this->GetNumber());
		PrintOperand(DefOp);
		SMP_msg("\n");
	}
	else { // found it
		// To change a set item, we must change a copy, remove the original item
		//  from the set, and then insert the modified copy of the item.
		SMPPhiFunction TempPhi = (*CurrPhi);
		const SMPInstr *FirstInst = (*(this->GetFirstInst()));
		TempPhi.SetRefType(index, Type, FirstInst);
		bool erased = this->ErasePhi(DefOp);
		assert(erased);
		updated = this->AddPhi(TempPhi);
		assert(updated);
	}
	return updated;
} // end of SMPBasicBlock::SetPhiUseType()

std::set<SMPPhiFunction, LessPhi>::iterator SMPBasicBlock::SetPhiDefMaybeAliased(const STARSOpndTypePtr &DefOp) {
	bool updated = false;
	set<SMPPhiFunction, LessPhi>::iterator CurrPhi;
	CurrPhi = this->FindPhi(DefOp);
	if (CurrPhi == this->PhiFunctions.end()) { // did not find
		SMP_msg("ERROR: SetPhiDefType for block %d did not find DefOp: ", this->GetNumber());
		PrintOperand(DefOp);
		SMP_msg("\n");
	}
	else { // found it
		// To change a set item, we must change a copy, remove the original item
		//  from the set, and then insert the modified copy of the item.
		SMPPhiFunction TempPhi = (*CurrPhi);
		TempPhi.SetDefMaybeAliased();
		bool erased = this->ErasePhi(DefOp);
		assert(erased);
		pair<set<SMPPhiFunction, LessPhi>::iterator, bool> InsertResult;
		InsertResult = this->PhiFunctions.insert(TempPhi);
		CurrPhi = InsertResult.first;
		updated = InsertResult.second;
		assert(updated);
	}
	return CurrPhi;
} // end of SMPBasicBlock::SetPhiDefMaybeAliased()

bool SMPBasicBlock::SetPhiRefMaybeAliased(const STARSOpndTypePtr &DefOp, std::size_t index) {
	bool updated = false;
	set<SMPPhiFunction, LessPhi>::iterator CurrPhi;
	CurrPhi = this->FindPhi(DefOp);
	if (CurrPhi == this->PhiFunctions.end()) { // did not find
		SMP_msg("ERROR: SetPhiUseType for block %d did not find DefOp: ", this->GetNumber());
		PrintOperand(DefOp);
		SMP_msg("\n");
	}
	else { // found it
		// To change a set item, we must change a copy, remove the original item
		//  from the set, and then insert the modified copy of the item.
		SMPPhiFunction TempPhi = (*CurrPhi);
		TempPhi.SetRefMaybeAliased(index);
		bool erased = this->ErasePhi(DefOp);
		assert(erased);
		updated = this->AddPhi(TempPhi);
		assert(updated);
	}
	return updated;
} // end of SMPBasicBlock::SetPhiRefMaybeAliased()

// Change the metadata type of the Phi function DEF corresponding to DefOp.
set<SMPPhiFunction, LessPhi>::iterator SMPBasicBlock::SetPhiDefMetadata(const STARSOpndTypePtr &DefOp, SMPMetadataType Status) {
	bool updated = false;
	set<SMPPhiFunction, LessPhi>::iterator CurrPhi;
	CurrPhi = this->FindPhi(DefOp);
	if (CurrPhi == this->PhiFunctions.end()) { // did not find
		SMP_msg("ERROR: SetPhiDefMetadata for block %d did not find DefOp: ", this->GetNumber());
		PrintOperand(DefOp);
		SMP_msg("\n");
	}
	else { // found it
		// To change a set item, we must change a copy, remove the original item
		//  from the set, and then insert the modified copy of the item.
		SMPPhiFunction TempPhi = (*CurrPhi);
		TempPhi.SetDefMetadata(Status);
		bool erased = this->ErasePhi(DefOp);
		assert(erased);
		pair<set<SMPPhiFunction, LessPhi>::iterator, bool> InsertResult;
		InsertResult = this->PhiFunctions.insert(TempPhi);
		CurrPhi = InsertResult.first;
		updated = InsertResult.second;
		assert(updated);
	}
	return CurrPhi;
} // end of SMPBasicBlock::SetPhiDefMetadata()

// Insert SCCP value for local name; change if already found.
map<int, struct STARS_SCCP_Const_Struct>::iterator SMPBasicBlock::InsertLocalConstValue(int DefHashValue, struct STARS_SCCP_Const_Struct NewConstEntry) {
	map<int, struct STARS_SCCP_Const_Struct>::iterator MapIter = this->FindLocalConstValue(DefHashValue);
	if (MapIter == this->GetLastLocalConstValueIter()) { // no old entry; insert
		pair<int, struct STARS_SCCP_Const_Struct> InsertPair(DefHashValue, NewConstEntry);
		pair<map<int, struct STARS_SCCP_Const_Struct>::iterator, bool> InsertResult = this->ConstantLocalDefs.insert(InsertPair);
		assert(InsertResult.second);
		MapIter = InsertResult.first;
	}
	else { // old entry found; update
		MapIter->second = NewConstEntry;
	}
	return MapIter;
} // end of SMPBasicBlock::InsertLocalConstValue()

// Six methods to get values into the maps of local reg/SSA to FG info.
//  For global names, see corresponding methods in SMPFunction.
void SMPBasicBlock::UpdateDefSignMiscInfo(int DefHashValue, unsigned short NewInfo) {
	map<int, struct FineGrainedInfo>::iterator MapIter;
	pair< map<int, struct FineGrainedInfo>::iterator, bool> MapResult;

	MapIter = this->LocalDefFGInfoBySSA.find(DefHashValue);
	if (MapIter == this->LocalDefFGInfoBySSA.end()) {
		// Not found; insert first.
		struct FineGrainedInfo NewFGInfo;
		NewFGInfo.SignMiscInfo = NewInfo;
		NewFGInfo.SizeInfo = 0;
		pair<int, struct FineGrainedInfo> MapItem(DefHashValue, NewFGInfo);
		MapResult = this->LocalDefFGInfoBySSA.insert(MapItem);
		assert(MapResult.second); // Was not previously found, insertion must work.
	}
	else { // found; just OR in the new bits.
		MapIter->second.SignMiscInfo |= NewInfo;
	}

	return;
} // end of SMPBasicBlock::UpdateDefSignMiscInfo()

void SMPBasicBlock::UpdateUseSignMiscInfo(int UseHashValue, unsigned short NewInfo) {
	map<int, struct FineGrainedInfo>::iterator MapIter;
	pair<map<int, struct FineGrainedInfo>::iterator, bool> MapResult;

	MapIter = this->LocalUseFGInfoBySSA.find(UseHashValue);
	if (MapIter == this->LocalUseFGInfoBySSA.end()) {
		// Not found; insert first.
		struct FineGrainedInfo NewFGInfo;
		NewFGInfo.SignMiscInfo = NewInfo;
		NewFGInfo.SizeInfo = 0;
		pair<int, struct FineGrainedInfo> MapItem(UseHashValue, NewFGInfo);
		MapResult = this->LocalUseFGInfoBySSA.insert(MapItem);
		assert(MapResult.second); // Was not previously found, insertion must work.
	}
	else { // found; just OR in the new bits.
		MapIter->second.SignMiscInfo |= NewInfo;
	}

	return;
} // end of SMPBasicBlock::UpdateUseSignMiscInfo()

void SMPBasicBlock::UpdateDefWidthTypeInfo(int DefHashValue, unsigned short NewInfo) {
	map<int, struct FineGrainedInfo>::iterator MapIter;
	pair<map<int, struct FineGrainedInfo>::iterator, bool> MapResult;

	MapIter = this->LocalDefFGInfoBySSA.find(DefHashValue);
	if (MapIter == this->LocalDefFGInfoBySSA.end()) {
		// Not found; insert first.
		struct FineGrainedInfo NewFGInfo;
		NewFGInfo.SignMiscInfo = 0;
		NewFGInfo.SizeInfo = NewInfo;
		pair<int, struct FineGrainedInfo> MapItem(DefHashValue, NewFGInfo);
		MapResult = this->LocalDefFGInfoBySSA.insert(MapItem);
		assert(MapResult.second); // Was not previously found, insertion must work.
	}
	else { // found; just OR in the new bits.
		MapIter->second.SizeInfo |= NewInfo;
	}

	return;
} // end of SMPBasicBlock::UpdateDefWidthTypeInfo()

void SMPBasicBlock::UpdateUseWidthTypeInfo(int UseHashValue, unsigned short NewInfo) {
	map<int, struct FineGrainedInfo>::iterator MapIter;
	pair<map<int, struct FineGrainedInfo>::iterator, bool> MapResult;

	MapIter = this->LocalUseFGInfoBySSA.find(UseHashValue);
	if (MapIter == this->LocalUseFGInfoBySSA.end()) {
		// Not found; insert first.
		struct FineGrainedInfo NewFGInfo;
		NewFGInfo.SignMiscInfo = 0;
		NewFGInfo.SizeInfo = NewInfo;
		pair<int, struct FineGrainedInfo> MapItem(UseHashValue, NewFGInfo);
		MapResult = this->LocalUseFGInfoBySSA.insert(MapItem);
		assert(MapResult.second); // Was not previously found, insertion must work.
	}
	else { // found; just OR in the new bits.
		MapIter->second.SizeInfo |= NewInfo;
	}

	return;
} // end of SMPBasicBlock::UpdateUseWidthTypeInfo()

void SMPBasicBlock::UpdateDefFGInfo(int DefHashValue, struct FineGrainedInfo NewFG) {
	map<int, struct FineGrainedInfo>::iterator MapIter;
	pair<map<int, struct FineGrainedInfo>::iterator, bool> MapResult;

	MapIter = this->LocalDefFGInfoBySSA.find(DefHashValue);
	if (MapIter == this->LocalDefFGInfoBySSA.end()) {
		// Not found; insert it.
		pair<int, struct FineGrainedInfo> MapItem(DefHashValue, NewFG);
		MapResult = this->LocalDefFGInfoBySSA.insert(MapItem);
		assert(MapResult.second); // Was not previously found, insertion must work.
	}
	else { // found; just put in the new bits.
		MapIter->second.SignMiscInfo |= NewFG.SignMiscInfo;
		MapIter->second.SizeInfo |= NewFG.SizeInfo;
	}

	return;
} // end of SMPBasicBlock::UpdateDefFGInfo()

void SMPBasicBlock::UpdateUseFGInfo(int UseHashValue, struct FineGrainedInfo NewFG) {
	map<int, struct FineGrainedInfo>::iterator MapIter;
	pair<map<int, struct FineGrainedInfo>::iterator, bool> MapResult;

	MapIter = this->LocalUseFGInfoBySSA.find(UseHashValue);
	if (MapIter == this->LocalUseFGInfoBySSA.end()) {
		// Not found; insert it.
		pair<int, struct FineGrainedInfo> MapItem(UseHashValue, NewFG);
		MapResult = this->LocalUseFGInfoBySSA.insert(MapItem);
		assert(MapResult.second); // Was not previously found, insertion must work.
	}
	else { // found; just put in the new bits.
		MapIter->second.SignMiscInfo |= NewFG.SignMiscInfo;
		MapIter->second.SizeInfo |= NewFG.SizeInfo;
	}

	return;
} // end of SMPBasicBlock::UpdateUseFGInfo()

// Reset the signedness bits to zero for DEF.
void SMPBasicBlock::ClearDefSignedness(int DefHashValue) {
	map<int, struct FineGrainedInfo>::iterator MapIter;

	MapIter = this->LocalDefFGInfoBySSA.find(DefHashValue);
	if (MapIter != this->LocalDefFGInfoBySSA.end()) {
		MapIter->second.SignMiscInfo &= (~FG_MASK_SIGNEDNESS_BITS);
	}
	return;
} // end of SMPBasicBlock::ClearDefSignedness()

// initialize InArgPointerCopies container
void SMPBasicBlock::InitPointerCopies(std::size_t LocalSSALimit) {
	size_t LocalNameCount = this->LocalNames.size();
	for (size_t index = 0; index < LocalNameCount; ++index) {
		STARSBitSet TempBitSet;
		TempBitSet.AllocateBits(LocalSSALimit);
		this->InArgPointerCopies.push_back(TempBitSet);
	}
	return;
} // end of SMPBasicBlock::InitPointerCopies()

// Set in local InArgPointerCopies
void SMPBasicBlock::SetInArgPointerCopy(const STARSOpndTypePtr &CurrOp, int SSANum) {
	STARSOpndSetIter NameIter = this->LocalNames.find(CurrOp);
	if ((NameIter != this->LocalNames.end()) && (0 <= SSANum)) {
		unsigned int LocalNameIndex = ExtractGlobalIndex(*NameIter);
		assert(LocalNameIndex < this->LocalNames.size());
		this->InArgPointerCopies[LocalNameIndex].SetBit((size_t) SSANum);
	}
	return;
}

// Ensure that the Phi DEFs have all the fine grained info from all Phi USEs.
void SMPBasicBlock::PropagatePhiFGInfo(void) {
	set<SMPPhiFunction, LessPhi>::iterator CurrPhi;
	size_t PhiUseIndex;
	struct FineGrainedInfo UseFG;

	for (CurrPhi = this->GetFirstPhi(); CurrPhi != this->GetLastPhi(); ++CurrPhi) {
		STARSOpndTypePtr PhiOp = CurrPhi->GetAnyOp();
		int DefSSANum = CurrPhi->GetDefSSANum();
		if (!MDIsGeneralPurposeReg(PhiOp))   // !!!!****!!!! Add stack locations also
			continue;
		int DefHashValue = HashGlobalNameAndSSA(PhiOp, DefSSANum);
		for (PhiUseIndex = 0; PhiUseIndex < CurrPhi->GetPhiListSize(); ++PhiUseIndex) {
			int UseSSANum = CurrPhi->GetUseSSANum(PhiUseIndex);
			int UseHashValue = HashGlobalNameAndSSA(PhiOp, UseSSANum);
			UseFG = this->GetFunc()->GetUseFGInfo(UseHashValue);
#if 1
			if ((UseFG.SignMiscInfo == 0) && (UseFG.SizeInfo == 0)) {
				// If no info for USE, get info from DEF for that USE.
				UseFG = this->GetFunc()->GetDefFGInfo(UseHashValue);
			}
#endif
			this->GetFunc()->UpdateDefFGInfo(DefHashValue, UseFG);
		}
	}

	return;
} // end of SMPBasicBlock::PropagatePhiFGInfo()

// Find consecutive DEFs of the same name within the block, both having
//  live metadata, that have the same type, making the second DEF
//  redundant metadata.
bool SMPBasicBlock::FindRedundantLocalMetadata(bool SafeFunc) {
	bool changed = false;
	bool CallInst;
	bool UseFP = this->GetFunc()->UsesFramePointer();
	bool EligibleOperand;
	bool LeafFunc = this->GetFunc()->IsLeaf();
	STARSOpndSetIter NameIter;
	vector<SMPInstr *>::iterator InstIter;
	SMPInstr *CurrInst;
	set<DefOrUse, LessDefUse>::iterator CurrDef;
	size_t LocalNameLimit = this->LocalNames.size();
	size_t GlobalNameLimit = this->GetFunc()->NumGlobalNames();
	vector<SMPOperandType> LocalTypes;
	vector<SMPOperandType> GlobalTypes;

	// Initialize the most recently seen type for each possible DEF.
	for (size_t i = 0; i < LocalNameLimit; ++i)
		LocalTypes.push_back(UNINIT);
	for (size_t i = 0; i < GlobalNameLimit; ++i)
		GlobalTypes.push_back(UNINIT);

	for (InstIter = this->GetFirstInst(); InstIter != this->GetLastInst(); ++InstIter) {
		CurrInst = (*InstIter);
		STARS_ea_t InstAddr = CurrInst->GetAddr();
		if (CurrInst->IsMarkerInst())
			continue; // skip SSA marker instruction; has no normal DEFs
		CallInst = ((CurrInst->GetDataFlowType() == CALL)
			|| (CurrInst->GetDataFlowType() == INDIR_CALL));
		for (CurrDef = CurrInst->GetFirstDef(); CurrDef != CurrInst->GetLastDef(); ++CurrDef) {
			STARSOpndTypePtr DefOp = CurrDef->GetOp();
			// We limit our analysis to register DEFs, or stack memory DEFs
			//  in SAFE functions. ***!!!*** Could add o_mem direct accesses
			//  in SAFE functions.
			EligibleOperand = ((DefOp->IsRegOp()) 
				&& !(MDIsStackOrFramePointerReg(DefOp, UseFP)));
			if (!EligibleOperand && (! DefOp->IsRegOp()) && SafeFunc) {
				// Make some exceptions for SAFE functions.
				EligibleOperand = MDIsDirectStackAccessOpnd(DefOp, UseFP)
					|| (LeafFunc && !MDIsIndirectMemoryOpnd(DefOp, UseFP));
			}
			if (!EligibleOperand)
				continue;
			// If the metadata is not live for the DEF it is irrelevant
			//  to redundancy analysis. We make an exception for the DEFs
			//  of caller-saved regs at a CALL instruction, because we have
			//  no idea what might have happened to the metadata in the callee.
			int TempInt = (int) CurrDef->GetMetadataStatus();
			TempInt &= (int) DEF_METADATA_USED;
			if ((0 == TempInt) && (!(CallInst && MDIsCallerSavedReg(DefOp))))
				continue;
			bool GlobalName = false;
			NameIter = this->GetFunc()->FindGlobalName(DefOp);
			unsigned int NameIndex;
			SMPOperandType CurrType = CurrDef->GetType();
			SMPOperandType OldType;
			if (NameIter != this->GetFunc()->GetLastGlobalName()) {
				// global name
				GlobalName = true;
				NameIndex = ExtractGlobalIndex(*NameIter);
				OldType = GlobalTypes.at(NameIndex); 
			}
			else { // local name
				NameIter = this->FindLocalName(DefOp);
				assert(NameIter != this->GetLastLocalName());
				NameIndex = ExtractGlobalIndex(*NameIter);
				OldType = LocalTypes.at(NameIndex); 
			}
			if (IsEqType(CurrType, OldType)) {
				// types agree; only redundant if NUMERIC or pointers with
				//  the same referent, which we cannot determine statically,
				//  so we stick to NUMERIC redundancy only.
				if (IsNumeric(CurrType)) {
					// Metadata is redundant; profiler-dependent or not?
					bool ProfDerived = IsProfDerived(CurrType) || IsProfDerived(OldType);
					if (ProfDerived) {
						CurrDef = CurrInst->SetDefMetadata(DefOp, DEF_METADATA_PROF_REDUNDANT);
					}
					else {
						CurrDef = CurrInst->SetDefMetadata(DefOp, DEF_METADATA_REDUNDANT);
					}
					changed = true;
				}
			} // end if current type same as old type
			else {
				// Update latest type seen.
				if (GlobalName)
					GlobalTypes[NameIndex] = CurrType;
				else
					LocalTypes[NameIndex] = CurrType;
			}
		} // end for all DEFs in current instruction
	} // end for all instructions

	LocalTypes.clear();
	GlobalTypes.clear();
	return changed;
} // end of SMPBasicBlock::FindRedundantLocalMetadata()

// Does block infinitely loop back to itself?
bool SMPBasicBlock::IsInfiniteSelfLoop(void) {
	bool InfiniteLoop = false;
	size_t NumSucc = this->GetNumSuccessors();
	if (1 == NumSucc) {
		// Is the only successor this block, and there are no function calls?
		int MyNum = this->GetNumber();
		list<SMPBasicBlock *>::iterator SuccIter = this->GetFirstSucc();
		int SuccNum = (*SuccIter)->GetNumber();
		if ((SuccNum == MyNum) && (!(this->HasCallInstruction() || this->HasIndirectJump()))) {
			InfiniteLoop = true;
		}
	}
	return InfiniteLoop;
} // end of SMPBasicBlock::IsInfiniteSelfLoop()

// Does block end with a conditional branch?
bool SMPBasicBlock::HasConditionalBranch(void) const {
	vector<SMPInstr *>::const_reverse_iterator InstIter = this->InstVec.crbegin();
	SMPInstr *LastInst = (*InstIter);

	bool HasCondBranch = (LastInst->GetDataFlowType() == COND_BRANCH);

	return HasCondBranch;
} // end of SMPBasicBlock::HasConditionalBranch()

// Writes to stack or has indirect write
bool SMPBasicBlock::HasMemoryWrite(void) const {
	bool HasMemWrite = this->HasIndirectMemWrite();
	if (!HasMemWrite) {
		STARSOpndSetConstIter KillIter = this->KillSet.cbegin();
		while (KillIter != this->KillSet.cend()) {
			STARSOpndTypePtr KillOp = (*KillIter);
			if (KillOp->IsMemOp()) {
				HasMemWrite = true;
				break;
			}
			++KillIter;
		}
	}
	return HasMemWrite;
} // end of SMPBasicBlock::HasMemoryWrite()

// found in local InArgPointerCopies?
bool SMPBasicBlock::IsInArgPointerCopy(const STARSOpndTypePtr &CurrOp, int SSANum) const {
	bool InArgCopy = false;
	STARSOpndSetIter NameIter = this->LocalNames.find(CurrOp);
	if ((NameIter != this->LocalNames.end()) && (0 <= SSANum)) {
		unsigned int LocalNameIndex = ExtractGlobalIndex(*NameIter);
		assert(LocalNameIndex < this->LocalNames.size());
		InArgCopy = this->InArgPointerCopies[LocalNameIndex].GetBit((size_t) SSANum);
	}
	return InArgCopy;
}

// We don't care if we overflow computing some DEFs, because of how they are used.
//  Return true when we detect one of these cases.
bool SMPBasicBlock::IsBenignOverflowDEF(const STARSOpndTypePtr &DefOp, int DefSSANum, STARS_ea_t DefAddr, bool UnderflowOpcode, int &IdiomCode) {
	bool benign = false;
	vector<SMPInstr *>::iterator InstIter;
	set<DefOrUse, LessDefUse>::iterator UseIter, DefIter;
	bool RegDef = (DefOp->IsRegOp());
	struct FineGrainedInfo DefFGInfo;
	bool LocalDefName = this->IsLocalName(DefOp);
	bool UseFP = this->GetFunc()->UsesFramePointer();

	if (RegDef) {
		unsigned int DefHashValue = HashGlobalNameAndSSA(DefOp, DefSSANum);
		if (LocalDefName) {
			// Local name, find in basic block maps.
			DefFGInfo = this->GetDefFGInfo(DefHashValue);
		}
		else { // Global name, find in global maps.
			DefFGInfo = this->GetFunc()->GetDefFGInfo(DefHashValue);
		}
	}
	else if (MDIsDirectStackAccessOpnd(DefOp, UseFP)) {
		bool success = this->GetFunc()->MDGetFGStackLocInfo(DefAddr, DefOp, DefFGInfo);
		assert(success);
	}
	else {
		return false; // we can only handle register and stack DefOp types.
	}
	bool DefIsUnsigned = (FG_MASK_UNSIGNED == (DefFGInfo.SignMiscInfo & FG_MASK_SIGNEDNESS_BITS));
	size_t DefBitWidth = LargestBitWidthFromMask(DefFGInfo.SizeInfo);

	// First, we find the USEs of the DEF.
	if (RegDef) {
		bool FoundDefAddrInst = false;
		for (InstIter = this->GetFirstInst(); InstIter != this->GetLastInst(); ++InstIter) {
			SMPInstr *CurrInst = (*InstIter);
			size_t InstAddr = CurrInst->GetAddr();
			if (CurrInst->IsMarkerInst())
				continue;
			if (InstAddr == DefAddr) {
				FoundDefAddrInst = true;
			}
			else if (FoundDefAddrInst) {
				UseIter = CurrInst->FindUse(DefOp);
				if (UseIter != CurrInst->GetLastUse()) {
					if (UseIter->GetSSANum() == DefSSANum) {
						// Found use of same SSA name.
						if (CurrInst->IsSubRegUsedAsShiftCount(DefOp)) {
							// A subreg of DefOp is used as a shift counter.
							//  We don't care if we overflowed DefOp, because
							//  we don't mind tossing upper bits anyway, as we
							//  need an 8-bit shift count for this USE.
							benign = true;
							IdiomCode = 1;
						}
						else if (INDIR_JUMP == CurrInst->GetDataFlowType()) {
							// Lots of false positives on indirect jumps via switch tables, because if
							//  any switch index has a negative number, then the compiler makes the "base"
							//  address be the address of "case 0:" rather than the beginning address, and then
							//  uses negative offsets to reach the "case -1:" and "case -2:" etc. addresses.
							//  So, STARS infers CODEPOINTER type => UNSIGNED, but we have unsigned+negative offset
							//  which looks like an unsigned addition overflow. We need to change the signedness
							//  to UNKNOWNSIGN when the annotation is emitted. This is not a truly benign case,
							//  so we just use the IdiomCode to signal what is going on and leave benign set to false.
							//  In the cases that occur later, we will continue to process this code just in case
							//  a truly benign code pattern is detected that will override this IdiomCode.
							// NOTE: We could make this case less likely to produce false negatives by getting info
							//  from IDA Pro to see if we have a switch table with negative offsets.
							IdiomCode = 17;
						}
						break; // For now, stop at first USE.
					}
				}
			}
		}
	}
	if (benign) {
		return benign;
	}

	SMPInstr *DefInst = this->GetFunc()->GetInstFromAddr(DefAddr);

	// For load-effective-address instructions, we can have two source operands that are not
	//  the same as the DEF operand, e.g. lea eax,[edx+ebx] is  eax := edx+ebx; which is a 
	//  compiler optimization to produce a three-address instruction on a machine that is otherwise
	//  only capable of two-address instructions. For either of the source operands, we want to
	//  suppress annotations if the ultimate source was any special operation, such as bitwise operations
	//  or zero-extended moves.
	STARS_ea_t SourceInstAddr = STARS_BADADDR;
	int UseSSANum = SMP_SSA_UNINIT;
	bool SpecialSourceFound = false;
	if (DefInst->MDIsLoadEffectiveAddressInstr()) {
		for (UseIter = DefInst->GetFirstUse(); UseIter != DefInst->GetLastUse(); ++UseIter) {
			STARSOpndTypePtr UseOp = UseIter->GetOp();
			if (MDIsDataFlowOpnd(UseOp, UseFP)) {
				UseSSANum = UseIter->GetSSANum();
				this->GetFunc()->ResetProcessedBlocks();
				SpecialSourceFound = DefInst->IsOpSourceSpecial(UseOp, UseSSANum, false, SourceInstAddr);
				if (SpecialSourceFound) {
					benign = true;   // meat-axe approach to eliminate false positives at the risk of false negatives
					IdiomCode = 32;
					break;
				}
			}
		}
		if (DefInst->FindUse(DefOp) == DefInst->GetLastUse()) {
			// We have the three-address case where the DEF is not USEd. Analyses below could be misled
			//  by assuming that the DEF is also a USE, as is the case for two-address non-lea opcodes.
			return benign;  // No further analysis of Load-effective-address instructions
		}
	}

	SMPInstr *SourceInst = NULL;
	SourceInstAddr = STARS_BADADDR;
	UseSSANum = SMP_SSA_UNINIT;
	SpecialSourceFound = false;
	UseIter = DefInst->FindUse(DefOp); // DEF is also USE in addition, subtraction, etc.
	if (MDIsDataFlowOpnd(DefOp, UseFP) && (UseIter != DefInst->GetLastUse())) {
		UseSSANum = UseIter->GetSSANum();
		this->GetFunc()->ResetProcessedBlocks();
		SpecialSourceFound = DefInst->IsOpSourceSpecial(DefOp, UseSSANum, false, SourceInstAddr);
		if (SpecialSourceFound) {
			SourceInst = this->GetFunc()->GetInstFromAddr(SourceInstAddr);
		}
	}

	// Second case: Subtraction of two values, each of which is just a condition code.
	//  This can happen in optimized code that tests for which of two condition codes
	//  has been set, e.g.:
	//  [arithmetic operation or comparison]
	//  seta al  ; if "above" flag then al := 1 else al := 0
	//  setb bl  ; if "below" flag then bl := 1 else bl := 0
	//  mov ecx,eax  ; make copy of eax (including al) into ecx
	//  sub al,bl ; al := al - bl
	//  The resulting value in al will be 1 if the above flag was set, -1 if below, 0 if neither.
	//  We see the seta/setb as implying unsigned values, which is true, but if al becomes -1
	//   we do not want to consider this to be an underflow.
	//  If the source of the minuend and subtrahend are one condition code and one small immediate,
	//   then we also want to suppress underflow false positives.
	//  The same logic applies to overflow opcodes: lea edx,[eax+eax-1] is hard to classify as
	//   potential underflow or overflow. If eax comes from a condition code (seta eax), the values
	//   will be small and no overflow check is needed.
	if (!benign) {
		if (UnderflowOpcode && MDIsDataFlowOpnd(DefOp, UseFP)) {
			// Prepare for possible recursive traversals through the function blocks.
			this->GetFunc()->ResetProcessedBlocks();
			if (SpecialSourceFound && DefInst->IsOpSourceConditionCode(DefOp, UseSSANum)) {
				// If we had a decrement opcode, we are done, and underflow is benign.
				if (DefInst->IsDecrementRTL()) {
					benign = true;
				}
				else { // For subtraction, need to see if BOTH operands came from condition codes (flags).
					STARSOpndTypePtr UseOp = CloneIfNecessary(DefInst->GetUseOnlyAddSubOp(), UseFP);
					if (MDIsDataFlowOpnd(UseOp, UseFP)) {
						CanonicalizeOpnd(UseOp);
						UseIter = DefInst->FindUse(UseOp);
						int SubtrahendSSANum = UseIter->GetSSANum();
						this->GetFunc()->ResetProcessedBlocks();
						benign = DefInst->IsOpSourceConditionCode(UseOp, SubtrahendSSANum);
					}
					else if (UseOp->IsImmedOp()) {
						uintptr_t ImmVal = UseOp->GetImmedValue();
						benign = (ImmVal <= 7); // small value is intentionally setting up bit pattern of mostly 1-bits
					}
				}
				if (benign) {
					IdiomCode = 3;
				}
			}
			// Prepare for possible recursive traversals through the function blocks.
			this->GetFunc()->ResetProcessedBlocks();
			if (!benign && SpecialSourceFound && (DefInst->IsOpSourceZeroExtendedMove(DefOp, UseSSANum, false))) {
				// If we had a decrement opcode or small subtraction, we are done, and underflow is benign.
				STARSOpndTypePtr SubOp = nullptr;
				STARS_uval_t ConstValue;
				bool ConstSub = DefInst->IsSubtractionOfConstant(SubOp, ConstValue);
				if (ConstSub && (ConstValue <= 7)) {
					benign = true;
					IdiomCode = 29;
				}
				else { // For subtraction, need to see if BOTH operands came from zero-extended moves for IDIOM 12.
					STARSOpndTypePtr UseOp = CloneIfNecessary(DefInst->GetUseOnlyAddSubOp(), UseFP);
					if (MDIsDataFlowOpnd(UseOp, UseFP)) {
						CanonicalizeOpnd(UseOp);
						UseIter = DefInst->FindUse(UseOp);
						UseSSANum = UseIter->GetSSANum();
						this->GetFunc()->ResetProcessedBlocks();
						benign = DefInst->IsOpSourceZeroExtendedMove(UseOp, UseSSANum, false);
						if (benign) {
							IdiomCode = 12;
						}
					}
				}
			}
			// Third case: Subtract original value from shifted value, unsigned. If the shift reduced the original value,
			//  you will have an underflow. Right shifts always reduce the unsigned value, and left shifts reduce the unsigned
			//  value in many cases (but not for signed values).
			else if (!benign) {
				// Check for UNSIGNED.
				if (DefIsUnsigned) {
					// Does the subtraction use a value from a shift?
					UseIter = DefInst->FindUse(DefOp);
					if (UseIter != DefInst->GetLastUse()) {
						int UseSSANum = UseIter->GetSSANum();
						STARS_ea_t UseDefAddr = this->GetDefAddrFromUseAddr(DefOp, DefAddr, UseSSANum, LocalDefName);
						if (this->IsAddrInBlock(UseDefAddr)) {
							// Limit to the simple case in which UseDefAddr is in our block, not Phi function or function entry marker inst.
							SMPInstr *UseDefInst = this->GetFunc()->GetInstFromAddr(UseDefAddr);
							if (UseDefInst->MDIsShiftOrRotate()) {
								// Find USE that got shifted.
								UseIter = UseDefInst->FindUse(DefOp);
								if (UseIter != UseDefInst->GetLastUse()) {
									UseSSANum = UseIter->GetSSANum();
									// At this point, we might still have sequence:
									// mov edi,eax
									// shl edi,5
									// sub edi,eax
									//
									// We have backtracked from the subtraction to the USE of EDI in the shift. Now we need to
									//  find the move that is the DEF of this USE of EDI, i.e. we need to find EAX.
									UseDefAddr = this->GetDefAddrFromUseAddr(DefOp, UseDefAddr, UseSSANum, LocalDefName);
									if ((UseDefAddr != this->GetFirstAddr()) && (UseDefAddr < STARS_PSEUDO_ID_MIN)) {
										SMPInstr *MoveInst = this->GetFunc()->GetInstFromAddr(UseDefAddr);
										if (MoveInst->MDIsMoveInstr()) {
											STARSOpndTypePtr MoveSourceOp = CloneIfNecessary(MoveInst->GetMoveSource(), UseFP);
											// Now for the $64,000 question: Is MoveSourceOp the same operand as
											//  is subtracted in DefInst?
											// NOTE: We should probably limit this check to register or direct stack operands.
											//  We are currently matching MoveSourceOp to the subtrahend operand on immediate
											//  value cases, e.g. 0x80a7884 in sudoku.exe with immediate value 1.
											UseIter = DefInst->FindUse(MoveSourceOp);
											if (UseIter != DefInst->GetLastUse()) {
												// Found it. The value that got shifted later had the unshifted
												//  original value subtracted from the shifted value, which will
												//  produce a benign unsigned underflow quite often.
												benign = true;
												IdiomCode = 16;
											}
										}
									}
								}
							}
						}
					}
				}
			}
		} // end if (UnderflowOpcode && MDIsDataFlowOpnd())
		// Fourth case: lea reg, [basereg+indexreg*scalefactor+offset], where
		//  basereg and indexreg both got their values from condition codes, so the
		//  computational values are too small for non-benign overflow or underflow.
		else if (DefInst->MDIsLoadEffectiveAddressInstr()) {
			STARSOpndTypePtr UseOp = DefInst->GetLeaMemUseOp();
			int BaseReg, IndexReg;
			uint16_t ScaleFactor;
			STARS_ea_t offset;
			MDExtractAddressFields(UseOp, BaseReg, IndexReg, ScaleFactor, offset);
			bool SafeBaseReg = (BaseReg == STARS_x86_R_none);
			bool SafeIndexReg = (IndexReg == STARS_x86_R_none);
			STARSOpndTypePtr SearchOp = nullptr;
			if (!SafeBaseReg) {
				SearchOp = DefInst->MakeRegOpnd((STARS_regnum_t) BaseReg);
				UseIter = DefInst->FindUse(SearchOp);
				assert(UseIter != DefInst->GetLastUse());
				UseSSANum = UseIter->GetSSANum();
				this->GetFunc()->ResetProcessedBlocks();
				SafeBaseReg = DefInst->IsOpSourceConditionCode(SearchOp, UseSSANum);
			}
			if (SafeBaseReg && (!SafeIndexReg)) {
				// We avoid redundant work when BaseReg == IndexReg.
				if (IndexReg == BaseReg) {
					SafeIndexReg = true;
				}
				else {
					SearchOp = DefInst->MakeRegOpnd((STARS_regnum_t) IndexReg);
					UseIter = DefInst->FindUse(SearchOp);
					assert(UseIter != DefInst->GetLastUse());
					UseSSANum = UseIter->GetSSANum();
					this->GetFunc()->ResetProcessedBlocks();
					SafeIndexReg = DefInst->IsOpSourceConditionCode(SearchOp, UseSSANum);
				}
			}
			if (SafeBaseReg && SafeIndexReg) { // We could check for huge offset values here (corner case) !!!!****!!!!
				benign = true;
				IdiomCode = 8;
			}
		}
	}

#if STARS_AGGRESSIVE_BENIGN_POINTER_ARITHMETIC
	if (!benign && UnderflowOpcode && DefIsUnsigned) {
		// Search for a sequence of SUB then ADD operations that
		//  are applied to a POINTER type.
		InstIter = this->GetInstIterFromAddr(DefAddr);
		assert(InstIter != this->GetLastInst());
		DefIter = DefInst->FindDef(DefOp);
		assert(DefIter != DefInst->GetLastDef());
#if 0  // Make it apply to all UNSIGNED; POINTERs are already IDIOM 18
		SMPOperandType DefType = DefIter->GetType();
		if (IsDataPtr(DefType)) {
#endif
			// Restrict pattern to POINTER -= [operand1]; POINTER += [operand2]
			//  in consecutive instructions. Both instructions could overflow, while
			//  producing a valid result after canceling each other out.
			++InstIter;
			if (InstIter != this->GetLastInst()) {
				SMPInstr *DefInst2 = (*InstIter);
				if (DefInst2->MDIsAddition()) {
					UseIter = DefInst2->FindUse(DefOp);
					if (UseIter != DefInst2->GetLastUse()) {
						// POINTER DefOp is USE (and also DEF, which we did not check explicitly)
						//  in addition after being re-DEFed by the original subtraction inst.
						benign = true;
						IdiomCode = 23;
						DefInst2->SetSuppressNumericAnnotation();
					}
				}
			}
#if 0
		}
#endif
		if (!benign) {
			// Look for optimized range check:
			//   sub reg1,const1
			//   cmp reg1,const2
			//  If this sequence is followed by an UNSIGNED condition code set:
			//   setbe cl
			//  then it appears that reg1 is UNSIGNED. Actually, this is just an optimized way
			//  to determine if reg1 falls in the range of const1..(const1+const2). The fact that
			//  we "underflow" the "unsigned" subtraction for values of reg1 less than const1 is
			//  irrelevant.
			STARS_uval_t SubConstant;
			STARSOpndTypePtr DummyOp = nullptr;
			if (DefInst->IsSubtractionOfConstant(DummyOp, SubConstant)) {
				// find out if DefOp a.k.a. DummyOp is used in a comparison.
				InstIter = this->GetInstIterFromAddr(DefAddr);
				assert(InstIter != this->GetLastInst());
				++InstIter;
				while (InstIter != this->GetLastInst()) {
					SMPInstr *DefInst2 = (*InstIter);
					STARS_uval_t CmpConstant;
					if (DefInst2->MDIsCompareToPositiveConstant(DummyOp, CmpConstant)) {
						UseIter = DefInst2->FindUse(DefOp);
						if (UseIter != DefInst2->GetLastUse()) {
							if (UseIter->GetSSANum() == DefSSANum) {
								IdiomCode = 31;
								benign = true;
							}
						}
						break;  // stop after first comparison inst.
					}
					else {
						DefIter = DefInst2->FindDef(DefOp);
						if (DefIter != DefInst2->GetLastDef()) { // re-DEF
							break;
						}
					}
					++InstIter;
				}
			}
		}
	}
#endif

	if (!benign) {
		// Bitwise NOT is likely to produce an overflow/underflow for a subsequent add/subtract.
#if 0
		if (RegDef) {
#endif
			// DEF is also USE in addition or subtraction ...
			if (UseIter != DefInst->GetLastUse()) {   // ... but not always on lea opcode
				if (SpecialSourceFound && SourceInst->MDIsBitwiseNotOpcode()) {
					benign = true;
					IdiomCode = 21;
				}
			}
#if 0
		}
#endif
#if 0
		else if (MDIsDirectStackAccessOpnd(DefOp, this->GetFunc()->UsesFramePointer())){
			// We need the canonicalized stack operand as found in DEFs and USEs.
			UseOp = DefInst->MDGetMemUseOp(); // DEF is also USE in addition or subtraction ...
			UseIter = DefInst->FindUse(UseOp);
			if (UseIter != DefInst->GetLastUse()) {   // ... but not always on lea opcode
				UseSSANum = UseIter->GetSSANum();
				if (SourceInst->MDIsBitwiseNotOpcode()) {
					benign = true;
					IdiomCode = 21;
				}
			}
		}
#endif
	}

	if (!benign) {
		// Look for bitwise NOT idiom on USE-only addend or subtrahend.
		STARSOpndTypePtr UseOp = CloneIfNecessary(DefInst->GetUseOnlyAddSubOp(), UseFP);
		if (UseOp->IsRegOp()) {
			CanonicalizeOpnd(UseOp);
			UseIter = DefInst->FindUse(UseOp); 
			if (UseIter != DefInst->GetLastUse()) {
				UseSSANum = UseIter->GetSSANum();
				this->GetFunc()->ResetProcessedBlocks();
				if (DefInst->IsOpSourceBitwiseNot(UseOp, UseSSANum)) {
						benign = true;
						IdiomCode = 21;
				}
			}
		}
		else if (MDIsDirectStackAccessOpnd(UseOp, this->GetFunc()->UsesFramePointer())) {
			// We need the canonicalized stack operand as found in DEFs and USEs.
			UseOp = DefInst->MDGetMemUseOp();
			assert((nullptr != UseOp) && (! UseOp->IsVoidOp()));
			UseIter = DefInst->FindUse(UseOp);
			assert(UseIter != DefInst->GetLastUse());
			UseSSANum = UseIter->GetSSANum();
			this->GetFunc()->ResetProcessedBlocks();
			if (DefInst->IsOpSourceBitwiseNot(UseOp, UseSSANum)) {
				benign = true;
				IdiomCode = 21;
			}
		}
	}

	if (!benign && !UnderflowOpcode) {
		// IDIOM 27: Chain of ADD/OR/AND/XOR/NOT on same reg in a basic block.
		size_t OperationCounter = 0;
		// Find start of chain.
		InstIter = this->GetInstIterFromAddr(DefAddr);
		assert(InstIter != this->GetLastInst());
		vector<SMPInstr *>::reverse_iterator RevInstIter(InstIter);
		STARS_ea_t FirstAddrFound = (*InstIter)->GetAddr();
		while (RevInstIter != this->GetRevInstEnd()) {
			SMPInstr *SearchInst = (*RevInstIter);
			if (SearchInst->FindDef(DefOp) != SearchInst->GetLastDef()) {
				// SearchInst DEFs the reg we are concerned with.
				if (SearchInst->MDIsAddition() || SearchInst->MDIsBitwiseAndOpcode()
					|| SearchInst->MDIsBitwiseNotOpcode() || SearchInst->MDIsBitwiseOrOpcode()
					|| SearchInst->MDIsHashingArithmetic()) {
					FirstAddrFound = SearchInst->GetAddr();
				}
				else {
					break;
				}
			}
			else {
				break;
			}
			++RevInstIter;
		}
		// Go forward and count qualifying operations.
		InstIter = this->GetInstIterFromAddr(FirstAddrFound);
		assert(InstIter != this->GetLastInst());
		size_t BitOperationCounter = 0;
		while (InstIter != this->GetLastInst()) {
			SMPInstr *SearchInst = (*InstIter);
			if (SearchInst->FindDef(DefOp) != SearchInst->GetLastDef()) {
				if ((SearchInst->GetAddr() > DefAddr) && (SearchInst->FindUse(DefOp) == SearchInst->GetLastUse())) {
					break; // re-DEF
				}
				if (SearchInst->MDIsBitwiseAndOpcode() || SearchInst->MDIsBitwiseNotOpcode() 
					|| SearchInst->MDIsBitwiseOrOpcode() || SearchInst->MDIsHashingArithmetic()) {
					++BitOperationCounter;
					++OperationCounter;
				}
				else if (SearchInst->MDIsAddition()) {
					++OperationCounter;
				}
				else {
					break;
				}
			}
			++InstIter;
		}
		if ((BitOperationCounter > 1) && (OperationCounter > 2)) {
			benign = true;
			IdiomCode = 27;
		}
	}

	if (!benign && !UnderflowOpcode && DefIsUnsigned) {
		// Search for a sequence of ADD, OR, and AND operations that
		//  use constant operands that add up to an UNSIGNED OVERFLOW.
		InstIter = this->GetInstIterFromAddr(DefAddr);
		assert(InstIter != this->GetLastInst());
		uint32_t ImmedValue = (uint32_t) (*InstIter)->MDGetImmedUse();
		// NOTE: If ImmedValue is very large, we catch this elsewhere as IDIOM 11.
		if (ImmedValue > 0) {
			vector<SMPInstr *>::reverse_iterator RevInstIter(InstIter);
			uint32_t ConstLimit = 0xffffffff - ImmedValue;
			while (RevInstIter != this->GetRevInstEnd()) {
				SMPInstr *SearchInst = (*RevInstIter);
				DefIter = SearchInst->FindDef(DefOp);
				if (DefIter != SearchInst->GetLastDef()) { // found previous DEF
					if (SearchInst->MDIsBitwiseAndOpcode() || SearchInst->MDIsBitwiseOrOpcode() || SearchInst->MDIsAddition()) {
						ImmedValue = (uint32_t) SearchInst->MDGetImmedUse();
						if (0 < ImmedValue) {
							if (ImmedValue >= ConstLimit) {
								benign = true;
								IdiomCode = 22;
								break; // no need to search further
							}
							else {
								ConstLimit -= ImmedValue; // set new limiting value
							}
						}
						else {
							break; // not the pure chain we are searching for
						}
					}
					else {
						break; // not the pure chain we are searching for
					}
				}
				++RevInstIter;
			}
		}
	}

	if (!benign) {
		STARSOpndTypePtr UseOp = DefInst->GetUseOnlyAddSubOp();
		size_t UseBitWidth = 8 * UseOp->GetByteWidth();
		// IDIOM 28: Normal width arithmetic produces value that is then involved in subword arithmetic.
		//  Ignore overflow or underflow on the subword arithmetic.
		if ((UseOp->IsRegOp()) && (UseBitWidth < MD_NORMAL_MACHINE_BITWIDTH)) { // UseOp is candidate for IDIOM 28.
			STARSOpndTypePtr UseOp2 = CloneIfSubwordReg(UseOp);
			CanonicalizeOpnd(UseOp2);
			set<DefOrUse, LessDefUse>::iterator UseIter = DefInst->FindUse(UseOp2);
			assert(UseIter != DefInst->GetLastUse());
			int UseSSANum = UseIter->GetSSANum();
			STARS_ea_t UseDefAddr = this->GetDefAddrFromUseAddr(UseOp2, DefAddr, UseSSANum, this->IsLocalName(UseOp2));
			// Stick with simple case in which an instruction was found in the block.
			if (this->IsAddrInBlock(UseDefAddr)) {
				vector<SMPInstr *>::iterator InstIter = this->GetInstIterFromAddr(UseDefAddr);
				SMPInstr *UseDefInst = (*InstIter);
				// If UseOp was DEFed by addition or subtraction, examine the bit width of the operation.
				STARSOpndTypePtr UseDefDEFOperand = UseDefInst->GetDefUseAddSubOp();
				if (UseDefDEFOperand->IsRegOp()) {
					size_t UseDefBitWidth = 8 * UseDefDEFOperand->GetByteWidth();
					if (UseDefBitWidth > UseBitWidth) {
						// Current instruction uses a truncated portion of the previous DEF.
						//  The current add or subtract might overflow or underflow, but this
						//  seems to be intentional, as the code is not likely to be written to
						//  keep subregisters within certain bounds.
						IdiomCode = 28;
						benign = true;
					}
				}
			}
		}
		if (!benign && (DefOp->IsRegOp())) {
			STARSOpndTypePtr OriginalDefOp = DefInst->GetDefUseAddSubOp();
			size_t OriginalDefBitWidth = 8 * OriginalDefOp->GetByteWidth();
			if (OriginalDefBitWidth < MD_NORMAL_MACHINE_BITWIDTH) { // DefOp is candidate for IDIOM 28.
				set<DefOrUse, LessDefUse>::iterator UseIter = DefInst->FindUse(DefOp);
				assert(UseIter != DefInst->GetLastUse());
				int DefUseSSANum = UseIter->GetSSANum();
				STARS_ea_t DefUseDefAddr = this->GetDefAddrFromUseAddr(DefOp, DefAddr, DefUseSSANum, this->IsLocalName(DefOp));

				// Stick with simple case in which an instruction was found in the block.
				if (this->IsAddrInBlock(DefUseDefAddr)) {
					vector<SMPInstr *>::iterator InstIter = this->GetInstIterFromAddr(DefUseDefAddr);
					SMPInstr *DefUseDefInst = (*InstIter);
					if (DefUseDefInst->IsMarkerInst()) {
					    // SSA Marker always looks like full width operand, but can be misleading.
					    IdiomCode = 26;
					    benign = true;
					}
					else {
					    // If DefOp was DEFed by addition or subtraction, examine the bit width of the operation.
					    STARSOpndTypePtr DefUseDefDEFOperand = DefUseDefInst->GetDefUseAddSubOp();
					    if (DefUseDefDEFOperand->IsRegOp()) {
						size_t DefUseDefBitWidth = 8 * DefUseDefDEFOperand->GetByteWidth();
						if (DefUseDefBitWidth > OriginalDefBitWidth) {
						    // Current instruction uses a truncated portion of the previous DEF.
						    //  The current add or subtract might overflow or underflow, but this
						    //  seems to be intentional, as the code is not likely to be written to
						    //  keep subregisters within certain bounds.
						    IdiomCode = 28;
						    benign = true;
						}
					    }
					}
				}
			}
		}
	}

	return benign;
} // end of SMPBasicBlock::IsBenignOverflowDEF()

// Is subword of DefOp written later, before TruncAddr?
bool SMPBasicBlock::IsDEFWrittenWithTruncation(const STARSOpndTypePtr &DefOp, int DefSSANum, STARS_ea_t DefAddr, STARS_ea_t TruncAddr, bool PhiDEF) {
	assert(DefOp->IsRegOp());
	STARSOpndTypePtr SearchOp = CloneIfNecessary(DefOp, this->GetFunc()->UsesFramePointer());
	CanonicalizeOpnd(SearchOp);
	vector<SMPInstr *>::iterator InstIter;
	bool FoundSubwordWrite = false;

	InstIter = this->GetInstIterFromAddr(DefAddr);
	if (!PhiDEF) {
		// DefAddr is just first addr in block with PhiDEF, else it is real DefAddr that should be skipped.
		++InstIter;
	}
	while (InstIter != this->GetLastInst()) {
		// Search for subword write of DefOp/DefSSANum.
		SMPInstr *CurrInst = (*InstIter);
		STARS_ea_t InstAddr = CurrInst->GetAddr();
		if (InstAddr == TruncAddr)
			break;  // else we will find the apparent truncation we started with!

		// Re-DEF of DefOp terminates search unless it is a right shift.
		//  A right shift often precedes a subword write, e.g.:
		//  edx := ...
		//  ecx := edx;   // need to recurse into ecx chain here
		//  ecx := ecx >> 8;  // don't terminate ecx chain search yet
		//  mem := cl (subword of ecx) // found subword write
		set<DefOrUse, LessDefUse>::iterator DefIter = CurrInst->FindDef(SearchOp);
		if (DefIter != CurrInst->GetLastDef()) { // found re-DEF
			if (CurrInst->MDIsShiftRight()) {
				// Reset DefSSANum and DefAddr and recurse.
				FoundSubwordWrite = this->IsDEFWrittenWithTruncation(DefOp, DefIter->GetSSANum(), InstAddr, TruncAddr, false);
			}
			break; // we always terminate after re-DEF; either we recursed or we failed our search.
		}

		set<DefOrUse, LessDefUse>::iterator UseIter = CurrInst->FindUse(SearchOp);
		if (UseIter != CurrInst->GetLastUse()) { // Used, so instruction might be relevant.
			if (CurrInst->MDIsMoveInstr()) { // register DefOp might be copied somewhere
				STARSOpndTypePtr MoveSourceOp = CloneIfSubwordReg(CurrInst->GetMoveSource());
				CanonicalizeOpnd(MoveSourceOp);
				if (IsEqOp(MoveSourceOp, SearchOp)) { // SearchOp is being copied.
					// Subword write of DefOp is successful detection.
					if (CurrInst->MDIsReducedWidthMove()) { // Subword of SearchOp is being copied.
						FoundSubwordWrite = true;
						break;
					}
					else { // full-width move
						DefIter = CurrInst->GetFirstNonFlagsDef();
						STARSOpndTypePtr MoveDestOp = DefIter->GetOp();
						if (MoveDestOp->IsRegOp()) { // recurse into copy reg chain
							int MoveDestSSANum = DefIter->GetSSANum();
							FoundSubwordWrite = this->IsDEFWrittenWithTruncation(MoveDestOp, MoveDestSSANum, InstAddr, TruncAddr, false);
							if (FoundSubwordWrite)
								break; // success, else continue current chain
						}
						else { // copied to memory; do not recurse, but keep searching.
							;
						}
					}
				}
			} // end if Move instr
		} // end if USEd
		++InstIter;
	} // end while all insts

	return FoundSubwordWrite;
} // end of SMPBasicBlock::IsDEFWrittenWithTruncation()

// We don't care if we truncate computing some DEFs, because of how they are used.
//  Return true when we detect one of these cases.
bool SMPBasicBlock::IsBenignTruncationDEF(const STARSOpndTypePtr &DefOp, int DefSSANum, STARS_ea_t DefAddr, int &IdiomCode) {
	bool benign = false;
	vector<SMPInstr *>::iterator InstIter;
	set<DefOrUse, LessDefUse>::iterator UseIter, UseIter2, DefIter;
	STARSOpndTypePtr UseOp = nullptr, SearchOp = nullptr, UseOp2 = nullptr;
	unsigned short SignMask;
	bool FoundTruncationDef = false; // Found apparent truncation move
	SMPInstr *TruncationInst = NULL;
	size_t UseBitWidth = 0;   // Bit width involved in truncation move
	size_t BytesMasked;
	bool FoundSubregUse = false;   // Found next use of truncated DEF, which appears to make truncation benign
	int UseSSANum = -1;

	// First, we find the USEs of the DEF.
	for (InstIter = this->GetFirstInst(); InstIter != this->GetLastInst(); ++InstIter) {
		SMPInstr *CurrInst = (*InstIter);
		size_t InstAddr = CurrInst->GetAddr();
		if (InstAddr == DefAddr) {
			// In the truncating instruction, if we have a sub-register that was originally
			//  DEFed as its enclosing register, it appears to be a truncation. However, it
			//  could be that the rest of the register is used later, not discarded, e.g.:
			//   mov [ebp-12],ax  ; looks like EAX is being truncated to AX and stored
			//   shr eax,16       ; gets upper 16 bits into lower 16 bits
			//   mov [ebp-14],ax  ; store what used to be the upper 16 bits of EAX
			// The first instruction will trigger a CHECK TRUNCATION annotation that
			//  causes false positives. We need to analyze the context of the instruction
			//  to see that the whole register EAX was used, so no truncation occurred.
			//  The context analysis in the basic block will mark the second move as
			//  a "truncation" that should be ignored, so we can later avoid performing
			//  redundant analysis.
			if ((CurrInst->GetOptType() == 3) && (!CurrInst->MDIsSignedLoad(SignMask))) {
				// We have a regular load, not a sign-extended or zero-extended load.
				//  This is the kind of load that could look like a truncation, if the
				//  USE has smaller width than the DEF. Precise analysis of the width of
				//  the USE and DEF occurs in SMPInstr::EmitIntegerErrorAnnotations(), but
				//  we need only a good first cut here.
				UseOp = CurrInst->GetMoveSource();
				SearchOp = CloneIfSubwordReg(UseOp);
				if (UseOp->IsRegOp()) {
					CanonicalizeOpnd(SearchOp);
				}
				else {
					break;  // Not a sub-register, cannot fit the pattern
				}
				UseIter = CurrInst->FindUse(SearchOp);
				assert(UseIter != CurrInst->GetLastUse());
				UseSSANum = UseIter->GetSSANum();

				if (UseIter->DoesNotTruncate()) {
					benign = true; // already analyzed
					IdiomCode = 2;
					break;
				}

				FoundTruncationDef = true;
				TruncationInst = CurrInst;
				UseBitWidth = 8 * UseOp->GetByteWidth();
				if (MD_NORMAL_MACHINE_BITWIDTH <= UseBitWidth) {
					// Cannot be truncation; exit.
					break;
				}
			}
		}
		else if (FoundTruncationDef) {
			if (!FoundSubregUse) {
				// Find next USE of SearchOp.
				UseIter2 = CurrInst->FindUse(SearchOp);
				if (UseIter2 != CurrInst->GetLastUse()) {
					int UseSSANum2 = UseIter2->GetSSANum();
					if (UseSSANum != UseSSANum2) {
						break; // have re-DEFed; stop following SearchOp
					}
					// Found use of SearchOp+DefSSANum. Must be shift or rotate to fit pattern.
					if (CurrInst->MDIsShiftOrRotate()) {
						// If we shift or rotate by UseBitWidth bits, even if we don't make
						//  an exact full circle, we can pattern match a simpler pattern, e.g.:
						//  mov <memory>,dl
						//  shr edx,8
						//  loop back to process next byte in edx
						//  If we see this pattern, we go ahead and mark the move as non-truncating.
						//  Passing false as the second argument makes the check less restrictive.
						if (CurrInst->ShiftMakesUpperBitsLower(UseBitWidth, false)) {
							// At least it matches the less restrictive two-inst sequence.
							benign = true;
							IdiomCode = 2;
							// Set the NoTruncate flag in the first instruction for the sub-reg UseOp.
							TruncationInst->SetUseNoTruncate(SearchOp, true);
							if (CurrInst->ShiftMakesUpperBitsLower(UseBitWidth, true)) {
								FoundSubregUse = true;
							}
						}
					} // end if shift or rotate
					else if (CurrInst->MDIsSubregMaskInst(BytesMasked)) {
						// If we are masking off subreg regions, then we are operating on
						//  less than the full register after the truncation, so the truncation
						//  is almost certainly benign.
						benign = true;
						IdiomCode = 14;
						break; // Masking inst is redefining SearchOp
					}
				} // end if found SearchOp in CurrInst
				else {
					DefIter = CurrInst->FindDef(SearchOp);
					if (DefIter != CurrInst->GetLastDef()) {
						// SearchOp is not used as subreg yet, but is re-DEFed. Failed to match.
						break;
					}
				}
			}
			else {
				// Found apparent truncation, found shift that seems to make it benign.
				//  Now we need to see the shifted bits (former upper bits) get stored
				//  so we can mark both store instructions as benign.
				UseIter2 = CurrInst->FindUse(SearchOp);
				if (UseIter2 != CurrInst->GetLastUse()) {
					if ((CurrInst->GetOptType() == 3) && (!CurrInst->MDIsSignedLoad(SignMask))) {
						UseOp2 = UseIter2->GetOp();
						size_t UseBitWidth2 = 8 * UseOp2->GetByteWidth();
						if (UseBitWidth == UseBitWidth2) {
							// Entire pattern matched. Moved lower bits, shifted or rotated by
							//  stored bit width, then moved new lower bits.
							// Avoid redundant computation on sub-reg UseOp2 in 3rd instruction by setting its flag.
							CurrInst->SetUseNoTruncate(SearchOp, true);
						}
					}
					break; // Found whole pattern, or failed to find third inst in pattern.
				}
				else {
					DefIter = CurrInst->FindDef(SearchOp);
					if (DefIter != CurrInst->GetLastDef()) {
						// SearchOp is not written as subreg yet, but is re-DEFed. Failed to match.
						break;
					}
				}
			} // end if (!FoundSubregUse) ... else ...
		} // end if (InstAddr == DefAddr) ... else if (FoundTruncationDef) ...
	} // end for all insts in block

	if (benign)
		return true;

	SMPInstr *DefInst = this->GetFunc()->GetInstFromAddr(DefAddr);
	STARS_ea_t SourceInstAddr = STARS_BADADDR;
	SMPInstr *SourceInst = NULL;
	int DefUseSSANum = SMP_SSA_UNINIT;
	bool SpecialSourceFound = false;
	bool UseFP = this->GetFunc()->UsesFramePointer();
	UseIter = DefInst->FindUse(DefOp); // DEF is also USE in addition, subtraction, etc.
	if (MDIsDataFlowOpnd(DefOp, UseFP) && (UseIter != DefInst->GetLastUse())) {
		DefUseSSANum = UseIter->GetSSANum();
		this->GetFunc()->ResetProcessedBlocks();
		SpecialSourceFound = DefInst->IsOpSourceSpecial(DefOp, DefUseSSANum, true, SourceInstAddr);
		if (SpecialSourceFound) {
			SourceInst = this->GetFunc()->GetInstFromAddr(SourceInstAddr);
			if (SourceInst->IsReducedWidthDef()) {
				benign = true;
				IdiomCode = 24;
			}
		}
	}

	if (!benign && (nullptr != SearchOp) && (SearchOp->IsRegOp()) && (-1 != UseSSANum)) {
		// See if truncation use is a copy of another register that gets subword masked, which
		//  is a sign that the truncation use is also only valid for the lower bits.
		bool LocalName = this->IsLocalName(SearchOp);
		STARS_ea_t UseDefAddr = this->GetDefAddrFromUseAddr(SearchOp, DefAddr, UseSSANum, LocalName);
		if ((STARS_BADADDR != UseDefAddr) && (!STARS_IsBlockNumPseudoID(UseDefAddr))) {
			// We have a valid inst addr (not Phi block number). See if it is a reg to reg copy.
			SMPInstr *CopyInst = this->GetFunc()->GetInstFromAddr(UseDefAddr);
			if (CopyInst->MDIsMoveInstr()) {
				unsigned short SignMask;
				if (CopyInst->MDIsSignedLoad(SignMask)) {
					// Truncation use came from a zero-extended or sign-extended load. So,
					//  we know that the later truncation store is probably a benign truncation.
					IdiomCode = 14;
					benign = true;
				}
				else { // regular move. Is it reg to reg?
					UseOp2 = CloneIfSubwordReg(CopyInst->GetMoveSource());
					if (UseOp2->IsRegOp()) {
						// If UseOp2, the source of our truncation store USE operand,
						//  ever has a subreg masked off, then our truncation is probably
						//  benign.
						CanonicalizeOpnd(UseOp2);
						SMPBasicBlock *CopyBlock = CopyInst->GetBlock();
						UseIter2 = CopyInst->FindUse(UseOp2);
						assert(UseIter2 != CopyInst->GetLastUse());
						int DefSSANum2 = UseIter2->GetSSANum();
						// NOTE: UseDefAddr is not actually the DefAddr of UseOp2; it
						//  is the addr at which UseOp2 was moved into our truncation move
						//  USE, SearchOp. But we only want to see the USEs of UseOp2 that
						//  come after the reg to reg copy.
						benign = CopyBlock->IsDefMaskedOrShiftedRight(UseOp2, DefSSANum2, UseDefAddr);
						if (benign)
							IdiomCode = 14;
					}
				}
			}
		}
	}

#if 1
	if (!benign && (nullptr != SearchOp) && (SearchOp->IsRegOp()) && (-1 != UseSSANum)) {
		// See if truncation use is copied to another register that gets subword masked or right shifted, which
		//  is a sign that the truncation use is also only valid for the lower bits.
		vector<SMPInstr *>::iterator InstIter;
		for (InstIter = this->GetFirstInst(); InstIter != this->GetLastInst(); ++InstIter) {
			SMPInstr *CurrInst = (*InstIter);
			if (CurrInst->MDIsMoveInstr() && (!(CurrInst->HasDestMemoryOperand()))) {
				set<DefOrUse, LessDefUse>::iterator UseIter = CurrInst->FindUse(SearchOp);
				if ((UseIter != CurrInst->GetLastUse()) && (UseSSANum == UseIter->GetSSANum())) {
					// Found a copy of SearchOp+UseSSANum to another register.
					STARS_ea_t CopyAddr = CurrInst->GetAddr();
					set<DefOrUse, LessDefUse>::iterator DefIter = CurrInst->GetFirstNonFlagsDef();
					STARSOpndTypePtr CopyRegOp = DefIter->GetOp();
					int CopyRegSSANum = DefIter->GetSSANum();
					benign = this->IsDefMaskedOrShiftedRight(CopyRegOp, CopyRegSSANum, CopyAddr);
					if (benign)
						IdiomCode = 14;
				}
			}
		}
	}
#endif

	if (!benign && (nullptr != SearchOp) && (SearchOp->IsRegOp()) && (-1 != UseSSANum)) {
		// Check for case in which a copy of the DEF corresponding to the subword USE
		//  gets written as a subword. This occurs in patterns that write in pieces such as:
		//  edx := ...
		//  ecx : = edx; 
		//  write subword of ecx ; not a truncation
		//  ecx := edx; get original copy
		//  write different subword of ecx ; not a truncation
		//  etc.
		//  write subword of edx ; not really a truncation
		bool LocalName = this->IsLocalName(SearchOp);
		STARS_ea_t UseDefAddr = this->GetDefAddrFromUseAddr(SearchOp, DefAddr, UseSSANum, LocalName);
		if (STARS_BADADDR != UseDefAddr) {
			// Check for Phi DEF instead of inst DEF.
			bool PhiDEF = false;
			if (STARS_IsBlockNumPseudoID(UseDefAddr)) { // Phi DEF; UseDefAddr is pseudo-addr
				// For our purposes, we just want to search the block for subword writes, so we can pretend
				//  the DEF addr is right before the block.
				PhiDEF = true;
				SMPBasicBlock *DefBlock = this->GetFunc()->GetBlockByNum(STARS_GetBlockNumFromPseudoID(UseDefAddr));
				UseDefAddr = DefBlock->GetFirstAddr();
			}
			SMPInstr *UseDefInst = this->GetFunc()->GetInstFromAddr(UseDefAddr);
			benign = UseDefInst->GetBlock()->IsDEFWrittenWithTruncation(SearchOp, UseSSANum, UseDefAddr, DefAddr, PhiDEF);
			if (benign) 
				IdiomCode = 2;
		}
	}
#if SMP_MEASURE_NUMERIC_ANNOTATIONS
	if (benign) {
		++SuppressTruncationRegPiecesAllUsed;
	}
#endif

	return benign;
} // end of SMPBasicBlock::IsBenignTruncationDEF()

// Does DefOp+DefSSANum get subreg masked before it gets re-DEFined?
bool SMPBasicBlock::IsDefMaskedOrShiftedRight(const STARSOpndTypePtr &DefOp, int DefSSANum, STARS_ea_t DefAddr) {
	bool MaskingDetected = false;

	vector<SMPInstr *>::iterator InstIter;
	set<DefOrUse, LessDefUse>::iterator UseIter, DefIter;
	int UseSSANum;
	STARS_ea_t InstAddr;
	size_t BytesMasked;

	if (DefOp->IsRegOp()) {
		bool DefAddrInstFound = false;
		for (InstIter = this->GetFirstInst(); InstIter != this->GetLastInst(); ++InstIter) {
			SMPInstr *CurrInst = (*InstIter);
			InstAddr = CurrInst->GetAddr();
			if (CurrInst->IsMarkerInst())
				continue;
			if (InstAddr == DefAddr) {
				DefAddrInstFound = true;
			}
			else if (DefAddrInstFound) {
				UseIter = CurrInst->FindUse(DefOp);
				if (UseIter != CurrInst->GetLastUse()) {
					UseSSANum = UseIter->GetSSANum();
					if (UseSSANum == DefSSANum) {
						if (CurrInst->MDIsSubregMaskInst(BytesMasked)) {
							MaskingDetected = true;
							break;
						}
						else if (CurrInst->MDIsRightShift()) {
							STARSOpndTypePtr ShiftedOp, ShiftCounterOp;
							CurrInst->GetShiftOperands(ShiftedOp, ShiftCounterOp);
							if (ShiftCounterOp->IsImmedOp() && (0 == (ShiftCounterOp->GetImmedValue() % 8))) {
								// Constant shift count, multiple of 8 bits. Just like masking a byte.
								MaskingDetected = true;
								break;
							}
						}
					}
					else {
						// Must have been re-DEFed.
						break;
					}
				}
				DefIter = CurrInst->FindDef(DefOp);
				if (DefIter != CurrInst->GetLastDef()) {
					// Re-DEF. Did not find masking inst.
					break;
				}
			}
		}
	}

	return MaskingDetected;
} // end of SMPBasicBlock::IsDefMaskedOrShiftedRight()

// Is DEF later passed as an argument to a critical system or lib call? If so, return true and put
//  the integer error info string in SinkString. If not, leave SinkString empty and return false.
//  Recurse through successor blocks if needed to find all USEs of the DEF.
bool SMPBasicBlock::IsCriticalSink(const STARSOpndTypePtr &DefOp, int DefSSANum, std::string &SinkString, bool &FoundAnyCall) {
	// Call recursive helper that finds argument-passing USEs of the DEF.
	STARSOpndTypePtr AlternateDefOp = nullptr;
	bool FoundSink = this->UseHasCriticalSink(DefOp, DefSSANum, AlternateDefOp, SMP_SSA_UNINIT, SinkString, FoundAnyCall);

	return FoundSink;
} // end of SMPBasicBlock::IsCriticalSink()

// Can we find USE being passed to a critical system or library call?
bool SMPBasicBlock::UseHasCriticalSink(STARSOpndTypePtr DefOp, int DefSSANum, STARSOpndTypePtr DefOp2, int DefSSANum2, std::string &SinkString, bool &FoundAnyCall) {
	bool FoundSink = false;
	vector<SMPInstr *>::iterator InstIter;
	set<DefOrUse, LessDefUse>::iterator UseIter, DefIter;
	bool FoundArgPass = false;
	int SearchSSANum = DefSSANum;
	int SearchSSANum2 = DefSSANum2;
	STARS_ea_t InstAddr; // for debugging, breakpoints, etc.
	size_t ArgumentNumber = 7; // init to bad number

	// avoid infinite recursion
	if (this->IsProcessed()) {
		return false;
	}
	this->SetProcessed(true); 

	// We need to make the recursion robust by finding DefOp in Phi functions
	//  and altering the DefSSANum we are searching for.
	set<SMPPhiFunction, LessPhi>::iterator CurrPhi = this->FindPhi(DefOp);
	if (CurrPhi != this->GetLastPhi()) {
		// Found Phi function for DefOp. That means we will not be finding DefSSANum for
		//  DefOp in this block or its successors, because the Phi function will redefine
		//  to a new SSA number.
		size_t PhiNumUses = CurrPhi->GetPhiListSize();
		bool FoundPhiUse = false;
		for (size_t i = 0; i < PhiNumUses; ++i) {
			// Confirm that DefSSANum was actually a Phi USE in this block.
			if (CurrPhi->GetUseSSANum(i) == DefSSANum) {
				FoundPhiUse = true;
				break;
			}
		}
		if (FoundPhiUse) {
			SearchSSANum = CurrPhi->GetDefSSANum();  // new SSA name to search for is Phi DEF
		}
	}

	// Do the same for DefOp2/DefSSANum2.
	if ((DefOp2 != nullptr) && (! DefOp2->IsVoidOp())) {
		CurrPhi = this->FindPhi(DefOp2);
		if (CurrPhi != this->GetLastPhi()) {
			// Found Phi function for DefOp2. That means we will not be finding DefSSANum2 for
			//  DefOp2 in this block or its successors, because the Phi function will redefine
			//  to a new SSA number.
			size_t PhiNumUses = CurrPhi->GetPhiListSize();
			bool FoundPhiUse = false;
			for (size_t i = 0; i < PhiNumUses; ++i) {
				// Confirm that DefSSANum2 was actually a Phi USE in this block.
				if (CurrPhi->GetUseSSANum(i) == DefSSANum2) {
					FoundPhiUse = true;
					break;
				}
			}
			if (FoundPhiUse) {
				SearchSSANum2 = CurrPhi->GetDefSSANum();  // new SSA name to search for is Phi DEF
			}
		}
	}


	for (InstIter = this->GetFirstInst(); InstIter != this->GetLastInst(); ++InstIter) {
		SMPInstr *CurrInst = (*InstIter);
		InstAddr = CurrInst->GetAddr();
		// Two stage search: Find argument pass, then find next call instruction.
		if (!FoundArgPass) {
			DefIter = CurrInst->FindDef(DefOp);
			UseIter = CurrInst->FindUse(DefOp);
			if ((DefIter != CurrInst->GetLastDef()) && (DefIter->GetSSANum() == DefSSANum)) {
				// We are on the DEF instruction. See if it is an argument pass.
				if (CurrInst->MDIsArgumentPass(ArgumentNumber)) {
					FoundArgPass = true;
				}
			}
			else if (UseIter != CurrInst->GetLastUse()) {
				int UseSSANum = UseIter->GetSSANum();
				if (UseSSANum == SearchSSANum) {
					// Found a use that matches the DEF. Is it passed as an argument?
					if (CurrInst->MDIsArgumentPass(ArgumentNumber)) {
						FoundArgPass = true;
					}
					else if (CurrInst->MDIsMoveInstr() && (DefOp2 == nullptr)) {
						// DefOp is getting stored, but not as an argument pass.
						//  Hard to say if we should keep following DefOp or follow
						//  the stored copy. Let's try following the stored copy also if
						//  it is a register or stack location.
						DefIter = CurrInst->GetFirstNonFlagsDef();
						assert(DefIter != CurrInst->GetLastDef());
						if (MDIsDataFlowOpnd(DefIter->GetOp(), this->GetFunc()->UsesFramePointer())) {
							// Follow DefOp2 from now on in addition to DefOp.
							DefOp2 = DefIter->GetOp();
							SearchSSANum2 = DefIter->GetSSANum();
						}
					}
					else { // Stop searching for DefOp if it gets redefined.
						DefIter = CurrInst->FindDef(DefOp);
						if (DefIter != CurrInst->GetLastDef()) {
							// DefOp was USE and DEF in CurrInst.
							if (DefOp2 != nullptr) {
								// We have an alternate DEF to search for. Make it primary.
								DefOp = DefOp2;
								SearchSSANum = SearchSSANum2;
								DefOp2 = nullptr; // No more alternate.
								SearchSSANum2 = SMP_SSA_UNINIT;
							}
							else { // No alternate to search for; search has failed.
								DefOp = nullptr;
								SearchSSANum = SMP_SSA_UNINIT;
								break;
							}
						}
					}
				}
			}
			else if (nullptr != DefOp2) { // DefOp is not used; search for DefOp2
				UseIter = CurrInst->FindUse(DefOp2);
				if (UseIter != CurrInst->GetLastUse()) {
					int UseSSANum = UseIter->GetSSANum();
					if (UseSSANum == SearchSSANum2) {
						// Found a use that matches the DEF. Is it passed as an argument?
						if (CurrInst->MDIsArgumentPass(ArgumentNumber)) {
							FoundArgPass = true;
						}
						else if (CurrInst->MDIsMoveInstr()) { // DefOp2 is copied somewhere
							// We cannot follow a third location currently, but we can see
							//  if DefOp2 is redefining DefOp in this move, and get a new
							//  SearchSSANum for DefOp in that case.
							DefIter = CurrInst->FindDef(DefOp);
							if (DefIter != CurrInst->GetLastDef()) {
								// DefOp2 was copied to DefOp.
								SearchSSANum = DefIter->GetSSANum(); // new DefOp/DefSSANum pair to search for.
							}
						}
						else { // Stop searching for DefOp2 if it gets redefined.
							DefIter = CurrInst->FindDef(DefOp2);
							if (DefIter != CurrInst->GetLastDef()) {
								// DefOp2 was USE and DEF in CurrInst.
								DefOp2 = nullptr; // No more alternate.
								SearchSSANum2 = SMP_SSA_UNINIT;
							}
						}
					}
				}
			}
		}
		else { // found arg pass, need to find call
			if (CurrInst->GetDataFlowType() == CALL) {
				FoundAnyCall = true;
				string CalleeName = CurrInst->GetTrimmedCalledFunctionName();
				if (!(CalleeName.empty())) {
					GetSinkStringForCallName(CalleeName, SinkString);
					if (!(SinkString.empty())) {
						// ArgumentNumber needs to be 0 for malloc(), 0 or 1 for calloc(),
						//  and 1 for realloc.
						if (ArgumentNumber < 2) {
							bool FoundMalloc = (0 == CalleeName.compare("malloc"));
							bool FoundCalloc = (0 == CalleeName.compare("calloc"));
							bool FoundRealloc = (0 == CalleeName.compare("realloc"));
							if (((ArgumentNumber == 0) && (FoundMalloc || FoundCalloc))
								|| ((ArgumentNumber == 1) && (FoundCalloc || FoundRealloc))) {
								FoundSink = true;
							}
						}
					}
				}
				break;
			}
		}
	} // end for all instructions

	if (!FoundArgPass) {   // NOTE: Should track foundarg and foundcall separately  !!!!****!!!!
		if ((nullptr == DefOp2) || (!this->IsLiveOut(DefOp2))) {
			// Cannot search for DefOp2 if not LiveOut.
			DefOp2 = nullptr;
			SearchSSANum2 = SMP_SSA_UNINIT;
		}
		if ((nullptr == DefOp) || (!this->IsLiveOut(DefOp))) {
			// Search for DefOp2 instead of DefOp.
			DefOp = DefOp2;
			SearchSSANum = SearchSSANum2;
			DefOp2 = nullptr; // No more alternate.
			SearchSSANum2 = SMP_SSA_UNINIT;
		}

		if ((nullptr != DefOp) && (! DefOp->IsVoidOp())) {
			// Need to recurse through successor blocks to keep searching for arg pass.
			list<SMPBasicBlock *>::iterator SuccIter;
			for (SuccIter = this->GetFirstSucc(); SuccIter != this->GetLastSucc(); ++SuccIter) {
				if ((*SuccIter)->UseHasCriticalSink(DefOp, SearchSSANum, DefOp2, SearchSSANum2, SinkString, FoundAnyCall)) {
					FoundSink = true;
					break;
				}
			}
		}
	}

	return FoundSink;
} // end of SMPBasicBlock::UseHasCriticalSink()

// It is safe to emit signedness checks when the very next use of a stack DEF is as
//  a sign-extended or zero-extended load. Beyond a single basic block, we need to
//  do more analysis. This catches a simple case that is valid across all programs.
bool SMPBasicBlock::IsStackOpNextUsedWithSignedness(const STARSOpndTypePtr &StackDefOp, STARS_ea_t DefAddr, int SSANum) {
	bool SafeCheck = false;
	vector<SMPInstr *>::iterator InstIter;
	set<DefOrUse, LessDefUse>::iterator UseIter;
	STARS_ea_t InstAddr; // for debugging, breakpoints, etc.
	bool FoundDefInst = false;

	for (InstIter = this->GetFirstInst(); InstIter != this->GetLastInst(); ++InstIter) {
		SMPInstr *CurrInst = (*InstIter);
		InstAddr = CurrInst->GetAddr();
		// Two stage search: Find DEF instruction, then find next USE;
		if (!FoundDefInst) {
			FoundDefInst = (InstAddr == DefAddr);
		}
		else {
			UseIter = CurrInst->FindUse(StackDefOp);
			if (UseIter != CurrInst->GetLastUse()) {
				int UseSSANum = UseIter->GetSSANum();
				if (UseSSANum == SSANum) {
					// Found next USE of matching SSA name.
					// Is it an extended load?
					unsigned short DummySignMask;
					SafeCheck = CurrInst->MDIsSignedLoad(DummySignMask);
					break;
				}
			}
		}
	} // end for all instructions

	return SafeCheck;
} // end of SMPBasicBlock::IsStackOpNextUsedWithSignedness()

// Find inc/add that makes small negative back into positive. e.g.:
//  sbb edx,edx  ; unsigned edx becomes 0 or -1; SMPInstr is suppressing underflow false positive.
//  add edx,1    ; converts 0 or -1 to 1 or 0 (looks like unsigned overflow); catch this here.
//  We also catch the case in which a "not edx" or "and edx,..." or "or edx,..." instruction 
//  is interposed between the sbb and the add.
void SMPBasicBlock::SuppressAnnotOnSignChangingAddition(const STARSOpndTypePtr &DefOp, int DefSSANum, STARS_ea_t DefAddr) {
	vector<SMPInstr *>::iterator DefInstIter = this->GetInstIterFromAddr(DefAddr);
	int SearchSSANum = DefSSANum;
	set<DefOrUse, LessDefUse>::iterator UseIter, DefIter;

	++DefInstIter;
	while (DefInstIter != this->GetLastInst()) {
		SMPInstr *CurrInst = (*DefInstIter);
		STARS_ea_t InstAddr = CurrInst->GetAddr();
		if (CurrInst->MDIsSmallPositiveAddition()) {
			UseIter = CurrInst->FindUse(DefOp);
			if (UseIter != CurrInst->GetLastUse()) {
				if (SearchSSANum == UseIter->GetSSANum()) {
					// Found the increment or small addition that will cause false positives.
					CurrInst->SetSuppressNumericAnnotation();
					break;
				}
				else {
					// We must have moved on past the original def-use chain.
					break;
				}
			}
		}
		else if (CurrInst->MDIsBitwiseNotOpcode() || CurrInst->MDIsBitwiseAndOpcode() || CurrInst->MDIsBitwiseOrOpcode()) {
			UseIter = CurrInst->FindUse(DefOp);
			if (UseIter != CurrInst->GetLastUse()) {
				// Bitwise operation is performed on DefOp.
				if (SearchSSANum == UseIter->GetSSANum()) {
					// Reset the SearchSSANum that we expect to see in the small addition.
					DefIter = CurrInst->GetFirstNonFlagsDef();
					STARSOpndTypePtr NewDefOp = DefIter->GetOp();
					if (IsEqOp(DefOp, NewDefOp)) {
						SearchSSANum = DefIter->GetSSANum();
					}
					else {
						// Must have an instruction such as "or dontcarereg,ourreg" which
						//  does not re-DEF ourreg and thus the chain is no longer the pattern
						//  we are looking for.
						break;
					}
				}
				else {
					// We must have moved on past the original def-use chain.
					break;
				}
			}
		}
		++DefInstIter;
	}

	return;
} // end of SMPBasicBlock::SuppressAnnotOnSignChangingAddition()

// Last-minute prep before emitting numeric annotations
void SMPBasicBlock::AnalyzePrepForNumericAnnotations(void) {

	if (this->HasCallInstruction()) {
		vector<SMPInstr *>::iterator InstIter;
		bool FoundTrustedCall = false;
		STARS_ea_t FirstTrustedCallAddr = STARS_PSEUDO_ID_MIN;
		SMPInstr *CurrInst;
		STARS_ea_t InstAddr;
		unsigned int TaintArgPosBits = 0;
		bool FoundTaintWarningCall = false;
		STARS_ea_t FirstTaintSensitiveCallAddr = 0;
		STARS_ea_t SkipOverCallAddr = 0;
		string TaintFuncName;

		// See if we have any calls to system calls that produce trusted output or are taint sensitive.
		for (InstIter = this->GetFirstInst(); InstIter != this->GetLastInst(); ++InstIter) {
			CurrInst = (*InstIter);
			InstAddr = CurrInst->GetAddr();
			if (CurrInst->IsNumericTrustedSystemCall()) {
				FoundTrustedCall = true;
				if (STARS_PSEUDO_ID_MIN == FirstTrustedCallAddr) { // still initial value
					FirstTrustedCallAddr = InstAddr;
				}
			}
			else if (FoundTrustedCall) {
				// Already found a trusted call; suppress numeric annotations.
				CurrInst->SetSuppressNumericAnnotation();
			}
			// See if we make a call to a library function that needs signedness checks on
			//  unsigned incoming args.
			if (!FoundTaintWarningCall && ((CurrInst->GetDataFlowType() == CALL) || (CurrInst->GetDataFlowType() == INDIR_CALL))) {
				// Get name of function called.
				TaintFuncName = CurrInst->GetTrimmedCalledFunctionName();

				GetTaintWarningArgPositionsForCallName(TaintFuncName, TaintArgPosBits);
				if (0 < TaintArgPosBits) {
					// Set flag so we leave TaintArgPosBits alone and stop searching for the call.
					//  NOTE: This limits us to one taint-sensitive call per block.
					FoundTaintWarningCall = true;
					FirstTaintSensitiveCallAddr = InstAddr;
				}
				else {
					SkipOverCallAddr = InstAddr;  // track last call addr before taint sensitive call addr
				}
			}
		}
		if (FoundTrustedCall && this->GetFunc()->IsBlockInAnyLoop(this->GetNumber())) {
			// Overflows can occur as numeric values are used in next iteration of the
			//  loop, so we want to suppress annotations even for insts before the call.
			// FUTURE: Make more thorough by following def-use chains. As a first cut, we
			//  will just deal with the block that has the trusted call in it.
			for (InstIter = this->GetFirstInst(); InstIter != this->GetLastInst(); ++InstIter) {
				CurrInst = (*InstIter);
				InstAddr = CurrInst->GetAddr();
				if (CurrInst->IsMarkerInst())
					continue;
				if (InstAddr != FirstTrustedCallAddr) {
					CurrInst->SetSuppressNumericAnnotation();
				}
				else {
					break;
				}
			}
		}
		if (FoundTaintWarningCall) {
			bool SkipEarlierCall = (SkipOverCallAddr > 0);
			for (InstIter = this->GetFirstInst(); InstIter != this->GetLastInst(); ++InstIter) {
				CurrInst = (*InstIter);
				InstAddr = CurrInst->GetAddr();
				if (CurrInst->IsMarkerInst())
					continue;
				if (SkipEarlierCall) {
					if (InstAddr != SkipOverCallAddr)
						continue; // ignore SSA Marker, and don't want to find arguments for trusted numeric calls.
					else {
						SkipEarlierCall = false; // found it
						continue;
					}
				}
				if (InstAddr != FirstTaintSensitiveCallAddr) {
					size_t ArgumentNumber;
					if (CurrInst->MDIsArgumentPass(ArgumentNumber)) {
						if (0 != (TaintArgPosBits & (1 << ArgumentNumber))) {
							// ArgumentNumber matches a bit set in TaintArgPosBits.
							STARSOpndTypePtr ArgSourceOp = CloneIfNecessary(CurrInst->GetMoveSource(), this->GetFunc()->UsesFramePointer());
							CanonicalizeOpnd(ArgSourceOp);
							set<DefOrUse, LessDefUse>::iterator UseIter = CurrInst->FindUse(ArgSourceOp);
							if (UseIter != CurrInst->GetLastUse()) {
								int UseSSANum = UseIter->GetSSANum();
								this->GetFunc()->ResetProcessedBlocks(); // prevent infinite recursion in IsOpSourceByteSwap()
								STARS_ea_t ByteSwapAddr;
								if (CurrInst->IsOpSourceByteSwap(ArgSourceOp, UseSSANum, ByteSwapAddr)) {
									SMP_fprintf(global_STARS_program->GetAlarmFile(), "ALARM: Call to %s at %llx receives user-tainted argument at %llx which was tainted at %llx\n", 
										TaintFuncName.c_str(), (unsigned long long) FirstTaintSensitiveCallAddr, (unsigned long long) InstAddr, (unsigned long long) ByteSwapAddr);
									break;
								}
							}
						}
					}
				}
				else {
					break;
				}
			}
		}
	} // end if block has call
	else if (STARS_HASH_BLOCK_MIN_INSTRS <= this->InstVec.size()) {
		STARS_ea_t InstAddr;
		size_t HashingArithmeticCount = 0;
		for (vector<SMPInstr *>::iterator InstIter = this->GetFirstInst(); InstIter != this->GetLastInst(); ++InstIter) {
			SMPInstr *CurrInst = (*InstIter);
			InstAddr = CurrInst->GetAddr();
			if (CurrInst->MDIsHashingArithmetic()) {
				++HashingArithmeticCount;
			}
		}
		if (STARS_HASHING_OPERATIONS_THRESHOLD <= HashingArithmeticCount) {
			for (vector<SMPInstr *>::iterator InstIter = this->GetFirstInst(); InstIter != this->GetLastInst(); ++InstIter) {
				SMPInstr *CurrInst = (*InstIter);
				InstAddr = CurrInst->GetAddr();
				if (CurrInst->MDIsOverflowingOpcode() || CurrInst->MDIsUnderflowingOpcode() || CurrInst->MDIsLoadEffectiveAddressInstr()) {
					CurrInst->SetHashOperation();
				}
			}
		}
	}
	return;
} // end of SMPBasicBlock::AnalyzePrepForNumericAnnotations()

// Trace critical args for critical library calls back to InArgs, if possible. Return true if critical arg traced to InArg.
bool SMPBasicBlock::MarkTaintWarningInArgs(void) {
	bool TrackedToInArg = false;
	if (this->HasCallInstruction()) {
		bool CriticalCallFound = false;
		unsigned int TaintArgPosBits = 0;
		unsigned int TaintArgPositionsSeen = 0;
		for (vector<SMPInstr *>::const_reverse_iterator InstIter = this->GetRevInstCBegin(); InstIter != this->GetRevInstCEnd(); ++InstIter) {
			SMPInstr *CurrInst = (*InstIter);
			SMPitype FlowType = CurrInst->GetDataFlowType();
			bool CallInst = ((CALL == FlowType) || (INDIR_CALL == FlowType));
			if (CallInst) {
				string TaintFuncName = CurrInst->GetTrimmedCalledFunctionName();
				GetTaintWarningArgPositionsForCallName(TaintFuncName, TaintArgPosBits);
				CriticalCallFound = (0 < TaintArgPosBits); // found a critical call if any critical arg
			}
			else if (CriticalCallFound) { // working on args for a critical call
				size_t ArgumentNumber;
				STARS_ea_t InstAddr = CurrInst->GetAddr();
				if (CurrInst->MDIsArgumentPass(ArgumentNumber)) {
					unsigned int NewBitPos = (1 << ArgumentNumber);
					bool AlreadySeen = (TaintArgPositionsSeen & NewBitPos);
					TaintArgPositionsSeen |= NewBitPos;
					if (!AlreadySeen && (0 != (TaintArgPosBits & NewBitPos))) {
						// ArgumentNumber matches a bit set in TaintArgPosBits.
						STARSOpndTypePtr ArgSourceOp = CurrInst->GetMoveSource();
						ArgSourceOp = CloneIfNecessary(ArgSourceOp, this->GetFunc()->UsesFramePointer());
						if ((nullptr != ArgSourceOp) && ArgSourceOp->IsRegOp()) {
							CanonicalizeOpnd(ArgSourceOp);
							set<DefOrUse, LessDefUse>::iterator UseIter = CurrInst->FindUse(ArgSourceOp);
							if (UseIter != CurrInst->GetLastUse()) {
								int UseSSANum = UseIter->GetSSANum();
								bool LocalName = this->IsLocalName(ArgSourceOp);
								STARSOpndTypePtr SourceOp = nullptr;
								STARS_ea_t UltimateDefAddr = this->GetUltimateDefAddr(ArgSourceOp, InstAddr, UseSSANum, LocalName, true, false, SourceOp);
								if (STARS_IsSSAMarkerPseudoID(UltimateDefAddr)) {
									// traced to SSA Marker => InArg or LiveIn
									STARS_regnum_t RegNum = ArgSourceOp->GetReg();
									size_t ArgPos = 0;
									if (global_STARS_program->GetArgRegPosition(RegNum, ArgPos)) {
										TrackedToInArg = true;
										assert(ArgPos < (8 * sizeof(unsigned int))); // can set ArgPos bit in unsigned int
										this->GetFunc()->MarkTaintInArgReg(ArgPos);
									}
								}
							}
						}
					}
				}
			}
		}
	}
	return TrackedToInArg;
} // end of SMPBasicBlock::MarkTaintWarningInArgs()

// Unsigned branch should be treated as unknown signedness because of code idiom?
bool SMPBasicBlock::IsBranchSignednessUnreliable(void) {
	// There are two code patterns that make an unsigned branch an unreliable
	//  basis for inferring that the related operands were unsigned.
	//
	// First, we are looking for the pattern:
	// ja addr1
	// jb addr1
	// This is a stupid, inefficient way of saying jne addr1. The unsigned
	//  nature of the operands that produced the flags when compared cannot
	//  be reliably inferred, because this sequence is the same as a jne even
	//  for signed operands.
	//
	// Second, a typical pattern for optimizing compilers:
	// Source: if ((x >= 500) && (x <= 520)) where x is a signed int
	// Instead of generating two compares and two branches, only
	//  one branch is needed:
	//  sub ebx,500
	//  cmp ebx,20
	//  ja not_in_range
	// The jump-if-above is a trick to catch the less than 500 case (which
	//  is now a negative value in ebx) by pretending it is a large unsigned
	//  number, while also catching the greater than 520 case in the same
	//  conditional branch. One conditional branch is much faster than two
	//  in a pipelined machine, but it uses an unsigned conditional branch on
	//  signed data, which screws up our signedness inference.
	// Another variation is a small addition to get a range that spans zero to
	//  be all non-negative:
	//  add eax,1
	//  cmp eax,2  ; if ((eax < -1) || (eax > 1)) ...
	//  ja not_in_range
	//
	// Third, we look for a pattern such as:
	//  cmp eax,0xfffffffc
	//  ja loopexit
	// It is unlikely that an unsigned number is being compared to a value so
	//  close to the maximum unsigned value, with branching resulting. Instead,
	//  it is common for a decrementing counter to get cut off at a small negative
	//  value with this code pattern, giving a SIGNED var the false appearance of
	//  UNSIGNED.
	bool JumpAbove = false;
	bool JumpBelow = false;
	STARS_ea_t TargetAddr = STARS_BADADDR, BranchAddr = STARS_BADADDR;
	SMPInstr *BranchInst;
	unsigned short opcode;
	vector<SMPInstr *>::reverse_iterator InstRevIter;
	vector<SMPInstr *>::iterator InstIter;

	if (1 == this->InstVec.size()) {
		// Might be the second instruction in the sequence.
		InstIter = this->GetFirstInst();
		BranchInst = (*InstIter);
		BranchAddr = BranchInst->GetAddr();
		opcode = BranchInst->GetIDAOpcode();
		JumpAbove = (opcode == STARS_NN_ja);
		JumpBelow = (opcode == STARS_NN_jb);
		if (JumpAbove || JumpBelow) {
			// We look for the simple pattern of fall-through-only from first block
			//  to second block.
			if (1 == this->Predecessors.size()) {
				TargetAddr = BranchInst->GetJumpTarget();
				list<SMPBasicBlock *>::iterator PredIter = this->GetFirstPred();
				InstRevIter = (*PredIter)->GetRevInstBegin();
				BranchInst = (*InstRevIter);
				opcode = BranchInst->GetIDAOpcode();
				if (JumpBelow)
					JumpAbove = (opcode == STARS_NN_ja);
				else if (JumpAbove)
					JumpBelow = (opcode == STARS_NN_jb);
			}
		}
	}
	else {
		// Might be the first instruction in the sequence.
		InstRevIter = this->GetRevInstBegin();
		BranchInst = (*InstRevIter);
		BranchAddr = BranchInst->GetAddr();
		opcode = BranchInst->GetIDAOpcode();
		JumpAbove = (opcode == STARS_NN_ja);
		JumpBelow = (opcode == STARS_NN_jb);
		if (JumpAbove || JumpBelow) {
			// We look for the simple pattern of fall-through-only from first block
			//  to second block, plus conditional branch target.
			if (2 == this->Successors.size()) {
				TargetAddr = BranchInst->GetJumpTarget();
				list<SMPBasicBlock *>::const_iterator SuccIter = this->GetFallThroughSucc();
				if (SuccIter != this->GetLastSucc()) {
					InstIter = (*SuccIter)->GetFirstInst();
					BranchInst = (*InstIter);
					opcode = BranchInst->GetIDAOpcode();
					if (JumpBelow)
						JumpAbove = (opcode == STARS_NN_ja);
					else if (JumpAbove)
						JumpBelow = (opcode == STARS_NN_jb);
				}
			}
		}
	}
	if (JumpAbove && JumpBelow && (BranchInst->GetJumpTarget() == TargetAddr))
		return true;

#if STARS_DETECT_OPTIMIZED_RANGE_CHECK
	// Now, search for the second and third cases.
	bool OptimizedRangeComparison = false;
	bool TooLargeCompareConstant = false;
	InstRevIter = this->GetRevInstBegin();
	BranchInst = (*InstRevIter);
	// Find the flags USE.
	STARSOpndTypePtr UseOp = BranchInst->MakeRegOpnd(X86_FLAGS_REG); // set up a dummy op for searching
	set<DefOrUse, LessDefUse>::iterator UseIter = BranchInst->FindUse(UseOp);
	assert(UseIter != BranchInst->GetLastUse());
	int UseSSANum = UseIter->GetSSANum();
	bool LocalFlags = this->IsLocalName(UseOp);
	STARS_ea_t DefAddr = this->GetDefAddrFromUseAddr(UseOp, BranchInst->GetAddr(), UseSSANum, LocalFlags);
	// Phi DEFs are rarely or never found as the source of the flags for a branch, so keep it simple.
	if (!STARS_IsBlockNumPseudoID(DefAddr)) { // DefAddr is an inst, not a Phi DEF.
		// In order to fit the pattern, we need to see a comparison to a positive immediate constant.
		SMPInstr *CompareInst = this->GetFunc()->GetInstFromAddr(DefAddr);
		STARSOpndTypePtr CompareOperand = nullptr;
		STARS_uval_t CompareConstant, SubtractConstant;
		if (CompareInst->MDIsCompareToPositiveConstant(CompareOperand, CompareConstant)) {
			// For the third pattern, we just need to have a really large CompareConstant.
			if (CompareConstant >= 0xfffffff0) {
				TooLargeCompareConstant = true;
			}
			else {
				// To complete the second pattern, we need CompareOperand to have been DEFed by a 
				//  subtraction of a constant, or the addition of a small positive constant.
				if (CompareOperand->IsRegOp()) {
					CanonicalizeOpnd(CompareOperand);
					UseIter = CompareInst->FindUse(CompareOperand);
					assert(UseIter != CompareInst->GetLastUse());
					UseSSANum = UseIter->GetSSANum();
					SMPBasicBlock *CompareBlock = CompareInst->GetBlock(); // just in case it's not this block
					bool LocalCompName = CompareBlock->IsLocalName(CompareOperand);
					STARS_ea_t SubtractAddr = CompareBlock->GetDefAddrFromUseAddr(CompareOperand, DefAddr, UseSSANum, LocalCompName);
					if (!STARS_IsBlockNumPseudoID(SubtractAddr)) {
						SMPInstr *SubtractInst = this->GetFunc()->GetInstFromAddr(SubtractAddr);
						STARSOpndTypePtr MinuendOp = nullptr;
						if (SubtractInst->IsSubtractionOfConstant(MinuendOp, SubtractConstant)) {
							OptimizedRangeComparison = true;
						}
						else if (SubtractInst->MDIsSmallPositiveAddition()) {
							OptimizedRangeComparison = true;
						}
					}
				}
			}
		}
	}

	return (OptimizedRangeComparison || TooLargeCompareConstant);
#else
	return false;
#endif


} // end of SMPBasicBlock::IsBranchSignednessUnreliable()

// Find stack adjustment code, if any, after CallAddr.
STARS_sval_t SMPBasicBlock::ComputeStackAdjustmentAfterCall(STARS_ea_t CallAddr) {
	STARS_sval_t AdjustmentBytes = 0;
	STARS_sval_t PriorAdjustmentBytes = 0; // stack deltas before the call
	vector<SMPInstr *>::iterator InstIter;
	STARS_ea_t InstAddr; // for debugging, breakpoints, etc.
	bool FoundCallInst = false;
	bool FirstBlock = (this->GetFirstAddr() == this->GetFunc()->GetFirstFuncAddr());

	for (InstIter = this->GetFirstInst(); InstIter != this->GetLastInst(); ++InstIter) {
		SMPInstr *CurrInst = (*InstIter);
		InstAddr = CurrInst->GetAddr();
		// Two stage search: Find CALL instruction, then find next stack adjustment.
		STARS_sval_t CurrentDelta = CurrInst->FindStackAdjustment();
		SMPitype DataFlowType = CurrInst->GetDataFlowType();
		if (!FoundCallInst) {
			FoundCallInst = (InstAddr == CallAddr);
			if (!FoundCallInst && !FirstBlock) {
				PriorAdjustmentBytes += CurrentDelta;
				if (JUMP <= DataFlowType) {
					// Reset PriorAdjustmentBytes and exit. Basic block is too confusing
					//  to analyze if multiple calls happen. We cannot tell if stack adjustments
					//  between the first call and the second call are related to the first call
					//  or the second call.
					PriorAdjustmentBytes = 0;
					break;
				}
			}
		}
		else {
			if (JUMP > DataFlowType) {
				if (SMP_STACK_DELTA_ERROR_CODE == ((STARS_uval_t) CurrentDelta)) {
					// An error will soon occur, no need for further computations.
					AdjustmentBytes = 0;
					SMP_msg("INFO: Stack adjustment set to zero due to error code at %llx\n", (unsigned long long) InstAddr);
					break;
				}
				else if (SMP_STACK_POINTER_BITWISE_AND_CODE == ((STARS_uval_t) CurrentDelta)) {
					; // Ignore AND with stack pointer for now
				}
				else if (0 != CurrentDelta) {
					SMP_msg("INFO: Found stack adjustment of %ld bytes at %llx\n", 
						(long) CurrentDelta, (unsigned long long) InstAddr);
					AdjustmentBytes += CurrentDelta;
				}
				// NOTE: ****!!!!****!!!! We want to find changes in stack direction and terminate
				//  in the future, in case we see code such as:
				//
				//  call foo
				//  pop eax  ; adjust stack after call to foo, grab return arg and put in eax
				//  push ebx ; caller-save before call to bar
				//  call bar
			}
			else if ((RETURN == DataFlowType) 
				|| ((!this->GetFunc()->HasExplicitReturnInstruction()) && ((INDIR_JUMP == DataFlowType) || (JUMP == DataFlowType)))) {
				// We have been counting up adjustment bytes that precede a RETURN. Those
				//  adjustments are expected before a RETURN and have nothing to do with
				//  the function call at CallAddr. If the function has no explicit return instructions, we will
				//  conservatively treat jumps as probable tail calls, which will later be given the RETURN
				//  data flow type but have not yet been analyzed and marked.
				AdjustmentBytes = 0;
				SMP_msg("INFO: Resetting stack adjustment to zero at return inst at %llx\n", 
					(unsigned long long) InstAddr);
				break;
			}
			else {
				// We want to break at calls after CallAddr, and other instructions that
				//  hit this break would terminate the block anyway (e.g. jumps, branches).
				break;
			}
		}
	}

	if (AdjustmentBytes != 0) {
		// PriorAdjustmentBytes is usually negative (making stack space before the call)
		//  and AdjustmentBytes is usually positive (returning stack space after the call).
		//  Adding them gets the net adjustment after the call.
		AdjustmentBytes += PriorAdjustmentBytes;
	}
	return AdjustmentBytes;
} // end of SMPBasicBlock::ComputeStackAdjustmentAfterCall()

// Find the push operand that passed the fall-through address to a fixed call
STARS_ea_t SMPBasicBlock::FindFixedCallFallThrough(void) const {
	STARS_ea_t FallThroughAddr = STARS_BADADDR;

	for (vector<SMPInstr *>::const_reverse_iterator RevInstIter = this->GetRevInstCBegin(); RevInstIter != this->GetRevInstCEnd(); ++RevInstIter) {
		SMPInstr *CurrInst = (*RevInstIter);
		if (CurrInst->IsFixedCallPush()) {
			bool success = false;
			STARS_InstructionID_Set_t ReferencedIDs = CurrInst->GetReferencedIDs(success);
			if (success) {
				if (1 == ReferencedIDs.size()) {
					FallThroughAddr = ReferencedIDs.cbegin()->GetIDWithinFile();
				}
				else {
					SMP_msg("ERROR: Fixed call has %zu fall-through addrs.\n", ReferencedIDs.size());
				}
			}
			break;
		}
	}

	return FallThroughAddr;
} // end of SMPBasicBlock::FindFixedCallFallThrough()

// Find inst after CallAddr; grab from successor if needed
STARS_ea_t SMPBasicBlock::FindCallInstFallThrough(STARS_ea_t CallAddr) const {
	STARS_ea_t FallThroughAddr = STARS_BADADDR;
	vector<SMPInstr *>::const_iterator CallInstIter = this->GetInstConstIterFromAddr(CallAddr);
	SMPInstr *CallInst = (*CallInstIter);
	assert(DEFAULT != CallInst->GetDataFlowType());

	if (CallInst->IsFixedCallJump()) {
		FallThroughAddr = this->FindFixedCallFallThrough();
	}
	else if (CallInst->IsTailCall() || CallInst->IsBranchToOtherFunc()) {
		SMP_msg("WARNING: Returning BADADDR for tail call fall-through at %llx\n", (unsigned long long) CallAddr);
	}
	else { // regular call
		++CallInstIter;
		if (CallInstIter != this->GetLastConstInst()) {
			// Simple case: Next inst is inside the block.
			FallThroughAddr = (*CallInstIter)->GetAddr();
		}
		else { // CallInst was last inst in block.
			list<SMPBasicBlock *>::const_iterator SuccIter = this->GetFallThroughSucc();
			if (SuccIter != this->GetLastConstSucc()) {
				FallThroughAddr = (*SuccIter)->GetFirstAddr();
			}
			else {
				FallThroughAddr = STARS_LIVEIN_PSEUDO_ID; // signal error case as opposed to tail call case
			}
		}
	}

	return FallThroughAddr;
} // end of SMPBasicBlock::FindCallInstFallThrough()

// Is reg DefOp+DefSSANum (popped from WorkList) at DefAddr used as address reg or as source operand in memory write?
bool SMPBasicBlock::IsDefUsedInUnsafeMemWrite(list<pair<pair<STARSOpndTypePtr, int>, STARS_ea_t> > &WorkList, const STARSOpndTypePtr &DefOp, int DefSSANum, STARS_ea_t DefAddr) {
	bool FoundMemWriteUse = false;
	bool InvariantDEF = false;
	assert(DefOp->IsRegOp());
	if (!(this->IsProcessed())) {
		this->SetProcessed(true);
		bool ReDEF = false;
		vector<SMPInstr *>::iterator InstIter;
		if (STARS_IsLiveInPseudoID(DefAddr)) { // live into block
			InstIter = this->GetFirstInst();
			// Search for Phi function DEF.
			set<SMPPhiFunction, LessPhi>::iterator PhiIter = this->FindPhi(DefOp);
			if (PhiIter != this->GetLastPhi()) {
				DefSSANum = PhiIter->GetDefSSANum(); // get new SSANum to search for
			}
		}
		else {
			InstIter = this->GetInstIterFromAddr(DefAddr);
			SMPInstr *DefInst = (*InstIter);
			set<DefOrUse, LessDefUse>::iterator DefIter = DefInst->FindDef(DefOp);
			assert(DefIter != DefInst->GetLastDef());
			InvariantDEF = DefIter->IsInvariantForAllLoops();
			++InstIter;
		}
		while (InstIter != this->GetLastInst()) {
			SMPInstr *CurrInst = (*InstIter);
			set<DefOrUse, LessDefUse>::iterator UseIter = CurrInst->FindUse(DefOp);
			set<DefOrUse, LessDefUse>::iterator DefIter;
			if (UseIter != CurrInst->GetLastUse()) {
				// Is it used as an address reg or SourceOp in a memory write?
				DefIter = CurrInst->GetFirstNonFlagsDef();
				if (DefIter == CurrInst->GetLastDef()) {
					++InstIter;
					continue;
				}
				STARSOpndTypePtr NewDefOp = DefIter->GetOp();
				assert(nullptr != NewDefOp);
#if 1
				bool DangerousWrite = (!InvariantDEF) && CurrInst->IsUnsafeMemWrite();
#else
				bool DangerousWrite = CurrInst->HasDestMemoryOperand();
				if (DangerousWrite) {
					// Analyze whether it is truly dangerous.
					if (MDIsDirectStackAccessOpnd(NewDefOp, this->GetFunc()->UsesFramePointer())) {
						if (this->GetFunc()->OutArgsRegionComputed() && this->GetFunc()->IsInOutgoingArgsRegion(NewDefOp)) {
							DangerousWrite = false;
						}
					}
					else if (!(this->GetFunc()->IsBlockInAnyLoop(this->GetNumber()))) {
						// Without looping, the only danger is having an index register rather than
						//  a direct write through DefOp, or storing DefOp itself.
						if (IsMemOperand(NewDefOp)) { 
							int BaseReg, IndexReg;
							uint16_t ScaleFactor;
							STARS_ea_t Offset;
							MDExtractAddressFields(NewDefOp, BaseReg, IndexReg, ScaleFactor, Offset);
							if ((BaseReg == DefOp->GetReg()) && ((IndexReg == STARS_x86_R_none) || (MD_STACK_POINTER_REG == IndexReg))) {
								// Direct write through DefOp reg. Only danger is the unlikely case that DefOp appears twice,
								//   e.g. mov [eax],eax in which case it is hard to track eax later in memory.
								DangerousWrite = CurrInst->IsNonAddressReg(DefOp);
							}
						}
					}
				}
#endif
				SMPitype FlowType = CurrInst->GetDataFlowType();
				if (DangerousWrite || CurrInst->MDIsPushInstr() || (INDIR_CALL == FlowType)) {
					// 1. USEd in a memory write inst, so must be either address reg or source of write or of arithmetic into mem dest.
					// 2. USE in indirect call, so who knows if it is used in callee mem write. Safe to assume it is.
					FoundMemWriteUse = true;
					break;
				}
				else {
					if ((FlowType == DEFAULT) && (NewDefOp->IsRegOp()) && CurrInst->IsNonAddressReg(DefOp)) {
						// USEd to create a new DEF by arithmetic or by a copy.
						int NewDefSSANum = DefIter->GetSSANum();
						pair<STARSOpndTypePtr, int> NewSearchItem(NewDefOp, NewDefSSANum);
						pair<pair<STARSOpndTypePtr, int>, STARS_ea_t> NewWorkListItem(NewSearchItem, CurrInst->GetAddr());
						WorkList.push_back(NewWorkListItem);
					}
				}
			} // end if USE

			// See if CurrInst is a re-DEF of DefOp.
			DefIter = CurrInst->FindDef(DefOp);
			if (DefIter != CurrInst->GetLastDef()) { // re-DEF
				// No more searching on this chain, will search NewDefSSANum when we reach that item in the WorkList.
				ReDEF = true;
				break;
			}
			++InstIter;
		} // end while InstIter not at end of block

		// Are we done, or do we need to recurse into successor blocks?
		if (!FoundMemWriteUse && (!ReDEF)) { // have to keep searching
			list<SMPBasicBlock *>::iterator SuccIter = this->GetFirstSucc();
			while (SuccIter != this->GetLastSucc()) {
				SMPBasicBlock *SuccBlock = (*SuccIter);
				if (SuccBlock->IsLiveIn(DefOp)) {
					STARS_ea_t NewDefAddr = STARS_LIVEIN_PSEUDO_ID;
					FoundMemWriteUse = SuccBlock->IsDefUsedInUnsafeMemWrite(WorkList, DefOp, DefSSANum, NewDefAddr);
					if (FoundMemWriteUse)
						break;
				}
				++SuccIter;
			}
		}
	} // end if block not yet processed.

	return FoundMemWriteUse;
} // end of SMPBasicBlock::IsDefUsedInUnsafeMemWrite()

// Emit SPARK Ada short circuit expr for entire expression tree.
void STARSCondExpr::EmitSPARKAdaShortCircuitExpr(FILE *BodyFile) {
	SMPoperator CurrOper = this->GetOperator();
	assert((SMP_BITWISE_NOT == CurrOper) || (SMP_LOGICAL_AND == CurrOper) || (SMP_LOGICAL_OR == CurrOper));
	if (CurrOper == SMP_BITWISE_NOT) {
		// Special case: Unary NOT as prefix operator, before we emit LeftExpr
		SMP_fprintf(BodyFile, "not (");
	}

	// Emit LeftExpr
	if (this->NegateLeftExpr) {
		SMP_fprintf(BodyFile, "not (");
	}
	if (nullptr == this->GetLeftExpr()) {
		SMPBasicBlock *LeftBlock = this->MyCFGBlock->GetFunc()->GetBlockByNum((size_t) this->GetLeftBlockNum());
		if (!this->NegateLeftExpr) {
			SMP_fprintf(BodyFile, " (");
		}
		LeftBlock->EmitSPARKAdaForExpression(BodyFile);
#if 1
		if (!this->NegateLeftExpr)
#endif
			SMP_fprintf(BodyFile, ") ");
	}
	else {
		// Recurse into LeftExpr
		this->GetLeftExpr()->EmitSPARKAdaShortCircuitExpr(BodyFile);
	}
	if (this->NegateLeftExpr) { // close "not" parenthesis
		SMP_fprintf(BodyFile, ") ");
	}

	// Emit operator
	if (SMP_LOGICAL_AND == CurrOper) {
		SMP_fprintf(BodyFile, "and then ");
	}
	else if (SMP_LOGICAL_OR == CurrOper) {
		SMP_fprintf(BodyFile, "or else ");
	}
	else {  // NOT
		assert(this->GetRightExpr() == nullptr);
		assert(this->GetRightBlockNum() == SMP_BLOCKNUM_UNINIT);
		SMP_fprintf(BodyFile, ") "); // close parenthesis
		return; // no right expr
	}

	// Emit right expr
	if (nullptr == this->GetRightExpr()) {
		SMPBasicBlock *RightBlock = this->MyCFGBlock->GetFunc()->GetBlockByNum((size_t) this->GetRightBlockNum());
		SMP_fprintf(BodyFile, " (");
		RightBlock->EmitSPARKAdaForExpression(BodyFile);
		SMP_fprintf(BodyFile, ") ");
	}
	else {
		// Recurse into RightExpr
		this->GetRightExpr()->EmitSPARKAdaShortCircuitExpr(BodyFile);
	}

	return;
} // end of STARSCondExpr::EmitSPARKAdaShortCircuitExpr()



